/*
 * main.c -
 *  main entry point
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"
#include "types.h"
#include "graph2d.h"

#include <stdio.h>
#include <stdlib.h>  /* malloc() */
#include <varargs.h>

#include "gamemem.h"
#include "game.h"
#include "gui.h"
#include "main.h"
#include "matrix.h"
#include "option.h"

#include "event.h"

#include "sound.h"

#include "hud.h"
#include "mfd.h"


/*
 * Global variables
 */
int Screen_width  = 320;
int Screen_height = 240;


int    Argc;
char **Argv;


int main (int argc, char *argv [])
{
	bounding_box bMessages;
  int parm;

  /*
   * Set the command-line options that may be mangled by the GUI
   */
  Argc = argc;
  Argv = argv;

	Mem_init();
	Evt_init();

  /*
   * The order of initialisation is important. First GUI_init must be called,
   *  as the underlying GUI might want to take some of the command-line options
   *  for itself. Then Opt_parse() for TSSHP's own options, and only then may
   *  we call GUI_init_screen() as the screen dimensions have now been
   *  finalised
   */
  GUI_init (&Argc, Argv);

  parm = Opt_parse (Argc, Argv, 1);
  if (parm < Argc)
  {
    fprintf (stderr, "Warning: non-option option `%s':"
	             " it and subsequent options will be ignored\n",
	     Argv [parm]);
  }

  Screen_width  = Option_screen_width;
  Screen_height = Option_screen_height;

  Game_init ();

	Sound_init();       /* has to go after resources are loaded so we know if
                         we've got logs or not                              */
  GUI_init_screen ();

	HUD_init(Screen_width, Screen_height);
	MFD_init(Screen_width, Screen_height);

	bMessages.top = (Screen_height % 200);
	bMessages.left = (Screen_width % 300) * 3 / 2;
	bMessages.right = Screen_width - bMessages.left;
	bMessages.bottom = Screen_height / 2;
	MSG_init(bMessages);

  if (!Option_underworld) /* tries to load a System Shock resource */
  {
		MFD_engage_mode(C_MFD_MODE_WORLD);

    MFD_set_HW_installed(C_MFD_HW_BIO, true);
    MFD_set_HW_installed(C_MFD_HW_SENSA, true);
    MFD_set_HW_installed(C_MFD_HW_COMPASS, true);

    MFD_set_active_id(C_MFD_SIDE_RIGHT, 
                      mfd_set_id(F_SMFD_MAP, F_SMFD_MSIDE, 0x0000), 
                      true);
    MFD_show_media_relative(C_MFD_MEDIA_EMAIL, 27);  /* "Look at you, Hacker ..." */
  }

  HUD_show_cursor(0);	/* show standard cursor */

  GUI_loop ();

  /*
   * Not supposed to get here ...
   */
  printf ("main.c(): warning: GUI_loop() returned! I'm confused ...\n");
  return 0;

}




void *Alloc_init (size_t size, const char *funcname, const char *desc)
{
  void *stuff = malloc (size);
  if (stuff == NULL)
  {
    fprintf (stderr, "%s(): couldn't allocate space for %s\n", funcname, desc);
    Cant_happen ("Out of memory!");
  }
# if MEM_INIT_MESSAGES
  printf ("%s(): %d bytes allocated for %s\n", funcname, size, desc);
# endif
  return stuff;
}



/*
 * Error handling
 */
void Cant_happen (const char *message)
{
  /*
   * Couldn't resist it 8-)
   */
  fprintf (stderr, "\nRIP shack, killed by a system shock\n\n");
  fprintf (stderr, "%s\n\n", message);
  fprintf (stderr, "Salt the Fries!\n\n");
  exit (1);
}


/*
 * End of os/linux/main.c
 */

/*
 * Emacs magic
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
