/*
 * main.c -
 *  main entry point
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */

/* ff1: now i got it. min/max are so called Language Extensions of */
/*			VC -- as we have to let them enabled for this file (see		 */
/*			Project settings) because of the windows header,					 */
/*			the define for min/max can be taken from there.						 */
#include "defines.h"

#include <stdio.h>
#include <stdlib.h> 
#include <assert.h>

#include <windows.h>

#include "types.h"
#include "event.h"
#include "game.h"
#include "gamemem.h"
#include "graph2d.h"
#include "gui.h"
#include "hud.h"
#include "main.h"
#include "matrix.h"
#include "mfd.h"
#include "option.h"
#include "palette.h"
#include "resource.h"
#include "shade.h"
#include "sound.h"
#include "timer.h"


/*
 * local functions
 */ 

void exit_shutdown(void);	/* gets called at exit() */


/*
 * Global variables
 */
HINSTANCE ghInstance = NULL;

int Screen_width  = 320;
int Screen_height = 240;

#if defined(GUI_DIRECTDRAW)
/* fake a command line */
int    Argc = 0;
char	*Argv[] = 
{
	"dshack",
	NULL
};
#else
int		 Argc = 0;
char **Argv = NULL;
#endif

#if defined(GUI_DIRECTDRAW)
int WINAPI WinMain(
    HINSTANCE hInstance,			/* handle to current instance		*/
    HINSTANCE hPrevInstance,	/* handle to previous instance	*/
    LPSTR lpCmdLine,					/* pointer to command line			*/
    int nCmdShow 							/* show state of window					*/
   )
#else
int main (int argc, char *argv []) 
#endif
{
	bounding_box bMessages;
  int parm;

#if defined(GUI_DIRECTDRAW)
	/* used in directx.c */
	ghInstance = hInstance;
	/* ff1: get this command parsing here sometime... */
	/*			but will we still do it later on?					*/
#else
  Argc =  argc;
  Argv =  argv; 
#endif

	Mem_init();
	Evt_init();
  GUI_init (&Argc, Argv);

  /*
   * Command-line options. Must do this firstest
   */
  parm = Opt_parse (Argc, Argv, 1);
  if (parm < Argc)
  {
    fprintf (stderr, "Warning: non-option option `%s':"
	           " it and subsequent options will be ignored\n",
						 Argv [parm]);
  }

  Screen_width  = Option_screen_width;
  Screen_height = Option_screen_height;

#if defined(GUI_DIRECTDRAW)
	/* have to do that currently until i've got out the bugs... */
	Screen_width = 320;
	Screen_height = 240;
/* /that/ i call a teaser ;) */
/*
	Screen_width  = 1024;
	Screen_height = 768;
*/
#endif

  Game_init ();

	Sound_init();       /* has to go after resources are loaded so we know if
                         we've got logs or not                              */
  GUI_init_screen ();

	HUD_init((uint16)Screen_width, (uint16)Screen_height);
	MFD_init(Screen_width, Screen_height);

	bMessages.top = (Screen_height * 15) / 100;
	bMessages.left = (Screen_width * 10) / 100;
	bMessages.right = Screen_width - bMessages.left;
	bMessages.bottom = Screen_height / 2;
	MSG_init(bMessages);

  if (!Option_underworld)
  {
		MFD_engage_mode(C_MFD_MODE_WORLD);

    MFD_set_HW_installed(C_MFD_HW_BIO, true);
    MFD_set_HW_installed(C_MFD_HW_SENSA, true);
    MFD_set_HW_installed(C_MFD_HW_COMPASS, true);
/*
    MFD_set_active_id(C_MFD_SIDE_LEFT, 
											mfd_set_id_s(F_SMFD_DATA, F_SMFD_DCONT, 0x0003),
											true);																 
*/
    MFD_set_active_id(C_MFD_SIDE_RIGHT, 
                      mfd_set_id(F_SMFD_MAP, F_SMFD_MSIDE, 0x0000), 
                      true);
/*    MFD_show_media_relative(C_MFD_MEDIA_EMAIL, 10); */ /* "Damn ..." */

    MFD_show_media_relative(C_MFD_MEDIA_EMAIL, 27);  /* "Look at you, Hacker ..." */
/*
		MFD_set_active_id(C_MFD_SIDE_LEFT, mfd_set_id_s(F_SMFD_DATA, F_SMFD_DKEYPAD, 0x1234), true);
		MFD_set_active_id(C_MFD_SIDE_RIGHT, mfd_set_id_s(F_SMFD_DATA, F_SMFD_DKEYPAD, 0x1234), true);
*/
		
  }

	atexit(exit_shutdown);

	GUI_loop ();

  /*
   * Not supposed to get here ...
   */
  MessageBox (NULL, "main.c(): warning",
              "GUI_loop() returned! I'm confused ...", MB_OK);
  return 0;

}

void exit_shutdown(void)
{
	Game_deinit();
	MFD_deinit();
	MSG_deinit();
	HUD_deinit();
  Res_free_all();
	Timer_shutdown();
	Sound_deinit();
	Evt_deinit();
	Pal_deinit();
	Mem_deinit();
  printf("bye-bye from shack\n");
}

/*
 * Error handling
 */
void Cant_happen (const char *message)
{

  /*
   * Put up a MessageBox to make sure the user is alerted
   */
#ifndef _DEBUG
  MessageBox (NULL, message, "Salt the Fries!", MB_OK);
#else
	/* ff1: sometimes it's usefull to preserve the current state... */
	if (MessageBox(NULL, message, "Salt the Fries! (Cancel for debug)",
								 MB_OKCANCEL) == IDCANCEL)
	{
		assert(false);
	}
#endif

  /*
   * Couldn't resist it 8-)
   */
  fprintf (stderr, "\nRIP shack, killed by a system shock\n\n");
  fprintf (stderr, "%s\n\n", message);
  fprintf (stderr, "Salt the Fries!\n\n");
  exit (1);
}


void *Alloc_init (size_t size, const char *funcname, const char *desc)
{
  void *stuff = malloc (size);
  if (stuff == NULL)
  {
    fprintf (stderr, "%s(): couldn't allocate space for %s\n", funcname, desc);
    Cant_happen ("Out of memory!");
  }
# if MEM_INIT_MESSAGES
  printf ("%s(): %lu bytes allocated for %s\n", funcname, size, desc);
# endif
  return stuff;
}



/*
 * End of os/win32/main.c
 */

/*
 * Emacs magic
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
