/*
 *  sound_core.c - Core Sound Routines
 */

/*
 * File:    sound/sound_core.c
 * Purpose: Audio handling
 * Prefix:  Sound_
 * Notes:	Audio Handling!
 * Music:   Hybrid, "Kid 2000"
 *
 * This file by Bryan Nolen 2001-2002 bryan@r3v3ng.net
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */

#include <stdio.h>
#include <stdlib.h>
#include "gamemem.h"
#include <math.h>
#include <string.h>										/* strncpy() - msgs */

#include "defines.h"
#include "types.h"
#include "option.h"
#include "event.h"

#include "sound.h"
#include "sound_core.h"

#include "main.h"											/* Can't_happen */
#include "resource.h"


/*
 * jim - from game.c
 * This is set if citalog.res has been found i.e. we're running off the CD and
 *  have audio logs
 */
extern bool Have_audio_logs;

/*
 * And similarly for vidmail
 */
extern bool Have_video_mail;

/*
 * These are the main event handlers
 */
void sound_event_play(HEVENTCLASS hClass, HEVENT hEvent, void *pData);
void sound_event_stop(HEVENTCLASS hClass, HEVENT hEvent, void *pData);
int sound_play_log(short log_num, int mediaType);
int sound_play_fx_2d(int fx_num);

void Sound_init(void)
{
	/*
	 * Lets register us to grab those events!
	 */
	HEVENTCLASS hClass = Evt_get_class("MFD");

	/*
	 * jim - don't register for logs and emails if we don't have the audio
	 */
	if (Have_audio_logs)
	{
	  Evt_register(hClass, Evt_get_event(hClass, "PlayLog"), sound_event_play);
	  Evt_register(hClass, Evt_get_event(hClass, "PlayEMail"), sound_event_play);
	}

	/*
	 * jim - vmail events shouldn't happen at all if we don't have
	 *  vidmail.res, but just to be on the anal side ;)
	 */
	if (Have_video_mail)
	{
	  Evt_register(hClass, Evt_get_event(hClass, "PlayVMail"), sound_event_play);
	}
	Evt_register(hClass, Evt_get_event(hClass, "TargetNew"), sound_event_play);
	Evt_register(hClass, Evt_get_event(hClass, "PlayFX"), sound_event_play);
	printf("Sound: Events Registered\n");
	/*
	 * unused, but should be....
	**  Evt_register(hClass, Evt_get_event(hClass, "StopLog"), sound_event_stop);
	**  Evt_register(hClass, Evt_get_event(hClass, "StopEMail"), sound_event_stop);	
	 */
 
	/*
	 * Ok - This is a major change, now the driver specific methods
	 * feature the keyword _driver_ so that we can use a runtime linked
	 * library should we choose to do so.
	 */
	sound_driver_init();
}

void Sound_deinit(void)
{
	/*
	 * Lets register us to grab those events!
	 */
	HEVENTCLASS hClass = Evt_get_class("MFD");
	Evt_unregister(hClass, Evt_get_event(hClass, "PlayLog"), sound_event_play);
	Evt_unregister(hClass, Evt_get_event(hClass, "PlayEMail"), sound_event_play);
	Evt_unregister(hClass, Evt_get_event(hClass, "PlayVMail"), sound_event_play);	
	Evt_unregister(hClass, Evt_get_event(hClass, "PlayFX"), sound_event_play);
	Evt_unregister(hClass, Evt_get_event(hClass, "TargetNew"), sound_event_play);
	printf("Sound: Events UnRegistered\n");
	/*
	 * unused, but should be....
	**  Evt_unregister(hClass, Evt_get_event(hClass, "StopLog"), sound_event_stop);
	**  Evt_unregister(hClass, Evt_get_event(hClass, "StopEMail"), sound_event_stop);	
	 */

	sound_driver_deinit();
}

void sound_event_play(HEVENTCLASS hClass, HEVENT hEvent, void *pData)
{
/*
 * This should only be thrown by the MFD
 */
	if (hClass == Evt_get_class("MFD"))
	{
		if (hEvent==Evt_get_event(hClass, "PlayLog")) sound_play_log((short)(pData), SOUND_TYPE_LOG);
		else if (hEvent==Evt_get_event(hClass, "PlayEMail")) sound_play_log((short)(pData), SOUND_TYPE_EMAIL);
		else if (hEvent==Evt_get_event(hClass, "PlayVMail")) sound_play_fx_2d(SOUND_FX_VMAIL);
		else if (hEvent==Evt_get_event(hClass, "TargetNew")) sound_play_fx_2d(SOUND_FX_NEWTARGET);
		else if (hEvent==Evt_get_event(hClass, "PlayFX")) sound_play_fx_2d((short)(pData));
	}
}

void sound_event_stop(HEVENTCLASS hClass, HEVENT hEvent, void *pData)
{
/*
 * Routines to stop log playing will go here....
 */
 UNUSED(hClass);
 UNUSED(hEvent);
 UNUSED(pData);
}

int sound_play_log(short log_num, int mediaType)
{
/*
 * Log playing goes here, 
 * EMails are also handled here.....
 * Remember actual log chunk played is [log_num+300]
 */
  block_handle bh;
  size_t size;
  size_t offset;
  unsigned char *buffer;
  int Sample_rate = -1;

if (mediaType == SOUND_TYPE_LOG) (log_num += 2488);
else if (mediaType == SOUND_TYPE_EMAIL) (log_num += 2441);

#ifdef SOUND_DEBUG
printf ("\nSound: ... Playing log num %i right about now (%i)...\n", (log_num+300),log_num );
#endif

  if((Res_load_chunk_by_id (log_num+300, &size)) != NULL) 
	{bh = Res_load_chunk_by_id (log_num+300, &size);}
  else
  {
    fprintf (stderr, "Sound: %i: no such audio log chunk\n", (log_num+300) );
    return 1;
  }

  buffer = *bh;
  
  /*
   * We ought to test this by checking the chunk type, that's what it's for
   */
  if (strncmp ((char *) buffer, "MOVI", 4) == 0)
  {
    /*
     * Audiolog in some sort of movie format. This is organised into chunks
     *  but I'm not bothering with that for our purposes. The chunkdir is at
     *  offset 0x0400 and the 32-bit word at offset 0x0008 is its length.
     * Skip the header
     */
    offset = 0x400 + buffer [8] + (buffer [9] << 8) +
      (buffer [10] << 16) + (buffer [11] << 24);

    if (Sample_rate < 0)
    {
      /*
       * If sample rate hasn't been explicitly overwritten by the user, set it
       *  from the file. It's the 32-bit word at offset 0x0026
       */
      Sample_rate = buffer [38] + (buffer [39] << 8) +
		(buffer [40] << 16) + (buffer [41] << 24);
    }
    buffer += offset;
    size -= offset;
  }
  else
  {
	/*
	 * We only want to play logs
	 */
	return 1;
  }

  if (Sample_rate < 0)
  {
    /*
     * Default sample rate
     */
    Sample_rate = 22050;
  }

  /*
   * Got the chunk. Now initialise the sound system
   */

#ifdef SOUND_DEBUG
  printf ("Sound: sample rate set to %dHz\n", Sample_rate);
#endif


/* ************  Driver Specifics ************************* */

sound_driver_play(size, buffer, Sample_rate);

/* ************ End Driver Specifics *********************** */
return 0;

}

int sound_play_fx_2d(int fx_num)
{
/*
 * FX playing goes here
 */
  block_handle bh;
  size_t size;
  unsigned char *buffer;
  int Sample_rate = -1;

  if((Res_load_chunk_by_id (fx_num, &size)) != NULL) 
	{bh = Res_load_chunk_by_id (fx_num, &size);}
  else
  {
    fprintf (stderr, "Sound: %i: no such fx chunk\n", (fx_num) );
    return 1;
  }
  
#ifdef SOUND_DEBUG
  printf ("\nSound: ... Playing FX num %i right about now ...\n", (fx_num));
#endif

  buffer = *bh;
  /*
   * We ought to test this by checking the chunk type, that's what it's for
   */
  if (strncmp ((char *) buffer, "Crea", 4) == 0)
  {
    if (Sample_rate < 0)
    {
      Sample_rate = 1000000 / (256 - buffer[30]);
      buffer += 32;
    }
  }
  else
  {
	/*
	 * We only want to play FX
	 */
	return 1;
  }

  if (Sample_rate < 0)
  {
    /*
     * Default sample rate
     */
    Sample_rate = 22050;
  }

  /*
   * Got the chunk. Now initialise the sound system
   */

#ifdef SOUND_DEBUG
  printf ("Sound: sample rate set to %dHz\n", Sample_rate);
#endif


/* ************  Driver Specifics ************************* */

sound_driver_play(size, buffer, Sample_rate);

/* ************ End Driver Specifics *********************** */

return 0;
}

