/*
 * dummy.c -
 *  Dummy Sound routines
 */

/*
 * File:    sound/dummy.c
 * Purpose: Audio handling
 * Prefix:  Sound_
 * Notes:   Providing a framework, for now
 * Music:   Olive, "Creature of Comfort"
 *
 * This file by Bryan Nolen 2001 bryan@bryann.ath.cx
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <math.h>
#include <string.h>										/* strncpy() - msgs */

#include "defines.h"
#include "types.h"
#include "option.h"

#include "event.h"
#include "sound.h"
#include "sound_core.h"
					
#include "main.h"											/* Can't_happen */


/*
 * These are the main event handlers
 */
void sound_event_play(HEVENTCLASS hClass, HEVENT hEvent, void *pData);
void sound_event_stop(HEVENTCLASS hClass, HEVENT hEvent, void *pData);

int sound_play_log(short log_num, int mediaType);

int sound_play_fx_2d(int fx_num);
 
void Sound_init(void)
{
	/*
	 * Lets register us to grab those events!
	 */
	HEVENTCLASS hClass = Evt_get_class("MFD");
	Evt_register(hClass, Evt_get_event(hClass, "PlayLog"), sound_event_play);
	Evt_register(hClass, Evt_get_event(hClass, "PlayEMail"), sound_event_play);
	Evt_register(hClass, Evt_get_event(hClass, "PlayVMail"), sound_event_play);
	printf("Sound: Events Registered\n");
	/*
	 * unused, but should be....
	**  Evt_register(hClass, Evt_get_event(hClass, "StopLog"), sound_event_stop);
	**  Evt_register(hClass, Evt_get_event(hClass, "StopEMail"), sound_event_stop);	
	**	Evt_register(hClass, Evt_get_event(hClass, "PlayFX"), sound_event_play);
	 */
}

void Sound_deinit(void)
{
	/*
	 * Lets register us to grab those events!
	 */
	HEVENTCLASS hClass = Evt_get_class("MFD");
	Evt_unregister(hClass, Evt_get_event(hClass, "PlayLog"), sound_event_play);
	Evt_unregister(hClass, Evt_get_event(hClass, "PlayEMail"), sound_event_play);
	Evt_unregister(hClass, Evt_get_event(hClass, "PlayVMail"), sound_event_play);
	printf("Sound: Events UnRegistered\n");
	/*
	 * unused, but should be....
	**  Evt_unregister(hClass, Evt_get_event(hClass, "StopLog"), sound_event_stop);
	**  Evt_unregister(hClass, Evt_get_event(hClass, "StopEMail"), sound_event_stop);	
	**	Evt_unregister(hClass, Evt_get_event(hClass, "PlayFX"), sound_event_play);
	 */
}

void sound_event_play(HEVENTCLASS hClass, HEVENT hEvent, void *pData)
{
/*
 * This should only be thrown by the MFD
 */
	if (hClass == Evt_get_class("MFD"))
	{
		if (hEvent==Evt_get_event(hClass, "PlayLog")) sound_play_log((int)(pData), SOUND_TYPE_LOG);
		else if (hEvent==Evt_get_event(hClass, "PlayEMail")) sound_play_log((int)(pData), SOUND_TYPE_EMAIL);
		else if (hEvent==Evt_get_event(hClass, "PlayVMail")) sound_play_fx_2d(SOUND_FX_VMAIL);
	}
}

void sound_event_stop(HEVENTCLASS hClass, HEVENT hEvent, void *pData)
{
/*
 * Routines to stop log playing will go here....
 */
 UNUSED(hClass);
 UNUSED(hEvent);
 UNUSED(pData);
}

int sound_play_log(short log_num, int mediaType)
{
/*
 * Log playing goes here, 
 * EMails are also handled here.....
 * Remember actual log chunk played is [log_num+300]
 */
if (mediaType == SOUND_TYPE_LOG) (log_num += 2488);
else if (mediaType == SOUND_TYPE_EMAIL) (log_num += 2441);
printf ("\nSound: ... Will be playing log num %i right about now (%i)...\n", (log_num+300),log_num );
return 1;
}

int sound_play_fx_2d(int fx_num)
{
/*
 * FX playing goes here 
 */
printf ("\nSound: ... Will be playing FX num %i right about now ...\n", (fx_num) );
return 1;
}



