/*
 *  sdl_driver_sdl.c - SDL Sound Routines
 *  Sound routines
 */

/*
 * File:    sound/sdl_driver_sdl.c
 * Purpose: Audio handling
 * Prefix:  Sound_
 * Notes:	SDL Audio Handling!
 * Music:   Olive, "Outlaw"
 *
 * This file by Bryan Nolen 2001 bryan@r3v3ng.net
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */

#include <stdio.h>
#include <stdlib.h>
#include "gamemem.h"
#include <math.h>
#include <string.h>										/* strncpy() - msgs */

#include "defines.h"
#include "types.h"
#include "option.h"
#include "event.h"

#include "sound.h"
#include "sound_core.h"
					
#include "main.h"											/* Can't_happen */
#include "resource.h"

#include <SDL.h>

/*
 * GLOBAL
 */
#define MAX_SDL_SOUNDS 10
#define VOL_PER_SOUND SDL_MIX_MAXVOLUME / 2
sound_play_t sound_playing[MAX_SDL_SOUNDS];

/*
 *  This holds the state of initalization of the sound card
 *  if it is 0 then the card has not been initalized and so do
 *  not play any sounds.
 *  This is to clear a known bug with SDL ..... that I missed :)
 */
int sound_init_state=0;

/*
 * Functional Prototypes!!
 */
void sound_sdl_callback (void *data, byte_t *stream, int len);


/*
 * jim - global AudioSpec for the actual hardware setup
 */
SDL_AudioSpec Hardware_spec;


int sound_driver_init(void)
{
  SDL_AudioSpec desired;

  /*
   * jim - bit of a problem, we need to initialise the GUI side of the SDL
   *  first, at GUI init time (very early) but can't call SDL_Init() twice.
   * For the moment, assume that if we're using SDL audio we're using SDL
   *  video as well (it's a pretty odd setup if this isn't the case) and so
   *  we can use SDL_InitSubSystem() for the audio
   */
    if( SDL_InitSubSystem(SDL_INIT_AUDIO) != 0 )
    {
        printf("Sound: SDL_InitSubSystem failed (%s)\n", SDL_GetError());
        return(1);
    }

    desired.freq       = 22050;
    desired.format     = AUDIO_U16;                        
    desired.channels   = 2;                                          
    desired.callback   = sound_sdl_callback;  
    desired.userdata   = NULL;             
    desired.samples    = 1024;

    if( SDL_OpenAudio( &desired, &Hardware_spec ) < 0 )
    {
        printf("Sound: SDL_OpenAudio failed (%s)\n", SDL_GetError());
        return(1);
    }

#ifdef SOUND_DEBUG
    printf("Sound: sound_driver_init() Done\n");
#endif
	/*
	 * we got what we want, unpause the sound sys
	 */
	SDL_PauseAudio(0);

	sound_init_state = 1;
	return 0;
}

int sound_driver_deinit(void)
{
	int i = 0;
/*
 * Close it down...
 */
	SDL_CloseAudio();
/*
 * Free used memory
 */ 
    for (i = 0; i < MAX_SDL_SOUNDS; i++) 
    {
		free(sound_playing[i].sound_data.sound_buf);
	}

    sound_init_state = 0;

	return 0;
}


int sound_driver_play( size_t size, unsigned char * buf, int samrate )
{

int Sample_rate = samrate;
unsigned char * buffer = buf;
sound_data_t sound_data;
SDL_AudioCVT  wav_cvt;
int sound_len=0;
int i;


/*
 * Only proceed if we have been initalized...
 */
    if( sound_init_state==0 )
    {
        printf("Sound: System initalise failed - no play possible\n");
        return(1);
    }


/*
 * ok, now we need to pass the driver the buffer(which contains the data), Sample_rate and size
 */

    /* Find an empty sound slot. */
    for (i = 0; i < MAX_SDL_SOUNDS; i++) 
    {
		if (sound_playing[i].sound_active == 0) break;
    }

    if (i == MAX_SDL_SOUNDS)
	return 1;

    SDL_LockAudio();

if (buffer != NULL)
{
	if (SDL_BuildAudioCVT(&wav_cvt, AUDIO_U8, 1, Sample_rate,
          Hardware_spec.format, Hardware_spec.channels, Hardware_spec.freq) > 0)
    {
		wav_cvt.buf=(byte_t *)malloc(size*wav_cvt.len_mult);
		wav_cvt.len=size;
		memcpy(wav_cvt.buf, buffer, size);
		SDL_ConvertAudio(&wav_cvt);
		sound_len = (wav_cvt.len*wav_cvt.len_ratio);
	    sound_data.sound_buf = wav_cvt.buf;
	    sound_data.sound_len = (int)sound_len;  
	}
	else
	{
		#ifdef SOUND_DEBUG
		  printf ("Sound: SDL_BuildAudioCVT failed, using raw buffer\n");
		#endif
		sound_data.sound_buf = buffer;
		sound_data.sound_len = size;
	}
		#ifdef SOUND_DEBUG
		  printf ("Sound: new sound playing, len is %i\n",sound_data.sound_len);
		#endif

	sound_playing[i].sound_active = 1;
    sound_playing[i].sound_data = sound_data;
    sound_playing[i].sound_pos = 0;
}   

    SDL_UnlockAudio();

return 0;
}


void sound_sdl_callback(void *userdata, byte_t *stream, int len)
{
    int i=0;
    UNUSED (userdata);
	memset(stream, 0, len);
    for (i = 0; i < MAX_SDL_SOUNDS; i++) 
    {
	  if (sound_playing[i].sound_active) 
	  {
	    byte_t *sound_buf;
	    int sound_len;

		sound_buf = sound_playing[i].sound_data.sound_buf;
	    sound_buf += sound_playing[i].sound_pos;

	    if ((sound_playing[i].sound_pos + len) > sound_playing[i].sound_data.sound_len) 
	    {
			sound_len = (int)sound_playing[i].sound_data.sound_len - (int)sound_playing[i].sound_pos;
	    } else 
	    { 
	    	sound_len = len; 
	    }
		
	    SDL_MixAudio(stream, sound_buf, sound_len, VOL_PER_SOUND);

	    sound_playing[i].sound_pos += len;

	    if (sound_playing[i].sound_pos >= sound_playing[i].sound_data.sound_len) 
	    {
			sound_playing[i].sound_active = 0;
	    }
	  }
    }
}

