/*
 * accel.h -
 *  Interface for accelerated rendering functions
 */

/*
 * File:    src/accel.h
 * Purpose: Low-level rendering API
 * Prefix:  Acc_
 * Note:    Refers to functions in accel/
 * Music:   
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

#ifndef ACCEL_H
#define ACCEL_H

/*
 * #includes
 */
#include "types.h"
#include "matrix.h" /* naughty ... */




/*
 * #defines and macros
 */

/*
 * Bitfields for accelerated back-ends
 */
#define ACCEL_SOFTWARE 0x01
#define ACCEL_OPENGL   0x02
#define ACCEL_DIRECT3D 0x04




#ifdef __cplusplus
extern "C"
{
#endif

void Acc_init (int backends);

struct tag_camera;
void Acc_setup_scene (struct tag_camera *c);
void Acc_end_scene   (void);

/*
 * Interface functions. These are used as functions, but are actually function
 *  pointers set to the relevant low-level rendering routines by
 *  Acc_setup_scene() .
 */
struct tag_line_seg;
struct tag_clip_region;
extern void (*Acc_draw_line) (struct tag_line_seg *l,
			      const struct tag_clip_region *frustrum);
/*
 * Note: the high-level frag / model drawing functions are provided in case the
 *  accelerated back-end wants to cache the object e.g. with an OpenGL display
 *  list. The accelerated function should call the software renderer to do the
 *  drawing proper, which will then call the accelerated interface for the
 *  individual primitives within it
 */
struct tag_ss_bitmap_header;
extern void (*Acc_draw_fragmented_sprite) (vec *pos, mat *orn, 
					   struct tag_ss_bitmap_header *bmp,
					   byte *zarray, int frag_size);
extern void (*Acc_draw_fragment) (vec *pos, uint32 rgb, int frag_size);


/*
 * Low-level surface drawing function. The bitmap here is probably misplaced,
 *  since textures will have been downloaded to the hardware, but it's needed
 *  by the software renderer at the moment so it stands
 */
struct tag_surface;
extern void (*Acc_draw_surface) (struct tag_surface *s);

/*
 * Low-level sprite drawing function. This also takes a surface, but is
 *  allowed to assume that it is in view space and z values are constant over
 *  it, and that its texture is aligned with the screen
 */
extern void (*Acc_draw_sprite) (struct tag_surface *s,
				struct tag_ss_bitmap_header *bmp);

/*
 * Draw 2D images in the framebuffer. the data and pointer provided
 * are valid for this call (= frame). either the underlying renderer
 * copies the values or uses it only once.
 */
struct tag_hudLayer;
struct tag_camera;

extern void (*Acc_draw_overlay) (struct tag_hudLayer *pLayer,
																 struct tag_camera *pCam);

#ifdef __cplusplus
}
#endif

#endif /* ACCEL_H */
