/*
 * anim.h -
 *  Animation control definitions
 */

/*
 * File:    src/anim.h
 * Purpose: Animation defines and externs
 * Prefix:  Anim_
 * Note:    
 * Music:   Van Halen, "Fair Warning"
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */


#ifndef ANIM_H
#define ANIM_H



/*
 * Animation flags
 */
#define AF_PAUSED        4  /* animation is paused, don't advance it  */
#define AF_BACKWARDS     2  /* animation is currently going backwards */
#define AF_BIDIRECTIONAL 1  /* animation repeats backwards            */



struct tag_map_object;  /* referred to by anim_control */
struct tag_anim_control;

typedef void (*anim_frame_hook) (struct tag_anim_control *);

typedef struct tag_anim_control
{
  double delay;    /* delay between frames, in seconds */
  double elapsed;  /* time since last frame update     */
  int num_frames;  /* number of distinct frames        */
  int flags;       /* AF_ constants above              */
  struct tag_map_object *obj;
  anim_frame_hook advance_hook;
} anim_control;


typedef struct tag_uw_animobj
{
  int stuff1;
  short base_frame;
  short num_frames;
} uw_animobj;


/*
 * Function prototypes
 */
void Anim_init (void);
void Anim_deinit(void);
void Anim_init_level (int level);
anim_control *Anim_add (double delay, struct tag_map_object *obj,
			int nframes, int flags);
void Anim_frame (double elapsed);
void Anim_advance (anim_control *a);
uw_animobj *Anim_get_uw_info (int objid);



#endif /* ANIM_H */
