/*
 * bitmap.h -
 *  System Shock bitmaps
 */

/*
 * File:    src/bitmap.h
 * Purpose: Defines and prototypes for generic bitmap handling
 * Prefix:  Bmp_
 * Note:
 * Music:   Judas Priest, "Screaming For Vengeance"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */


#ifndef BITMAP_H
#define BITMAP_H



/*
 * #includes
 */
#include "types.h"




/*
 * #defines
 */

/*
 * jim -use this idiom for multi-statement macros
 */
#define	Bmp_set_box(Head, left, top, right, bottom)	do \
{                                                      \
  (Head).x1 = (left);												           \
  (Head).y1 = (top);										               \
  (Head).x2 = (right);					                       \
  (Head).y2 = (bottom);                                \
} while (0)

#define	Bmp_reset_box(Head) do \
{                              \
  (Head).x1 = (Head).y1 = 0;	 \
  (Head).x2 = (Head).width;		 \
  (Head).y2 = (Head).height;   \
} while (0)


/*
 * Bitmap chunks
 */
#define BITMAP_CHUNK_SPRITE    1350
#define BITMAP_CHUNK_UW_SPRITE   20  /* objects.gr in my list */
#define BITMAP_CHUNK_UW_SWITCH   29
#define BITMAP_CHUNK_UW_DECAL    30
#define BITMAP_CHUNK_UW_TMAP     33  /* wall textures ... off list */


typedef struct tag_ss_bitmap_header
{

  int chunk;
  int subchunk;

  int type;

  int width;
  int height;

  int x1;
  int y1;
  int x2;
  int y2;

  unsigned char *bits;

} ss_bitmap_header;




#ifdef __cplusplus
extern "C"
{
#endif


int Bmp_get_num_bmps (int chunk);
bool Bmp_load_bitmap_header (ss_bitmap_header *bmp, int chunk, int subchunk);
bool Bmp_load_bitmap_bits (ss_bitmap_header *bmp, unsigned char *bits,
			   size_t bufsize);
ss_bitmap_header *Bmp_load_bitmap (int chunk, int subchunk);
ss_bitmap_header *Bmp_create (int width, int height);
ss_bitmap_header *Bmp_clone (ss_bitmap_header *pHead);
void Bmp_free_bitmap (ss_bitmap_header *bmp);
void Bmp_unpack_uw_rle (ss_bitmap_header *bmp, byte *pack, int wsize,
			byte *auxpal);


#ifdef __cplusplus
}
#endif



#endif /* BITMAP_H */


/*
 * End of src/bitmap.h
 */

/*
 * Follows some emacs magic to get VC++-formatted code to display right (and
 *  hopefully vice versa)
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
