/*
 * bsp.h -
 *  Binary Space Partitioning
 */

/*
 * File:    src/bsp.h
 * Purpose: Defines and types for BSP trees
 * Prefix:  BSP_
 * Note:    
 * Music:   
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */


#ifndef BSP_H
#define BSP_H


/*
 * #includes
 */
#include "clip.h" /* plane_t (bad! naughty!) */
#include "vec.h"


/*
 * A node in a BSP tree
 */
typedef struct tag_bsp_node
{

  /*
   * The partitioning plane
   */
  plane_t split_plane;

  /*
   * Binary children (NULL for a leaf node)
   */
  struct tag_bsp_node *front;
  struct tag_bsp_node *back;

  /*
   * Tile pointer, valid for a leaf node
   * This may be NULL even if the node is a leaf, indicating that the node
   *  contains no drawable tiles. This will never be the case for a BSP tree
   *  within a sector in the "real world" but might be for cyberspace.
   */
  struct tag_map_tile *tile;

  /*
   * For a quick-and-dirty clip test. The basic shape of a tile is a cuboid, so
   *  it makes sense to use an ellipsoid concentric, or rougly so, with the
   *  tile and perform a stretch in the z direction when clipping
   */
  vec centre;
  double radius;
  double z_scale;

} bsp_node;


#ifdef __cplusplus
extern "C"
{
#endif

/*
 * Function prototypes
 */
void BSP_init (void);
void BSP_exit (void);
void BSP_init_level (void);

#ifdef __cplusplus
}
#endif



/*
 * Global variables
 */
extern bsp_node *BSP_root;


#endif /* BSP_H */
