/*
 * container.c -
 *  Storage and retrieval of container data
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "types.h"
#include "gamemem.h"
#include "container.h"
#include "main.h"     /* Cant_happen() */
#include "map.h"
#include "object.h"
#include "resource.h"



/*
 * Debug
 */
#define CONTAINER_MESSAGES 0


#define CONTAINER_ARCHIVE_SIZE 21



/*
 * Local typedefs and structs
 */



container *Container_list;
int Num_containers;


/*
 * Local function prototypes
 */



/*
 * Load the container table for a level. Called from Obj_init_level()
 */
void Cnr_init_level (int level)
{

  block_handle bh;
  size_t size;
  int i;
  unsigned char *fp;

  /*
   * Load up the container list
   */
  bh = Res_load_chunk_by_id (4000 + 100*level + MAP_CHUNK_CONTAINER, &size);
  if ((bh == NULL) || (*bh == NULL))
  {
    Cant_happen ("Cnr_init_level(): couldn't load container data");
  }
  if (size % CONTAINER_ARCHIVE_SIZE != 0)
  {
    Cant_happen ("Cnr_init_level(): not exact number of containers");
  }
  Num_containers = size / CONTAINER_ARCHIVE_SIZE;
  Container_list = malloc (Num_containers * sizeof *Container_list);
  if (Container_list == NULL)
  {
    Cant_happen ("Cnr_init_level(): couldn't allocate space for list");
  }

# if MEM_INIT_MESSAGES
  printf ("Cnr_init_level(): %d bytes allocated for container list\n",
	  Num_containers * sizeof *Container_list);
# endif

  fp = (unsigned char *) *bh;
  for (i = 0; i < Num_containers; ++i)
  {
    Container_list [i].map_index    = *fp + 256 * fp [1]; fp += 2;
    Container_list [i].prev         = *fp + 256 * fp [1]; fp += 2;
    Container_list [i].next         = *fp + 256 * fp [1]; fp += 2;
    Container_list [i].content [0]  = *fp + 256 * fp [1]; fp += 2;
    Container_list [i].content [1]  = *fp + 256 * fp [1]; fp += 2;
    Container_list [i].content [2]  = *fp + 256 * fp [1]; fp += 2;
    Container_list [i].content [3]  = *fp + 256 * fp [1]; fp += 2;
    Container_list [i].width        = *fp++;
    Container_list [i].height       = *fp++;
    Container_list [i].depth        = *fp++;
    Container_list [i].top_texture  = *fp++;
    Container_list [i].side_texture = *fp++;
    fp += 2;
  }

  Res_free_chunk_by_id (4000 + 100*level + MAP_CHUNK_CONTAINER);

}

void Cnr_deinit(void)
{
	free(Container_list);
	Container_list = NULL;
}

/*
 * Retrieve a container, with error checking
 */
container *Cnr_get_container (map_object *mo)
{

  int ci = Cnr_check_index (mo);

  /*
   * Ought to check that the container links back to the correct map object,
   *  but it's slightly tricky as it stands
   */
  return &(Container_list [ci]);

}

container *Cnr_get_container_i(int nIndex)
{
	assert(nIndex >= 0 && nIndex < Num_containers);

  return &(Container_list[nIndex]);
}

int Cnr_check_index (map_object *mo)
{
  int ci = mo->class_index;
  if (mo->class != OBJCLASS_CONTAINER)
  {
    Cant_happen ("Cnr_check_index(): non-container object");
  }
  if ((ci < 1) || (ci >= Num_containers))
  {
    Cant_happen ("Cnr_check_index(): container index out of range");
  }
  return ci;
}


