/*
 * control.c -
 *  controlling your avatar for fun and profit
 */

/*
 * File:    src/control.c
 * Purpose: Dispatching of controls to the game engine
 * Prefix:  Con_
 * Note:    
 * Music:   Definitely NOT any Metallica! Remember kids, if you rip Metallica
 *           MP3s you are going to HELL.
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 *   modified and extended by Christian Haas 2001 christian.haas@eunet.at
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <string.h>

#include "types.h"
#include "event.h"
#include "graph2d.h"
#include "control.h"
#include "game.h"
#include "gamemem.h"
#include "gamestring.h"
#include "graph2d.h"
#include "hud.h"
#include "main.h"
#include "mfd.h"
#include "render.h"


#ifdef _DEBUG
/*
#define CON_DEBUG 
*/
#endif

#define	C_MAX_EXCLUSIONS	4											/* see Actions */

const char *MovementNames[C_MOVE_AMOUNT] =
{
	"MoveNone",
	"MoveMouseX",
	"MoveMouseY",
	"MoveMouseWheel",
	"MoveJoyX",
	"MoveJoyY"
};

/* should be sorted by id... */
const KeyName KeyNames[] =
{
	{ 0x0008, "Backspace" },
	{ 0x0009, "Tab" },
	{ 0x000D, "Return" },
	{ 0x0013, "Pause" },
	{ 0x001B, "Escape" },
	{ 0x0100, "Keypad0" },
	{ 0x0101, "Keypad1" },
	{ 0x0102, "Keypad2" },
	{ 0x0103, "Keypad3" },
	{ 0x0104, "Keypad4" },
	{ 0x0105, "Keypad5" },
	{ 0x0106, "Keypad6" },
	{ 0x0107, "Keypad7" },
	{ 0x0108, "Keypad8" },
	{ 0x0109, "Keypad9" },
	{ 0x010A, "KeypadPeriod" },
	{ 0x010B, "KeypadDiv" },
	{ 0x010C, "KeypadMul" },
	{ 0x010D, "KeypadMinus" },
	{ 0x010E, "KeypadPlus" },
	{ 0x010F, "KeypadEnter" },

	{ 0x0110, "F1" },
	{ 0x0111, "F2" },
	{ 0x0112, "F3" },
	{ 0x0113, "F4" },
	{ 0x0114, "F5" },
	{ 0x0115, "F6" },
	{ 0x0116, "F7" },
	{ 0x0117, "F8" },
	{ 0x0118, "F9" },
	{ 0x0119, "F10" },
	{ 0x011A, "F11" },
	{ 0x011B, "F12" },

	{ 0x0120, "Insert" },
	{ 0x0121, "Home" },
	{ 0x0122, "PgUp" },
	{ 0x0123, "Delete" },
	{ 0x0124, "End" },
	{ 0x0125, "PgDn" },
	{ 0x0126, "Left" },
	{ 0x0127, "Up" },
	{ 0x0128, "Right" },
	{ 0x0129, "Down" },

	/* jim - added bonus ones */
	{ KM_PRINTSCRN,  "PrintScrn"  },
	{ KM_SCROLLLOCK, "ScrollLock" },
	{ KM_BREAK,      "Break"      },
	{ KM_NUMLOCK,    "NumLock"    },
	{ KM_CAPSLOCK,   "CapsLock"   },

	{ KM_MENU,       "Menu"       },
	{ KM_HELP,       "Help"       },
	{ KM_EXECUTE,    "Execute"    },

	{ KM_LSHIFT,     "LShift"     },
	{ KM_RSHIFT,     "RShift"     },
	{ KM_LCTRL,      "LControl"   },
	{ KM_RCTRL,      "RControl"   },
	{ KM_LALT,       "LAlt"       },
	{ KM_RALT,       "RAlt"       },
	{ KM_FN,         "Fn"         },

	{ KM_MB_BUTTON0, "MouseLeft"	},
	{ KM_MB_BUTTON1, "MouseCenter"	},
	{ KM_MB_BUTTON2, "MouseRight"	},
	{ KM_MB_BUTTON3, "MouseBut4"	},
	{ KM_MB_BUTTON4, "MouseBut5"	},
	{ KM_MB_BUTTON5, "MouseBut6"	},
	{ KM_MB_BUTTON6, "MouseBut7"	},
	{ KM_MB_BUTTON7, "MouseBut8"	},
	{ KM_MB_NONE,		 "MouseNone"	},

	{ KM_JOY_BUTTON0, "JoyLeft"	},
	{ KM_JOY_BUTTON1, "JoyCenter"	},
	{ KM_JOY_BUTTON2, "JoyRight"	},
	{ KM_JOY_BUTTON3, "JoyBut4"	},
	{ KM_JOY_BUTTON4, "JoyBut5"	},
	{ KM_JOY_BUTTON5, "JoyBut6"	},
	{ KM_JOY_BUTTON6, "JoyBut7"	},
	{ KM_JOY_BUTTON7, "JoyBut8"	},

	{ 0xFFFF, "" }
};

const char *OperationNames[OIDAmount] =
{
	/* movement */
	"Strafing",
	"Walking",
	/* orientation */
	"Orientation Side",
	"Orientation Height",
	"Leaning",
	"Standing",
	/* system */
	"Cursor Movement X",
	"Cursor Movement Y"
};

const char *ActionNames[AIDAmount] = 
{
	/* system */
	"Switch to Action Mode",
	"Switch to User Mode 1",
	"Switch to User Mode 2",
	"Switch to User Mode 3",
	/* movement */
	"Run Forward",
	"Walk Forward",
	"Walk Backward",
	"Turn Right",
	"Turn Left",
	"Strafe Right",
	"Strafe Left",
	"Jump",
	/* viewpose */
	"Lean Left",
	"Lean Off",
	"Lean Right",
	"Look Up",
	"Look Straight",
	"Look Down",
	"Stand Up",
	"Duck",
	"Crawl",
	/* interaction */
	"Select",
	"Activate",
	"Grab",
	/* shortcuts */
	"Next Weapon",
	"Reload Weapon",
	/* hardware */
	"Biomonitor",
	"Sensaround",
	"Lantern",
	"Shield",
	"Infrared",
	"Compass",
	"Data Reader",
	"Motion Booster",
	"Jump Jets",
	"Bigmap",
	"Bigmap ZoomIn",
	"Bigmap ZoomOut",
	/* MFD cards */
	"LeftMFD Weapon",
	"LeftMFD Ammo",
	"LeftMFD Item",
	"LeftMFD Map",
	"LeftMFD TargetList",
	"LeftMFD TargetInfo",
	"LeftMFD Data",
	"CenterMFD Weapons",
	"CenterMFD Hardware",
	"CenterMFD Inventory",
	"CenterMFD Software",
	"RightMFD Weapon",
	"RightMFD Ammo",
	"RightMFD Item",
	"RightMFD Map",
	"RightMFD TargetList",
	"RightMFD TargetInfo",
	"RightMFD Data",
	/* system */
	"Screenshot",
	"Quit",
	"DebugOutput"
};

/* types of control settings																								*/
#define	C_CON_OFF			0x00	/* not a type, but a state											*/
#define	C_CON_SYNCH		0x01	/* 'on' only while active												*/
#define	C_CON_SWITCH	0x02	/* 'on' at 1st activation, 'off' at 2nd					*/
#define	C_CON_TRIGGER	0x03	/* 'on' at every activation											*/
#define	C_CON_AMOUNT	0x04

const char *ControlTypes[C_CON_AMOUNT] =
{
	"OFF",
	"SYNC",
	"SWITCH",
	"TRIGGER"
};

typedef struct tag_ModeDescription
{
	char					 mdName[31];										/* human readable name	*/
	bool					 mdReadable;										/* assigns readable?		*/
	bool					 mdWriteable;										/* assigns settable?		*/
	bool					 mdActive;											/* mode active?					*/
} ModeDescription;

/* an AssignmentMode holds all the assignments for one key									*/
typedef struct tag_AssignmentMode
{
	uint16										 amAssignments;
	struct tag_Assignment		 **amAssigns;
} AssignmentMode;

typedef struct tag_OperationMode
{
	uint16										 omAssignments;
	struct tag_OpAssignment	 **omAssigns;
} OperationMode;

/* Trigger as a binary state (active / not active)													*/
/*   a trigger has some assignments dependand on active Mode								*/
typedef struct tag_Trigger
{
	bool											 tState;
	uint32										 tDownTick;
	uint32										 tUpTick;

	struct tag_AssignmentMode	 tModes[C_CON_MODE_AMOUNT];
} Trigger;

typedef struct tag_Movement
{
	int												 mPos;
	uint32										 mLastTick;

	struct tag_OperationMode	 mModes[C_CON_MODE_AMOUNT];
} Movement;

/* one assignment has a base key and 0..n modifier which have to						*/
/*   be active to fire the assignment	(take action)													*/
typedef struct tag_Assignment
{
	uint16										 aTrigger;
	uint16										 aModifier[C_CON_MAX_MODIFIER];

	byte											 aType;
	uint16										 aActionId;
} Assignment;

typedef struct tag_OpAssignment
{
	uint16										 oaMovement;
	uint16										 oaModifier[C_CON_MAX_MODIFIER];

	uint16										 oaOperationId;
} OpAssignment;

/* an action can be started by several assignments and also has							*/
/*   some exclusions (ie: you cant run forward when walking back)						*/
typedef struct tag_BinaryAction
{
	byte											 baActive[C_CON_MODE_AMOUNT];

	ss_action_id							 baExclusions[C_MAX_EXCLUSIONS];

	struct tag_AssignmentMode	 baModes[C_CON_MODE_AMOUNT];
} BinaryAction;

typedef struct tag_Operation
{
/*	byte											 oActive[C_CON_MODE_AMOUNT]; */

	int												 oCurPos;

	struct tag_OperationMode	 oModes[C_CON_MODE_AMOUNT];
} Operation;

bool	con_read_config(void);
void	con_write_config(void);
void	con_init_basics(void);
int		con_find_mode(char *strMode);
int		con_find_type(char *strType);
uint16 con_find_key(char *strKey);
uint16 con_find_move(char *strMove);
uint16 con_mod_to_key(uint32 nModifier);
uint32 con_key_to_mod(uint16 nKey);
void	con_exclude_action(ss_action_id idAction1, ss_action_id idAction2);
byte	con_count_bits32(uint32 nMask);
void	con_free_modes(void);
bool	con_add_operation(byte nMode, ss_operation_id nId,
												uint16 nMove, uint32 nModifier);
bool	con_add_action_key(byte nMode, ss_action_id nId,
												 uint16 nKey, uint32 nModifier, byte nType);
Assignment *con_find_best_assignment(AssignmentMode *pMode);
OpAssignment *con_find_best_op_assignment(OperationMode *pMode);
void	con_update_actions(byte nMode);
void	con_execute_trigger(uint16 nKey);
void	con_stop_trigger(uint16 nKey);
void	con_execute_movement(uint16 nMove, int rel);
void	con_action_execute(ss_action_id nId, bool bStart);
void	con_operation_execute(ss_operation_id nId, int rel);

ModeDescription ModeDescriptions[C_CON_MODE_AMOUNT] =
{
	{ "Writing",				false,	false,	false },
	{ "Menu",						false,	false,	false },
	{ "User Mode 1",		true,		true,		false },
	{ "User Mode 2",		true,		true,		false },
	{ "User Mode 3",		true,		true,		false },
	{ "Action Mode",		true,		true,		false },
	{ "Global",					true,		false,	true  }
};

/* summary of the keyboard, mouse and the actions														*/
Trigger				 gTrigger[C_KM_AMOUNT];
Movement			 gMovements[C_MOVE_AMOUNT];
BinaryAction	 gActions[AIDAmount];
Operation			 gOperations[OIDAmount];
uint32				 nCurModifier = 0x00000000;

/* current 'body status' of the player																			*/
ss_controlset	 CurSet;	

extern int		 Option_underworld;
extern char		 Option_con_cfg[256];

void Con_init(int nWidth, int nHeight)
{
	bool bConfigFromFile = false;

	memset(&CurSet, 0x00, sizeof(ss_controlset));
	memset(gTrigger, 0x00, sizeof(Trigger) * C_KM_AMOUNT);
	memset(gMovements, 0x00, sizeof(Movement) * C_MOVE_AMOUNT);
	memset(gActions, 0x00, sizeof(BinaryAction) * AIDAmount);
	memset(gOperations, 0x00, sizeof(Operation) * OIDAmount);
	/* no need for init basics here, since it's done either in con_read or		*/
	/* Con_reset...																														*/
	/*	con_init_basics(); */

	if (strlen(Option_con_cfg))
		bConfigFromFile = con_read_config();

	if (!bConfigFromFile)
		Con_reset_defaults();

	gOperations[OIDMoveCursorX].oCurPos = nWidth / 2;
	gOperations[OIDMoveCursorY].oCurPos = nHeight / 2;
}

void Con_deinit(void)
{
	con_write_config();
	con_free_modes();
}

const char *Con_get_keyname(uint16 nKey)
{
	int i;
	const char *strName = NULL;

	for (i = 0; KeyNames[i].knId != 0xFFFF && !strName; i++)
	{
		if (KeyNames[i].knId == nKey)
			strName = KeyNames[i].knName;
	}

	return strName;
}

const char *Con_get_mode_name(byte nMode)
{
	const char *strName = NULL;

	if (nMode < C_CON_MODE_AMOUNT)
		strName = ModeDescriptions[nMode].mdName;

	return strName;
}

bool		 Con_is_mode_readable(byte nMode)
{
	bool rCode = false;

	if (nMode < C_CON_MODE_AMOUNT)
		rCode = ModeDescriptions[nMode].mdReadable;

	return rCode;
}

bool		 Con_is_mode_writable(byte nMode)
{
	bool rCode = false;

	if (nMode < C_CON_MODE_AMOUNT)
		rCode = ModeDescriptions[nMode].mdWriteable;

	return rCode;
}

void		 Con_activate_mode(byte nMode, bool bActive)
{
	if (nMode < C_CON_MODE_AMOUNT)
	{
		if (ModeDescriptions[nMode].mdActive != bActive)
		{
#ifdef CON_DEBUG 
			printf("Con_activate_mode() Mode <%s>: %s\n", 
								 ModeDescriptions[nMode].mdName, 
								 (bActive ? "activated" : "deactivated"));
#endif
			ModeDescriptions[nMode].mdActive = bActive;

      /*
       * jim - casting a byte (nMode) to a pointer makes me (and gcc) twitchy.
       *  You just know that some idiot is going to try to dereference it
       *  somewhere down the line (and it'll probably be me).
       * Far better to pass a pointer to ModeDescriptions [nMode]; if you need
       *  the number, add it to the mode description struct.
       */
			if (bActive)
			{
				Evt_fire_event(Evt_get_event(Evt_get_class("CONTROL"), "ModeActivate"),
											 (void *)nMode, false);
				con_update_actions(nMode);
			}
			else
				Evt_fire_event(Evt_get_event(Evt_get_class("CONTROL"), "ModeDeactivate"),
											 (void *)nMode, false);
		}
	}
}

void		 Con_activate_user_mode(byte nMode)
{
	byte nModeBuf;

	if (nMode >= C_CON_MODE_USER1 && nMode <= C_CON_MODE_USER3)
	{
		for (nModeBuf = C_CON_MODE_USER1; nModeBuf <= C_CON_MODE_USER3; nModeBuf++)
			if (nMode != nModeBuf)
				Con_activate_mode(nModeBuf, false);
		Con_activate_mode(nMode, true);
	}
}

bool		 Con_is_mode_active(byte nMode)
{
	bool rCode = false;

	if (nMode < C_CON_MODE_AMOUNT)
		rCode = ModeDescriptions[nMode].mdActive;

	return rCode;
}

void con_free_modes(void)
{ 
	point ptCursor;
	int i, j, k;

	for (i = 0; i < AIDAmount; i++)
	{
		for (j = 0; j < C_CON_MODE_AMOUNT; j++)
		{
			for (k = 0; k < gActions[i].baModes[j].amAssignments; k++)
				free(gActions[i].baModes[j].amAssigns[k]);
			free(gActions[i].baModes[j].amAssigns);
		}
	}
	memset(gActions, 0x00, sizeof(BinaryAction) * AIDAmount);
	for (i = 0; i < OIDAmount; i++)
	{
		for (j = 0; j < C_CON_MODE_AMOUNT; j++)
		{
			for (k = 0; k < gOperations[i].oModes[j].omAssignments; k++)
				free(gOperations[i].oModes[j].omAssigns[k]);
			free(gOperations[i].oModes[j].omAssigns);
		}
	}
	memset(gOperations, 0x00, sizeof(Operation) * OIDAmount);
	HUD_get_cursor_pos(&ptCursor);
	gOperations[OIDMoveCursorX].oCurPos = ptCursor.x;
	gOperations[OIDMoveCursorY].oCurPos = ptCursor.y;
	for (i = 0; i < C_KM_AMOUNT; i++)	
	{
		for (j = 0; j < C_CON_MODE_AMOUNT; j++)
		{
/* ff1: the actions don't have to be freed in the keys since they were			*/
/*			were already deleted in the Actions																	*/
/*
			for (k = 0; k < gTrigger[i].tModes[j].amAssignments; k++)
				free(gTrigger[i].tModes[j].amAssigns[k]);
*/
			free(gTrigger[i].tModes[j].amAssigns);
		}
	}
	memset(gTrigger, 0x00, sizeof(Trigger) * C_KM_AMOUNT);
	for (i = 0; i < C_MOVE_AMOUNT; i++)
	{
		for (j = 0; j < C_CON_MODE_AMOUNT; j++)
		{
			free(gMovements[i].mModes[j].omAssigns);
		}
	}
	memset(gMovements, 0x00, sizeof(Movement) * C_MOVE_AMOUNT);
	memset(&CurSet, 0x00, sizeof(ss_controlset));
}

void	con_init_basics(void)
{
	int i, j;

	/* init exclusions - actions that cant be together */
	for (i = 0; i < (int)AIDAmount; i++)
		for (j = 0; j < C_MAX_EXCLUSIONS; j++)
			gActions[i].baExclusions[j] = AIDAmount;
	con_exclude_action(AIDBack, AIDRun);
	con_exclude_action(AIDBack, AIDWalk);
	con_exclude_action(AIDRun, AIDWalk);
	con_exclude_action(AIDRight, AIDLeft);
	con_exclude_action(AIDStrafeRight, AIDRight);
	con_exclude_action(AIDStrafeLeft, AIDLeft);
	con_exclude_action(AIDLeanLeft, AIDLeanRight);
	con_exclude_action(AIDLeanRight, AIDLeanCenter);
	con_exclude_action(AIDLeanCenter, AIDLeanLeft);
	con_exclude_action(AIDLookUp, AIDLookDown);
	con_exclude_action(AIDLookStraight, AIDLookUp);
	con_exclude_action(AIDLookDown, AIDLookUp);
	con_exclude_action(AIDStand, AIDDuck);
	con_exclude_action(AIDDuck, AIDCrawl);
	con_exclude_action(AIDCrawl, AIDStand);
	con_exclude_action(AIDBigmapZoomIn, AIDBigmapZoomOut);
	/* set global, unsettable configs */
	con_add_operation(C_CON_MODE_GLOBAL, OIDMoveCursorX, MOVE_MOUSEX, KM_MOD_NONE);
	con_add_operation(C_CON_MODE_GLOBAL, OIDMoveCursorY, MOVE_MOUSEY, KM_MOD_NONE);

	con_add_action_key(C_CON_MODE_GLOBAL, AIDScreenshot, KM_F12, KM_MOD_NONE, C_CON_TRIGGER);
	con_add_action_key(C_CON_MODE_GLOBAL, AIDQuit, KM_ESCAPE, KM_MOD_NONE, C_CON_TRIGGER);
	con_add_action_key(C_CON_MODE_GLOBAL, AIDQuit, 'x', KM_MOD_LCTRL, C_CON_TRIGGER);
	con_add_action_key(C_CON_MODE_GLOBAL, AIDDebug, '-', KM_MOD_NONE, C_CON_TRIGGER);
}

void	con_update_actions(byte nMode)
{
	AssignmentMode *pMode = NULL;
	BinaryAction *pAction = NULL;
	Assignment *pAssignment = NULL;
	int i;

	for (i = 0; i < AIDAmount; i++)
	{
		pAction = &(gActions[i]);
		pMode = &(pAction->baModes[nMode]);
		if (pAction->baActive[nMode])
		{
			if (!con_find_best_assignment(pMode) &&
					pAction->baActive[nMode] != C_CON_SWITCH)
			{
				pAction->baActive[nMode] = C_CON_OFF;
				con_action_execute(i, false);
			}
		}
		else
		{
			pAssignment = con_find_best_assignment(pMode);
			if (pAssignment)
			{
				pAction->baActive[nMode] = pAssignment->aType;
				con_action_execute(i, true);
			}
		}
	}
}

int		con_find_mode(char *strMode)
{
	int nMode = -1, i;

	for (i = 0; i < C_CON_MODE_AMOUNT && nMode < 0; i++)
		if (!strcmp(strMode, ModeDescriptions[i].mdName))
			nMode = i;

	return nMode;
}

int		con_find_type(char *strType)
{
	int nType = -1, i;

	for (i = 0; i < C_CON_AMOUNT && nType < 0; i++)
		if (!strcmp(strType, ControlTypes[i]))
			nType = i;

	return nType;
}

uint16	con_find_key(char *strKey)
{
	uint16 nKey = 0x0000, i;

	if (strlen(strKey) == 1)
		nKey = strKey[0];
	else
		for (i = 0; KeyNames[i].knId != 0xFFFF && !nKey; i++)
			if (!strcmp(strKey, KeyNames[i].knName))
				nKey = KeyNames[i].knId;

	return nKey;
}

uint16 con_find_move(char *strMove)
{
	uint16 nMove = 0x0000, i;

	for (i = 0; i < C_MOVE_AMOUNT && !nMove; i++)
		if (!strcmp(strMove, MovementNames[i]))
			nMove = i;

	return nMove;
}

bool	con_read_config(void)
{
	FILE *pFile = fopen(Option_con_cfg, "r");
	int nLine = 0, nParsed = 0, nLength, i, nParts,
			nCurAction = AIDAmount, nNewAction, 
			nCurOperation = OIDAmount, nNewOperation,
			nMode, nType;
	uint16 nKey, nMove;
	uint32 nMods;
	char strLine[256], strBuf[256], **pLine, **pKeys, **pMoves;
	bool rCode = false, bParsed = false;

	if (pFile)
	{
		con_free_modes();
		con_init_basics();
		fseek(pFile, 0, SEEK_SET);

		memset(strLine, 0x00, 256);
		fgets(strLine, 255, pFile);
		while (!feof(pFile))
		{
			bParsed = false;
			nLine++;
			nLength = strlen(strLine);
			if (nLength > 0 && (strLine[nLength - 1] == 0x0A || strLine[nLength - 1] == 0x0D))
				strLine[nLength - 1] = 0x00;
			if (nLength > 1 && (strLine[nLength - 2] == 0x0A || strLine[nLength - 2] == 0x0D))
				strLine[nLength - 2] = 0x00;
			Str_trim_sides(strLine, ' ');
			nLength = strlen(strLine);
			if (nLength && strLine[0] != ';')
			{	/* ignore empty lines and comments */
				nNewAction = AIDAmount;
				for (i = 0; i < AIDAmount && nNewAction == AIDAmount; i++)
				{
					sprintf(strBuf, "[%s]", ActionNames[i]);
					if (!strcmp(strBuf, strLine))
						nNewAction = i;
				}
				if (nNewAction < AIDAmount)
				{	/* next action block found */
					nCurAction = nNewAction;
					nCurOperation = OIDAmount;
					nParsed++;
					bParsed = true;
				}
				else
				{
					nNewOperation = OIDAmount;
					for (i = 0; i < OIDAmount && nNewOperation == OIDAmount; i++)
					{
						sprintf(strBuf, "[%s]", OperationNames[i]);	
						if (!strcmp(strBuf, strLine))
							nNewOperation = i;
					}
					if (nNewOperation < OIDAmount)
					{	/* new movement block found */
						nCurOperation = nNewOperation;
						nCurAction = AIDAmount;
						nParsed++;
						bParsed = true;
					}
				}

				if (!bParsed)
				{
					if (nCurAction < AIDAmount)
					{	/* action block is active */
						pLine = Str_extract(strLine, ";");
						nParts = 0;
						while (pLine[nParts]) /* count tokens */
							nParts++;
						if (nParts == 3 || nParts == 4)
						{	/* Mode;Type;Combo -- ';' may add an empty line */
							if ((nMode = con_find_mode(pLine[0])) >= 0 &&
									(nType = con_find_type(pLine[1])) >= 0)
							{
								nMods = nKey = 0x0000;
								if (nParts == 4)
								{	/* this means only one thing */
									nKey = ';';
									pKeys = Str_extract(pLine[3], "+");
								}
								else
								{
									pKeys = Str_extract(pLine[2], "+");
									if (pKeys[0])
										nKey = con_find_key(pKeys[0]);
								}
								if (nKey)
								{	/* got the main key, now go for modifier */
									for (i = 0; i < C_CON_MAX_MODIFIER && pKeys[i]; i++)
										if (i || nParts == 4)	/* skip first when != ';'*/
											nMods |= con_key_to_mod(con_find_key(pKeys[i]));
/*
printf("got <%s> %04X in <%s> as <%s> for <%s>\n",
			 pKeys[0], nKey, pLine[0], pLine[1], ActionNames[nCurAction]);
*/
									bParsed = true;
									if (ModeDescriptions[nMode].mdWriteable)
									{	/* only settable please */
										bParsed = con_add_action_key((byte)nMode, nCurAction, 
																								 nKey, nMods, (byte)nType);
									}
								}
								for (i = 0; pKeys[i]; i++)
									free(pKeys[i]);
								free(pKeys);
							}
						}
						for (i = 0; pLine[i]; i++)
							free(pLine[i]);
						free(pLine);
					}
					if (nCurOperation < OIDAmount)
					{	/* movement block is active */
						pLine = Str_extract(strLine, ";");
						nParts = 0;
						while (pLine[nParts]) /* count tokens */
							nParts++;
						if (nParts == 2)		
						{
							if ((nMode = con_find_mode(pLine[0])) >= 0)
							{
								nMods = nMove = 0x0000;
								pMoves = Str_extract(pLine[1], "+");
								if (pMoves[0])
									nMove = con_find_move(pMoves[0]);
								if (nMove)
								{
									for (i = 0; i < C_CON_MAX_MODIFIER && pMoves[i + 1]; i++)
										nMods |= con_key_to_mod(con_find_key(pMoves[i + 1]));
/*
printf("got <%s> %04X in <%s> for <%s>\n",
			 pMoves[0], nMove, pLine[0], OperationNames[nCurOperation]);
*/
									bParsed = true;
									if (ModeDescriptions[nMode].mdWriteable)
									{	/* only settable please */
										bParsed = con_add_operation((byte)nMode, nCurOperation, 
																								 nMove, nMods);
									}
								}
								for (i = 0; pMoves[i]; i++)
									free(pMoves[i]);
								free(pMoves);
							}
						}
						for (i = 0; pLine[i]; i++)
							free(pLine[i]);
						free(pLine);
					}
					if (!bParsed)
						printf("con_read_config() Couldn't parse Line %d\n", nLine);
					else
						nParsed++;
				}
			}
			memset(strLine, 0x00, 256);
			fgets(strLine, 255, pFile);
		}
		if (nParsed)
			rCode = true;
		fclose(pFile);
	}
	else
		printf("con_read_config() Couldn't read Config File <%s>\n"
					 "                  will use defaults and write it at end\n",
					 Option_con_cfg);

	return rCode;
}

void	con_write_config(void)
{
	FILE *pFile = fopen(Option_con_cfg, "w");
	const char *strKeyName = NULL, *strMoveName = NULL;
	char strKey[2] = { 0x00, 0x00 };
	int i, j, k, l, nAssigns;

	if (!pFile && strcmp(Option_con_cfg, "control.cfg"))
		pFile = fopen("control.cfg", "w");
	if (pFile)
	{
		fprintf(pFile, "; Control config file V0.1\n;\n");
		fprintf(pFile, "; a comment has a ';' at the first column\n");
		fprintf(pFile, "; at the bottom of the file is a list of all special Keys,\n");
		fprintf(pFile, "; Modifier and Moves\n");
		fprintf(pFile, "; !! This file is written during each exit of game !!\n;\n");

		fprintf(pFile, "\n;;;;;;;;;\n; TRIGGER\n");
		for (i = 0; i < AIDAmount; i++)
		{
			nAssigns = 0;
			fprintf(pFile, "[%s]\n", ActionNames[i]);
/* only for debugging */
/*
			fprintf(pFile, "; EXCL:");
			for (j = 0; j < C_MAX_EXCLUSIONS; j++)
			{
				if (gActions[i].baExclusions[j] < AIDAmount)
				{
					fprintf(pFile, "%s%s", (j ? ";" : ""), 
									ActionNames[gActions[i].baExclusions[j]]);
				}
			}
			fprintf(pFile, "\n");
*/
			for (j = 0; j < C_CON_MODE_AMOUNT; j++)
			{
				for (k = 0; k < gActions[i].baModes[j].amAssignments &&
										ModeDescriptions[j].mdReadable; k++)
				{	
					nAssigns++;
					strKeyName = Con_get_keyname(gActions[i].baModes[j].amAssigns[k]->aTrigger);
					if (!strKeyName)
					{
						strKey[0] = (char)(gActions[i].baModes[j].amAssigns[k]->aTrigger);
						strKeyName = strKey;
					}
					fprintf(pFile, "%s;%s;%s", 
									ModeDescriptions[j].mdName,
									ControlTypes[gActions[i].baModes[j].amAssigns[k]->aType],
									strKeyName);
					for (l = 0; l < C_CON_MAX_MODIFIER; l++)
					{
						if (gActions[i].baModes[j].amAssigns[k]->aModifier[l])
						{
							strKeyName = Con_get_keyname(gActions[i].baModes[j].amAssigns[k]->aModifier[l]);
							if (strKeyName)
								fprintf(pFile, "+%s", strKeyName);
						}
					}
					fprintf(pFile, "\n");
				}
			}
			if (!nAssigns)
				fprintf(pFile, "\n");
		}
		fprintf(pFile, "\n;;;;;;;;;;\n; MOVEMENT\n");
		for (i = 0; i < OIDAmount; i++)
		{
			nAssigns = 0;
			fprintf(pFile, "[%s]\n", OperationNames[i]);
			for (j = 0; j < C_CON_MODE_AMOUNT; j++)
			{
				for (k = 0; k < gOperations[i].oModes[j].omAssignments &&
										ModeDescriptions[j].mdReadable; k++)
				{	
					nAssigns++;
					strMoveName = MovementNames[gOperations[i].oModes[j].omAssigns[k]->oaMovement];
					fprintf(pFile, "%s;%s", 
									ModeDescriptions[j].mdName, strMoveName);
					for (l = 0; l < C_CON_MAX_MODIFIER; l++)
					{
						if (gOperations[i].oModes[j].omAssigns[k]->oaModifier[l])
						{
							strKeyName = Con_get_keyname(gOperations[i].oModes[j].omAssigns[k]->oaModifier[l]);
							if (strKeyName)
								fprintf(pFile, "+%s", strKeyName);
						}
					}
					fprintf(pFile, "\n");
				}
			}
			if (!nAssigns)
				fprintf(pFile, "\n");
		}
		/* for convenience, write all special key names...											*/
		fprintf(pFile, "\n\n; Next is the list of all possible Special Keys and Modifier\n"
									 "; (other than the ASCII keys)\n;\n");
		for (i = 0; KeyNames[i].knId != 0xFFFF; i++)
		{
			fprintf(pFile, "; %s", KeyNames[i].knName);
			if (con_key_to_mod(KeyNames[i].knId))
				fprintf(pFile, " (as Modifier)");
			fprintf(pFile, "\n");
		}
		fprintf(pFile, "\n\n; Next is the list of all possible Movements\n");
		fprintf(pFile, "; (Not all are currently supported)\n;\n");
		for (i = 1; i < C_MOVE_AMOUNT; i++)
		{
			fprintf(pFile, "; %s\n", MovementNames[i]);
		}

		fclose(pFile);
	}
	else
	{
		Cant_happen("con_write_config() Couldn't access config file!");
	}
}

void	con_exclude_action(ss_action_id idAction1, ss_action_id idAction2)
{
	int i;
	bool bDone = false;

	for (i = 0; i < C_MAX_EXCLUSIONS && !bDone; i++)
		if (gActions[idAction1].baExclusions[i] == idAction2 ||
				gActions[idAction1].baExclusions[i] == AIDAmount)
		{
			gActions[idAction1].baExclusions[i] = idAction2;
			bDone = true;
		}
	bDone = false;
	for (i = 0; i < C_MAX_EXCLUSIONS && !bDone; i++)
		if (gActions[idAction2].baExclusions[i] == idAction1 ||
				gActions[idAction2].baExclusions[i] == AIDAmount)
		{
			gActions[idAction2].baExclusions[i] = idAction1;
			bDone = true;
		}
}

byte	con_count_bits32(uint32 nMask)
{	/* does one have a faster way in mind? */
	byte nBits = 0;

	if (nMask & 0x00000001)	nBits++;
	if (nMask & 0x00000002)	nBits++;
	if (nMask & 0x00000004)	nBits++;
	if (nMask & 0x00000008)	nBits++;
	if (nMask & 0x00000010)	nBits++;
	if (nMask & 0x00000020)	nBits++;
	if (nMask & 0x00000040)	nBits++;
	if (nMask & 0x00000080)	nBits++;
	if (nMask & 0x00000100)	nBits++;
	if (nMask & 0x00000200)	nBits++;
	if (nMask & 0x00000400)	nBits++;
	if (nMask & 0x00000800)	nBits++;
	if (nMask & 0x00001000)	nBits++;
	if (nMask & 0x00002000)	nBits++;
	if (nMask & 0x00004000)	nBits++;
	if (nMask & 0x00008000)	nBits++;
	if (nMask & 0x00010000)	nBits++;
	if (nMask & 0x00020000)	nBits++;
	if (nMask & 0x00040000)	nBits++;
	if (nMask & 0x00080000)	nBits++;
	if (nMask & 0x00100000)	nBits++;
	if (nMask & 0x00200000)	nBits++;
	if (nMask & 0x00400000)	nBits++;
	if (nMask & 0x00800000)	nBits++;
	if (nMask & 0x01000000)	nBits++;
	if (nMask & 0x02000000)	nBits++;
	if (nMask & 0x04000000)	nBits++;
	if (nMask & 0x08000000)	nBits++;
	if (nMask & 0x10000000)	nBits++;
	if (nMask & 0x20000000)	nBits++;
	if (nMask & 0x40000000)	nBits++;
	if (nMask & 0x80000000)	nBits++;

	return nBits;
}

uint16 con_mod_to_key(uint32 nModifier)
{
	uint16 nKey = 0x0000;

	switch (nModifier)
	{
    case	KM_MOD_LSHIFT:	nKey = KM_LSHIFT;			break;
    case	KM_MOD_RSHIFT:	nKey = KM_RSHIFT;			break;
    case	KM_MOD_LCTRL:		nKey = KM_LCTRL;			break;
    case	KM_MOD_RCTRL:		nKey = KM_RCTRL;			break;
    case	KM_MOD_LALT:		nKey = KM_LALT;				break;
    case	KM_MOD_RALT:		nKey = KM_RALT;				break;
    case	KM_MOD_CAPS:		nKey = KM_CAPSLOCK;		break;
    case	KM_MOD_NUM:			nKey = KM_NUMLOCK;		break;
    case	KM_MOD_SCROLL:	nKey = KM_SCROLLLOCK;	break;
		case	KM_MOD_LMOUSE:	nKey = KM_MB_LEFT;		break;
		case	KM_MOD_CMOUSE:	nKey = KM_MB_CENTER;	break;
		case	KM_MOD_RMOUSE:	nKey = KM_MB_RIGHT;		break;
		case	KM_MOD_MOUSE4:	nKey = KM_MB_BUTTON4;	break;
		case	KM_MOD_MOUSE5:	nKey = KM_MB_BUTTON5;	break;
		case	KM_MOD_MOUSE6:	nKey = KM_MB_BUTTON6;	break;
		case	KM_MOD_MOUSE7:	nKey = KM_MB_BUTTON7;	break;
		case	KM_MOD_JOY0:	nKey = KM_JOY_BUTTON4;	break;
		case	KM_MOD_JOY1:	nKey = KM_JOY_BUTTON5;	break;
		case	KM_MOD_JOY2:	nKey = KM_JOY_BUTTON6;	break;
		case	KM_MOD_JOY3:	nKey = KM_JOY_BUTTON7;	break;
		case	KM_MOD_JOY4:	nKey = KM_JOY_BUTTON4;	break;
		case	KM_MOD_JOY5:	nKey = KM_JOY_BUTTON5;	break;
		case	KM_MOD_JOY6:	nKey = KM_JOY_BUTTON6;	break;
		case	KM_MOD_JOY7:	nKey = KM_JOY_BUTTON7;	break;
		default:							break;
	}

	return nKey;
}

uint32 con_key_to_mod(uint16 nKey)
{
	uint32 nMod = 0x00000000;

	switch (nKey)
	{
    case	KM_LSHIFT:			nMod = KM_MOD_LSHIFT;		break;
    case	KM_RSHIFT:			nMod = KM_MOD_RSHIFT;		break;
    case	KM_LCTRL:				nMod = KM_MOD_LCTRL;		break;
    case	KM_RCTRL:				nMod = KM_MOD_RCTRL;		break;
    case	KM_LALT:				nMod = KM_MOD_LALT;			break;
    case	KM_RALT:				nMod = KM_MOD_RALT;			break;
    case	KM_CAPSLOCK:		nMod = KM_MOD_CAPS;			break;
    case	KM_NUMLOCK:			nMod = KM_MOD_NUM;			break;
    case	KM_SCROLLLOCK:	nMod = KM_MOD_SCROLL;		break;
		case	KM_MB_LEFT:			nMod = KM_MOD_LMOUSE;		break;
		case	KM_MB_CENTER:		nMod = KM_MOD_CMOUSE;		break;
		case	KM_MB_RIGHT:		nMod = KM_MOD_RMOUSE;		break;
		case	KM_MB_BUTTON4:	nMod = KM_MOD_MOUSE4;		break;
		case	KM_MB_BUTTON5:	nMod = KM_MOD_MOUSE5;		break;
		case	KM_MB_BUTTON6:	nMod = KM_MOD_MOUSE6;		break;
		case	KM_MB_BUTTON7:	nMod = KM_MOD_MOUSE7;		break;
		case	KM_JOY_BUTTON0:	nMod = KM_MOD_JOY0;			break;
		case	KM_JOY_BUTTON1:	nMod = KM_MOD_JOY1;			break;
		case	KM_JOY_BUTTON2:	nMod = KM_MOD_JOY2;			break;
		case	KM_JOY_BUTTON3:	nMod = KM_MOD_JOY3;			break;
		case	KM_JOY_BUTTON4:	nMod = KM_MOD_JOY4;			break;
		case	KM_JOY_BUTTON5:	nMod = KM_MOD_JOY5;			break;
		case	KM_JOY_BUTTON6:	nMod = KM_MOD_JOY6;			break;
		case	KM_JOY_BUTTON7:	nMod = KM_MOD_JOY7;			break;
		default:							break;
	}

	return nMod;
}

bool con_is_assignment_firing(Assignment *pAssignment)
{
	int i;
	bool rCode = false;

	if (gTrigger[pAssignment->aTrigger].tState)
	{	/* at least main trigger has to be down */
		rCode = true; /* assume modifier true */
		for (i = 0; i < C_CON_MAX_MODIFIER && rCode; i++)
			if (pAssignment->aModifier[i])
			{
				if(!(gTrigger[pAssignment->aModifier[i]].tState))
					rCode = false;
			}
	}
	
	return rCode;
}

bool con_is_op_assignment_firing(OpAssignment *pAssignment)
{
	int i;
	bool rCode = false;

	rCode = true; /* assume modifier true */
	for (i = 0; i < C_CON_MAX_MODIFIER && rCode; i++)
		if (pAssignment->oaModifier[i])
		{
			if(!(gTrigger[pAssignment->oaModifier[i]].tState))
				rCode = false;
		}
	
	return rCode;
}

Assignment *con_find_best_assignment(AssignmentMode *pMode)
{
	Assignment *pAssignment = NULL;
	int i, j, nMods = -1, nBuf;
	
	for (i = 0; i < pMode->amAssignments; i++)
	{
		if (gTrigger[pMode->amAssigns[i]->aTrigger].tState)
		{	/* at least main trigger has to be down */
			nBuf = 0;
			for (j = 0; j < C_CON_MAX_MODIFIER && nBuf >= 0; j++)
				if (pMode->amAssigns[i]->aModifier[j])
				{
					if (gTrigger[pMode->amAssigns[i]->aModifier[j]].tState)
						nBuf++;
					else
						nBuf = -1;
				}
			if (nBuf > nMods)
				pAssignment = pMode->amAssigns[i];
		}
	}

	return pAssignment;
}

OpAssignment *con_find_best_op_assignment(OperationMode *pMode)
{
	OpAssignment *pAssignment = NULL;
	int i, j, nMods = -1, nBuf;
	
	for (i = 0; i < pMode->omAssignments; i++)
	{
		nBuf = 0;
		for (j = 0; j < C_CON_MAX_MODIFIER && nBuf >= 0; j++)
			if (pMode->omAssigns[i]->oaModifier[j])
			{
				if (gTrigger[pMode->omAssigns[i]->oaModifier[j]].tState)
					nBuf++;
				else
					nBuf = -1;
			}
		if (nBuf > nMods)
			pAssignment = pMode->omAssigns[i];
	}

	return pAssignment;
}

void Con_reset_defaults(void)
{
	/* first free all memory */
	con_free_modes();
	con_init_basics();
	/* assign mouse motions */
	con_add_operation(C_CON_MODE_PLAY, OIDLookSide, MOVE_MOUSEX, KM_MOD_RMOUSE);
	con_add_operation(C_CON_MODE_PLAY, OIDLookHeight, MOVE_MOUSEY, KM_MOD_RMOUSE);
	con_add_operation(C_CON_MODE_PLAY, OIDMoveSide, MOVE_MOUSEX, KM_MOD_LALT);
	con_add_operation(C_CON_MODE_PLAY, OIDLookHeight, MOVE_MOUSEY, KM_MOD_LALT);
	con_add_operation(C_CON_MODE_PLAY, OIDLookSide, MOVE_MOUSEX, KM_MOD_LSHIFT);
	con_add_operation(C_CON_MODE_PLAY, OIDMoveWalk, MOVE_MOUSEY, KM_MOD_LSHIFT);

	con_add_operation(C_CON_MODE_USER1, OIDMoveCursorX, MOVE_MOUSEX, KM_MOD_NONE);
	con_add_operation(C_CON_MODE_USER1, OIDMoveCursorY, MOVE_MOUSEY, KM_MOD_NONE);

	con_add_operation(C_CON_MODE_PLAY, OIDLookSide, MOVE_MOUSEX, KM_MOD_NONE);
	con_add_operation(C_CON_MODE_PLAY, OIDLookHeight, MOVE_MOUSEY, KM_MOD_NONE);
	/* now assign keys */
	con_add_action_key(C_CON_MODE_PLAY, AIDUserMode1, KM_TAB, KM_MOD_NONE, C_CON_TRIGGER);
	con_add_action_key(C_CON_MODE_USER1, AIDPlayMode, KM_TAB, KM_MOD_NONE, C_CON_TRIGGER);

	con_add_action_key(C_CON_MODE_PLAY, AIDSelect, KM_MB_LEFT, KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDGrab, KM_MB_RIGHT, KM_MOD_NONE, C_CON_SYNCH);

	con_add_action_key(C_CON_MODE_PLAY, AIDRun, 's', KM_MOD_LSHIFT, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDRun, KM_UP, KM_MOD_LSHIFT, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDWalk, 's', KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDWalk, KM_UP, KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDBack, 'x', KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDBack, KM_DOWN, KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDLeft, 'a', KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDLeft, KM_LEFT, KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDRight, 'd', KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDRight, KM_RIGHT, KM_MOD_NONE, C_CON_SYNCH);

	con_add_action_key(C_CON_MODE_PLAY, AIDLeanLeft, 'q', KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDLeanLeft, KM_LEFT, KM_MOD_LCTRL, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDLeanCenter, 'w', KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDLeanRight, 'e', KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDLeanRight, KM_RIGHT, KM_MOD_LCTRL, C_CON_SYNCH);

	con_add_action_key(C_CON_MODE_PLAY, AIDStrafeLeft, 'z', KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDStrafeLeft, KM_LEFT, KM_MOD_LALT, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDStrafeRight, 'c', KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDStrafeRight, KM_RIGHT, KM_MOD_LALT, C_CON_SYNCH);

	con_add_action_key(C_CON_MODE_PLAY, AIDLookUp, 'r', KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDLookStraight, 'f', KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDLookDown, 'v', KM_MOD_NONE, C_CON_SYNCH);

	con_add_action_key(C_CON_MODE_PLAY, AIDStand, 't', KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDDuck, 'g', KM_MOD_NONE, C_CON_SYNCH);
	con_add_action_key(C_CON_MODE_PLAY, AIDCrawl, 'b', KM_MOD_NONE, C_CON_SYNCH);

	if (!Option_underworld)
	{	/* integrate SS controls */
		con_add_action_key(C_CON_MODE_PLAY, AIDBigmap, '`', KM_MOD_NONE, C_CON_SYNCH);
		con_add_action_key(C_CON_MODE_PLAY, AIDBigmap, '`', KM_MOD_LSHIFT, C_CON_SWITCH);

		con_add_action_key(C_CON_MODE_PLAY, AIDBigmapZoomIn, KM_KP_PLUS, KM_MOD_NONE, C_CON_TRIGGER);
		con_add_action_key(C_CON_MODE_PLAY, AIDBigmapZoomOut, KM_KP_MINUS, KM_MOD_NONE, C_CON_TRIGGER);
		con_add_action_key(C_CON_MODE_PLAY, AIDSensaround, '3', KM_MOD_NONE, C_CON_SYNCH);
		con_add_action_key(C_CON_MODE_PLAY, AIDSensaround, '3', KM_MOD_LALT, C_CON_SWITCH);
		con_add_action_key(C_CON_MODE_PLAY, AIDCompass, '7', KM_MOD_NONE, C_CON_SWITCH);

		con_add_action_key(C_CON_MODE_PLAY, AIDLMFDWeapon, KM_F01, KM_MOD_NONE, C_CON_TRIGGER);
		con_add_action_key(C_CON_MODE_PLAY, AIDLMFDItem, KM_F02, KM_MOD_NONE, C_CON_TRIGGER);
		con_add_action_key(C_CON_MODE_PLAY, AIDLMFDMap, KM_F03, KM_MOD_NONE, C_CON_TRIGGER);
		con_add_action_key(C_CON_MODE_PLAY, AIDLMFDTargetList, KM_F04, KM_MOD_NONE, C_CON_TRIGGER);
		con_add_action_key(C_CON_MODE_PLAY, AIDLMFDData, KM_F05, KM_MOD_NONE, C_CON_TRIGGER);
		con_add_action_key(C_CON_MODE_PLAY, AIDRMFDWeapon, KM_F06, KM_MOD_NONE, C_CON_TRIGGER);
		con_add_action_key(C_CON_MODE_PLAY, AIDRMFDItem, KM_F07, KM_MOD_NONE, C_CON_TRIGGER);
		con_add_action_key(C_CON_MODE_PLAY, AIDRMFDMap, KM_F08, KM_MOD_NONE, C_CON_TRIGGER);
		con_add_action_key(C_CON_MODE_PLAY, AIDRMFDTargetList, KM_F09, KM_MOD_NONE, C_CON_TRIGGER);
		con_add_action_key(C_CON_MODE_PLAY, AIDRMFDData, KM_F10, KM_MOD_NONE, C_CON_TRIGGER);
	}
}

void con_assignment_to_mode(AssignmentMode *pMode, 
														Assignment *pAssignment)
{
	pMode->amAssignments++;
	pMode->amAssigns = (Assignment **)realloc(pMode->amAssigns,
																						sizeof(Assignment *) * 
																						pMode->amAssignments);
	pMode->amAssigns[pMode->amAssignments - 1] = pAssignment;
}

void con_op_assignment_to_mode(OperationMode *pMode,
															 OpAssignment *pAssignment)
{
	pMode->omAssignments++;
	pMode->omAssigns = (OpAssignment **)realloc(pMode->omAssigns,
																							sizeof(OpAssignment *) * 
																							pMode->omAssignments);
	pMode->omAssigns[pMode->omAssignments - 1] = pAssignment;
}

void con_assignment_to_trigger(byte nMode, Assignment *pAssignment, 
															 uint16 nKey)
{
	AssignmentMode *pMode = NULL;

	pMode = &(gTrigger[nKey].tModes[nMode]);
	con_assignment_to_mode(pMode, pAssignment);
}

void con_assignment_to_movement(byte nMode, OpAssignment *pAssignment,
																uint16 nMove)
{
	OperationMode *pMode;

	pMode = &(gMovements[nMove].mModes[nMode]);
	con_op_assignment_to_mode(pMode, pAssignment);
}

void con_assignment_to_action(byte nMode, Assignment *pAssignment,
															ss_action_id nId)
{
	AssignmentMode *pMode = NULL;

	pMode = &(gActions[nId].baModes[nMode]);
	con_assignment_to_mode(pMode, pAssignment);
}

void con_assignment_to_operation(byte nMode, OpAssignment *pAssignment,
																 ss_operation_id nId)
{
	OperationMode *pMode = NULL;

	pMode = &(gOperations[nId].oModes[nMode]);
	con_op_assignment_to_mode(pMode, pAssignment);
}

bool	con_add_operation(byte nMode, ss_operation_id nId,
												uint16 nMove, uint32 nModifier)
{
	OpAssignment *pAssignment = NULL;
	uint16 i, nModKey;
	int j;
	bool rCode = false;

	if (nMode < C_CON_MODE_AMOUNT && nMove && 
			nMove < C_MOVE_AMOUNT && nId < OIDAmount)
	{
		/* make up assignment */
		pAssignment = (OpAssignment *)malloc(sizeof(OpAssignment));
		memset(pAssignment, 0x00, sizeof(OpAssignment));
		/* set up keys */
		pAssignment->oaMovement = nMove;
		con_assignment_to_movement(nMode, pAssignment, nMove);
		for (i = 0, j = 0; i < 32 && j < C_CON_MAX_MODIFIER; i++)
		{
			nModKey = con_mod_to_key((uint32)(nModifier & (0x00000001 << i)));
			if (nModKey)
			{
				pAssignment->oaModifier[j++] = nModKey;
				/* con_assignment_to_trigger(nMode, pAssignment, nModKey); */
			}
		}
		/* id */
		pAssignment->oaOperationId = nId;
		/* assign action */
		con_assignment_to_operation(nMode, pAssignment, nId);
		rCode = true;
	}	

	return rCode;
}

bool con_add_action_key(byte nMode, ss_action_id nId,
												uint16 nKey, uint32 nModifier, byte nType)
{
	Assignment *pAssignment = NULL;
	uint16 i, nModKey;
	int j;
	bool rCode = false;

	if (nMode < C_CON_MODE_AMOUNT && nKey && 
			nKey < C_KM_AMOUNT && nId < AIDAmount)
	{
		/* make up assignment */
		pAssignment = (Assignment *)malloc(sizeof(Assignment));
		memset(pAssignment, 0x00, sizeof(Assignment));
		/* set up keys */
		pAssignment->aTrigger = nKey;
		con_assignment_to_trigger(nMode, pAssignment, nKey);
		for (i = 0, j = 0; i < 32 && j < C_CON_MAX_MODIFIER; i++)
		{
			nModKey = con_mod_to_key((uint32)(nModifier & (0x00000001 << i)));
			if (nModKey)
			{
				pAssignment->aModifier[j++] = nModKey;
				con_assignment_to_trigger(nMode, pAssignment, nModKey);
			}
		}
		/* type, id */
		pAssignment->aType = nType;
		pAssignment->aActionId = nId;
		/* assign action */
		con_assignment_to_action(nMode, pAssignment, nId);
		rCode = true;
	}	

	return rCode;
}

void Con_key_activate(uint16 nKey, uint32 dwTick)
{
	if (nKey < C_KM_AMOUNT)
	{	/* safety first */
		if (!gTrigger[nKey].tDownTick)
			gTrigger[nKey].tDownTick = dwTick;
		gTrigger[nKey].tUpTick = 0;
		gTrigger[nKey].tState = true;
		con_execute_trigger(nKey);

		if (nKey != KM_MB_NONE)
		{
			Con_key_activate(KM_MB_NONE, dwTick);
			Con_key_deactivate(KM_MB_NONE, dwTick);
		}
	}
}

void Con_key_deactivate(uint16 nKey, uint32 dwTick)
{
	if (nKey < C_KM_AMOUNT)
	{	/* safety first */
		if (gTrigger[nKey].tState)
		{	/* deactivate although not activated ?? */
			gTrigger[nKey].tDownTick = 0;
			gTrigger[nKey].tUpTick = dwTick;
			gTrigger[nKey].tState = false;
			con_stop_trigger(nKey);

			if (nKey != KM_MB_NONE)
			{
				Con_key_activate(KM_MB_NONE, dwTick);
				Con_key_deactivate(KM_MB_NONE, dwTick);
			}
		}
	}
}

void Con_key_modifier(uint32 fModifier, bool bAdd)
{
	if (bAdd)
	{
		nCurModifier |= fModifier;
		Con_key_activate(con_mod_to_key(fModifier), 0x00000000);
	}
	else
	{
		nCurModifier &= ~fModifier;
		Con_key_deactivate(con_mod_to_key(fModifier), 0x00000000);
	}
#ifdef CON_DEBUG
	printf("Con_key_modifier() Modifier now %08lX\n", nCurModifier);
#endif
}

uint32	Con_get_modifier(void)
{
	return nCurModifier;
}

void Con_mouse_activate(uint16 nButton, point ptMouse, uint32 dwTick)
{
	nCurModifier |= nButton;
	Con_key_activate(nButton, dwTick);
#ifdef CON_DEBUG
	printf("Con_mouse_activate() %04X at %03d.%03d\n",
				 nButton, ptMouse.x, ptMouse.y);
#else
	UNUSED(ptMouse);
#endif
}

void Con_move(uint16 nMove, int rel, uint32 dwTick)
{
	if (nMove < C_MOVE_AMOUNT)
	{
		gMovements[nMove].mPos += rel;
		gMovements[nMove].mLastTick = dwTick;
		con_execute_movement(nMove, rel);

		Con_key_activate(KM_MB_NONE, dwTick);
		Con_key_deactivate(KM_MB_NONE, dwTick);
	}
}

void Con_mouse_deactivate(uint16 nButton, point ptMouse, uint32 dwTick)
{
	nCurModifier &= ~nButton;

	Con_key_deactivate(nButton, dwTick);
#ifdef CON_DEBUG
	printf("Con_mouse_deactivate() %04X at %03d.%03d\n",
				 nButton, ptMouse.x, ptMouse.y);
#else
	UNUSED(ptMouse);
#endif
}

void	con_execute_trigger(uint16 nKey)
{
	Trigger *pTrigger = NULL;
	AssignmentMode *pMode = NULL;
/*  	Assignment *pAssignment = NULL; */
	int i, j;
	byte nMode;
	bool bFinished = false;

	pTrigger = &gTrigger[nKey];
	for (nMode = 0; nMode < C_CON_MODE_AMOUNT && !bFinished; nMode++)
	{
		pMode = &pTrigger->tModes[nMode];

		for (i = 0; i < pMode->amAssignments && Con_is_mode_active(nMode); i++)
		{
			if (con_is_assignment_firing(pMode->amAssigns[i]) &&
					pMode->amAssigns[i] == 
					con_find_best_assignment(&gActions[pMode->amAssigns[i]->aActionId].baModes[nMode]) &&
					pMode->amAssigns[i] == con_find_best_assignment(pMode))
			{
				bFinished = true;
				if (gActions[pMode->amAssigns[i]->aActionId].baActive[nMode] ==
						C_CON_SYNCH && pMode->amAssigns[i]->aType != C_CON_SYNCH)
				{	/* synchronized get overruled */
					gActions[pMode->amAssigns[i]->aActionId].baActive[nMode] = pMode->amAssigns[i]->aType;
				}
				else
				{
					if (!(gActions[pMode->amAssigns[i]->aActionId].baActive[nMode]))
					{
#ifdef CON_DEBUG
						printf("con_execute_trigger() Starting Action <%s> as %02X\n",
									 ActionNames[pMode->amAssigns[i]->aActionId],
									 pMode->amAssigns[i]->aType);
#endif
						if (pMode->amAssigns[i]->aType == C_CON_SYNCH)
							for (j = 0; j < C_MAX_EXCLUSIONS; j++)
								if (gActions[pMode->amAssigns[i]->aActionId].baExclusions[j] != AIDAmount)
								{
									gActions[gActions[pMode->amAssigns[i]->aActionId].baExclusions[j]].baActive[nMode] = C_CON_OFF;
									con_action_execute(gActions[pMode->amAssigns[i]->aActionId].baExclusions[j], false);
								}
						con_action_execute(pMode->amAssigns[i]->aActionId, true);
						gActions[pMode->amAssigns[i]->aActionId].baActive[nMode] = 
								pMode->amAssigns[i]->aType;
					}
					else
					{
						if (pMode->amAssigns[i]->aType == C_CON_SWITCH &&
								gActions[pMode->amAssigns[i]->aActionId].baActive[nMode] == C_CON_SWITCH)
						{
#ifdef CON_DEBUG
							printf("con_execute_trigger() Stopping Action <%s>\n",
										 ActionNames[pMode->amAssigns[i]->aActionId]);
#endif
							con_action_execute(pMode->amAssigns[i]->aActionId, false);
							gActions[pMode->amAssigns[i]->aActionId].baActive[nMode] = C_CON_OFF;
						}
					}
				}
			}
		}
	}
}

void	con_execute_movement(uint16 nMove, int rel)
{
	Movement *pMove = NULL;
	OperationMode *pMode = NULL;
	int i; 
	byte nMode;
	bool bFinished = false;

	pMove = &gMovements[nMove];
	for (nMode = 0; nMode < C_CON_MODE_AMOUNT && !bFinished; nMode++)
	{
		pMode = &pMove->mModes[nMode];

		for (i = 0; i < pMode->omAssignments && Con_is_mode_active(nMode); i++)
		{
			if (con_is_op_assignment_firing(pMode->omAssigns[i]) &&
					con_find_best_op_assignment(pMode) == pMode->omAssigns[i])
			{
				gOperations[pMode->omAssigns[i]->oaOperationId].oCurPos += rel;
				con_operation_execute(pMode->omAssigns[i]->oaOperationId, rel);
				bFinished = true;
			}
		}
	}
}

void	con_stop_trigger(uint16 nKey)
{
	Trigger *pTrigger = NULL;
	AssignmentMode *pMode = NULL, *pModeBuf = NULL;
	int i, j, k;
	byte nMode;

	pTrigger = &gTrigger[nKey];
	for (nMode = 0; nMode < C_CON_MODE_AMOUNT; nMode++)
	{
		pMode = &pTrigger->tModes[nMode];

		for (i = 0; i < pMode->amAssignments && Con_is_mode_active(nMode); i++)
		{
			if (pMode->amAssigns[i]->aType == C_CON_SYNCH &&
					gActions[pMode->amAssigns[i]->aActionId].baActive[nMode] == C_CON_SYNCH)
			{	/* synchronized stop at release */
				if (!con_find_best_assignment(&(gActions[pMode->amAssigns[i]->aActionId].baModes[nMode])))
				{
#ifdef CON_DEBUG
					printf("con_stop_trigger() Stopping Action <%s>\n",
								 ActionNames[pMode->amAssigns[i]->aActionId]);
#endif
					gActions[pMode->amAssigns[i]->aActionId].baActive[nMode] = C_CON_OFF;
					con_action_execute(pMode->amAssigns[i]->aActionId, false);
					for (j = 0; j < C_MAX_EXCLUSIONS; j++)
					{
						if (gActions[pMode->amAssigns[i]->aActionId].baExclusions[j] < AIDAmount)
						{
							pModeBuf = &(gActions[gActions[pMode->amAssigns[i]->aActionId].baExclusions[j]].baModes[nMode]);
							for (k = 0; k < pModeBuf->amAssignments; k++)
							{	/* excluded synchronized can be restarted */
								if (pModeBuf->amAssigns[k]->aType == C_CON_SYNCH &&
										con_is_assignment_firing(pModeBuf->amAssigns[k]))
								{
#ifdef CON_DEBUG
									printf("con_stop_trigger() Reactivating Action <%s>\n",
												 ActionNames[pModeBuf->amAssigns[k]->aActionId]);
#endif
									con_action_execute(pModeBuf->amAssigns[k]->aActionId, true);
									gActions[pModeBuf->amAssigns[k]->aActionId].baActive[nMode] = pModeBuf->amAssigns[k]->aType;
								}
							}
						}
					}
				}
			}
			else
				if (gActions[pMode->amAssigns[i]->aActionId].baActive[nMode] != C_CON_SWITCH)
					gActions[pMode->amAssigns[i]->aActionId].baActive[nMode] = C_CON_OFF;
		}
	}
}

/*
 * execute an action. function doesn't think - it's only smart.
 */
void	con_action_execute(ss_action_id nId, bool bStart)
{
	uint32 nCardId[3] = { 0, 0, 0 };
	int i;

	switch (nId)
	{	
		/* system */
		case AIDPlayMode:
		{
			if (bStart)
			{
				Con_activate_mode(C_CON_MODE_USER1, false);
				Con_activate_mode(C_CON_MODE_USER2, false);
				Con_activate_mode(C_CON_MODE_USER3, false);
			}
			break;
		}
		case AIDUserMode1:
		{
			Con_activate_mode(C_CON_MODE_USER2, false);
			Con_activate_mode(C_CON_MODE_USER3, false);
			Con_activate_mode(C_CON_MODE_USER1, bStart);
			break;
		}
		case AIDUserMode2:
		{
			Con_activate_mode(C_CON_MODE_USER1, false);
			Con_activate_mode(C_CON_MODE_USER3, false);
			Con_activate_mode(C_CON_MODE_USER2, bStart);
			break;
		}
		case AIDUserMode3:
		{
			Con_activate_mode(C_CON_MODE_USER1, false);
			Con_activate_mode(C_CON_MODE_USER2, false);
			Con_activate_mode(C_CON_MODE_USER3, bStart);
			break;
		}
    /* movement */
    case AIDRun:				CurSet.bControl.run = (bStart ? 2 : 0); break;
    case AIDWalk:				CurSet.bControl.forwards = (bStart ? 1 : 0); break;
    case AIDBack:				CurSet.bControl.back = (bStart ? 1 : 0);		 break;
    case AIDRight:			CurSet.bControl.right = (bStart ? 1 : 0);		 break;
    case AIDLeft:				CurSet.bControl.left = (bStart ? 1 : 0);		 break;
    case AIDStrafeRight: CurSet.bControl.stepleft = (bStart ? 1 : 0);	 break;
    case AIDStrafeLeft: CurSet.bControl.stepright = (bStart ? 1 : 0);	 break;
    case AIDJump:				CurSet.bControl.jump = (bStart ? 1 : 0);			break;
    /* viewpose */
    case AIDLeanLeft:
    case AIDLeanCenter:
    case AIDLeanRight:
				printf("con_action_execute() Implement me! <%s>\n", ActionNames[nId]);
				break;
		case AIDLookUp:				CurSet.bControl.look = (bStart ? +2 : +1);	break;
		case AIDLookStraight:	CurSet.bControl.look = 0;									break;
		case AIDLookDown:			CurSet.bControl.look = (bStart ? -2 : -1);	break;
    case AIDStand:
    case AIDDuck:
    case AIDCrawl:
				printf("con_action_execute() Implement me! <%s>\n", ActionNames[nId]);
				break;
    /* interaction */
    case AIDSelect:
		{
			if (!HUD_select_at_cursor(bStart))
			{	/* HUD didn't eat that event, ask world */

			}
			break;
		}
    case AIDActivate:
		{
			if (!HUD_activate_at_cursor(bStart))
			{	/* HUD didn't eat that event, ask world */

			}
			break;
		}
    case AIDGrab:
		{
			if (!HUD_grab_at_cursor(bStart))
			{	/* HUD didn't eat that event, ask world */

			}
			break;
		}
		/* shortcuts */
		case AIDNextWeapon:
		case AIDReloadWeapon:
    /* hardware */
    case AIDBiomonitor:
		{
			nCardId[0] = mfd_set_id(F_SMFD_DATA, F_SMFD_DBIO, 0x0000);
			break;
		}
    case AIDSensaround:
		{
			if (!Option_underworld)		
				MFD_show_sensaround(bStart);
			break;
		}
    case AIDLantern:
    case AIDShield:
    case AIDInfrared:
			printf("con_action_execute() Implement me! <%s>\n", ActionNames[nId]);
			break;
    case AIDCompass:
		{												
			MFD_show_compass(bStart);
			break;
		}
    case AIDDataReader:
    case AIDMotionBooster:
    case AIDJumpJet:			 
			printf("con_action_execute() Implement me! <%s>\n", ActionNames[nId]);
			break;
    case AIDBigmap:
		{
			if (!Option_underworld)		
				MFD_show_bigmap(bStart);
			break;
		}
    case AIDBigmapZoomIn:
		{
			if (MFD_is_bigmap_visible() && bStart)
				MFD_set_zoom_bigmap(MFD_get_zoom_bigmap() + 5);
			break;
		}
    case AIDBigmapZoomOut:
		{
			if (MFD_is_bigmap_visible() && bStart)
				MFD_set_zoom_bigmap(MFD_get_zoom_bigmap() - 5);
			break;
		}
    /* mfd cards */
    case AIDLMFDWeapon:			nCardId[0] = F_SMFD_WEAPON;	break;
    case AIDLMFDAmmo:				/*nCardId[0] = F_SMFD_WEAPON;*/ break;
    case AIDLMFDItem:				nCardId[0] = F_SMFD_ITEM;		break;
    case AIDLMFDMap:				nCardId[0] = F_SMFD_MAP;		break;
    case AIDLMFDTargetList:	nCardId[0] = F_SMFD_TARGET;	break;
    case AIDLMFDTargetInfo:	/*nCardId[0] = F_SMFD_WEAPON;*/	break;
    case AIDLMFDData:				nCardId[0] = F_SMFD_DATA;		break;

    case AIDCMFDWeapon:			nCardId[1] = F_CMFD_WEAPON;	break;
    case AIDCMFDHardware:		nCardId[1] = F_CMFD_HW;			break;
    case AIDCMFDInventory:	nCardId[1] = F_CMFD_INVENT;	break;
    case AIDCMFDSoftware:		nCardId[1] = F_CMFD_SW;			break;

    case AIDRMFDWeapon:			nCardId[2] = F_SMFD_WEAPON;	break;
    case AIDRMFDAmmo:				/*nCardId[2] = F_SMFD_WEAPON;*/	break;
    case AIDRMFDItem:				nCardId[2] = F_SMFD_ITEM;		break;
    case AIDRMFDMap:				nCardId[2] = F_SMFD_MAP;		break;
    case AIDRMFDTargetList: nCardId[2] = F_SMFD_TARGET;	break;
    case AIDRMFDTargetInfo: /*nCardId[2] = F_SMFD_WEAPON;*/	break;
    case AIDRMFDData:				nCardId[2] = F_SMFD_DATA;		break;
		/* system */
		case AIDScreenshot:
		{
			if (bStart)
				Game_screenshot();
			break;
		}
    case AIDQuit:						CurSet.quit = (bStart ? 1 : 0); break;
		case AIDDebug:
		{
			if (bStart)
			{
				Game_debug_output();
			}
			break;
		}
		case AIDAmount:
		default:
			Cant_happen("con_action_execute() Can't execute unknown Action");
			break;
	}
	for (i = 0; i < 3; i++)
	{
		if (nCardId[i])
		{
			/* ff1: this is wrong. cards can show different things; */
			if (mfd_get_card(MFD_get_active_id(i)) != mfd_get_card(nCardId[i]))
				MFD_set_active_id(i, nCardId[i], true);
			else
				MFD_set_active_id(i, F_MFD_NONE, true);

		}
	}
}

/*
 * execute an operation. function doesn't think - it's only smart.
 */

void	con_operation_execute(ss_operation_id nId, int rel)
{
	point ptCursor;

	HUD_get_center_pos(&ptCursor);
	switch (nId)	
	{
		case OIDMoveSide:
		{
			CurSet.rControl.stepleft = max(rel, 0);
			CurSet.rControl.stepright = abs(min(rel, 0));
			break;
		}
		case OIDMoveWalk:
		{
			CurSet.rControl.forwards = abs(min(rel, 0));
			CurSet.rControl.back = max(rel, 0);
			break;
		}
		case OIDLookSide:
		{
			CurSet.rControl.left = abs(min(rel, 0));
			CurSet.rControl.right = max(rel, 0);
			break;
		}
		case OIDLookHeight:
		{
			CurSet.rControl.height = rel;
			break;
		}
		case OIDPose:
			printf("con_action_execute() Implement me! <%p>\n", OperationNames[nId]);
			break;
		case OIDMoveCursorX:
		{	/* cursor movement causes it to snap back to it's pos */
			HUD_get_cursor_pos(&ptCursor);
			ptCursor.x = gOperations[OIDMoveCursorX].oCurPos;
			break;
		}
		case OIDMoveCursorY:
		{
			HUD_get_cursor_pos(&ptCursor);
			ptCursor.y = gOperations[OIDMoveCursorY].oCurPos;
			break;
		}
		case OIDAmount:
		default:
			Cant_happen("con_operation_execute() Can't execute unknown Action");
			break;
	}
	HUD_set_cursor_pos(ptCursor);
}

/*
 * Take a snapshot of the current control set. Keyboard and mouse handling
 *  might still be going on in parallel with other processing and we don't
 *  want our control set changing in the middle of some physics
 */
void Con_snapshot (ss_controlset *cs, bool bResetRelative)
{
	memcpy(cs, &CurSet, sizeof(ss_controlset));
	if (bResetRelative)
		memset(&(CurSet.rControl), 0x00, sizeof(RelativeControl));
}



/*
 * End of src/control.c
 */

/*
 * Follows some emacs magic to get VC++-formatted code to display right (and
 *  hopefully vice versa)
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
