/*
 * dark.h -
 *  Definitions for a Dark Sucker
 */


#ifndef DARK_H
#define DARK_H




/*
 * #includes
 */
#include "vec.h"



/*
 * Dark Sucker flags
 */
#define DS_FREE     0x0001  /* is on free list, mainly for debugging     */
#define DS_DYNAMIC  0x0002  /* recalculate lighting every frame (or few) */
#define DS_RELATIVE 0x0004  /* rel_pos is valid, else use obj pos        */
#define DS_AVATAR   0x0008  /* always applied, implies DYNAMIC           */


/*
 * The dark sucker xref table, analagous to the map object xref table
 */
struct tag_dark_sucker;
struct tag_map_tile;
struct tag_map_object;

typedef struct tag_sucker_xref
{
  struct tag_dark_sucker *ds;
  struct tag_map_tile *tile;
  int sucker_chain;            /* next tile this sucker, also freelist */
  int tile_chain;              /* next sucker this tile                */
} sucker_xref;


typedef struct tag_dark_sucker
{
  struct tag_map_object *obj;   /* base object for positioning        */
  vec    rel_pos;               /* relative to obj if DS_RELATIVE set */
  vec    position;              /* absolute, set at start of frame    */
  float  intensity;
  float  max_radius;            /* tile search terminates here        */
  uint16 flags;                 /* DS_ flags above                    */
  int    tile_index;            /* sucker xref no. in xref list       */
  struct tag_dark_sucker *next; /* singly linked list                 */
} dark_sucker;



/*
 * Temporary fudge for Underworld
 */
#define UW_LANTERN_INTENSITY 150000.f




#ifdef __cplusplus
extern "C"
{
#endif


struct tag_surface;
struct tag_map_tile;
struct tag_map_object;
struct Vector;

void Dark_init (void);
void Dark_exit (void);
dark_sucker *Dark_new_sucker (void);
void Dark_free_sucker (dark_sucker *ds);

dark_sucker *Dark_add_static_sucker (struct tag_map_object *obj,
				     float intensity, float max_radius);

void Dark_frame (void);

void Dark_do_gouraud (const struct tag_map_tile *tile, struct tag_surface *s);
float Dark_do_plane (const struct tag_map_tile *tile, const vec *v,
		     const vec *normal);
float Dark_do_point  (const struct tag_map_tile *tile, const vec *v);

void Dark_apply_dynamic_sucker (const dark_sucker *ds, struct tag_surface *s);
float Dark_get_sucker_plane (const dark_sucker *ds, const Vector *v,
			     const Vector *normal);
float Dark_get_sucker_point (const dark_sucker *ds, const Vector *v);

#ifdef __cplusplus
}
#endif




/*
 * External global variables defined in dark.c
 */



#endif /* DARK_H */
