/*
 * defines.h -
 *  TSSHP global defines
 */

/*
 * File:    src/defines.h
 * Purpose: Compile-time configuration and platform-specific defines
 * Prefix:
 * Note:    Always include this file first!
 * Music:   Life of Agony, "Ugly"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * The bool type. This is a built-in type in C++ and is defined in <stdbool.h>
 *  in C99 (allegedly). Otherwise we need define it here
 */
#if !defined (__cplusplus) && !defined (bool)
#define bool  int    /* or char or short, whatever */
#define false 0
#define true  1
#endif



/*
 * Unified DEBUG definitions
 */
#if defined (_DEBUG)
#define __DEBUG__ 1
#undef NDEBUG
#elif defined (__DEBUG__)
#define _DEBUG 1
#undef NDEBUG
#endif



/*
 * min/max defines. Not ANSI standard
 */
#if (!defined (__STDC__ ) || defined (__BORLANDC__))
#define MINMAX_IN_STDLIB 1
#else
#define MINMAX_IN_STDLIB 0
#endif


#define hi(w)	((byte)((w >> 8) & 0x00FF))
#define lo(w) ((byte)(w & 0x00FF))

#define sgn(i) ((i > 0) ? 1 : ((i < 0) ? -1 : 0))


/*
 * An `unused' macro to squash compiler `parameter is unused' warning. Don't
 *  overuse this
 */
#ifdef __GNUC__
#define UNUSED(x) x=x
#elif (defined(__BORLANDC__)) /* for Borland we use -w-par */
#define UNUSED(x)
#else
#define UNUSED(x) x
#endif

/*
 * Turn off/on some (useless) warnings of MS VC...
 */
#ifdef _MSC_VER
/* nonstandard: single line comments							*/
#pragma warning( error: 4001 )
/* macro redefinition	(because of our min/max)		*/
#pragma warning( disable : 4005 )
/* overflow in floating-point constant arithmetic */
#pragma warning( disable : 4056 )
/* conditional expression is constant							*/
#pragma warning( disable : 4127 )
/* variable is initialized but not referenced			*/
#pragma warning( 3 : 4189 )
/* 'identifier' was previously defined as a type	*/
#pragma warning( disable : 4224 )	
/* conversion x to y: possible loss of data ;)		*/
#pragma warning( disable : 4244 )
#endif

/*
 * Do we compile in support for 3D APIs (OpenGL, Direct3D etc)?
 * Note that defining this doesn't mean that they will be used at run-time ...
 *  this is dependent on the global boolean API_3d
 */
#ifndef SUPPORT_3DAPI
#define SUPPORT_3DAPI 0
#endif /* SUPPORT_3DAPI */



/*
 * Mouse support?
 */
#if (!defined(__WIN32__))	
#define MOUSE 1
#endif


/*
 * Lighting support is now ALWAYS present ... it's a vital part of all these
 *  games
 */



/*
 * Print a summary of memory allocation on startup
 */
#define MEM_INIT_MESSAGES 1




/*
 * Use quadtrees rather than BSP trees for cyberspace rendering
 */
#define QUADTREES 1



/*
 * Use the Z-buffer for sprites as well as for 3-D objects
 */
#ifndef Z_BUFFER_SPRITES
#define Z_BUFFER_SPRITES 0
#endif




/*
 * Screen bit depth
 */
#ifndef FF1							 /* err - did that one slip through? ;) */
/* #define BITS_PER_PIXEL 8  laptop development ... */
#endif

#if !defined (BITS_PER_PIXEL)
#define BITS_PER_PIXEL 32      /* default 32bpp */
#endif




/*
 * For floating-point compares. The actual value of this doesn't matter that
 *  much
 */
#define A_VERY_SMALL_THING 1.e-8




/*
 * For texture extraction and screenshots
 */
/* #define HAVE_PNG 1 */

#if !defined (HAVE_PNG)
# if defined (__Linux__)
#define HAVE_PNG 1
# else
#define HAVE_PNG 0
# endif /* defined (__Linux__) */
#endif /* !defined (HAVE_PNG) */




/*
 * Available accelerated back-ends
 */
#if !defined (HAVE_OPENGL)
#define HAVE_OPENGL 0
#endif
