/*
 * event.c -
 *  event system
 *
 * This file by Christian Haas 2001 christian.haas@eunet.at
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>								

#include "types.h"
#include "gamemem.h"
#include "event.h"

/*
#define	D_EVENT_DEBUG
*/

#define	H2CL(hClass)	((EventClass *)(hClass))
#define	H2EV(hEvent)	((Event*)(hEvent))

typedef struct tag_EventListener
{
	struct tag_EventListener	*pelPrev;
	struct tag_EventListener	*pelNext;

	EVENT_LISTENER						 elFunc;
} EventListener;

struct tag_EventClass;

typedef struct tag_Event
{
	struct tag_EventClass	*peOwner;
	struct tag_Event			*pePrev;
	struct tag_Event			*peNext;

	char									eEventName[C_MAX_EVT_EVENTNAME + 1];

	EventListener					*peListenerFirst;
	EventListener					*peListenerLast;

} Event;

typedef struct tag_EventSignal
{
	struct tag_EventSignal	*pesPrev;
	struct tag_EventSignal	*pesNext;

	Event										*pesEvent;
	void										*pesData;
	bool										 esAutoDel;
} EventSignal;

typedef struct tag_EventClass
{
	/* links into this list				*/
	struct tag_EventClass	*pecPrev;
	struct tag_EventClass	*pecNext;
	/* the name of this class			*/
	char									 ecClassName[C_MAX_EVT_CLASSNAME + 1];
	/* all listener of this class	*/	
	EventListener					*pecListenerFirst;
	EventListener					*pecListenerLast;
	/* events in this class				*/
	Event									*pecEventFirst;
	Event									*pecEventLast;	
	/* buffered signals						*/
	EventSignal						*pecSignalFirst;
	EventSignal						*pecSignalLast;
} EventClass;

EventClass		*pClassFirst = NULL;
EventClass		*pClassLast = NULL;

EventClass *evt_find_class(const char *strEventClass)
{
	EventClass *pClass = pClassFirst, *pFound = NULL;

	while (pClass && !pFound)
	{
		if (!strcmp(pClass->ecClassName, strEventClass))
			pFound = pClass;
		pClass = pClass->pecNext;
	}

	return pFound;
}

Event *evt_find_event(EventClass *pClass, const char *strEvent)
{
	Event *pEvent = pClass->pecEventFirst, *pFound = NULL;

	while (pEvent && !pFound)
	{
		if (!strcmp(pEvent->eEventName, strEvent))
			pFound = pEvent;
		pEvent = pEvent->peNext;
	}

	return pFound;
}

EventListener *evt_find_listener(EventListener *pListenerFirst,
																 EVENT_LISTENER evtListener)
{
	EventListener *pListener = pListenerFirst, *pFound = NULL;

	while (pListener)
	{
		if (pListener->elFunc == evtListener)
			pFound = pListener;
		pListener = pListener->pelNext;
	}

	return pFound;
}

void evt_remove_listener(HEVENTCLASS hClass, HEVENT hEvent,
												 EventListener *pListenerStart)
{
	EventRemover er;
	EventListener *pListener = pListenerStart, *pListenerNext;

	er.erEventType = C_EVT_INTERN_REMOVE;
	er.erClass = hClass;
	er.erEvent = hEvent;
#ifdef D_EVENT_DEBUG
	printf("+++ evt_remove_listener() Removing of Class %p, Event %p\n",
				 hClass, hEvent);
#endif
	while (pListener)
	{
		pListenerNext = pListener->pelNext;
		pListener->elFunc(NULL, NULL, &er);
		free(pListener);
		pListener = pListenerNext;
	}
#ifdef D_EVENT_DEBUG
	printf("--- evt_remove_listener()\n");
#endif
}

void				Evt_init(void)
{

}

void				Evt_deinit(void)
{
	EventClass *pClass = pClassFirst, *pClassNext;

	while (pClass)	
	{
		pClassNext = pClass->pecNext;
		Evt_remove_class(pClass);
		pClass = pClassNext;
	}
	pClassFirst = pClassLast = NULL;
}

HEVENTCLASS	Evt_get_class(const char *strEventClass)
{
	EventClass *pClass;

	pClass = evt_find_class(strEventClass);
	if (!pClass)
	{
		pClass = (EventClass *)malloc(sizeof(EventClass));
		memset(pClass, 0x00, sizeof(EventClass));
		strncpy(pClass->ecClassName, strEventClass, 
						C_MAX_EVT_CLASSNAME);
		pClass->pecPrev = pClassLast;
		if (pClassLast)
			pClassLast->pecNext = pClass;
		pClassLast = pClass;
		if (!pClassFirst)
			pClassFirst = pClass;
#ifdef D_EVENT_DEBUG
		printf("### Evt_get_class() Class <%s> created\n",
					 strEventClass);
#endif
	}

	return pClass;
}

bool				Evt_remove_class(HEVENTCLASS hEventClass)
{
	EventClass *pClass = H2CL(hEventClass);
	EventSignal *pSignal, *pSignalNext;
	Event *pEvent, *pEventNext;
	bool rCode = false;

	if (pClass)
	{
#ifdef D_EVENT_DEBUG
		printf("+++ Evt_remove_class() Removing Class <%s>\n",
					 pClass->ecClassName);
#endif
		pSignal = pClass->pecSignalFirst;
		while (pSignal)
		{
			pSignalNext = pSignal->pesNext;
			if (pSignal->esAutoDel)
				free(pSignal->pesData);
			free(pSignal);
			pSignal = pSignalNext;
		}
		pEvent = pClass->pecEventFirst;
		while (pEvent)
		{
			pEventNext = pEvent->peNext;
			Evt_remove_event(pEvent);
			pEvent = pEventNext;
		}
		evt_remove_listener(pClass, NULL, pClass->pecListenerFirst);
		if (pClass->pecPrev)
			pClass->pecPrev = pClass->pecNext;
		else
			pClassFirst = pClass->pecNext;
		if (pClass->pecNext)
			pClass->pecNext = pClass->pecPrev;
		else
			pClassLast = pClass->pecPrev;
		free(pClass);
		rCode = true;
#ifdef D_EVENT_DEBUG
		printf("--- Evt_remove_class()\n");
#endif
	}

	return rCode;
}

HEVENT			Evt_get_event(HEVENTCLASS hEventClass, const char *strEvent)
{
	EventClass *pClass = H2CL(hEventClass);
	Event *pEvent;

	pEvent = evt_find_event(pClass, strEvent);
	if (!pEvent)
	{
		pEvent = (Event *)malloc(sizeof(Event));
		memset(pEvent, 0x00, sizeof(Event));
		pEvent->peOwner = pClass;
		strncpy(pEvent->eEventName, strEvent, 
						C_MAX_EVT_EVENTNAME);
		pEvent->pePrev = pClass->pecEventLast;
		if (pClass->pecEventLast)
			pClass->pecEventLast->peNext = pEvent;
		pClass->pecEventLast = pEvent;
		if (!pClass->pecEventFirst)
			pClass->pecEventFirst = pEvent;
#ifdef D_EVENT_DEBUG
		printf("### Evt_get_event() Event <%s> in Class <%s> created\n",
					 strEvent, pClass->ecClassName);
#endif
	}

	return pEvent;
}

bool				Evt_remove_event(HEVENT hEvent)
{
	Event *pEvent;
	EventClass *pClass;
	bool rCode = false;

	pEvent = H2EV(hEvent);
	pClass = pEvent->peOwner;
	if (pClass && pEvent)
	{
#ifdef D_EVENT_DEBUG
		printf("+++ Evt_remove_event() Removing Event <%s> in Class <%s>\n",
					 pEvent->eEventName, pClass->ecClassName);
#endif
		evt_remove_listener(pClass, pEvent, pEvent->peListenerFirst);
		if (pClass->pecEventFirst == pEvent)
			pClass->pecEventFirst = pEvent->peNext;
		else
			pEvent->pePrev->peNext = pEvent->peNext;
		if (pClass->pecEventLast == pEvent)
			pClass->pecEventLast = pEvent->pePrev;
		else
			pEvent->peNext->pePrev = pEvent->pePrev;
		free(pEvent);
#ifdef D_EVENT_DEBUG
		printf("--- Evt_remove_event()\n");
#endif
	}

	return rCode;
}

bool				Evt_register(HEVENTCLASS hEventClass, HEVENT hEvent, 
												 EVENT_LISTENER evtListener)
{
	EventClass *pClass;
	Event *pEvent;
	EventListener *pListener;
	bool rCode = false;

	pClass = H2CL(hEventClass);
	if (pClass)
	{
#ifdef D_EVENT_DEBUG
		printf("### Evt_register() Adding Listener to Event %p "
					 "in Class %p\n", hEvent, hEventClass);
#endif
		pListener = (EventListener *)malloc(sizeof(EventListener));
		memset(pListener, 0x00, sizeof(EventListener));
		pListener->elFunc = evtListener;
		if (hEvent)
		{
			pEvent = H2EV(hEvent);
			pListener->pelPrev = pEvent->peListenerLast;
			if (pEvent->peListenerLast)
				pEvent->peListenerLast->pelNext = pListener;
			pEvent->peListenerLast = pListener;
			if (!pEvent->peListenerFirst)
				pEvent->peListenerFirst = pListener;
		}
		else
		{
			pListener->pelPrev = pClass->pecListenerLast;
			if (pClass->pecListenerLast)
				pClass->pecListenerLast->pelNext = pListener;
			pClass->pecListenerLast = pListener;
			if (!pClass->pecListenerFirst)
				pClass->pecListenerFirst = pListener;
		}
		rCode = true;
	}

	return rCode;
}

void				Evt_unregister(HEVENTCLASS hEventClass, HEVENT hEvent,
													 EVENT_LISTENER evtListener)
{
	EventClass *pClass;
	Event *pEvent;
	EventListener *pListener, **pListenerFirst, **pListenerLast;

	pClass = H2CL(hEventClass);
	if (pClass)
	{
#ifdef D_EVENT_DEBUG
		printf("### Evt_unregister() Removing Listener from Event %p "
					 "in Class %p\n", hEvent, hEventClass);
#endif
		if (hEvent)
		{
			pEvent = H2EV(hEvent);
			pListenerFirst = &(pEvent->peListenerFirst);
			pListenerLast = &(pEvent->peListenerLast);
		}
		else
		{
			pListenerFirst = &(pClass->pecListenerFirst);
			pListenerLast = &(pClass->pecListenerLast);
		}
		pListener = evt_find_listener(*pListenerFirst, evtListener);
		if (pListener)
		{
			if (*pListenerFirst == pListener)
				*pListenerFirst = pListener->pelNext;
			else
				pListener->pelPrev->pelNext = pListener->pelNext;
			if (*pListenerLast == pListener)
				*pListenerLast = pListener->pelPrev;
			else
				pListener->pelNext->pelPrev = pListener->pelPrev;
			free(pListener);
		}
	}
}

int					evt_call_listener(EventListener *pFirst, 
															HEVENTCLASS hClass,
															HEVENT hEvent, void *pData)
{
	EventListener *pCur = pFirst;
	int nCalled = 0;

	while (pCur)
	{
		if (pCur->elFunc)
		{
#ifdef D_EVENT_DEBUG
			printf("### evt_call_listener() Calling Listener of Event %p "
						 "in Class %p with %p\n", hEvent, hClass, pData);
#endif
			pCur->elFunc(hClass, hEvent, pData);
			nCalled++;
		}
		pCur = pCur->pelNext;
	}

	return nCalled;
}

int					Evt_pulse_event(HEVENT hEvent, void *pData)
{
	EventClass *pClass;
	Event *pEvent;
	int nCalled = 0;

	pEvent = H2EV(hEvent);
	if (pEvent)
	{
		pClass = pEvent->peOwner;
#ifdef D_EVENT_DEBUG
		printf("+++ Evt_pulse_event() Pulsing Event <%s> "
					 "in Class <%s> with %p\n", 
					 pEvent->eEventName, pClass->ecClassName, pData);
#endif
		nCalled = evt_call_listener(pClass->pecListenerFirst, 
																pClass, hEvent, pData);
		nCalled += evt_call_listener(pEvent->peListenerFirst,
																 pClass, hEvent, pData);
#ifdef D_EVENT_DEBUG
		printf("--- Evt_pulse_event()\n");
#endif
	}

	return nCalled;
}

void				Evt_fire_event(HEVENT hEvent, void *pData, bool bAutoDel)
{
	EventSignal *pSignal;
	EventClass *pClass;
	Event *pEvent;

	pEvent = H2EV(hEvent);
	if (pEvent)
	{
		pClass = pEvent->peOwner;
#ifdef D_EVENT_DEBUG
		printf("### Evt_fire_event() Firing Event <%s> "
					 "in Class <%s> with %p (%d)\n", 
					 pEvent->eEventName, pClass->ecClassName, 
					 pData, bAutoDel);
#endif
		pSignal = (EventSignal *)malloc(sizeof(EventSignal));
		memset(pSignal, 0x00, sizeof(EventSignal));
		pSignal->pesEvent = pEvent;
		pSignal->pesData = pData;
		pSignal->esAutoDel = bAutoDel;
		pSignal->pesPrev = pClass->pecSignalLast;
		if (pClass->pecSignalLast)
			pClass->pecSignalLast->pesNext = pSignal;
		pClass->pecSignalLast = pSignal;
		if (!pClass->pecSignalFirst)
			pClass->pecSignalFirst = pSignal;
	}
}

void				Evt_work(void)
{
	EventSignal *pSignal, *pSignalNext;
	EventClass *pClass = pClassFirst;

/*
#ifdef D_EVENT_DEBUG
	printf("+++ Evt_work()\n");
#endif
*/
	while (pClass)
	{
		pSignal = pClass->pecSignalFirst;
		while (pSignal)
		{
			Evt_pulse_event(pSignal->pesEvent, pSignal->pesData);

			pSignalNext = pSignal->pesNext;
			if (pSignal->esAutoDel)
				free(pSignal->pesData);
			free(pSignal);	
			pSignal = pSignalNext;
		}
		pClass->pecSignalFirst = pClass->pecSignalLast = NULL;
		pClass = pClass->pecNext;
	}
/*
#ifdef D_EVENT_DEBUG
	printf("--- Evt_work()\n");
#endif
*/
}

const char *Evt_get_event_name(HEVENT hEvent)
{
	return H2EV(hEvent)->eEventName;
}

const char *Evt_get_class_name(HEVENTCLASS hEventClass)
{
	return H2CL(hEventClass)->ecClassName;
}

HEVENTCLASS Evt_get_class_from_event(HEVENT hEvent)
{
	return H2EV(hEvent)->peOwner;
}

