/*																																	*/
/* graph2d.h -																											*/
/*  Stuff which is useful for general 2d graphics and effects				*/
/*																																	*/
/*	firefreak's annotations:																				*/
/*	ok, i added most of the functions/effects - well, 							*/
/*	currently ALL of them :) -- i tried to comment as good and			*/
/*	as accurate as possible since most of the functions							*/
/*	are little helper functions but can make a WOW effect						*/
/*	when used right - at least, that's my intention in them :)			*/
/*	--																															*/
/*	i would also like to use this intro for some 'oral' definitions */
/*	so that we talk the same language about the later things:				*/
/*																																	*/
/*	trans byte: the fourth byte to a rgb value.											*/
/*	trans mask: an 'image' consisting of 'offsets' - not trans bytes*/
/*		only when clipped, they can be used as trans bytes						*/
/*		is for special effects																				*/
/*	a shade: in context of transparency a shade is a level of				*/
/*		transparency intensity. is the result of a trans byte					*/
/*		(see shade.h for further info)																*/
/*																																	*/
/*	pcolor: a value of 4 (four) bytes containing color information	*/
/*		plus a trans byte																							*/
/*	pixel: as noun: a picture element																*/
/*				 as pointer: adressing palette indizes when in 8bit or		*/
/*										 16bit mode; rgb values in 24bit and 32bit		*/
/*																																	*/
/*	-> a pixel must NOT be a pcolor! <-															*/
/*																																	*/
/*	a render function: with the help of transparency, the function  */
/*		draws a certain pattern under respect of the previous pixel		*/
/*		trans bytes are generally lost and not saved in the target		*/
/*		image. render functions have the prefix G2Dr_									*/
/*																																	*/
/*	a copy function: such a function sets the pixels in a						*/
/*		certain pattern regardless of the previous pixel, but					*/
/*		preserves the trans bytes when transferring to the target			*/
/*		image. copy functions have the prefix G2Dc_										*/
/*																																	*/
/*	a mask function: a function that alters a	trans mask in a				*/
/*		pattern. mask functions have the prefix G2Dm_									*/		
/*																																	*/
/*	G2Dr_* : all render functions																		*/
/*	G2Dc_* : all copy functions																			*/
/*	G2Dm_* : all mask functions																			*/
/*																																	*/
/*	source converter: converts an image (probably of lower bitrate)	*/
/*		into a true color image buffer																*/
/*	target converter: converts an true color buffer into an image		*/
/*		(probably of lower bitrate)																		*/
/*																																	*/
/*	an effect cycle: several operations and/or effects on/in an			*/
/*		image buffer. a cycle starts with a source converter and ends	*/
/*		with a target converter. buffer image allocation functions		*/
/*		can be included																								*/

/*
 * Sentry
 */
#ifndef GRAPH2D_H
#define GRAPH2D_H


#include "types.h" /* this is allowed */ 


/* quite a simple define, but can shrink some text...								*/
#define	getPixelOff(x, cx, y) (x + (cx * y))

/* a bounding box. or RECT as microsoft calls it :)									*/
typedef struct
{
  int top;
  int left;
  int bottom;
  int right;
} bounding_box;

/* helper defines for boxes																					*/
#define getRectWidth(r)	((r).right - (r).left)
#define getRectHeight(r) ((r).bottom - (r).top)
#define getRectCenterx(r) ((r).left + (((r).right - (r).left) / 2))
#define getRectCentery(r) ((r).top + (((r).bottom - (r).top) / 2))
#define isRectEqual(r1, r2)							\
				((r1).left		== (r2).left	&&	\
				 (r1).top			== (r2).top		&&	\
				 (r1).right		== (r2).right &&	\
				 (r1).bottom	== (r2).bottom		\
				)
#define isPointinRect(r, x, y)					\
				(r.left <= x && r.right > x &&	\
				 r.top <= y && r.bottom > y)
/* and some helpfer for g2darea																			*/
#define	resetArea(gb)																\
				{																						\
					(gb).bBoxArea.left = 0;										\
					(gb).bBoxArea.top = 0;										\
					(gb).bBoxArea.right = (gb).bSizeImg.cx;		\
					(gb).bBoxArea.bottom = (gb).bSizeImg.cy;	\
				}

#define convertHeadToArea(h, gb)																\
				{																												\
					(gb).bSizeImg.cx = (gb).bBoxArea.right = (h).width;		\
					(gb).bSizeImg.cy = (gb).bBoxArea.bottom = (h).height;	\
					(gb).bBoxArea.left = (gb).bBoxArea.top = 0;						\
					(gb).bImg.bit = (h).bits;															\
					(gb).nType = G2DPal;																	\
				}

/* jim -gcc was bitching about missing braces round the unions in G2D_NULLAREA
   so I added them in */
#define	D_G2D_NULLAREA									\
				{ G2DAmount, { NULL }, { 0, 0 }, { 0, 0, 0, 0 }, { 0 }, NULL, 0 }

/* that's the point																									*/
typedef struct tag_point
{
  int x;
  int y;
} point;


/*
 * A 2D vector, occasionally useful
 */
typedef struct
{
  double x;
  double y;
} vec2d;


/*
 * Analagous arithmetic
 */
#define vec2_subtract(d,s1,s2) do { \
  (d).x = (s1).x - (s2).x; (d).y = (s1).y - (s2).y; } while (0)

/* a size, offset, distance,...																			*/
typedef struct
{
	unsigned short		cx;
	unsigned short		cy;
} size;

/* the base structure for all 2d operations.									      */
/* note that the functions differently use the pxl/clr/msk members  */
/*		and the bClr/bTrn members																			*/
/* bBoxArea.right/.bottom are EXCLUDING															*/

typedef enum tag_G2DType
{
	G2DPal,
	G2DPixel,
	G2DTrue,
	G2DMask,

	G2DAmount
} G2DType;

typedef struct tag_g2dbase
{
	G2DType				 nType;
	union											/* pixel != pcolor (may)								*/
	{													/* (i can't point that out too often)		*/
		pixel				*pxl;				/* an output image											*/
		pcolor			*clr;				/* a buffer image												*/
		int					*msk;				/* a trans mask													*/
		byte				*bit;				/* a paletted image											*/
	} bImg;

	size					 bSizeImg;	/* size of image												*/
	bounding_box	 bBoxArea;	/* defined box for operation						*/
	union
	{
		pcolor			 clr;				/* (main) color + trans to use					*/
		int					 trn;				/* transparency altering value					*/
	} bOp;
	pcolor				*pPal;			/* palette															*/
	byte					 nPalSize;	/* palette entries - 1									*/
} g2dbase;
 
#ifdef __cplusplus
extern "C"
{
#endif

/* an interesting warning VC threw at me :)                         */
struct tag_ss_bitmap_header; 

/* zoom size orig into size target without loosing its scale				*/
void	G2D_zoom_rect(const size *orig, const size *target, 
										size *result);

/* init the effects subsystem																				*/
void	G2D_init(void);

void	G2D_reset_base(g2dbase *pG2DB, bool bFreeBuffer);
void	G2D_resize_base(g2dbase *pG2DB, uint16 nWidth, uint16 nHeight);

/* initialize a base.	COLOR, PIXEL or TRANS MASK										*/
/* after return this is set:																				*/
/* bImg.??  : a pointer that can hold the image											*/
/* bSizeImg : given width and height																*/
/* bBoxArea : 0.0 width.height																			*/
/* bClr			: 0x00000000																						*/
/*																																	*/
/* return value: true if memory could be allocated, false otherwise	*/
/*							 (when false the pointer is NULL; all other members */
/*								are set either way)																*/
/* note: the entire structure gets set and there's no 'checking' or	*/
/*			 'freeing' of a previous buffer															*/
/*       the new buffer is NOT set - contains random memory chunk   */
bool	G2D_init_clrbase(g2dbase *pG2DB, uint16 width, uint16 height);
bool	G2D_init_pxlbase(g2dbase *pG2DB, uint16 width, uint16 height);
bool	G2D_init_mskbase(g2dbase *pG2DB, uint16 width, uint16 height);
bool	G2D_init_bitbase(g2dbase *pG2DB, uint16 width, uint16 height);

/* convert a ss_bitmap into a true color base and back using the		*/
/* given palette - the target struct has to be initialized 					*/
/* properly, _including_ the image pointer!													*/
/* (basically: bits -> clr / clr -> bits)														*/
/* DON'T FORGET THE TRANS BYTES IN THE PALETTE TABLE!								*/
/*																																	*/
/* note: since we use the ipal.dat conversion and no rgb						*/
/*			 approximation, pPal is not used in base_to_head						*/
void	G2Dr_head_to_base(const g2dbase *pG2DB, 
											  const struct tag_ss_bitmap_header *pHead,
											  pcolor *pPal);
void	G2Dc_head_to_base(const g2dbase *pG2DB, 
											  const struct tag_ss_bitmap_header *pHead,
											  pcolor *pPal, bool ignoreNull);
void	G2Dc_pal_to_clr(g2dbase *pDest, g2dbase *pSrc);
void	G2D_base_to_head(const struct tag_ss_bitmap_header *pHead, 
											 const g2dbase *pG2DB, pcolor *pPal);

/* convert a true color image to output bpp image	and back					*/
/* (only if we aren't in true color) - the target struct has to			*/
/* be initialized properly, _including_ the image pointer!					*/
/* (basically: pxl -> clr / clr -> pxl)															*/
/* DON'T FORGET THE TRANS BYTES IN THE PALETTE TABLE!								*/
/*																																	*/
/* note: since we use the ipal.dat conversion and no rgb						*/
/*			 approximation, pPal is not used in color_to_pixel					*/
#if BITS_PER_PIXEL < 32
void	G2D_pixel_to_color(const g2dbase *pTarget, const g2dbase *pSrc,
												 pcolor *pPal);
void	G2D_color_to_pixel(const g2dbase *pTarget, const g2dbase *pSrc,
												 pcolor *pPal);
#else /* c_merge uses memcpy; is faster															*/
#define G2D_pixel_to_color(t, s, p) G2Dc_merge_areas(t, s)
#define G2D_color_to_pixel(t, s, p) G2Dc_merge_areas(t, s)
#endif

/* set all trans bytes of the area																	*/
/* can be used to make a bridge between G2Dr_* and 'render merge'		*/
void	G2D_set_shades(const g2dbase *pG2DB, byte sh);
/* add a certain value to the trans bytes.													s*/
void	G2D_add_shades(const g2dbase *pG2DB, int nOffset);

/* put the trans mask into the according trans bytes of the area		*/
/* it is advised that the trans mask should be clipped - otherwise	*/
/* unexpected results may happen																		*/
void	G2D_set_mask(const g2dbase *pTarget, const g2dbase *pMask);

/* the Render, Copy and Mask functions:															*/
/* G2Dr_*: renders the area (-> colors are shaded)									*/
/* G2Dc_*: copy (into) the area (-> colors are NOT shaded, shades		*/
/*															 are preserved)											*/
/*																																	*/
/*	 NOTE THAT THOSE TWO FUNCTIONS USE TRUE COLOR PROCESSING				*/
/*	 (always the bImg.clr member is used AND EXPECTED)							*/
/*																																	*/
/*   so the typical process will be:																*/
/*   G2D_pixel_to_color(buf, screen, pal)														*/
/*   G2D[cr]_* (buf)																								*/
/*   ...																														*/
/*   G2D_color_to_pixel(screen, buf, pal)														*/
/*																																	*/
/* G2Dm_*: 'alters' the trans mask. bTrn marks an /offset/ for the	*/
/*		trans bytes. those operations are cumulative, so don't forget	*/
/*		to clip the mask afterwards																		*/
/*																																	*/
/* CAUTION:																													*/
/* G2Dr_* functions are basically used for areas that DON'T get			*/
/* 'render merge'd into another one afterwards (copy merge possible)*/
/* since the render (better: shade) operations result in a finished	*/
/* pcolor with the trans byte = 0x00; so render merge would ignore	*/
/* those pixels.																										*/
/* as a solution, use mask functions or more buffers								*/

/* clip a trans mask. sets every offset below 0 to 0 and above 255	*/
/* to 255																														*/
void	G2Dm_clip_mask(const g2dbase *pG2DB);

/* draw map.																												*/
#define	F_DRAWMAP_MODEL		0x01
#define	F_DRAWMAP_RADIUS	0x02
#define	F_DRAWMAP_ELEV		0x04
#define	F_DRAWMAP_SEC			0x08
#define F_DRAWMAP_CRIT		0x10
#define	F_DRAWMAP_RAD			0x20
#define	F_DRAWMAP_BIO			0x40
#define	F_DRAWMAP_NORMBG	0x80

void	 G2D_draw_map(const g2dbase *pG2DB, 
										double orientX, double orientY, 
										point center,
										byte zoom, byte radius, byte flags);

/* 'shift' and/or 'roll' an area																		*/
/* line 0 becomes line off.cy (% height)														*/
/* column 0 becomes column off.cx (% width)													*/
void	G2Dc_offset_area(const g2dbase *pG2DB, size off);

/* enclose an area																									*/
/* fx/fy: percentage about how much of the according side should be	*/
/*				enclosed starting from the edges													*/
/*				(100 makes a solid box, 1 will probably lead to one pixel */
/*				 per edge)																								*/
void	G2Dr_enclose_area(const g2dbase *pG2DB, byte fx, byte fy);
void	G2Dc_enclose_area(const g2dbase *pG2DB, byte fx, byte fy);
void	G2Dm_enclose_area(const g2dbase *pG2DB, byte fx, byte fy);

/* merge src into dest.																							*/
void	G2Dr_merge_areas(const g2dbase *pTarget, const g2dbase *pSrc);
void	G2Dc_merge_areas(const g2dbase *pTarget, const g2dbase *pSrc);
void	G2Dm_merge_areas(const g2dbase *pTarget, const g2dbase *pSrc);
void	G2Dc_merge_heads(const struct tag_ss_bitmap_header *pTarget,
											 const struct tag_ss_bitmap_header *pSrc);

/* stretch src into target																					*/
/* this code can overlook some [in] pixel when shrinking (currently)*/
/* but this is not so important since most graphics get stretched 	*/
/* instead of shrinked																							*/
void	G2Dr_stretch_area(const g2dbase *pTarget, const g2dbase *pSrc);
void	G2Dc_stretch_area(const g2dbase *pTarget, const g2dbase *pSrc);
void	G2Dm_stretch_area(const g2dbase *pTarget, const g2dbase *pSrc);
void	G2Dr_stretch_head(const g2dbase *pTarget, 
												const struct tag_ss_bitmap_header *pHead,
												const pcolor *pPal);

/* fill area with a color																						*/
void	G2Dr_fill_area(const g2dbase *pG2DB);
void	G2Dc_fill_area(const g2dbase *pG2DB);
void	G2Dm_fill_area(const g2dbase *pG2DB);

/* draw a box																												*/
void	G2Dr_draw_box(const g2dbase *pG2DB, bounding_box b);
void	G2Dc_draw_box(const g2dbase *pG2DB, bounding_box b);

/* draw a line																											*/
void	G2Dr_draw_line(const g2dbase *pG2DB, point p1, point p2);
void	G2Dc_draw_line(const g2dbase *pG2DB, point p1, point p2);

/* draw a circle																										*/
void G2Dr_draw_circle(const g2dbase *pG2DB, point center,
											uint16 rad, double start, double end);
void G2Dc_draw_circle(const g2dbase *pG2DB, point center,
											uint16 rad, double start, double end);
  
/* do_static: draws static																					*/
/* pG2DB:		the pointer describing the area													*/
/* center:	the center line (y)																			*/
/* height:	the entire height																				*/
/* dens_x:	percentage how filled a line is at max									*/
/* dens_y:	percentage how the last line is filled max					    */
/* shades:	how many additional shades (to the one given in pG2DB)  */
/*					the pixel should have at random.												*/
/*					(between 0x00 and the one of pG2DB)											*/
/*																																	*/
/* notes:																														*/
/* the center line is drawn at dens_x percentage; toward the border,*/
/* parts of dens_y are applied.																			*/
/*																																	*/
/* examples:																												*/
/* dens_x: 100, dens_y: 100 : a solid block													*/
/* dens_x: 100, dens_y:  50 : solid center and borders at 50%				*/
/* dens_x:  50, dens_y: 100 : a block filled 50%										*/
/* dens_x:   0, dens_y: ??? : nothing but speed loss ;)							*/

void	G2Dr_do_static(const g2dbase *pG2DB, uint16 center, uint16 height,
										 byte dens_x, byte dens_y, byte shades);
void	G2Dc_do_static(const g2dbase *pG2DB, uint16 center, uint16 height,
										 byte dens_x, byte dens_y, byte shades);
void	G2Dm_do_static(const g2dbase *pG2DB, uint16 center, uint16 height,
										 byte dens_x, byte dens_y, byte shades);

/* write_string: write a string																			*/
/* pStrings:	array of strings																			*/
/* nStrings:	amount in string array																*/
/* nFont:			the font that is to be used	(only mono fonts work)		*/
/* flags:			how to render strings; see flags desc below					  */
/* vSpace:		pixels that are between lines; negative values				*/
/*						possible (but think before to do ;)										*/
/* hSpace:    maximum horizontal width															*/
/*																																	*/
/* notes:																														*/
/* write_string creates a bitmap with indizes 0x00, 0x01 and 0x02		*/
/* (= void, border and character).																	*/
/* strings are parsed according to resource format:									*/
/* character 0x02 marks a soft hyphen that gets honoured if the word*/
/*   crosses the hSpace																							*/
/* character 0x0A marks a newline - the string array does not define*/
/*   this (exception: FORCE_NL)!																		*/

#define F_WSTR_DRAW_BORDER	0x01	/* force to prerender border			*/
#define	F_WSTR_RIGHT_ALIGN	0x02	/* align text right								*/
#define	F_WSTR_CENTER_ALIGN	0x04	/* align text centered						*/
#define	F_WSTR_FORCE_NL			0x08	/* force newlines	after each str	*/

struct tag_ss_bitmap_header *G2D_write_string(char * const *pStrings,
																				int nStrings, int nFont, 
																				byte flags,	int vSpace, 
																				int hSpace);

#endif /* sentry */


/*
 * End of src/graph2d.h
 */

/*
 * Emacs magic
 *
 * Local variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
