/*
 * hud.h -
 *  defines and prototypes for the hud
 */

/*
 * File:    src/hud.h
 * Purpose: TSSHP HUD System
 * Prefix:
 * Note:    should be easy to port to other projects.
 * Music:   Nostrum "Brainchild"
 */


#ifndef	HUDSYSTEM
#define	HUDSYSTEM

#define	C_HUD_MOUSE_HOVER			0x00
#define	C_HUD_MOUSE_SELECT		0x01
#define	C_HUD_MOUSE_ACTIVATE	0x02
#define	C_HUD_MOUSE_GRAB			0x03

#define	P_HUD_HOOK_MOUSE	byte action, bool bStart, uint16 x, uint16 y
#define	P_HUD_HOOK_RENDER	hudLayer **pLayer
#define	P_HUD_HOOK_DEST		void

/*                 */
/* usefull defines */
/*                 */

/* init a hudLayer struct and assign a pG2DB */
#define	D_HUD_NULLLAYER(pG2DB)		\
				{ NULL, { 0, 0 }, { 0, 0 }, (pG2DB) }

/* helper for the single linked list of hudLayer */
#define	appendHudLayer(pLayer, Layer)	\
				{															\
					(Layer).pNext = NULL;				\
					*pLayer = &(Layer);					\
					pLayer = &((Layer).pNext);	\
				}


/*													*/
/* message area definitions */
/*													*/

#define	C_MSG_MAXLEN		256
#define	C_MSG_MAXLINE		10
#define	C_MSG_STDTIME		2000						/* 2 seconds */

/* flags for edit_message */
#define	F_MSG_MSK_FLAG	0x01
#define	F_MSG_MSK_TEXT	0x02
#define	F_MSG_MSK_TIME	0x04
#define	F_MSG_MSK_FONT	0x08
#define	F_MSG_MSK_COLOR	0x10
/* flags for msg properties */
#define	F_MSG_CENTER		0x01
#define	F_MSG_RIGHT			0x02
#define	F_MSG_STATIC		0x04

#ifdef __cplusplus
extern "C"
{
#endif

typedef struct tag_hudLayer
{
	/* hudLayer are linked together, single linked */
	struct tag_hudLayer *pNext;
	
	/* position within the hudArea; 0:0 based	*/
	point								 ptPos;
	/* size of the layer. pG2DB will be zoomed into it! */
	size								 szSize;
	g2dbase							*pG2DB;
} hudLayer;

typedef struct tag_hudHooks
{
	bool								 (*hhMouseEvent)(P_HUD_HOOK_MOUSE);
	void								 (*hhRender)(P_HUD_HOOK_RENDER);
	bool								 (*hhDestroy)(P_HUD_HOOK_DEST);
} hudHooks;

typedef struct
{
/* identity																	*/
	char								 haName[31];
/* visual																		*/
	bool								 haVisible;
	bounding_box				 haArea;
/* entries																	*/
	hudHooks						 haHooks;
} hudArea;
			  
typedef struct tag_hudDragDrop
{
	struct ss_bitmap_header		*hddImage;

	uint16							 hddDataType;
	uint32							 hddDataSize;	
	void								*hddData;
} hudDragDrop;

void	HUD_init(uint16 width, uint16 height);
void	HUD_deinit(void);

void	HUD_reset_layer(hudLayer *pLayer, g2dbase *pG2DB);

bool	HUD_show_cursor(int nCursor);
int		HUD_get_cursor(void);
void	HUD_move_cursor(int relx, int rely);
void	HUD_set_cursor_pos(point ptCursor);
void	HUD_get_cursor_pos(point *pptCursor);
void	HUD_get_center_pos(point *pptCenter);
void	HUD_set_cursor_shade(byte bShade);
void	HUD_lock_cursor(bounding_box *pBox);
void	HUD_unlock_cursor(void);

bool	HUD_select_at_cursor(bool bStart);
bool	HUD_activate_at_cursor(bool bStart);
bool	HUD_grab_at_cursor(bool bStart);

void	HUD_set_description(char *strDesc, pcolor cFG, pcolor cBG,
													bounding_box *pArea);

bool	HUD_register_area(hudArea *pHA);
short HUD_find_area(hudArea *pHA);
bool	HUD_unregister_area(hudArea *pHA);

struct tag_camera; /* jim -don't need to put typedef here */
short HUD_render(struct tag_camera *pCam);

void	HUD_clear(void);


/* Messages																														*/
void	MSG_init(bounding_box boxArea);
void	MSG_deinit(void);
short MSG_add_message(byte bFlags, const char *pMsg, short nTime,
											int nFont, pcolor clrFg, pcolor clrBg);
bool	MSG_edit_message(short nId, byte bMask, byte bFlags, 
											 const char *pMsg, short nTime,
											 int nFont, pcolor clrFg, pcolor clrBg);
bool	MSG_del_message(short nId);

#ifdef __cplusplus
}
#endif


#endif  /*HUDSYSTEM*/
