/*
 * item.c -
 *  Items class 8
 */

/*
 * File:    src/item.c
 * Purpose: Class-specific routines for items, class 8
 * Prefix:  Item_
 * Note:    
 * Music:   
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */


/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <stdio.h>

#include "item.h"
#include "object.h"


#if __DEBUG__
#define ITEM_MESSAGES 0
#endif


#define ITEM_NUM_CYBER 12


/*
 * Colour scheme for the cyber items class 8/5
 */
byte Cyber_item_colour [12] [6];


/*
 * Local function prototypes
 */




/*
 * File offset for class 8 in objprop.dat
 */
#define OBJPROP_OFFSET_ITEM 0x05ab


/*
 * Load up item data from objprop.dat / objects.dat
 * We are passed the file with its pointer at the beginning of the item
 *  properties, and must leave it at the end for the next class (switches)
 */
void Item_load_objprop (FILE *propfile)
{

  int i; /* misc. loop index */

  /*
   * Trust No One ...
   */
  assert (ftell (propfile) == OBJPROP_OFFSET_ITEM);

  /*
   * First comes the generic item properties, 2 bytes per item, but with
   *  nothing in them
   */
  fseek (propfile, 160, SEEK_CUR);

  /*
   * In fact, the only table with any information in it is that of the cyber
   *  subclass, which contains auxiliary palettes
   */
  fseek (propfile,  8, SEEK_CUR); /* 0 junk         */
  fseek (propfile, 10, SEEK_CUR); /* 1 debris       */
  fseek (propfile, 15, SEEK_CUR); /* 2 corpses      */
  fseek (propfile,  6, SEEK_CUR); /* 3 items        */
  fseek (propfile, 12, SEEK_CUR); /* 4 access cards */

  for (i = 0; i < 12; ++i)
  {
    fread (Cyber_item_colour [i], 1, 6, propfile);
#   if ITEM_MESSAGES
    printf ("colour %02x %02x %02x %02x %02x %02x\n",
	    Cyber_item_colour [i] [0], Cyber_item_colour [i] [1],
	    Cyber_item_colour [i] [2], Cyber_item_colour [i] [3],
	    Cyber_item_colour [i] [4], Cyber_item_colour [i] [5]);
#   endif
  }

  fseek (propfile,  9, SEEK_CUR); /* 6 stains       */
  fseek (propfile, 16, SEEK_CUR); /* 7 quest items (2 bytes each) */

}



byte *Item_get_colour (map_object *obj)
{

  assert (obj->class == OBJCLASS_ITEM);

  /*
   * Only cyberspace objects have colour schemes (as colour schemes aren't
   *  used except for c/space)
   */
  if (obj->subclass != ITEM_SUBCLASS_CYBER) return NULL;

  assert (obj->type < ITEM_NUM_CYBER);
  return Cyber_item_colour [obj->type];      

}
