/*
 * map.c -
 *  Stuff to do with the level map
 */

/*
 * File:    src/map.c
 * Purpose: Map loading and rendering
 * Prefix:  Map_
 * Note:    This file is too big and complex!
 * Music:   Def Leppard, "Pyromania"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



/*
 * #includes
 */
#include "defines.h"

#include <float.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>		/* memset() */

#include "gamemem.h"
#include "bitmap.h"
#include "bsp.h"
#include "clip.h"
#include "game.h"
#include "graph2d.h"
#include "main.h"
#include "map.h"
#include "object.h"
#include "option.h"
#include "palette.h"
#include "quadtree.h"
#include "render.h"
#include "resource.h"
#include "shade.h"
#include "surface.h"
#include "texture.h"


/*
 * Debugging
 */
#define MAP_MESSAGES 0
surface *Debug_check_surface = NULL;




/*
 * Got to get this sorted somewhere
 */
#ifndef M_PI
# define M_PI		3.14159265358979323846	/* pi */
#endif



#define NORTH     0
#define NORTHEAST 1
#define EAST      2
#define SOUTHEAST 3
#define SOUTH     4
#define SOUTHWEST 5
#define WEST      6
#define NORTHWEST 7


/*
 * Global variables
 */

/*
 * Map info chunk
 */
map_info Map_info;

/*
 * Master tile map
 */
map_tile *Tile;

#define MAP_VERTICES 8192
#define MAP_SURFACES 32768
int Num_map_vertices;
map_vertex  *Vertex;
int Num_map_surfaces;
map_surface *Surface;

       int Height_scale;
       int Level_height;



/*
 * Start points for the levels. These aren't really from the game
 */
static point Start_point [16] =
{ { 0x1E80, 0x1280 },
  { 0x1E80, 0x1680 },
  { 0x0F80, 0x2880 },
  { 0x2580, 0x1980 },
  { 0x1B80, 0x1B80 },
  { 0x2080, 0x1C80 },
  { 0x2A80, 0x2A80 },
  { 0x1A80, 0x2280 },
  { 0x1880, 0x2480 },
  { 0x2380, 0x2A80 },
  { 0x1F80, 0x1F80 },
  { 0x2080, 0x1680 },
  { 0x1880, 0x2280 },
  { 0x2080, 0x2880 },
  { 0x0880, 0x2F80 },
  { 0x1F80, 0x1F80 } };

static point Start_point_uw [9] =
{ { 0x2000, 0x0280 },
  { 0x0880, 0x2680 },
  { 0x2a80, 0x3c80 },
  { 0x2980, 0x0280 },
  { 0x1d80, 0x2f80 },
  { 0x2b80, 0x1c80 },
  { 0x1380, 0x0c80 },
  { 0x3c80, 0x0280 },
  { 0x1F80, 0x1F80 } };

static point Start_point_uw2 [80] =
{ { 0x1380, 0x3080 },                /* Britannia 1 (castle)   */
  { 0x0B80, 0x2A80 },                /* Britannia 2 (cellars)  */
  { 0x0E80, 0x3580 },                /* Britannia 3 (sewers)   */
  { 0x1A80, 0x1280 },                /* Britannia 4            */
  { 0x2C80, 0x2180 },                /* Britannia 5 (gem)      */
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x2180, 0x2080 },                /* Prisons 1 (cellar)     */
  { 0x1C80, 0x1F80 },                /* Prisons 2              */
  { 0x2480, 0x2380 },                /* Prisons 3              */
  { 0x1B80, 0x2080 },                /* Prisons 4              */
  { 0x1C80, 0x1F80 },                /* Prisons 5              */
  { 0x2180, 0x2480 },                /* Prisons 6              */
  { 0x1C80, 0x1F80 },                /* Prisons 7              */
  { 0x2580, 0x2080 },                /* Prisons 8              */
  { 0x1B80, 0x2280 },                /* Killorn Keep 1         */
  { 0x1280, 0x2080 },                /* Killorn Keep 2         */
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x1280, 0x2880 },                /* Ice Caverns 1          */
  { 0x0680, 0x2080 },                /* Ice Caverns 2          */
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x1F80, 0x1F80 },                /* Talorus 1              */
  { 0x0F80, 0x2080 },                /* Talorus 2              */
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0480, 0x2780 },                /* Academy 1              */
  { 0x0180, 0x1E80 },                /* Academy 2              */
  { 0x1B80, 0x1D80 },                /* Academy 3              */
  { 0x0180, 0x1E80 },                /* Academy 4              */
  { 0x1480, 0x2980 },                /* Academy 5              */
  { 0x1880, 0x2680 },                /* Academy 6              */
  { 0x1680, 0x1D80 },                /* Academy 7              */
  { 0x1080, 0x2780 },                /* Academy 8              */
  { 0x2080, 0x2080 },                /* Loth's Tomb 1          */
  { 0x2080, 0x3580 },                /* Loth's Tomb 2          */
  { 0x2B80, 0x1B80 },                /* Loth's Tomb 3          */
  { 0x1E80, 0x3280 },                /* Loth's Tomb 4          */
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x3B80, 0x1380 },                /* Pits of Carnage 1      */
  { 0x1580, 0x0480 },                /* Pits of Carnage 2      */
  { 0x2E80, 0x3880 },                /* Pits of Carnage 3      */
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0A80, 0x3A80 },                /* Ethereal Void 1        */
  { 0x2580, 0x1380 },                /* Ethereal Void 2        */
  { 0x2B80, 0x0B80 },                /* Ethereal Void 3        */
  { 0x0280, 0x1580 },                /* Ethereal Void (yellow) */
  { 0x2080, 0x0680 },                /* Ethereal Void (main)   */
  { 0x0000, 0x0000 },
  { 0x1980, 0x3180 },                /* Ethereal Void          */
  { 0x1380, 0x3080 },
  { 0x2080, 0x3580 },                /* Loth's Tomb?           */
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 },
  { 0x0000, 0x0000 } };




/*
 * Local function prototypes
 */
void try_recurse_wall (map_tile *tile, int n, const bounding_box *box, 
		       surface *s);
int find_overlap (point pt [4]);

int add_tile_vertex (map_tile *t, const map_vertex *v);
int add_tile_surface (map_tile *t, const map_surface *ms);
map_tile *in_direction (int t, int d);
map_surface *find_back_surface (const map_tile *t, const map_surface *s);
void check_aperture (map_surface *ms1, map_surface *ms2);
void add_floor (map_tile *t, int fake_type, bool ceiling);

extern Vector Avatar_pos;



void Map_init (void)
{

  Tile = Alloc_init (64*64 * sizeof *Tile, "Map_init", "tile map");
  Vertex = Alloc_init (MAP_VERTICES * sizeof *Vertex, "Map_init",
		       "vertex map");
  Surface = Alloc_init (MAP_SURFACES * sizeof (map_surface), "Map_init",
			"surface map");

  BSP_init ();
}

void Map_deinit(void)
{
  free(Surface);
  Surface = NULL;
  free(Vertex);
  Vertex = NULL;
  free(Tile);
  Tile = NULL;

  BSP_exit ();
}

/*
 * Northwest going clockwise, as usual
 */
static const short_vec Tile_corner [TILE_NUM_TYPES] [4] =
{ 
  { {0, 1, 0}, {1, 1, 0}, {1, 0, 0}, {0, 0, 0} },  /* TILE_SOLID       */
  { {0, 1, 0}, {1, 1, 0}, {1, 0, 0}, {0, 0, 0} },  /* TILE_OPEN        */
  { {1, 1, 0}, {1, 0, 0}, {0, 0, 0}, {1, 1, 0} },  /* TILE_DIAGONAL_SE */
  { {1, 0, 0}, {0, 0, 0}, {0, 1, 0}, {1, 0, 0} },  /* TILE_DIAGONAL_SW */
  { {0, 0, 0}, {0, 1, 0}, {1, 1, 0}, {0, 0, 0} },  /* TILE_DIAGONAL_NW */
  { {0, 1, 0}, {1, 1, 0}, {1, 0, 0}, {0, 1, 0} },  /* TILE_DIAGONAL_NE */
  { {0, 1, 1}, {1, 1, 1}, {1, 0, 0}, {0, 0, 0} },  /* TILE_SLOPE_N     */
  { {0, 1, 0}, {1, 1, 1}, {1, 0, 1}, {0, 0, 0} },  /* TILE_SLOPE_E     */
  { {0, 1, 0}, {1, 1, 0}, {1, 0, 1}, {0, 0, 1} },  /* TILE_SLOPE_S     */
  { {0, 1, 1}, {1, 1, 0}, {1, 0, 0}, {0, 0, 1} },  /* TILE_SLOPE_W     */
  { {0, 1, 1}, {1, 1, 1}, {1, 0, 0}, {0, 0, 1} },  /* TILE_VALLEY_NW   */
  { {0, 1, 1}, {1, 1, 1}, {1, 0, 1}, {0, 0, 0} },  /* TILE_VALLEY_NE   */
  { {0, 1, 0}, {1, 1, 1}, {1, 0, 1}, {0, 0, 1} },  /* TILE_VALLEY_SE   */
  { {0, 1, 1}, {1, 1, 0}, {1, 0, 1}, {0, 0, 1} },  /* TILE_VALLEY_SW   */
  { {0, 1, 0}, {1, 1, 0}, {1, 0, 1}, {0, 0, 0} },  /* TILE_RIDGE_SE    */
  { {0, 1, 0}, {1, 1, 0}, {1, 0, 0}, {0, 0, 1} },  /* TILE_RIDGE_SW    */
  { {0, 1, 1}, {1, 1, 0}, {1, 0, 0}, {0, 0, 0} },  /* TILE_RIDGE_NW    */
  { {0, 1, 0}, {1, 1, 1}, {1, 0, 0}, {0, 0, 0} },  /* TILE_RIDGE_NE    */
};


/*
 * Wall list for normal and diagonal tiles
 */
static const int Walls_present [5] [4] =
{
  { NORTH, EAST,  SOUTH, WEST },  /* square */
  { EAST, SOUTH, NORTHWEST, 0 },  /* TILE_DIAGONAL_SE */
  { NORTHEAST, SOUTH, WEST, 0 },  /* TILE_DIAGONAL_SW */
  { NORTH, SOUTHEAST, WEST, 0 },  /* TILE_DIAGONAL_NW */
  { NORTH, EAST, SOUTHWEST, 0 }   /* TILE_DIAGONAL_NE */
};



int uw_type [16] =
{
  TILE_SOLID,
  TILE_OPEN,
  TILE_DIAGONAL_SE,
  TILE_DIAGONAL_SW,
  TILE_DIAGONAL_NE,
  TILE_DIAGONAL_NW,
  TILE_SLOPE_N,
  TILE_SLOPE_S,
  TILE_SLOPE_E,
  TILE_SLOPE_W,
  TILE_SOLID,
  TILE_SOLID,
  TILE_SOLID,
  TILE_SOLID,
  TILE_SOLID,
  TILE_SOLID
};


/*
 * A probably-should-be-more-general macro
 */
#define U32(p,n) ((p)[(n)]+((p)[(n)+1]<<8)+((p)[(n)+2]<<16)+((p)[(n)+3]<<24))

/*
 * New geometry:
 * Initialise each tile: flags, type etc.
 * For each tile:
 *  If it is solid, skip it. The idea of rendering SOLID tiles as transparent
 *   didn't pan out. Sorry.
 *  OTHERWISE
 *  Add the 6 or 8 basic vertices
 *  Add the 3 or 4 basic walls: solid at this stage but with "behind" pointers
 *   if appropriate
 *  For each wall with a facing wall (i.e. whose "behind" tile has been loaded)
 *   check the aperture and amend, as discussed elsewhere
 *  Add the floor and ceiling, 1 or 2 surfaces each
 */
void Map_init_level (int level)
{

  size_t size;
  block_handle bh;
  unsigned char *tp;
  int i;
  int n;
  int next;
  const short_vec *corner;

  static int tile_top [4];
  static int tile_bot [4];

  map_surface ts;      /* Temp. surface */
  map_surface *bs;     /* Behind surface */
  map_vertex tv;       /* Temp vertex */
  /*  int org_vertex [8];  Original tile vertices (duplicates allowed) */
  int nv;              /* No. starting tile vertices */
  double m;
  map_tile *tex_tile;

  /*
   * Underworldly things
   */
  unsigned char uwraw [0x4000];
  unsigned long offs;
  unsigned char *uw2raw = NULL;
  size_t uw2size;

  if (Option_underworld)
  {
    FILE *levfile = fopen (UW_MAP_FILE, "rb");
    if (levfile == 0)
    {
      Cant_happen ("Map_init_level(): failed to open UW level file");
    }

    if (Underworld2)
    {
      uw2raw = Res_load_uw2_map (levfile, level, 0, &uw2size);
      tp = uw2raw;
    }
    else
    {
      if (fgetc (levfile) != 0x87)
      {
	Cant_happen ("Map_init_level(): not Underworld mapfile");
      }
      fseek (levfile, 0x02 + 4 * level, SEEK_SET);
      READ_LE_LONG (levfile, offs);
      fseek (levfile, offs, SEEK_SET);
      fread (uwraw, 1, 0x4000, levfile);
      tp = uwraw;
    }
    fclose (levfile);

    /*
     * Fill in the map info struct
     */
    Map_info.xsize        = 64;
    Map_info.ysize        = 64;
    Map_info.log_xsize    = 6;
    Map_info.log_ysize    = 6;
    Map_info.height_shift = 3;
    Map_info.cyberspace   = false;
  }
  else
  {

    /*
     * Get the map info chunk
     */
    bh = Res_load_chunk_by_id (4000 + 100*level + MAP_CHUNK_INFO, &size);
    if ((bh == NULL) || (*bh == NULL))
    {
      Cant_happen ("Map_init_level(): failed to load map info chunk");
    }
    if (size != 0x3A)
    {
      Cant_happen ("Map_init_level(): map info size mismatch");
    }
    tp = *bh;

    Map_info.xsize        = U32 (tp,  0);
    Map_info.ysize        = U32 (tp,  4);
    Map_info.log_xsize    = U32 (tp,  8);
    Map_info.log_ysize    = U32 (tp, 12);
    Map_info.height_shift = U32 (tp, 16);
    Map_info.cyberspace   = U32 (tp, 24);

    /*
     * Get the tile map chunk
     */
    bh = Res_load_chunk_by_id (4000 + 100*level + MAP_CHUNK_TILE, &size);
    if ((bh == NULL) || (*bh == NULL))
    {
      Cant_happen ("Map_init_level(): failed to load tile map");
    }
    if (size != 0x10000)
    {
      Cant_happen ("Map_init_level(): tile map size mismatch");
    }
    tp = *bh;
  }

  /*
   * Check that the map info is well-formed
   */
  if (Map_info.xsize != Map_info.ysize)
  {
    Cant_happen ("Map_init_level(): map is not square");
  }
  if ((Map_info.xsize != (1u<<Map_info.log_xsize)) ||
      (Map_info.ysize != (1u<<Map_info.log_ysize)))
  {
    Cant_happen ("Map_init_level(): map size mismatch");
  }

  /*
   * Height scale is stored as log(2) (no. height units per texture height)
   */
  Height_scale = 256 >> Map_info.height_shift;
  Level_height = 32 * Height_scale;

  for (i = 0; i < 64*64; ++i)
  {

    Tile [i].rflags = 0;

    if (Option_underworld)
    {
      offs = *tp + (tp [1] << 8) + (tp [2] << 16) + (tp [3] << 24); tp += 4;
      Tile [i].type    = uw_type [offs & 0x0F];
      Tile [i].floor   = (short)(((offs & 0xF0) >> 3) * Height_scale);
      Tile [i].ceiling = Level_height;
      Tile [i].slope   = 2 * Height_scale;
      Tile [i].index   = (offs  & 0xFFC00000) >> 22;
      Tile [i].texture = (offs & 0x003F0000) >> 16;
      offs = ((offs & 0x00003C00) >> 10);
      if (!Underworld2)
      {
	offs += 48;
      }
      Tile [i].texture += offs << 6;
      Tile [i].flags   = 0x0f0f0000 | TF_SLOPE_FLOOR;
    }
    else
    {
      Tile [i].type    = *tp++;

      /*
       * Aha! They aren't signed after all. Looks like the top bits are to do
       *  with radiation and biohazard
       * Bits 5 and 6 look to be orientation
       */
			/* ff1: and i wondered why they weren't to be found - you ANDed */
			/*			with 0x60 so exactly those bits were lost ;) furthermore*/
			/*			i realigned them, so they both are now in the first byte*/
			/*			if this is ok then i'll apply the automap to it					*/
      Tile [i].rflags |= (*tp & 0xE0); /* << 1; */
      Tile [i].floor   = (*tp++ & 0x1f) * Height_scale;
      Tile [i].rflags |= (*tp & 0xE0) >> 4; /* >> 1; */
      Tile [i].ceiling = Level_height - (*tp++ & 0x1f) * Height_scale;
      Tile [i].slope   = *tp++ * Height_scale;
      Tile [i].index   = tp [0] + (tp [1] << 8);   tp += 2;
      Tile [i].texture = tp [0] + (tp [1] << 8);   tp += 2;
      Tile [i].flags   = tp [0] + (tp [1] << 8)
	+ (tp [2] << 16) + (tp [3] << 24);         tp += 4;
      Tile [i].state [0] = *tp++;
      Tile [i].state [1] = *tp++;
      Tile [i].state [2] = *tp++;
      Tile [i].state [3] = *tp++;
    }

    Tile [i].x = i % 64;
    Tile [i].y = i / 64;
    Tile [i].frame  = 0;
    Tile [i].num_vertices = 0;
    Tile [i].first_surface = -1;

    Tile [i].dark_index = -1;

  }
# if MAP_MESSAGES
  printf ("Map_init_level(): tile map loaded\n");
# endif

  /*
   * Add the tile's vertices to the map
   */
  for (i = 0; i < 64*64; ++i)
  {

    map_tile *tile = &(Tile [i]);
    if (Tile [i].type == TILE_SOLID)
    {
      /*
       * Solid tiles don't add any vertices, but get 4 "ignore" walls so that
       *  something should be drawn if the viewpoint gets into one
       */
/*        ts.flags        = MSF_PORTAL | SF_NOCLIP; */
/*        ts.num_vertices = 0; */
/*        ts.behind = in_direction (i, NORTH); add_tile_surface (tile, &ts); */
/*        ts.behind = in_direction (i, EAST ); add_tile_surface (tile, &ts); */
/*        ts.behind = in_direction (i, SOUTH); add_tile_surface (tile, &ts); */
/*        ts.behind = in_direction (i, WEST ); add_tile_surface (tile, &ts); */
      continue;
    }

    corner = &(Tile_corner [tile->type] [0]);

    for (n = 0; n < 4; ++n)
    {
      tile_top [n] = tile->ceiling;
      tile_bot [n] = tile->floor;

      /*
       * Slopes. The falls-through here are deliberate
       */
      if (tile->type >= TILE_SLOPE_BASE)
      {
	switch (tile->flags & TF_WHICH_SLOPE)
	{
	case TF_SLOPE_OPPOSITE:
	  tile_top [n]    -= corner [n].z * tile->slope;
	case TF_SLOPE_FLOOR:
	  tile_bot [n]    += corner [n].z * tile->slope;
	  break;
	case TF_SLOPE_BOTH:
	  tile_bot [n]    += corner [n].z * tile->slope;
	case TF_SLOPE_CEILING:
	  tile_top [n] -= tile->slope;
	  tile_top [n]    += corner [n].z * tile->slope;
	  break;
	}
      }
      if (tile->type != n + TILE_DIAGONAL_BASE)
      {
	int grid_x = (i % 64) + Tile_corner [0] [n].x;
	int grid_y = (i / 64) + Tile_corner [0] [n].y;
	tv.x = grid_x * TILE_GRID_SIZE;
	tv.y = grid_y * TILE_GRID_SIZE;
	tv.z = tile_top [n];
	if (Map_info.cyberspace)
	{
	  tv.u.rgb = Pal_index_to_rgb
	    (Tile [grid_x + 64 * grid_y].texture >> 8);
	}
	add_tile_vertex (tile, &tv);
	tv.z = tile_bot [n];
	if (Map_info.cyberspace)
	{
	  tv.u.rgb = Pal_index_to_rgb
	    (Tile [grid_x + 64 * grid_y].texture & 0xff);
	}
	add_tile_vertex (tile, &tv);
      }
    }

  }

  /*
   * Tile post-processing
   */
  for (i = 0; i < 64*64; ++i)
  {
    map_tile *tile = &(Tile [i]);
    int x, y;

    if (tile->type == TILE_SOLID)
    {
      continue;
    }

    /*
     * Add the basic walls
     */
    nv = Tile [i].num_vertices;
    for (n = 0; n < nv; n += 2)
    {
      int direction = Walls_present [(Tile [i].type >= TILE_SLOPE_BASE) ? 0 :
				    (Tile [i].type - 1)] [n / 2];
      next = (n + 2) % nv;
      ts.behind = in_direction (i, direction);

      /*
       * Set up the texture mapping for this face. The texture is given by the
       *  bottom 6 bits of the texture word for this tile or the one behind
       *  the face, depending on the TF_OWN_TEXTURE flag
       */
      tex_tile = ((Tile [i].flags & TF_OWN_TEXTURE) && (ts.behind != NULL)) ?
	ts.behind : &(Tile [i]);
      ts.texture = tex_tile->texture & 0x3F;

      /*
       * The texture anchor point is the top-left vertex offset by the texture
       *  offset (no, really!)
       * Level_height is added to force the initial texture offset in the v
       *  direction to be positive
       */
      ts.anchor.x = Vertex [Tile [i].vertex [n]].x;
      ts.anchor.y = Vertex [Tile [i].vertex [n]].y;
      ts.anchor.z = Level_height + 
  	Height_scale * (Tile [i].flags & TF_TEXTURE_OFFSET);

      /*
       * The texture `v' vector always points straight down
       */
      ts.v.x = 0.;  ts.v.y = 0.;  ts.v.z = -1.;

      /*
       * The `u' vector points horizontally along the wall
       */
      vec_subtract (ts.u, Vertex [tile->vertex [next]],
		    Vertex [tile->vertex [n]]);
      ts.u.z = 0;
      m = vec_magnitude2 (ts.u) / TILE_GRID_SIZE;
      vec_divide (ts.u, m);

      vec_cross (ts.normal, ts.v, ts.u);
      vec_normalise (ts.normal);

      ts.num_vertices = 4;
      ts.vertex [0] = Tile [i].vertex [n];
      ts.vertex [1] = Tile [i].vertex [next];
      ts.vertex [2] = Tile [i].vertex [next + 1];
      ts.vertex [3] = Tile [i].vertex [n + 1];
      ts.flags = SF_TEXTURED;                 /* default to texture-mapped */
      if (add_tile_surface (&(Tile [i]), &ts) >= 0)
      {
	bs = find_back_surface (&(Tile [i]),
				&(Surface [Tile [i].first_surface]));
	if (bs != NULL)
	{
	  check_aperture (&(Surface [Tile [i].first_surface]), bs);
	}
      }
    }

    /*
     * Now to add the floor
     */
    if ((tile->type < TILE_DSLOPE_BASE) ||
	((tile->flags & TF_WHICH_SLOPE) == TF_SLOPE_CEILING))
    {
      /*
       * Floor doesn't have a diagonal slope ... add a single flat surface
       */
      add_floor (tile, tile->type, false);
    }
    else
    {
      /*
       * Diagonal sloping floor ... we need to add 2 triangular surfaces
       */
      int dir = tile->type - TILE_DSLOPE_BASE;
      add_floor (tile, TILE_DIAGONAL_BASE + ((dir + 1) % 4), false);
      add_floor (tile, TILE_DIAGONAL_BASE + ((dir + 3) % 4), false);
    }

    /*
     * And similarly the ceiling
     */
    if ((tile->type < TILE_DSLOPE_BASE) ||
	((tile->flags & TF_WHICH_SLOPE) == TF_SLOPE_FLOOR))
    {
      /*
       * Ceiling doesn't have a diagonal slope ... add a single flat surface
       */
      add_floor (tile, tile->type, true);
    }
    else
    {
      /*
       * Diagonal sloping ceiling ... we need to add 2 triangular surfaces
       */
      int dir = tile->type - TILE_DSLOPE_BASE;
      add_floor (tile, TILE_DIAGONAL_BASE + ((dir + 1) % 4), true);
      add_floor (tile, TILE_DIAGONAL_BASE + ((dir + 3) % 4), true);
    }

    /*
     * A little bit of Underworld cuteness ... brighten lava
     */
    if (Map_terrain_type (tile) == TERRAIN_LAVA)
    {
      for (x = tile->x; x < tile->x + 2 && x < 64; ++x)
      {
	for (y = tile->y; y < tile->y + 2 && y < 64; ++y)
	{
  	  Tile [x + 64*y].flags &= 0xfff0ffff;
	  Tile [x + 64*y].flags |= 0x00040000;
	}
      }
    }

  }

  /*
   * Initialise anything else dependent on the level
   */
  if (!Map_info.cyberspace)
  {
    Shd_build_static_darkmap (Vertex, Num_map_vertices);
  }

# if QUADTREES
  Quad_init_level ();
# else
  BSP_init_level ();
# endif

# if MAP_MESSAGES
  printf ("Map_init_level(): tile tree initialised\n");
# endif

  printf ("Map_init_level(): %d map surfaces\n", Num_map_surfaces);
  printf ("Map_init_level(): %d map vertices\n", Num_map_vertices);

  /*
   * Set up the camera
   */
  if (Option_start_x >= 0)    Avatar_pos.x = Option_start_x;
  else if (Underworld2)       Avatar_pos.x = Start_point_uw2 [level].x;
  else if (Option_underworld) Avatar_pos.x = Start_point_uw [level].x;
  else                        Avatar_pos.x = Start_point [level].x;
  if (Option_start_y >= 0)    Avatar_pos.y = Option_start_y;
  else if (Underworld2)       Avatar_pos.y = Start_point_uw2 [level].y;
  else if (Option_underworld) Avatar_pos.y = Start_point_uw [level].y;
  else                        Avatar_pos.y = Start_point [level].y;

  if (Option_print_map)
  {

    /*
     * Flags testing
     */
    tp = *bh;
    for (i = 63; i >= 0; --i)
    {
      for (n = 0; n < 64; ++n)
      {
#       if 1
  	if (Tile [i * 64 + n].type == 0)
/*  	if (0) */
	{
	  /*  	printf ("%d %d\n", n, i); */
	  printf ("#");
	}
	else if (Tile [i * 64 + n].state [3])
/*  	else if (tp [16 * (i * 64 + n) + 1] & 0x80) */
	{
  	  printf ("x");
	}
	else
	{
	  printf (" ");
	}
#       else
	if (!(Tile [i * 64 + n].flags & 0x0f000000))
	{
	  printf (" ");
	}
	else
	{
	  printf ("%01lX", (Tile [i * 64 + n].flags & 0x0f000000) >> 24);
	}
#       endif
      }
      printf ("\n");
    }
  }

}



/*
 * Tile vertices used by each wall (clockwise from north, then floor, ceiling)
 */
int Wall_vertex [6] [4] =
{
  { 4, 5, 1, 0 },
  { 5, 6, 2, 1 },
  { 6, 7, 3, 2 },
  { 7, 4, 0, 3 },
  { 0, 1, 2, 3 },
  { 6, 5, 4, 7 }
};



map_tile *Map_get_tile(uint16 x, uint16 y)
{
  uint16	tile_x = x / TILE_GRID_SIZE;
  uint16	tile_y = y / TILE_GRID_SIZE;
	map_tile *pTile = NULL;
	
  if (tile_x < 64 && tile_y < 64)
		pTile = &(Tile[(tile_y << 6) | tile_x]);    

	return pTile;
}


/*
 * jim - this should probably replace the uint16 version above
 */
map_tile *Map_get_tile_v (const Vector *v)
{
  int tile_x = (int) v->x / TILE_GRID_SIZE;
  int tile_y = (int) v->y / TILE_GRID_SIZE;
  if ((tile_x < 0) || (tile_x > 63) || (tile_y < 0) || (tile_y > 63))
  {
    Cant_happen ("Map_get_tile_v(): point off map");
  }
  return &(Tile [tile_x + 64 * tile_y]);
}



int Map_floor_height (int x, int y)
{

  int tile_x = x / TILE_GRID_SIZE;
  int tile_y = y / TILE_GRID_SIZE;

  if ((tile_x < 0) || (tile_y < 0) || (tile_x > 63) || (tile_y > 63))
  {
    Cant_happen ("Map_floor_height(): point off map");
  }

  return Map_tile_height (&(Tile [tile_x + 64 * tile_y]),
			  x % TILE_GRID_SIZE, y % TILE_GRID_SIZE, false);

}






int Map_ceiling_height (int x, int y)
{

  int tile_x = x / TILE_GRID_SIZE;
  int tile_y = y / TILE_GRID_SIZE;

  if ((tile_x < 0) || (tile_y < 0) || (tile_x > 63) || (tile_y > 63))
  {
    Cant_happen ("Map_ceiling_height(): point off map");
  }

  return Map_tile_height (&(Tile [tile_x + 64 * tile_y]),
			  x % TILE_GRID_SIZE, y % TILE_GRID_SIZE, true);

}






int Map_tile_height (map_tile *tile, int x, int y, bool ceiling)
{

  int height = ceiling ? tile->ceiling : tile->floor;
  int slope  = tile->slope;

  if (ceiling)
  {
    if ((tile->flags & TF_WHICH_SLOPE) == TF_SLOPE_FLOOR)
    {
      return height;
    }
    else if ((tile->flags & TF_WHICH_SLOPE) == TF_SLOPE_OPPOSITE)
    {
      slope = -slope;
    }
    else if (tile->type >= TILE_SLOPE_BASE)
    {
      height -= slope;
    }
  }
  else if ((tile->flags & TF_WHICH_SLOPE) == TF_SLOPE_CEILING)
  {
    return height;
  }

  switch (tile->type)
  {

  case TILE_SOLID:
  case TILE_OPEN:
  case TILE_DIAGONAL_SE:
  case TILE_DIAGONAL_SW:
  case TILE_DIAGONAL_NW:
  case TILE_DIAGONAL_NE:
    /*
     * Flat tile
     */
    break;

  case TILE_SLOPE_N:
    height += (y * slope) / TILE_GRID_SIZE;
    break;

  case TILE_SLOPE_E:
    height += (x * slope) / TILE_GRID_SIZE;
    break;

  case TILE_SLOPE_S:
    height += ((TILE_GRID_SIZE - y) * slope) / TILE_GRID_SIZE;
    break;

  case TILE_SLOPE_W:
    height += ((TILE_GRID_SIZE - x) * slope) / TILE_GRID_SIZE;
    break;

  case TILE_VALLEY_NW:
    height += ((TILE_GRID_SIZE - x) > y) ?
      ((TILE_GRID_SIZE - x) * slope) / TILE_GRID_SIZE :
	(y * slope) / TILE_GRID_SIZE;
    break;

  case TILE_VALLEY_NE:
    height += (y > x) ?
      (y * slope) / TILE_GRID_SIZE : (x * slope) / TILE_GRID_SIZE;
    break;

  case TILE_VALLEY_SE:
    height += (x > (TILE_GRID_SIZE - y)) ?
      (x * slope) / TILE_GRID_SIZE :
	((TILE_GRID_SIZE - y) * slope) / TILE_GRID_SIZE;
    break;

  case TILE_VALLEY_SW:
    height += ((TILE_GRID_SIZE - y) > (TILE_GRID_SIZE - x)) ?
      ((TILE_GRID_SIZE - y) * slope) / TILE_GRID_SIZE :
	((TILE_GRID_SIZE - x) * slope) / TILE_GRID_SIZE;
    break;

  case TILE_RIDGE_SE:
    height += (x < (TILE_GRID_SIZE - y)) ?
      (x * slope) / TILE_GRID_SIZE :
	((TILE_GRID_SIZE - y) * slope) / TILE_GRID_SIZE;
    break;

  case TILE_RIDGE_SW:
    height += ((TILE_GRID_SIZE - y) < (TILE_GRID_SIZE - x)) ?
      ((TILE_GRID_SIZE - y) * slope) / TILE_GRID_SIZE :
	((TILE_GRID_SIZE - x) * slope) / TILE_GRID_SIZE;
    break;

  case TILE_RIDGE_NW:
    height += ((TILE_GRID_SIZE - x) < y) ?
      ((TILE_GRID_SIZE - x) * slope) / TILE_GRID_SIZE :
	(y * slope) / TILE_GRID_SIZE;
    break;

  case TILE_RIDGE_NE:
    height += (y < x) ?
      (y * slope) / TILE_GRID_SIZE : (x * slope) / TILE_GRID_SIZE;
    break;

  default:
    /*
     * Paranoia
     */
    Cant_happen ("Map_tile_height(): bogus tile type");

  }

  return height;

}




/*
 * Get the highest and lowest tile vertex z coords in top and bottom
 *  respectively
 * If the `accumulate' flag is set, top and bottom are assumed to contain a
 *  running maximum and minimum, and so only return a value if these are
 *  extended by the tile
 */
void Map_tile_max_height (const map_tile *tile, double *top, double *bottom,
			  bool accumulate)
{

  int i;
  double maxheight = -DBL_MAX;
  double minheight = DBL_MAX;

  for (i = 0; i < tile->num_vertices; ++i)
  {
    if (Vertex [tile->vertex [i]].z > maxheight)
    {
      maxheight = Vertex [tile->vertex [i]].z;
    }
    if (Vertex [tile->vertex [i]].z < minheight)
    {
      minheight = Vertex [tile->vertex [i]].z;
    }
  }

  if (accumulate)
  {
    if (maxheight > *top)    *top    = maxheight;
    if (minheight < *bottom) *bottom = minheight;
  }
  else
  {
    *top = maxheight;
    *bottom = minheight;
  }

}




int Map_terrain_type (const map_tile *tile)
{
  int floor_tex = tile->texture >> (Option_underworld ? 6 : 11);
  const texture_prop *texprop = Tex_get_prop (floor_tex);
  return texprop->type;
}




/*
 * Add a vertex to the map, checking for duplications. If a map tile is
 *  specified, only check that tile and its neighbours for duplications (i.e.
 *  assume the vertex is in that particular tile)
 */
int Map_add_vertex (const map_vertex *v, const map_tile *t)
{
  int x, y;
  int xstart, xend, ystart, yend;
  int i;

  ystart = t ? t->y - 1 : 0;  if (ystart <  0) ystart = 0;
  yend   = t ? t->y + 1 : 63; if (yend   > 63) yend   = 63;
  xstart = t ? t->x - 1 : 0;  if (xstart <  0) xstart = 0;
  xend   = t ? t->x + 1 : 63; if (xend   > 63) xend   = 63;

  for (y = ystart; y <= yend; ++y)
  {
    for (x = xstart; x <= xend; ++x)
    {
      const map_tile *t2 = &(Tile [x + 64 * y]);
      for (i = 0; i < t2->num_vertices; ++i)
      {
	if (vec_coincides (Vertex [t2->vertex [i]], *v))
	{
	  return t2->vertex [i];
	}
      }
    }
  }

  if (Num_map_vertices >= MAP_VERTICES)
  {
    Cant_happen ("Map_add_vertex(): too many vertices, increase MAP_VERTICES");
  }

  Vertex [Num_map_vertices] = *v;
  return Num_map_vertices++;

}




/*
 * Add a surface to the map (but don't link it into a specific tile)
 */
int Map_add_surface (const map_surface *ms)
{

  if (Num_map_surfaces >= MAP_SURFACES)
  {
    Cant_happen ("Map_add_surface(): too many surfaces, increase"
		 "  MAP_SURFACES");
  }
  Surface [Num_map_surfaces] = *ms;
  return Num_map_surfaces++;

}






/*
 * Find the overlapping triangle or quadrilateral, or return 0 if none. Points
 *  are clockwise from top left as ever
 * Assumes that x runs from 0 to 256
 */
int find_overlap (point pt [4])
{

  int right;
  int left;

  if ((pt [0].y >= pt [3].y) && (pt [1].y >= pt [2].y))
  {
    return 0;
  }

  if ((pt [0].y <= pt [3].y) && (pt [1].y <= pt [2].y))
  {
    /*
     * Clear
     */
  }
  else if (pt [0].y < pt [3].y)
  {

    /*
     * Left side is clear but right is not. 
     */
    left  = pt [3].y - pt [0].y;
    right = pt [1].y - pt [2].y;
    pt [1].x = (256 * left) / (left + right);
    right = pt [1].y - pt [0].y;
    pt [1].y = (pt [1].x * right) / 256;
    pt [2] = pt [1];

  }
  else
  {

    /*
     * Right side is clear but left is not. 
     */
    left  = pt [0].y - pt [3].y;
    right = pt [2].y - pt [1].y;
    pt [0].x = (256 * left) / (left + right);
    right = pt [1].y - pt [0].y;
    pt [0].y = (pt [0].x * right) / 256;
    pt [3] = pt [0];

  }

  return 1;

}



int dedupe_vertices (int num_vertices, int *vertex)
{
  int out_vertices = num_vertices;
  int in;
  int out;
  while (vertex [out_vertices - 1] == vertex [0])
  {
    --out_vertices;
  }
  for (in = 1; in < out_vertices; ++in)
  {
    if (vertex [in - 1] == vertex [in])
    {
      for (out = in + 1; out < out_vertices; ++out)
      {
	vertex [out - 1] = vertex [out];
      }
      --out_vertices;
    }
  }
  return out_vertices;
}

int add_tile_vertex (map_tile *t, const map_vertex *v)
{

  int nvert = Map_add_vertex (v, t);
  t->vertex [t->num_vertices++] = nvert;
  return nvert;

}

int add_tile_surface (map_tile *t, const map_surface *ms)
{

  int out;
  int nsurf = Map_add_surface (ms);

  /*
   * Check and remove duplicate vertices at this stage
   */
  out = dedupe_vertices (Surface [nsurf].num_vertices, Surface [nsurf].vertex);
  if (out < 3)
  {
    return -1;   /* surface must have at least 3 unique vertices */
  }
  Surface [nsurf].num_vertices = out;
  Surface [nsurf].tile_next = t->first_surface;
  t->first_surface = nsurf;
  return nsurf;

}

map_tile *in_direction (int t, int d)
{
  switch (d)
  {
  case NORTH:
    return (t / 64 != 63) ? &(Tile [t + 64]) : NULL;
  case EAST:
    return (t % 64 != 63) ? &(Tile [t +  1]) : NULL;
  case SOUTH:
    return (t / 64 !=  0) ? &(Tile [t - 64]) : NULL;
  case WEST:
    return (t % 64 !=  0) ? &(Tile [t -  1]) : NULL;
  default:
    return NULL;
  }
}

map_surface *find_back_surface (const map_tile *t, const map_surface *s)
{
  map_tile *bt = s->behind;
  int n;
  if (bt == NULL)
  {
    return NULL;
  }
  for (n = bt->first_surface; n != -1; n = Surface [n].tile_next)
  {
    if (Surface [n].behind == t)
    {
      return &(Surface [n]);
    }
  }
  return NULL;
}



#define SAMEXY(v1,v2) (                                           \
  (fabs (Vertex [v1].x - Vertex [v2].x) < A_VERY_SMALL_THING) &&  \
  (fabs (Vertex [v1].y - Vertex [v2].y) < A_VERY_SMALL_THING))

#define STRAIGHT_ABOVE(v1,v2) \
 ((Vertex [v1].z >= Vertex [v2].z) && SAMEXY (v1,v2))


void intersection (int v1, int v2, int a, int b, map_vertex *vret)
{
  double dz;
  int t;

  if (SAMEXY (v1, b))  /* a, b reversed */
  {
    t = a; a = b; b = t;
  }
  dz = Vertex [v1].z - Vertex [a].z;
  dz /= dz - (Vertex [v2].z - Vertex [b].z);
  vec_subtract (*vret, Vertex [v2], Vertex [v1]);
  vec_multiply (*vret, dz);
  vec_add (*vret, *vret, Vertex [v1]);
}

/*
 * Used by check_aperture
 */
int check_aperture_side (map_surface *ms1, map_surface *ms2, int *aperture)
{

  map_tile *t2 = ms1->behind;
  map_surface ts;
  map_vertex  tv;
  int nvert;

  int i;
  int next;
  int a;
  int b;
  int nreject;
  int reject [MAX_VERTEX_MAP] = {0};
  int scrap  [MAX_VERTEX_MAP];        /* scratch (intermediate) aperture */
  int scrap_vertices = 0;
  int ap_vertices = 0;

  /*
   * Check intersections with the top edge of the first surface. Here a is
   *  the left-hand upper vertex and b is the right-hand
   */
  if (SAMEXY (ms1->vertex [0], ms1->vertex [1]))
  {
    /*
     * It's triangular and the first vertex has been removed. Use the last and
     *  the first
     */
    a = ms1->vertex [ms1->num_vertices - 1];
    b = ms1->vertex [0];
  }
  else
  {
    a = ms1->vertex [0];
    b = ms1->vertex [1];
  }

  nreject = 0;
  for (i = 0; i < ms2->num_vertices; ++i)
  {
    if (STRAIGHT_ABOVE (ms2->vertex [i], a) ||
	STRAIGHT_ABOVE (ms2->vertex [i], b))
    {
      ++nreject;
      reject [i] = 1;
    }
  }
  if (nreject == ms2->num_vertices)
  {
    /*
     * The second face is entirely above the first; no aperture
     */
    return 0;
  }

  /*
   * Make a surface out of the rejects and a scratch aperture out of the
   *  accepts
   */
  ts = *ms2;
  ts.num_vertices = 0;
  for (i = 0; i < ms2->num_vertices; ++i)
  {
    next = (i + 1) % ms2->num_vertices;
    if (reject [i])
    {
      ts.vertex [ts.num_vertices++] = ms2->vertex [i];
      if (!reject [next])
      {
	if (SAMEXY (ms2->vertex [i], ms2->vertex [next]))
	{
	  nvert = a;
	}
	else
	{
	  /* New vertex needed */
	  intersection (ms2->vertex [i], ms2->vertex [next], a, b, &tv);
	  nvert = add_tile_vertex (t2, &tv);
	}
	ts.vertex [ts.num_vertices++] = nvert;
	scrap     [scrap_vertices++ ] = nvert;
      }
    }
    else 
    {
      scrap [scrap_vertices++] = ms2->vertex [i];
      if (reject [next])
      {
	if (SAMEXY (ms2->vertex [i], ms2->vertex [next]))
	{
	  nvert = b;
	}
	else
	{
	  /* New vertex needed */
	  intersection (ms2->vertex [i], ms2->vertex [next], a, b, &tv);
	  nvert = add_tile_vertex (t2, &tv);
	}
	ts.vertex [ts.num_vertices++] = nvert;
	scrap     [scrap_vertices++ ] = nvert;
      }
    }
  }
  scrap_vertices = dedupe_vertices (scrap_vertices, scrap);

  /*
   * We've been deliberately harsh about rejects, so that if the top edges or
   *  two of the top points coincide we have made a surface with zero area out
   *  of them. That doesn't matter however, as it will be rejected by
   *  add_tile_surface()
   */
  if (ts.num_vertices > 0)
  {
    /*
     * Add the new surface
     */
    add_tile_surface (t2, &ts);
  }

  /*
   * Now do the same with the bottom edge. Again a is the left-hand vertex and
   *  b is the right-hand
   */
  if (SAMEXY (ms1->vertex [ms1->num_vertices - 1],
	      ms1->vertex [ms1->num_vertices - 2]))
  {
    /*
     * It's triangular and the last vertex has been removed. Use the last and
     *  the first
     */
    a = ms1->vertex [0];
    b = ms1->vertex [ms1->num_vertices - 1];
  }
  else
  {
    a = ms1->vertex [ms1->num_vertices - 1];
    b = ms1->vertex [ms1->num_vertices - 2];
  }

  /*
   * This time we do rejects against the scratch aperture, and write to the
   *  final one
   */
  nreject = 0;
  for (i = 0; i < scrap_vertices; ++i)
  {
    if (STRAIGHT_ABOVE (a, scrap [i]) || STRAIGHT_ABOVE (b, scrap [i]))
    {
      ++nreject;
      reject [i] = 1;
    }
    else
    {
      reject [i] = 0;
    }
  }
  if (nreject == scrap_vertices)
  {
    /*
     * The second face is entirely above the first; no aperture
     */
    return 0;
  }

  /*
   * Make a surface out of the rejects and a scratch aperture out of the
   *  accepts
   */
  ts = *ms2;
  ts.num_vertices = 0;
  for (i = 0; i < scrap_vertices; ++i)
  {
    next = (i + 1) % scrap_vertices;
    if (reject [i])
    {
      ts.vertex [ts.num_vertices++] = scrap [i];
      if (!reject [next])
      {
	if (SAMEXY (scrap [i], scrap [next]))
	{
	  nvert = b;
	}
	else
	{
	  /* New vertex needed */
	  intersection (scrap [i], scrap [next], a, b, &tv);
	  nvert = add_tile_vertex (t2, &tv);
	}
	ts.vertex [ts.num_vertices++] = nvert;
	aperture  [ap_vertices++    ] = nvert;
      }
    }
    else 
    {
      aperture [ap_vertices++] = scrap [i];
      if (reject [next])
      {
	if (SAMEXY (scrap [i], scrap [next]))
	{
	  nvert = a;
	}
	else
	{
	  /* New vertex needed */
	  intersection (scrap [i], scrap [next], a, b, &tv);
	  nvert = add_tile_vertex (t2, &tv);
	}
	ts.vertex [ts.num_vertices++] = nvert;
	aperture  [ap_vertices++    ] = nvert;
      }
    }
  }
  ap_vertices = dedupe_vertices (ap_vertices, aperture);

  if (ts.num_vertices > 0)
  {
    /*
     * Add the new surface
     */
    add_tile_surface (t2, &ts);
  }

  return ap_vertices;

}

void check_aperture (map_surface *ms1, map_surface *ms2)
{

  int aperture [MAX_VERTEX_MAP];
  int aperture2 [MAX_VERTEX_MAP];
  int ap_vertices;
  int i;

  ap_vertices = check_aperture_side (ms1, ms2, aperture);
  if (ap_vertices == 0)
  {
    return;
  }
  if (check_aperture_side (ms2, ms1, aperture2) != ap_vertices)
  {
    /*
     * This REALLY can't happen ... I think
     */
    Cant_happen ("check_aperture(): aperture sides don't match");
  }

  if (ap_vertices < 3)
  {
    /*
     * Aperture must have at least 3 vertices! Yes, this is a nasty kludge
     */
    ms1->flags = 0;
    ms1->num_vertices = 0;
    ms2->flags = 0;
    ms2->num_vertices = 0;
  }
  else
  {

    ms1->flags = MSF_PORTAL;
    ms1->num_vertices = ap_vertices;
    for (i = 0; i < ap_vertices; ++i)
    {
      ms1->vertex [i] = aperture [ap_vertices - i - 1];
    }

    ms2->flags = MSF_PORTAL;
    ms2->num_vertices = ap_vertices;
    memcpy (ms2->vertex, aperture, ap_vertices * sizeof *aperture);

  }

}

/*
 * Add a floor or ceiling to a tile. Diagonal slopes are added in two stages
 *  with fake_type used to fake a diagonal tile with triangular floor/ceiling.
 */
void add_floor (map_tile *t, int fake_type, bool ceiling)
{
  Vector fake_vertex [4];
  Vector tv;
  map_surface ts;
  int n;
  int in;
  int new_vertex;
  int last = 3 - ((fake_type >= TILE_DIAGONAL_BASE) &&
		  (fake_type < TILE_SLOPE_BASE));
  int rot  = (t->rflags >> (ceiling ? 4 : 6)) & 3; /* texture rotation */
  double m;

  /*
   * Construct a fake quadrilateral for the texture mapping. In the case of a
   *  triangular tile or half of a diagonal slope we will need to add a vertex
   */
  ts.num_vertices = 0;
  in = !ceiling;          /* ceiling vertices are even, floor odd */
  for (n = 0; n < 4; ++n)
  {
    if (n + TILE_DIAGONAL_BASE == fake_type)
    {
      /*
       * This is the vertex that doesn't exist ... skip it and fill in later
       */
      if (t->type != fake_type)
      {
	in += 2;
      }
    }
    else
    {
      if (ceiling)
      {
	ts.vertex [last - ts.num_vertices++] = t->vertex [in];
      }
      else
      {
	ts.vertex [ts.num_vertices++] = t->vertex [in];
      }
      fake_vertex [n].x = Vertex [t->vertex [in]].x;
      fake_vertex [n].y = Vertex [t->vertex [in]].y;
      fake_vertex [n].z = Vertex [t->vertex [in]].z;
      in += 2;
    }
  }
  if ((fake_type >= TILE_DIAGONAL_BASE) && (fake_type < TILE_SLOPE_BASE))
  {
    /*
     * We need a fake vertex for the missing corner
     */
    new_vertex = fake_type - TILE_DIAGONAL_BASE;
    vec_add (tv, fake_vertex [(new_vertex + 3) % 4],
	     fake_vertex [(new_vertex + 1) % 4]);
    vec_subtract (fake_vertex [new_vertex], tv,
		  fake_vertex [(new_vertex + 2) % 4]);
  }

  /*
   * Now we can do the texture mapping
   */
  if (Option_underworld)
  {
    if (ceiling)
    {
      if (Underworld2) ts.texture = 32;
      else ts.texture = 48;
    }
    else ts.texture = t->texture >> 6;
  }
  else
  {
    ts.texture = (t->texture >> (ceiling ? 6 : 11)) & 0x1F;
  }
  ts.anchor  = fake_vertex [rot];
  vec_subtract (ts.u, fake_vertex [(rot + 1) % 4], fake_vertex [rot]);
  m = vec_magnitude2 (ts.u) / TILE_GRID_SIZE;
  vec_divide   (ts.u, m);
  vec_subtract (ts.v, fake_vertex [(rot + 3) % 4], fake_vertex [rot]);
  m = vec_magnitude2 (ts.v) / TILE_GRID_SIZE;
  vec_divide   (ts.v, m);

  if (ceiling) vec_cross (ts.normal, ts.u, ts.v);
  else vec_cross (ts.normal, ts.v, ts.u);
  vec_normalise (ts.normal);

  /*
   * Floors and ceilings can never be portals, of course
   */
  ts.flags  = SF_TEXTURED;
  ts.behind = NULL;

  add_tile_surface (t, &ts);

}
