/*
 * matrix.h -
 *  definitions and macros for matrix maths
 */


/* sentry */
#ifndef MATRIX_H
#define MATRIX_H


/*
 * #includes
 */
#include "vec.h"


/*
 * Across first, then down
 */
typedef struct tag_mat
{
  vec x;
  vec y;
  vec z;
} mat;

#define Matrix mat


#define Mat_transpose(dst,src) do { \
  (dst).x.x = (src).x.x;            \
  (dst).x.y = (src).y.x;            \
  (dst).x.z = (src).z.x;            \
  (dst).y.x = (src).x.y;            \
  (dst).y.y = (src).y.y;            \
  (dst).y.z = (src).z.y;            \
  (dst).z.x = (src).x.z;            \
  (dst).z.y = (src).y.z;            \
  (dst).z.z = (src).z.z;            \
} while (0)



#define Mat_multiply(dst,s1,s2) do {  \
  (dst).x.x = (s1).x.x * (s2).x.x     \
            + (s1).x.y * (s2).y.x     \
            + (s1).x.z * (s2).z.x;    \
  (dst).x.y = (s1).x.x * (s2).x.y     \
            + (s1).x.y * (s2).y.y     \
            + (s1).x.z * (s2).z.y;    \
  (dst).x.z = (s1).x.x * (s2).x.z     \
            + (s1).x.y * (s2).y.z     \
            + (s1).x.z * (s2).z.z;    \
  (dst).y.x = (s1).y.x * (s2).x.x     \
            + (s1).y.y * (s2).y.x     \
            + (s1).y.z * (s2).z.x;    \
  (dst).y.y = (s1).y.x * (s2).x.y     \
            + (s1).y.y * (s2).y.y     \
            + (s1).y.z * (s2).z.y;    \
  (dst).y.z = (s1).y.x * (s2).x.z     \
            + (s1).y.y * (s2).y.z     \
            + (s1).y.z * (s2).z.z;    \
  (dst).z.x = (s1).z.x * (s2).x.x     \
            + (s1).z.y * (s2).y.x     \
            + (s1).z.z * (s2).z.x;    \
  (dst).z.y = (s1).z.x * (s2).x.y     \
            + (s1).z.y * (s2).y.y     \
            + (s1).z.z * (s2).z.y;    \
  (dst).z.z = (s1).z.x * (s2).x.z     \
            + (s1).z.y * (s2).y.z     \
            + (s1).z.z * (s2).z.z;    \
} while (0)

#define mat_vec_multiply(dst,m,v) do { \
  (dst).x = vec_dot ((m).x, v);        \
  (dst).y = vec_dot ((m).y, v);        \
  (dst).z = vec_dot ((m).z, v);        \
} while (0)

#define mat_vec_multiply_transpose(dst,m,v) do {       	       	 \
  (dst).x = (m).x.x * (v).x + (m).y.x * (v).y + (m).z.x * (v).z; \
  (dst).y = (m).x.y * (v).x + (m).y.y * (v).y + (m).z.y * (v).z; \
  (dst).z = (m).x.z * (v).x + (m).y.z * (v).y + (m).z.z * (v).z; \
} while (0)


#endif /* sentry */
