/*
 * palette.h -
 *  Palettes
 */


#ifndef PALETTE_H
#define PALETTE_H


/*
 * #includes
 */
#include "types.h"




#ifdef __cplusplus
extern "C"
{
#endif



/*
 * Function prototypes
 */
void Pal_init (void);
void Pal_deinit(void);
void Pal_init_level (void);
void Pal_exit_level (void);
void Pal_load_palette (int offset);

/*
 * init a palette table with shades from one color to another. 
 * between the indizes iFrom (incl) and iTo (excl) the pcolor cFrom
 * gets shaded to cTo (including the trans byte!)
 */
void Pal_shade_palette(pcolor *pPal, int iFrom, int iTo,
											 pcolor cFrom, pcolor cTo);

/*
 * create a palette based on one main color (cFilter) and the 
 * intensity of the colors of another palette. 
 * the trans byte is not affected and will be taken from cFilter.
 * nGamma affects the calculated intensity of pPalIn.
 * eg: Pal_filter_palette(pPalBW, RGB_trans, 0xFFFFFFFF, 0x100, 0)
 *     creates a nice B/W view of tsshp :)
 */
void Pal_filter_palette(pcolor *pPalOut, const pcolor *pPalIn,
												pcolor cFilter, int nPalSize, int nGamma);

/*
 * copy a palette and set the trans bytes of the new palette
 */
void Pal_copy_palette(pcolor *pPalOut, const pcolor *pPalIn, byte nShade);

/*
 * Convert an RGB value to a palette index using a table lookup
 */
byte Pal_rgb_to_index (uint32 tc);

/*
 * We implement this one as a macro as it's so simple there's no point in
 *  having a complete function
 */
#define Pal_index_to_rgb(c) (RGB_trans [(c)])

/*
 * Convert an RGB value to a palette index using a distance test for each
 *  palette entry. Very slow. Should probably only be used to create lookup
 *  tables.
 */
byte Pal_approx_color (pcolor truecolor, pcolor *pPal);	


/*
 * Colour translation macros for pixels. These vary with bit depth
 */
#if BITS_PER_PIXEL == 8
#define Pal_pixel_to_rgb(c)   Pal_index_to_rgb (c)
#define Pal_rgb_to_pixel(c)   Pal_rgb_to_index (c)
#define Pal_index_to_pixel(c) (c)
#define Pal_pixel_to_index(c) (c)
#else
#define Pal_pixel_to_rgb(c)   (c)
#define Pal_rgb_to_pixel(c)   (c)
#define Pal_index_to_pixel(c) Pal_index_to_rgb (c)
#define Pal_pixel_to_index(c) Pal_rgb_to_index (c)
#endif



#ifdef __cplusplus
}
#endif




/*
 * Global variables
 */
extern int Palette [256] [3];
extern uint32 RGB_trans [256];


#endif /* PALETTE_H */
