/*
 * plot.h -
 *  Prototypes for low-level software plot routines
 */


#ifndef PLOT_H
#define PLOT_H




/*
 * #includes
 */
#include "types.h"
#include "matrix.h"


/*
 * A useful macro
 */
#define POS2PIXEL(x) (floor ((x) + 0.5))



#ifdef __cplusplus
extern "C"
{
#endif

/*
 * Single lines, flat-shaded and colour interpolated respectively, in line.c
 */
void Line_draw (const vec *v1, const vec *v2, pixel *bitmap, uint32 rgb);
void Line_draw_inter (const vec *v1, const vec *v2, pixel *bitmap,
		      uint32 rgb1, uint32 rgb2);

/*
 * "Fragments" objects, for cyberspace
 */
struct tag_ss_bitmap_header;
struct tag_surface;

void Plot_fragmented_sprite (vec *pos, mat *orn,
			     struct tag_ss_bitmap_header *bmp,
			     byte *zarray, int frag_size);
void Plot_fragment (vec *pos, uint32 rgb, int frag_size);
void Plot_surface (struct tag_surface *s);
void Plot_sprite (struct tag_surface *s, struct tag_ss_bitmap_header *bmp);

/*
 * Overlay images to be drawn
 */
struct tag_hudLayer;
struct tag_camera;

void Plot_overlay(struct tag_hudLayer *pLayer,
									struct tag_camera *pCam);

#ifdef __cplusplus
}
#endif



#endif /* PLOT_H */
