/*
 * render.h -
 *  Rendering frame
 */

/*
 * File:    src/render.h
 * Purpose: Renderer defines and externs
 * Prefix:  Ren_
 * Note:    No, there isn't a file with prefix "Stimpy_"
 * Music:   Van Halen, "1984"
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */


#ifndef RENDER_H
#define RENDER_H



/*
 * #includes
 */
#include "types.h"

#include "matrix.h"
#include "clip.h"  /* yes I know, but we need clipregion_t */



/*
 * A renderer camera. Note that this has got nothing to do with a System Shock
 *  camera object, except that a few of them have cameras attached
 */
#define CF_MAIN    1   /* is main view                            */
#define CF_3D_API  2   /* uses a 3D API (framebuffer may be NULL) */
#define CF_ZBUFFER 4   /* uses Z-buffering                        */

typedef struct tag_camera
{
  Vector position;
  Matrix orientation; /* world to camera coordinate transform */
  int window_width;
  int window_height;
  pixel *framebuffer; /* somewhere to draw stuff              */
  double screen_z;    /* screen projection distance           */
  int    flags;       /* CF_ flags above                      */
  clipregion_t view_frustrum;
} camera;



/*
 * Forward structure declarations for function prototypes
 */
struct tag_surface;

#ifdef __cplusplus
extern "C"
{
#endif

/*
 * Function prototypes
 */
void Ren_init (void);
void Ren_frame (double elapsed, pixel *framebuf);
void Ren_queue_camera (camera *c);
void Ren_scene (camera *c);
void Ren_transform_surface (struct tag_surface *s);

void Ren_spew (const char *format, ...);

#ifdef __cplusplus
}
#endif




/*
 * Global variables
 */
extern camera *Camera;

/*
 * Elapsed (renderer) time since game start
 */
extern double Render_frame_time;



#endif /* RENDER_H */
