/*
 * shade.h -
 *  defines and prototypes for 32bit color transparency
 */

/*
 * File:    shade.h
 * Purpose: 256 shades per colorpart
 * Notes:	this is my baby - think first before you start talking.
 *				i also added functionality to store shade info in the
 *				fourth (still) unused byte...
 */

#ifndef SHADING
#define SHADING

/* to have shading type-safe			*/
typedef short *			HSHADE;
typedef const HSHADE		HCSHADE;

/* define the colormasks				*/
#define	C_MASK_RED		0x00FF0000UL
#define	C_MASK_GREEN	0x0000FF00UL
#define	C_MASK_BLUE		0x000000FFUL
#define	C_MASK_TRANS	0xFF000000UL
/* ...and the shifting offsets		*/
#define	C_SHIFT_RED		0x10UL
#define	C_SHIFT_GREEN	0x08UL
#define	C_SHIFT_BLUE	0x00UL
#define	C_SHIFT_TRANS	0x18UL
/* now to get a colorpart				*/
#define	get_red(c)		((byte)((c) >> C_SHIFT_RED))
#define	get_green(c)	((byte)((c) >> C_SHIFT_GREEN))
#define	get_blue(c)		((byte)((c) >> C_SHIFT_BLUE))
#define	get_trans(c)	((byte)((c) >> C_SHIFT_TRANS))
/* and to reassemble it again			*/

#define	rgb(r, g, b)																\
				((uint32)(((uint32)(r) << C_SHIFT_RED)		|	\
									((uint32)(g) << C_SHIFT_GREEN)	|	\
									((uint32)(b) << C_SHIFT_BLUE)			\
				))


#define	rgbt(r, g, b, t)														\
				((uint32)(((uint32)(r) << C_SHIFT_RED)		|	\
									((uint32)(g) << C_SHIFT_GREEN)	|	\
									((uint32)(b) << C_SHIFT_BLUE)		|	\
									((uint32)(t) << C_SHIFT_TRANS)		\
				))


/* THIS is it								*/
#define	shade(sh, orig, newc)												\
				((byte)orig + sh[(short)(newc) - (short)(orig)])
#define	shade_color(sh, orig, newc)    	       	       	       	\
  rgb(shade(sh, get_red(orig),		get_red(newc)),		\
      shade(sh, get_green(orig),	get_green(newc)),	\
      shade(sh, get_blue(orig),		get_blue(newc))		\
					)

extern HCSHADE	shade_transparent;
extern HCSHADE	shade_opaque;

/*
 * jim - darkness table for lighting
 */
#define DARK_SHADES 64
extern pixel Dark_table [DARK_SHADES] [256];

#define FULL_BRIGHT     0
#define INKY_BLACKNESS (DARK_SHADES - 1)

/* but why? -jim #include "vector.h"  sorry ... */


#ifdef __cplusplus
extern "C"
{
#endif

/* but wouldn't work without these	*/
void		shade_init(void);

/*
 * a shade of 0x00 is full transparent (not visible),
 *				  0xFF gives an opaque color
 *
 *	sh: 0x00, orig: 0x3D, newc: 0x5A, result: 0x3D
 *	sh: 0xFF, orig: 0x3D, newc: 0x5A, result: 0x5A
 */
HSHADE	shade_get(byte s);

struct tag_map_vertex;
void Shd_build_static_darkmap (struct tag_map_vertex *vertex,
			       int num_vertices);
float Shd_get_dark_post   (int x, int y, double z);
float Shd_get_static_dark (double x, double y, double z);

#ifdef __cplusplus
}
#endif

#endif /*SHADING*/




/*
 * End of src/shade.h
 */

/*
 * Follows some emacs magic to get VC++-formatted code to display right (and
 *  hopefully vice versa)
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
