/*
 * softs.c -
 *  Software and logs class 6
 */

/*
 * File:    src/softs.c
 * Purpose: Class-specific routines for software / logs class 6
 * Prefix:  Sof_
 * Note:    
 * Music:   
 *
 * This file by Jim Cameron 2002 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */




/*
 * #includes
 */
#include "defines.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "fixture.h"
#include "main.h"    /* Cant_happen() */
#include "model.h"
#include "object.h"
#include "softs.h"


/*
 * Global variables
 */



/*
 * Local function prototypes
 */




/*
 * Scratch colour-scheme buffer for softs in c/space
 *  (Softs ... In ... SPAAAAAAAAAAAAAAAAAAACE!)
 * In fact, since softs are one solid colour only the first byte is used, but
 *  Mod_render() always expects a 6-byte colour scheme, and that is what it
 *  will get.
 */
static byte Softs_colour [6];


/*
 * A generic model for softs. Since all softs in c/space look the same modulo
 *  colour (and since there is no existing model template in obj3d.res), we
 *  can use this for all of them and use the existing mechanism for colour
 *  scheme.
 */
static const model Softs_model =
{
  8, /* num_vertices */
  {
    { -32, -48,  64 }, 
    {  32, -48,  64 },
    {  32, -48, -64 },
    { -32, -48, -64 },
    { -32,  48,  64 }, 
    {  32,  48,  64 },
    {  32,  48, -64 },
    { -32,  48, -64 }
  },
  {0},         /* dark   */
  { 0, 0, 0 }, /* centre */
  { 0, 0, 0 }, /* extent */
  0,           /* flags  */
  {
    6, /* num_surfaces */
    {
      { M3F_FLAT | M3F_AUX_SHADE, 0, 1, 4, { 0, 1, 2, 3 }, {  0, -1,  0 },
	{0,0,0}, {0,0,0}, {0,0,0} },
      { M3F_FLAT | M3F_AUX_SHADE, 0, 1, 4, { 4, 7, 6, 5 }, {  0,  1,  0 },
	{0,0,0}, {0,0,0}, {0,0,0} },
      { M3F_FLAT | M3F_AUX_SHADE, 0, 1, 4, { 1, 0, 4, 5 }, {  0,  0,  1 },
	{0,0,0}, {0,0,0}, {0,0,0} },
      { M3F_FLAT | M3F_AUX_SHADE, 0, 1, 4, { 2, 1, 5, 6 }, {  1,  0,  0 },
	{0,0,0}, {0,0,0}, {0,0,0} },
      { M3F_FLAT | M3F_AUX_SHADE, 0, 1, 4, { 3, 2, 6, 7 }, {  0,  0, -1 },
	{0,0,0}, {0,0,0}, {0,0,0} },
      { M3F_FLAT | M3F_AUX_SHADE, 0, 1, 4, { 0, 3, 7, 4 }, { -1,  0,  0 },
	{0,0,0}, {0,0,0}, {0,0,0} }
    },
    { { 0, 0, 0 }, 0 }, /* sort_plane (not used) */
    NULL,
    NULL
  }
};





/*
 * Softs colours, determined by subclass
 */
static const byte subclass_colour [] = { 0x3a, 0x4c, 0x77, 0x5b, 0x23 };


/*
 * Render a cyber-soft/log. This is only relevant in c/space since in realspace
 *  they appear as sprites
 */
void Sof_render_cyber (map_object *obj)
{

  int subclass = 0; /* "might be used uninitialised", blah! */

  if (obj->class == OBJCLASS_SOFTS)
  {
    /*
     * Actually I'm getting to be of the opinion that the "state" field is not
     *  a colour override, but denotes ICE defences
     */
/*      if (obj->state > 0) Softs_colour [0] = obj->state; */
    subclass = obj->subclass;
  }
  else if (obj->class == OBJCLASS_FIXTURE)
  {

    /*
     * I must confess to a certain sneaky admiration for a fudge like this.
     *  There are only enough softs slots for 15 objects, not enough for a
     *  c/space level. So objects of class 7 (fixtures: not normally used in
     *  c/space) are co-opted to replace them, and act as extra softs.
     */
    fixture *fix = Fix_get_fixture (obj);
    subclass = fix->flags;

  }
  else
  {
    printf ("%d/%d/%d\n", obj->class, obj->subclass, obj->type);
    Cant_happen ("Sof_render_cyber(): non-soft object");
  }

  /*
   * Softs colours are determined by subclass
   */
  assert (subclass < 5);
  Softs_colour [0] = subclass_colour [subclass];
  Mod_render (obj, &Softs_model, Softs_colour);

}

