/*
 * texture.h -
 *  Prototypes for dealing with textures
 */


#ifndef TEXTURE_H
#define TEXTURE_H


/*
 * #includes
 */




/*
 * Chunks where the various textures are to be found
 * These are obsolescent - use the defines from chunks.h
 */
#define TEXTURE_CHUNK_16    76
#define TEXTURE_CHUNK_32    77
#define TEXTURE_CHUNK_64   707
#define TEXTURE_CHUNK_128 1000


#define TEXPROP_FILE_SIZE 0x0b

typedef struct tag_texture_prop
{
  unsigned long stuff;
  bool  climbable;   /* can mantle up it (ladders / vines) */
  short starfield;   /* starfield control (for windows)    */
  unsigned short anim_group;
  short anim_no;     /* number within anim group           */
  int   type;        /* (Underworld) type                  */
} texture_prop;


struct tag_ss_bitmap_header;

typedef void (*tex_refresh_func) (struct tag_ss_bitmap_header *bmp);




/*
 * Terrain types (Underworld). Are these actually useful?
 */
#define TERRAIN_SOLID        0
#define TERRAIN_ANKH         2
#define TERRAIN_STAIRS_UP    3
#define TERRAIN_STAIRS_DOWN  4
#define TERRAIN_PIPE         5
#define TERRAIN_GRATING      6
#define TERRAIN_DRAIN        7
#define TERRAIN_PRINCESS     8
#define TERRAIN_WINDOW       9
#define TERRAIN_TAPESTRY    10
#define TERRAIN_DOOR        11
#define TERRAIN_WATER       16
#define TERRAIN_LAVA        32


/*
 * The last parameter to Tex_get_cache_block() determines whether entries are
 *  subchunks within a chunk or consecutive chunks
 */
#define TEXTURE_BLOCK_CHUNK    0
#define TEXTURE_BLOCK_SUBCHUNK 1
#define TEXTURE_BLOCK_UNIQUE   2 /* one-off, don't check */


#ifdef __cplusplus
extern "C"
{
#endif




void Tex_init (void);
void Tex_exit (void);
void Tex_init_level (int level_no);

int  Tex_get_cache_id    (int chunk, int subchunk);
int  Tex_get_cache_block (int chunk, int subchunk, int ntexs, int which);
void Tex_set_bitmap (int texid, struct tag_ss_bitmap_header *bmp);
int  Tex_add_cache (struct tag_ss_bitmap_header *bmp);
struct tag_ss_bitmap_header *Tex_get_texture (int texno);
const texture_prop *Tex_get_prop (int texid);
void Tex_set_refresh (int texid, tex_refresh_func hook);
struct tag_ss_bitmap_header *Tex_load_uw_texture (int texno);

#ifdef __cplusplus
}
#endif


#endif /* TEXTURE_H */
