/*
 * timer.c
 *  timing system
 *
 * This file by Christian Haas 2001 christian.haas@eunet.at
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "types.h"

#include "gamemem.h"
#include "timer.h"

/*
 * local structures
 */
typedef struct tag_TimerEntry
{
	char										 teName[31];
	TIMER_FUNCTION					 teTimerFunction;

	uint32									 teLastTime;
	uint32									 teCallTime;
	void										*tepData;
	
	struct tag_TimerEntry		*tepPrev;
	struct tag_TimerEntry		*tepNext;
} TimerEntry;

/*
 * local functions
 */

TimerEntry *timer_find_entry(char *strName);

/*
 * local variables
 */

struct tag_TimerEntry *tepFirst = NULL, *tepLast = NULL;


TimerEntry *timer_find_entry(char *strName)
{
	TimerEntry *pEntry = tepFirst, *pFound = NULL;

	while (pEntry && !pFound)
	{
		if (!strncmp(pEntry->teName, strName, 30))
			pFound = pEntry;

		pEntry = pEntry->tepNext;
	}

	return pFound;
}

void Timer_register_function(char *strName, uint32 msTime,
														 TIMER_FUNCTION pFunction, void *pData)
{
	TimerEntry *pEntry = malloc(sizeof(TimerEntry));

	memset(pEntry, 0x00, sizeof(TimerEntry));
	strncpy(pEntry->teName, strName, 30);
	pEntry->teTimerFunction = pFunction;
	pEntry->teCallTime = msTime;
	pEntry->tepData = pData;

	pEntry->tepPrev = tepLast;
	if (tepLast)
		tepLast->tepNext = pEntry;
	if (!tepFirst)
		tepFirst = pEntry;
	tepLast = pEntry;
}

void Timer_set_time(char *strName, uint32 msTime)
{
	TimerEntry *pEntry = timer_find_entry(strName);

	if (pEntry)
		pEntry->teCallTime = msTime;
}

void Timer_set_data(char *strName, void *pData)
{
	TimerEntry *pEntry = timer_find_entry(strName);

	if (pEntry)
		pEntry->tepData = pData;
}

void Timer_unregister_function(char *strName)
{
	TimerEntry *pEntry = timer_find_entry(strName);

	if (pEntry)
	{
		if (tepFirst == pEntry)
			tepFirst = pEntry->tepNext;
		if (tepLast == pEntry)
			tepLast = pEntry->tepPrev;
		if (pEntry->tepPrev)
			pEntry->tepPrev->tepNext = pEntry->tepNext;
		if (pEntry->tepNext)
			pEntry->tepNext->tepPrev = pEntry->tepPrev;
		free(pEntry);
	}
}

void Timer_work(uint32 now)
{
	TimerEntry *pEntry = tepFirst, *pEntryNext;
	uint32 last;

	while (pEntry)
	{
		pEntryNext = pEntry->tepNext;
		if (!pEntry->teLastTime || pEntry->teLastTime > now)
			pEntry->teLastTime = now;
		if ((now - pEntry->teLastTime) >= pEntry->teCallTime)
		{
			last = pEntry->teLastTime;
			pEntry->teLastTime = now;
			pEntry->teTimerFunction(now, last, pEntry->tepData);
		}

		pEntry = pEntryNext;
	}
}

void Timer_shutdown(void)
{
	TimerEntry *pEntry = tepLast, *pThis;

	tepFirst = NULL;
	tepLast = NULL;
	while (pEntry)
	{
		pThis = pEntry;
		pEntry = pEntry->tepPrev;
		free(pThis);
	}
}
