/*
 * vector.h -
 *  Definitions and macros to do with vectors
 */


#ifndef VECTOR_H
#define VECTOR_H


typedef struct tag_vec
{
  double x;
  double y;
  double z;
} vec;

/*
 * Backwards combatibility, but I was getting fed up of these capitals all
 *  over the place
 */
#define Vector vec



typedef struct
{
  short x;
  short y;
  short z;
} short_vec;



#define vec_add(dst,s1,s2) do { \
  (dst).x = (s1).x + (s2).x;    \
  (dst).y = (s1).y + (s2).y;    \
  (dst).z = (s1).z + (s2).z;    \
} while (0)


#define vec_subtract(dst,s1,s2) do { \
  (dst).x = (s1).x - (s2).x;         \
  (dst).y = (s1).y - (s2).y;         \
  (dst).z = (s1).z - (s2).z;         \
} while (0)

#define vec_multiply(v,s) do {\
  (v).x *= (s);               \
  (v).y *= (s);               \
  (v).z *= (s); } while (0)

#define vec_divide(v,s) do {\
  (v).x /= (s);         \
  (v).y /= (s);         \
  (v).z /= (s); } while (0)

#define vec_dot(s1,s2) \
  ((s1).x * (s2).x + (s1).y * (s2).y + (s1).z * (s2).z)

#define vec_magnitude2(s) vec_dot (s,s)

#define vec_magnitude(s) sqrt(vec_magnitude2 (s))

#define vec_cross(dst,s1,s2) do {              \
  (dst).x = (s1).y * (s2).z - (s1).z * (s2).y; \
  (dst).y = (s1).z * (s2).x - (s1).x * (s2).z; \
  (dst).z = (s1).x * (s2).y - (s1).y * (s2).x; \
} while (0)

/*
 * Don't like this, floating-point strict equality tests make me twitchy. Use
 *  vec_coincides() below
 */
#define vec_equal(s1, s2) \
	((s1).x == (s2).x && (s1).y == (s2).y && (s1).z == (s2).z)

#define vec_normalize(s) do {      \
  double vec_normalize_ts = vec_magnitude (s);   \
  if (vec_normalize_ts != 0.) vec_divide(s, vec_normalize_ts); \
	}	while(0)
#define vec_normalise(s) vec_normalize(s) /* pesky spelling! */

#define vec_coincides(v1, v2) (	        	    \
  (fabs ((v1).x - (v2).x) < A_VERY_SMALL_THING) &&  \
  (fabs ((v1).y - (v2).y) < A_VERY_SMALL_THING) &&  \
  (fabs ((v1).z - (v2).z) < A_VERY_SMALL_THING)    )

#endif
