/*
 * weapon.h -
 *  Weapons, grenades, bullets, clips
 */

/*
 * File:    src/weapon.h
 * Purpose: Class-specific defines for weaponry, classes 0-3
 * Prefix:  Wep_
 * Note:
 * Music:   
 *
 * This file by Jim Cameron 2001 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */



#ifndef WEAPON_H
#define WEAPON_H

/*
 * #includes
 */
#include "types.h"



#define NUM_WEAPONS    16
#define NUM_AMMO_CLIPS 15
#define NUM_EXPLOSIVES  8

/*
 * Weapon subclasses
 */
#define WEAPON_SUBCLASS_SEMIAUTO 0 /* pistol, dart, magnum, asslt rfl, riot */
#define WEAPON_SUBCLASS_FULLAUTO 1 /* flechette, Skorpion                   */
#define WEAPON_SUBCLASS_LAUNCHER 2 /* magpulse, railgun                     */
#define WEAPON_SUBCLASS_MELEE    3 /* pipe, laser rapier                    */
#define WEAPON_SUBCLASS_ENERGY   4 /* sparq, blaster, ion rfl               */
#define WEAPON_SUBCLASS_EN_LNCH  5 /* stungun, plasma rfl                   */


/*
 * Projectile subclasses
 */
#define PROJECTILE_SUBCLASS_TRACER  0
#define PROJECTILE_SUBCLASS_MISSILE 1


/*
 * Weapon properties.
 * This is a generic structure, containing all weapon information that isn't
 *  in the common objprop. Not all fields will be present for all weapon
 *  types
 */
typedef struct tag_weapon_prop
{

  byte   clipsub;     /* ammo clip subclass (class 1)         */
  byte   cliptypes;   /* ammo clip types accepted by weapon   */

  int    damage;      /* base damage, before modification     */
  short  offence;     /* tested vs. target defence value      */
  short  type;        /* I think, not sure how specified      */
  short  special;     /* 0x33 for magpulse, 7 for needle dart */
  int    stuff1;      /* No idea. Used for critter attacks    */
  int    penetration; /* tested vs. target armour value       */

  int    energ_rnds;  /* base energy usage / rounds per clip  */
  uint32 projectile;  /* projectile C/S/T value               */
  int    stuff2;      /* big, decimal                         */

} weapon_prop;



/*
 * Function prototypes
 */
#ifdef __cplusplus
extern "C"
{
#endif


void Wep_load_objprop (FILE *propfile);

struct tag_map_object;
byte *Wep_get_colour (struct tag_map_object *obj);

#ifdef __cplusplus
}
#endif



#endif /* WEAPON_H */
