/*
 * textex.c -
 *  System Shock texture extractor (PNG)
 *
 * This file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <stdlib.h>
#include "png.h"

#include "bitmap.h"
#include "main.h"
#include "palette.h"
#include "resource.h"


#define DEBUG_TEXTEX 0


/*
 * Testing
 */
/*  extern ss_bitmap_header *Bmp_load_thingy (int chunk); */
#define MONOCHROME 0


void TX_write_texture (int chunk, int subchunk)
{

  ss_bitmap_header *bmp;
  FILE *pngfile;
  png_structp png_ptr;
  png_infop info_ptr;
  png_colorp pngpalette;
  png_bytep *rows;
  char texname [256];
  int i;

  bmp = Bmp_load_bitmap (chunk, subchunk);
  if (bmp == NULL)
  {
    Cant_happen ("no such image");
  }

  /*
   * OK, we got it. Now for the PNG stuff. This is pretty much straight out
   *  of example.c
   */
  sprintf (texname, "tex%d-%d.png", chunk, subchunk);
  pngfile = fopen (texname, "wb");
  if (pngfile == NULL)
  {
    Cant_happen ("Couldn't open png file");
  }

  png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (png_ptr == NULL)
  {
    fclose (pngfile);
    Cant_happen ("png_create_write_struct() failed");
  }

  info_ptr = png_create_info_struct (png_ptr);
  if (info_ptr == NULL)
  {
    png_destroy_write_struct (&png_ptr, (png_infopp) NULL);
    fclose (pngfile);
    Cant_happen ("png_create_info_struct() failed");
  }

  if (setjmp (png_ptr->jmpbuf))
  {
    png_destroy_write_struct (&png_ptr, (png_infopp) NULL);
    fclose (pngfile);
    Cant_happen ("setjmp() failed");
  }

  png_init_io (png_ptr, pngfile);
# if MONOCHROME
  png_set_IHDR (png_ptr, info_ptr, bmp->width * 8, bmp->height, 1,
  		PNG_COLOR_TYPE_PALETTE, PNG_INTERLACE_NONE,
  		PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
# else
  png_set_IHDR (png_ptr, info_ptr, bmp->width, bmp->height, 8,
  		PNG_COLOR_TYPE_PALETTE, PNG_INTERLACE_NONE,
  		PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
# endif /* MONOCHROME */

  /*
   * The palette
   */
  pngpalette = (png_colorp) png_malloc (png_ptr, 256 * sizeof (png_color));
# if MONOCHROME
  for (i = 0; i < 2; ++i)
  {
    pngpalette [i].red   = 255 * i;
    pngpalette [i].green = 255 * i;
    pngpalette [i].blue  = 255 * i;
  }
# else
  for (i = 0; i < 256; ++i)
  {
    pngpalette [i].red   = Palette [i] [0];
    pngpalette [i].green = Palette [i] [1];
    pngpalette [i].blue  = Palette [i] [2];
  }
# endif
  png_set_PLTE (png_ptr, info_ptr, pngpalette, 256);

  /*
   * Write the stuff
   */
  png_write_info (png_ptr, info_ptr);

  rows = malloc (bmp->height * sizeof *rows);
  for (i = 0; i < bmp->height; ++i)
  {
    rows [i] = bmp->bits + i * bmp->width;
  }
  png_write_image (png_ptr, rows);
  free (rows);

  png_write_end (png_ptr, info_ptr);
  free (pngpalette);
  png_destroy_info_struct (png_ptr, &info_ptr);
  png_destroy_write_struct (&png_ptr, (png_infopp) NULL);

  /*
   * Like a good little texture extractor
   */
  fclose (pngfile);
  Bmp_free_bitmap (bmp);

}
