/*
 * viewlog.c -
 *  System Shock logfile viewer, on steroids :)
 * 
 * Written by Bryan Nolen 2001 bryann@swin.edu,au
 *
 * Based on a file by Jim Cameron 2000 jim@madeira.physiol.ucl.ac.uk
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 * 
 */

/*
 * #defines
 */
#define BN 1

#define STRINGS_START 0x0867
#define STRINGS_END 0x0883

#define WEAPONS_LIST 0x0024

#define TEXT_START 0x003C
#define TEXT_END 0x0046

#define EMAIL_START 0x0989
#define EMAIL_END 0x09A6

#define VMAIL_START 0x09B1
#define VMAIL_END 0x09B6

#define LOG_START 0x09B8
#define LOG_END 0x0A3D

#define CYBER_START 0x0A98
#define CYBER_END 0x0AA4

/*
 * #includes
 */
#include "../src/types.h"
#include "../src/resource.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <unistd.h>

/*
 * Function Prototypes
 */
int Init_Res();
void Print_Type_Header(int daChunk);
int Print_All_Subchunks(int daChunk, block_handle daBH, size_t daSize);
int Print_Subchunk(int daChunk, int daSubChunk, block_handle daBH, size_t daSize);
int Display_Usage();

void Cant_happen (const char *message);

/*
 * And begin....
 */
int main (int argc, char *argv [])
{
  int theChunkID;
  int subChunkID;
  block_handle bh;
  size_t size;

  if ((argc != 2) && (argc != 3) && (argc != 4)) /* we want 2, 3 or 4 args (arg0=name)*/
  {
    fprintf (stderr, "Not enough arguments given\n\n");
    Display_Usage(argv [0]);
    return 1;
  }

/*
 * Option Debugging Code.......
 fprintf (stderr, "Option Selected: %s\n argv 2 is : %s\n  argv 3 is : %s\n", argv [1], argv[2], argv[3]);
 */ 

/* Handle the command line arguments */

  if (strcasecmp (argv[1], "A") == 0) /* handle option A */
  {
/* Case: argv[1] = A && argv[2] != NULL && argv[3] == NULL */
    if ((argv [2] != NULL) && (argv [3] == NULL))
    {
      fprintf (stderr, "ViewLog - System Shock logfile viewer, on steroids\n");
      theChunkID = strtol (argv [2], NULL, 0);
      Init_Res(theChunkID, bh, size);
      Print_All_Subchunks(theChunkID, bh, size);
    }
/* Case: argv[1] = A && argv[2] != NULL && argv[3] != NULL */
    else if ((argv [2] != NULL) && (argv [3] != NULL))
    {
      fprintf (stderr, "ViewLog - System Shock logfile viewer, on steroids \n");
      theChunkID = strtol (argv [2], NULL, 0);
      subChunkID = strtol (argv [3], NULL, 0);
      Init_Res(theChunkID, bh, size);
      Print_Subchunk(theChunkID, subChunkID, bh, size);
    }      
/* Case: argv[1] = A, but argv[2] and argv[3] are NULL */
    else if (argv[2] == NULL)
    { 
      int l = STRINGS_START;
      fprintf (stderr, "ViewLog - System Shock logfile viewer, on steroids\n");
      fprintf (stderr, "\nNow printing all texts!!! between chunk %i & %i\n\n", STRINGS_START, CYBER_END);
      Init_Res();
      do {
          theChunkID = l;
	  Print_Type_Header(theChunkID);
          Print_All_Subchunks(theChunkID, bh, size);
	  l = l + 1;
      } while (l<CYBER_END);
    }
  }

  else if (strcasecmp (argv[1], "W") == 0) /* handle option W */
  {
      theChunkID = WEAPONS_LIST;
      Init_Res(theChunkID, bh, size);
      Print_All_Subchunks(theChunkID, bh, size);
  }

  else if (strcasecmp (argv[1], "L") == 0) /* handle option L */
  {
      int l = LOG_START;
      fprintf (stderr, "ViewLog - System Shock logfile viewer, on steroids\n");
      fprintf (stderr, "\nNow printing all LOGS between chunk %i & %i\n\n", LOG_START, LOG_END);
      Init_Res();
      do {
          theChunkID = l;
          Print_All_Subchunks(theChunkID, bh, size);
	  l = l + 1;
      } while (l<LOG_END);
  }

  else if (strcasecmp (argv[1], "E") == 0) /* handle option E */
  {
      int l = EMAIL_START;
      fprintf (stderr, "ViewLog - System Shock logfile viewer, on steroids\n");
      fprintf (stderr, "\nNow printing all EMAILs between chunk %i & %i\n\n", EMAIL_START, EMAIL_END);
      Init_Res();
      do {
          theChunkID = l;
          Print_All_Subchunks(theChunkID, bh, size);
	  l = l + 1;
      } while (l<EMAIL_END);
  }

  else if (strcasecmp (argv[1], "I") == 0) /* handle option I */
  {
      int l = EMAIL_START;
      fprintf (stderr, "ViewLog - System Shock logfile viewer, on steroids\n");
      fprintf (stderr, "\nNow printing Logs Index\n");
      Init_Res();
      do {
          theChunkID = l;
	  Print_Type_Header(theChunkID);
          Print_Subchunk(theChunkID, 1, bh, size);
	  l = l + 1;
      } while (l<CYBER_END);
  }

  else if (strcasecmp (argv[1], "S") == 0) /* handle option S */
  {
      int l = STRINGS_START;
      fprintf (stderr, "ViewLog - System Shock logfile viewer, on steroids\n");
      fprintf (stderr, "\nNow printing all GameStrings between chunk %i & %i\n\n", STRINGS_START, STRINGS_END);
      Init_Res();
      do {
          theChunkID = l;
          Print_All_Subchunks(theChunkID, bh, size);
	  l = l + 1;
      } while (l<STRINGS_END);
  }

  else if (strcasecmp (argv[1], "V") == 0) /* handle option V */
  {
      int l = VMAIL_START;
      fprintf (stderr, "ViewLog - System Shock logfile viewer, on steroids\n");
      fprintf (stderr, "\nNow printing all VMAILs between chunk %i & %i\n\n", VMAIL_START, VMAIL_END);
      Init_Res();
      do {
          theChunkID = l;
          Print_All_Subchunks(theChunkID, bh, size);
	  l = l + 1;
      } while (l<VMAIL_END);
  }

  else if (strcasecmp (argv[1], "C") == 0) /* handle option C */
  {
      int l = CYBER_START;
      fprintf (stderr, "ViewLog - System Shock logfile viewer, on steroids\n");
      fprintf (stderr, "\nNow printing all CyberSpace Messages between chunk %i & %i\n\n", CYBER_START, CYBER_END);
      Init_Res();
      do {
          theChunkID = l;
          Print_All_Subchunks(theChunkID, bh, size);
	  l = l + 1;
      } while (l<CYBER_END);
  }

  else if (strcasecmp (argv[1], "T") == 0) /* handle option T */
  {
      int l = TEXT_START;
      fprintf (stderr, "ViewLog - System Shock logfile viewer, on steroids\n");
      fprintf (stderr, "\nNow printing the text from scattered papers between chunk %i & %i\n\n", TEXT_START, TEXT_END);
      Init_Res();
      do {
          theChunkID = l;
          Print_All_Subchunks(theChunkID, bh, size);
	  l = l + 1;
      } while (l<TEXT_END);
  }

  else /* option not coded yet */
  {
        fprintf (stderr, "Option Still being written!\n");
	Display_Usage(argv [0]); 
  }

  return 0;
}

/*
 * Initialise the resources and find the chunk we are looking for
 */
int Init_Res()
{
  Res_init ();
  Res_add_resfile (Res_make_path (HD_DATA_PATH, "cybstrng.res"));
  return 0;
}

/*
 * Got the chunk(we think). Now deal with it all. 
 */
int Print_All_Subchunks(int daChunk, block_handle daBH, size_t daSize)
{
    int logchunk = daChunk;
    int numSubchunks;
    int i = 0;
    block_handle bh = daBH;
    size_t size = daSize;
    
  	numSubchunks = Res_get_numSubchunks_by_id(logchunk, &size);

    do {
	if((Res_load_subchunk(logchunk, i, &size)) != NULL)
	    {
	    	bh = Res_load_subchunk(logchunk, i, &size);
	    }
	else 
		{ 
		    fprintf (stderr, "%s: no such text log chunk %i\n", "viewlog", logchunk);
		    return 1; 
		}
    if (i==0) 
    	{
    		printf("chunk %i\n\t-====-\n",logchunk);
	    }
	
	printf ("sc %i: %s\n",i, *bh);
	i++;  
	
	} while (i<numSubchunks);
    printf("\t-====-\n");
    
    Res_free_chunk_by_id(logchunk);

  return 0;
}

/*
 * Got the chunk. Now deal with its subchunk 
 */
int Print_Subchunk(int daChunk, int daSubChunk, block_handle daBH, size_t daSize)
{
    int logchunk = daChunk;
    int i = daSubChunk;
    block_handle bh = daBH;
    size_t size = daSize;
    
    if((Res_load_subchunk(logchunk, i, &size)) != NULL)
	    {bh = Res_load_subchunk(logchunk, i, &size);}
    else { fprintf (stderr, "sc %i DOES NOT EXIST\n",i); return 1; }
    printf("chunk %i  ",logchunk);
    printf ("sc %i: %s\n",i, *bh);

    Res_free_chunk_by_id(logchunk);

  return 0;
}

void Print_Type_Header(int daChunk)
{
    int logchunk = daChunk;
    if (logchunk == LOG_START) { printf ("\nLogs Begin Now\n"); }
    if (logchunk == TEXT_START) { printf ("\nTexts From Papers Begin Now\n"); }
    if (logchunk == EMAIL_START) { printf ("\nEMails Begin Now\n"); }
    if (logchunk == VMAIL_START) { printf ("\nVMails Begin Now\n"); }
    if (logchunk == CYBER_START) { printf ("\nCyberSpace Messages Begin Now\n"); }
    if (logchunk == STRINGS_START) { printf ("\nGameStrings Begin Now\n"); }
}


/*
 * Error handling
 */
void Cant_happen (const char *message)
{
  fprintf (stderr, "%s\n\n", message);
  fprintf (stderr, "Salt the Fries!\n\n");
  exit (1);
}

/*
 * Display Usage Info
 */
int Display_Usage(const char *cmd_name)
{
    fprintf (stderr, "ViewLog - System Shock logfile viewer, on steroids \n");
    fprintf (stderr, "syntax: %s <type> (text log chunk no) (subchunk)\n", cmd_name);
    fprintf (stderr, "where <type> is one of the following::\n");
    fprintf (stderr, "\t L\t LOGS\n\t E\t E-Mails\n\t S\t GameStrings\n\t V\t V-Mails\n");
    fprintf (stderr, "\t C\t Cyber Space Messages\n\t T\t Text from scattered papers\n");
    fprintf (stderr, "\t A\t Any Type(see below)\n\t W\t The Weapons List!!\n\t I\t Logs Index(see below)\n");
    fprintf (stderr, " eg \t%s A 0x0989\t", cmd_name);    
    fprintf (stderr, "will print the entire contents of chunk 0x0989\n");
    fprintf (stderr, "\t%s L\t", cmd_name);
    fprintf (stderr, "\twill print the contents of ALL LOGS\n");
    fprintf (stderr, "\t%s I\t", cmd_name);
    fprintf (stderr, "\twill print the \"Logs Index\"\n");
    fprintf (stderr, "\t%s A\t", cmd_name);
    fprintf (stderr, "\twill print the contents of EVERY CHUNK \n(its good idea to pipe the output!!)\n");
    return 1;
} 

