/*
 * mfd.c -
 *  the mfds
 *
 * This file by Christian Haas 2001 christian.haas@eunet.at
 * System Shock copyright 1994 Looking Glass Technologies and ORIGIN Systems
 */

/*
 * #includes
 */
#include "defines.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <memory.h>
#include <math.h>
#include <string.h>										/* strncpy() - msgs */

#include "types.h"
#include "bitmap.h"
#include "chunks.h"
#include "container.h"
#include "graph2d.h"
#include "control.h"
#include "event.h"
#include "font.h"
#include "gamemem.h"
#include "gamestring.h"
#include "hud.h"
#include "main.h"											/* Can't_happen */
#include "map.h"
#include "matrix.h"
#include "mfd.h"
#include "object.h"
#include "option.h"
#include "palette.h"
#include "player.h"
#include "render.h"
#include "shade.h"
#include "timer.h"
#include "vec.h"

/*
 * We want pi!
 * (ff1: who doesn't ;)
 */
#ifndef M_PI
# define M_PI		3.14159265358979323846	/* pi */
#endif

#define	P_MFD_HOOK_INIT			HLAYER hLayer
#define	P_MFD_HOOK_MOUSE		HLAYER hLayer, byte action, bool bStart, uint16 x, uint16 y
#define	P_MFD_HOOK_RENDER		HLAYER hLayer, hudLayer **pLayer
#define P_MFD_HOOK_DEINIT		HLAYER hLayer

#define	SIDE2I(nSide)				(nSide / 2)
#define	I2SIDE(nIndex)			(nIndex ? C_MFD_SIDE_RIGHT : C_MFD_SIDE_LEFT)

/* layer functions */
void	mfd_layer_deinit(void);

/* targeting area functions*/
bool	mfd_rinit_targeting(size szScreen);
void	mfd_rdeinit_targeting(void);
void	mfd_timer_targeting(uint32 now, uint32 last, void *pData);
bool	mfd_mouse_targeting(P_HUD_HOOK_MOUSE);
void	mfd_render_targeting(P_HUD_HOOK_RENDER);
/* big map functions*/
void  mfd_map_refresh(void); /* jim -added to squatch a Borland warning */
bool	mfd_rinit_bigmap(size szScreen);
void	mfd_rdeinit_bigmap(void);
void	mfd_timer_bigmap(uint32 now, uint32 last, void *pData);
bool	mfd_mouse_bigmap(P_HUD_HOOK_MOUSE);
void	mfd_render_bigmap(P_HUD_HOOK_RENDER);
/* compass area functions*/
bool	mfd_rinit_compass(size szScreen);
void	mfd_rdeinit_compass(void);
void	mfd_render_compass(P_HUD_HOOK_RENDER);
/* general hw button functions */
bool	mfd_rinit_HW_but(size szScreen);
void	mfd_rdeinit_HW_but(void);
/* left hw buttons functions*/
bool	mfd_rinit_LHW_but(size szScreen);
bool	mfd_mouse_LHW_but(P_HUD_HOOK_MOUSE);
void	mfd_render_LHW_but(P_HUD_HOOK_RENDER);
/* right hw buttons functions*/
bool	mfd_rinit_RHW_but(size szScreen);
bool	mfd_mouse_RHW_but(P_HUD_HOOK_MOUSE);
void	mfd_render_RHW_but(P_HUD_HOOK_RENDER);
/* general (main) mfd functions */
bool	mfd_rinit_mfds(size szScreen);
void	mfd_rdeinit_mfds(void);
/* left mfd buttons functions*/
bool	mfd_mouse_LMFD_but(P_HUD_HOOK_MOUSE);
void	mfd_render_LMFD_but(P_HUD_HOOK_RENDER);
/* center mfd buttons functions*/
bool	mfd_mouse_CMFD_but(P_HUD_HOOK_MOUSE);
void	mfd_render_CMFD_but(P_HUD_HOOK_RENDER);
/* right mfd buttons functions*/
bool	mfd_mouse_RMFD_but(P_HUD_HOOK_MOUSE);
void	mfd_render_RMFD_but(P_HUD_HOOK_RENDER);
/* side/center mfd's */
void	mfd_timer_mfd(uint32 now, uint32 last, void *pData);
bool	mfd_mouse_lmfd(P_HUD_HOOK_MOUSE);
bool	mfd_mouse_rmfd(P_HUD_HOOK_MOUSE);
bool	mfd_mouse_cmfd(P_HUD_HOOK_MOUSE);
void	mfd_render_lmfd(P_HUD_HOOK_RENDER);
void	mfd_render_rmfd(P_HUD_HOOK_RENDER);
void	mfd_render_cmfd(P_HUD_HOOK_RENDER);

/* body status area functions*/
bool	mfd_rinit_bodystat(size szScreen);
void	mfd_rdeinit_bodystat(void);
void	mfd_timer_bodystat(uint32 now, uint32 last, void *pData);
void	mfd_render_bodystat(P_HUD_HOOK_RENDER);
/* viewpose area functions*/
bool	mfd_rinit_viewpose(size szScreen);
void	mfd_rdeinit_viewpose(void);
bool	mfd_mouse_viewpose(P_HUD_HOOK_MOUSE);
void	mfd_render_viewpose(P_HUD_HOOK_RENDER);
/* health area functions*/
bool	mfd_rinit_health(size szScreen);
void	mfd_rdeinit_health(void);
void	mfd_render_health(P_HUD_HOOK_RENDER);
/* video mail functions */
bool	mfd_rinit_vidmail(size szScreen);
void	mfd_rdeinit_vidmail(void);
void	mfd_timer_vidmail(uint32 now, uint32 last, void *pData);
bool	mfd_mouse_vidmail(P_HUD_HOOK_MOUSE);
void	mfd_render_vidmail(P_HUD_HOOK_RENDER);
bool	mfd_destroy_vidmail(P_HUD_HOOK_DEST);

/* sensaround           */
bool	mfd_rinit_SA(size szScreen);
void	mfd_rdeinit_SA(void);

/* local (helper) functions */
void	mfd_show_item(short nItem, bool bShow);
bool	mfd_load_bitmap(g2dbase *pBase, int nChunk, int nSubChunk);

typedef struct tag_MFD_Configuration
{
	/* MFDs */
	uint32					 mfdLastCards[3];
	uint32					 mfdLastIds[2][5];
	/* BigMap*/
	bool						 mfdBigMapAlignNorth;
	byte						 mfdBigMapZoom;
	/* SideMap */
	bool						 mfdSMapAlignNorth[2];
	uint32					 mfdSMapLastView[2];
	byte						 mfdSMapZoom[2];
	byte						 mfdSMapFlags[2];
} MFD_Configuration;

MFD_Configuration mfdConfig =
{
	/* MFDs */
	{ F_MFD_NONE, F_MFD_NONE, F_MFD_NONE },
	{ 
		{ 0x00000000, 0x00000000, F_SMFD_MSIDE, 0x00000000, 0x00000000 },
		{ 0x00000000, 0x00000000, F_SMFD_MMAP, 0x00000000, 0x00000000 }
	},
	/* MFD BigMap */
	false, 10,
	/* MFD SideMap */
	{ false, true }, { F_SMFD_MMAP, F_SMFD_MMAP }, { 4, 4 }, 
	{ F_DRAWMAP_MODEL | F_DRAWMAP_ELEV | F_DRAWMAP_SEC, F_DRAWMAP_MODEL | F_DRAWMAP_ELEV | F_DRAWMAP_SEC }
};

typedef struct tag_MFD_Runtime
{
	char				 *rtName;
	bool				(*rtInitFunction)(size szScreen);
	void				(*rtDeinitFunction)(void);
} MFD_Runtime;

MFD_Runtime mfd_ss_runtimes[11] =
{
	{ "Targeting",	mfd_rinit_targeting,	mfd_rdeinit_targeting },
	{ "ViewPose",		mfd_rinit_viewpose,		mfd_rdeinit_viewpose	},
	{ "BodyStat",		mfd_rinit_bodystat,		mfd_rdeinit_bodystat	},
	{ "Health",			mfd_rinit_health,			mfd_rdeinit_health		},
	{ "Compass",		mfd_rinit_compass,		mfd_rdeinit_compass		},
	{ "HW Buttons",	mfd_rinit_HW_but,			mfd_rdeinit_HW_but		},
	{ "Bigmap",			mfd_rinit_bigmap,			mfd_rdeinit_bigmap		},
	{ "VidMail",		mfd_rinit_vidmail,		mfd_rdeinit_vidmail		},
	{ "MFDs",				mfd_rinit_mfds,				mfd_rdeinit_mfds			},
	{ "SensAround",	mfd_rinit_SA,					mfd_rdeinit_SA				},
	
	{ NULL, NULL, NULL }
};

hudArea mfd_ss_items[C_MFD_AMOUNT] =
{
/* addons									*/
	{ "Target",					false, { 0, 0, 0, 0 }, 
	  { mfd_mouse_targeting, mfd_render_targeting, NULL } }, 
	{ "Big map",				false, { 0, 0, 0, 0},
	  { mfd_mouse_bigmap, mfd_render_bigmap, NULL } },
	{ "Compass",				false, { 0, 0, 0, 0 }, 
	  { NULL, mfd_render_compass, NULL } }, 
/* side mfd buttons						*/
	{ "Left MFD Buttons",	false, { 0, 0, 0, 0 }, 
	  { mfd_mouse_LMFD_but, mfd_render_LMFD_but, NULL } },  
	{ "Center MFD Buttons",	false, { 0, 0, 0, 0 }, 			 
	  { mfd_mouse_CMFD_but, mfd_render_CMFD_but, NULL } }, 
	{ "Right MFD Buttons",	false, { 0, 0, 0, 0 }, 
	  { mfd_mouse_RMFD_but, mfd_render_RMFD_but, NULL } }, 
/* side hardware buttons				*/
	{ "Left HW Buttons",		false, { 0, 0, 0, 0 }, 
	  { mfd_mouse_LHW_but, mfd_render_LHW_but, NULL } }, 
	{ "Right HW Buttons",	false, { 0, 0, 0, 0 }, 
	  { mfd_mouse_RHW_but, mfd_render_RHW_but, NULL } }, 
/* MFD_s themselves						*/
	{ "Center MFD",			false, { 0, 0, 0, 0 }, 
	  { mfd_mouse_cmfd, mfd_render_cmfd, NULL } }, 
	{ "Left MFD",				false, { 0, 0, 0, 0 }, 
	  { mfd_mouse_lmfd, mfd_render_lmfd, NULL } }, 
	{ "Right MFD",				false, { 0, 0, 0, 0 }, 
	  { mfd_mouse_rmfd, mfd_render_rmfd, NULL } }, 
/* stats				 						*/
	{ "Body Status",			false, { 0, 0, 0, 0 }, 
	  { NULL, mfd_render_bodystat, NULL } }, 
	{ "View/Pose",				false, { 0, 0, 0, 0 }, 
	  { mfd_mouse_viewpose, mfd_render_viewpose, NULL } }, 
	{ "Health",					false, { 0, 0, 0, 0 }, 
	  { NULL, mfd_render_health, NULL } },
/* vidmail			 						*/
	{ "Video Mail",			false, { 0, 0, 0, 0 }, 
	  { mfd_mouse_vidmail, mfd_render_vidmail, mfd_destroy_vidmail } }
};

bool bItemVisibility[C_MFD_MODE_AMOUNT][C_MFD_AMOUNT] =
{
	/* C_MFD_MODE_NONE */
	{ false, false, false, false, false, false, false, false,
		false, false, false, false, false, false, false },
	/* C_MFD_MODE_WORLD */
	{ true, true, true, true, true, true, true, true,
		true, true, true, true, true, true, true },
	/* C_MFD_MODE_CYBER */
	{ false, false, false, false, true, false, false, false,
		true, false, true, false, false, true, false }
};

byte nDetailLevel = C_MFD_DETAIL_AMOUNT - 1;	/* ;) */
byte nCurrentMode = C_MFD_MODE_NONE;

/* begin section main */

void MFD_init(int scrwidth, int scrheight)
{
	char strMessage[0x200];
	size szScreen;
	int i;

	szScreen.cx = (uint16)scrwidth;
	szScreen.cy = (uint16)scrheight;
	if (!Option_underworld)
	{ 
		i = 0;
		while (mfd_ss_runtimes[i].rtName)
		{
			if (mfd_ss_runtimes[i].rtInitFunction)
			{
#ifdef _DEBUG
				printf("MFD_Init(): Initializing %s\n", mfd_ss_runtimes[i].rtName);
#endif
				if (!mfd_ss_runtimes[i].rtInitFunction(szScreen))
				{
					sprintf(strMessage, "MFD_Init(): Failed to initialize %s",
									mfd_ss_runtimes[i].rtName);
					Cant_happen(strMessage);
				}
			}
			i++;
		}

		for(i = 0; i < C_MFD_AMOUNT; i++)					/* register all the areas */
			HUD_register_area(&mfd_ss_items[i]);
 /* MFD_show_vidmail(1); */
/* MFD_show_fragment(0); */
	}
}

void MFD_deinit(void)
{
	int i = 0;

	MFD_engage_mode(C_MFD_MODE_NONE);

	while (mfd_ss_runtimes[i].rtName)
		i++;
	while (i > 0)
	{
		i--;
		if (mfd_ss_runtimes[i].rtDeinitFunction != NULL)
		{
#ifdef _DEBUG
			printf("MFD_Deinit(): Shutting down %s\n", mfd_ss_runtimes[i].rtName);
#endif
			mfd_ss_runtimes[i].rtDeinitFunction();
		}
	}	
	mfd_layer_deinit();
}

void	MFD_set_detail(byte level)
{
	nDetailLevel = (byte)min(C_MFD_DETAIL_AMOUNT, level);
}

byte	MFD_get_detail()
{
	return nDetailLevel;
}

void mfd_show_item(short nItem, bool bShow)
{
	short i;

	if (nItem < C_MFD_AMOUNT)
	{
		if (nItem < 0)
		{
			for	(i = 0; i < C_MFD_AMOUNT; i++)
				mfd_ss_items[i].haVisible = bShow;	
		}
		else
			mfd_ss_items[nItem].haVisible = bShow;
	}
}

bool	MFD_is_item_visible(short nItem)
{
	bool rCode = false;

	if (nItem >= 0 && nItem < C_MFD_AMOUNT)
		rCode = mfd_ss_items[nItem].haVisible;

	return rCode;
}

void MFD_engage_mode(byte nMode)
{
	short i;

	if (nMode != nCurrentMode)
	{
		assert(nMode < C_MFD_MODE_AMOUNT);
		for (i = 0; i < C_MFD_AMOUNT; i++)
			mfd_show_item(i, bItemVisibility[nMode][i]);
		nCurrentMode = nMode;
	}
}

byte MFD_get_mode(void)
{
	return nCurrentMode;
}

bool	mfd_load_bitmap(g2dbase *pBase, int nChunk, int nSubChunk)
{
	ss_bitmap_header *pBmp = NULL;
	bool rCode = false;

	G2D_reset_base(pBase, false);
	pBmp = Bmp_load_bitmap(nChunk, nSubChunk);
	if (pBmp && pBmp->bits)
	{
		convertHeadToArea(*pBmp, *pBase);
		pBase->pPal = RGB_trans;
		pBase->nPalSize = 0xFF;

		rCode = true;
	}
#ifdef _DEBUG
	else
		printf("mfd_load_bitmap(): Failed to Load 0x%04X 0x%04X\n",
					 nChunk, nSubChunk);
#endif
	free(pBmp);

	return rCode;
}

/*                      */
/* targeting            */
/*                      */

/* begin section targeter */

/* after SPEED ms the target is minimal enclosed */
#define		C_MFD_TRGT_SPEED	400

enum
{
	C_SUB_ENEMY_HUMAN = 0x0000,
	C_SUB_ENEMY_TIGER,

	C_SUB_ENEMY_AVIAN = 0x0003,
	C_SUB_ENEMY_PLANT,
	C_SUB_ENEMY_ZEROG,

	C_SUB_ENEMY_INVIS = 0x0007, /* i'm guessing... */
	C_SUB_ENEMY_MUTANT = 0x0008,

	C_SUB_ENEMY_REPAIR = 0x000A,
	C_SUB_ENEMY_SERV,
	C_SUB_ENEMY_EXEC,

	C_SUB_ENEMY_HOPPER = 0x000E,
	C_SUB_ENEMY_SEC1,
	C_SUB_ENEMY_SEC2,
	C_SUB_ENEMY_AUTO,
	C_SUB_ENEMY_MAINT,
	C_SUB_ENEMY_FLYER,

	C_SUB_ENEMY_DRONE = 0x0015,
	C_SUB_ENEMY_WARR,
	C_SUB_ENEMY_ASS,
	C_SUB_ENEMY_ENFORC,
	C_SUB_ENEMY_ELITE,
	C_SUB_ENEMY_CORTEX,
	C_SUB_ENEMY_MUTCYB,

  C_SUB_ENEMY_CYBDOG = 0x001C,
  C_SUB_ENEMY_CGUARD,
  C_SUB_ENEMY_CREAVR,
  C_SUB_ENEMY_DYNICE,
  C_SUB_ENEMY_CYBKIL,
  C_SUB_ENEMY_CSHODN,
  C_SUB_ENEMY_CGUAR2,

  C_SUB_ENEMY_SHODAN = 0x0023,
  C_SUB_ENEMY_DIEGO

};

typedef struct tag_Enemy
{
	int								 eSubClass;		/* critter class */
	int								 eType;				/* critter type */

	char							*eName;				/* Nice readable name */

	int								 eSubChunk;		/* subchunk id for mfd */
	g2dbase						 eG2DB;				/* the mfd bitmap for this critter */
} Enemy;


Enemy ss_enemies[] = 
{
	{  0,  0, "Human Mutant",						C_SUB_ENEMY_HUMAN,	D_G2D_NULLAREA },	/* 14/0/0 */
	{  0,  1, "Tiger Gorilla Mutant",		C_SUB_ENEMY_TIGER,	D_G2D_NULLAREA },	/* 14/0/1 */

	{  0,  3, "Avian Flyer",						C_SUB_ENEMY_AVIAN,	D_G2D_NULLAREA },	/* 14/0/3 */
	{  0,  4, "Plant Mutant",						C_SUB_ENEMY_PLANT,	D_G2D_NULLAREA },	/* 14/0/4 */
	{  0,  5, "Zero Gravity Creature",	C_SUB_ENEMY_ZEROG,	D_G2D_NULLAREA },	/* 14/0/5 */

	{  0,  7, "Invisible Mutant",				C_SUB_ENEMY_INVIS,	D_G2D_NULLAREA },	/* 14/0/7 */
	{  0,  8, "Human Mutant",						C_SUB_ENEMY_MUTANT, D_G2D_NULLAREA },	/* 14/0/8 */

	{  1,  1, "Repair Bot",							C_SUB_ENEMY_REPAIR, D_G2D_NULLAREA },	/* 14/1/1 */
	{  1,  2, "ServBot",								C_SUB_ENEMY_SERV,		D_G2D_NULLAREA },	/* 14/1/2 */
	{  1,  3, "ExecBot",								C_SUB_ENEMY_EXEC,		D_G2D_NULLAREA },	/* 14/1/3 */

	{  1,  5, "Hopper",									C_SUB_ENEMY_HOPPER, D_G2D_NULLAREA },	/* 14/1/5 */
	{  1,  6, "Security-1 Bot",					C_SUB_ENEMY_SEC1,		D_G2D_NULLAREA },	/* 14/1/6 */
	{  1,  7, "Security-2 Bot",					C_SUB_ENEMY_SEC2,		D_G2D_NULLAREA },	/* 14/1/7 */
	{  1,  8, "Autobomb",								C_SUB_ENEMY_AUTO,		D_G2D_NULLAREA },	/* 14/1/8 */
	{  1,  9, "Maintenance Bot",				C_SUB_ENEMY_MAINT,	D_G2D_NULLAREA },	/* 14/1/9 */
	{  1, 10, "Flyer",									C_SUB_ENEMY_FLYER,	D_G2D_NULLAREA },	/* 14/1/10 */

	{  2,  0, "Cyborg Drone",						C_SUB_ENEMY_DRONE,	D_G2D_NULLAREA },	/* 14/2/0 */
	{  2,  1, "Cyborg Warrior",					C_SUB_ENEMY_WARR,		D_G2D_NULLAREA },	/* 14/2/1 */
	{  2,  2, "Cyborg Assassin",				C_SUB_ENEMY_ASS,		D_G2D_NULLAREA },	/* 14/2/2 */
	{  2,  3, "Cyborg Enforcer",				C_SUB_ENEMY_ENFORC, D_G2D_NULLAREA },	/* 14/2/3 */
	{  2,  4, "Elite Cyborg",						C_SUB_ENEMY_ELITE,	D_G2D_NULLAREA },	/* 14/2/4 */
	{  2,  5, "Cortex Reaver",					C_SUB_ENEMY_CORTEX, D_G2D_NULLAREA },	/* 14/2/5 */
	{  2,  6, "Mutated Cyborg",					C_SUB_ENEMY_MUTCYB, D_G2D_NULLAREA },	/* 14/2/6 */

  {  3,  0, "Cyberdog",               C_SUB_ENEMY_CYBDOG, D_G2D_NULLAREA }, /* 14/3/0 */
  {  3,  1, "Cyberguard",             C_SUB_ENEMY_CGUARD, D_G2D_NULLAREA }, /* 14/3/1 */
  {  3,  2, "Cyber Cortex Reaver",    C_SUB_ENEMY_CREAVR, D_G2D_NULLAREA }, /* 14/3/2 */
  {  3,  3, "Dynamic ICE",            C_SUB_ENEMY_DYNICE, D_G2D_NULLAREA }, /* 14/3/3 */
  {  3,  4, "Cyber Hunter Killer",    C_SUB_ENEMY_CYBKIL, D_G2D_NULLAREA }, /* 14/3/4 */
  {  3,  5, "SHODAN",                 C_SUB_ENEMY_CSHODN, D_G2D_NULLAREA }, /* 14/3/5 */
  {  3,  6, "Cyberguard",             C_SUB_ENEMY_CGUAR2, D_G2D_NULLAREA }, /* 14/3/6 */

  {  4,  0, "SHODAN",                 C_SUB_ENEMY_SHODAN, D_G2D_NULLAREA }, /* 14/4/0 */
  {  4,  1, "Cyborg of Edward Diego", C_SUB_ENEMY_DIEGO,  D_G2D_NULLAREA }, /* 14/4/1 */

	{ -1, -1, NULL,											-1,									D_G2D_NULLAREA }		
};

typedef struct tag_Target
{
	/* basic (rendering) stuff */
	struct tag_map_object		*tpObj;								/* the object of interest */
	bool										 tBoxIsSet;						/* valid coordinates ?		*/
	bounding_box						 tBox;								/* the box on the screen	*/
	bounding_box						 tOrigBox;						/* original box of the bmp*/
	bool										 tRendered;						/* target rendered				*/
	/* event related */
  bool                     tEventFired;         /* TargetNew fired?				*/
	/* effect stuff */
	uint32									 tStartTime;					/* first tick rendered		*/
	uint32									 tLastTime;						/* last (cur) time 				*/

	bool										 tSelected;
	/* output data */
	g2dbase									 tgbBracet;
	g2dbase									 tgbHighlight;
	hudLayer								 thlBracets[8];
	hudLayer								 thlHighlight;
	/* list links */
	struct tag_Target				*tpNext;							/* double linked list			*/
	struct tag_Target				*tpPrev;
} Target;

Target		*pTargetListStart = NULL;							/* list entries						*/
Target		*pTargetListEnd = NULL;
Target		*pTargetLastAccess = NULL;						/* last registered target	*/

bool			 bTrgInit = false;

Enemy *mfd_find_enemy(struct tag_map_object *pObj)
{
	Enemy *pEnemy = NULL;
	int i = 0;
	
	while (ss_enemies[i].eSubClass != -1 && !pEnemy)
	{
		if (pObj->class == 0x0E &&
				ss_enemies[i].eSubClass == pObj->subclass &&
				ss_enemies[i].eType == pObj->type)
			pEnemy = &(ss_enemies[i]);
		i++;
	}

  if (pEnemy == NULL)
  {
    Cant_happen ("mfd_find_enemy(): unknown critter type");
  }
	return pEnemy;
}

void mfd_target_destroy(Target *pTarget)
{
	G2D_reset_base(&pTarget->tgbBracet, true);
	G2D_reset_base(&pTarget->tgbHighlight, true);
	free(pTarget);
}

Target *mfd_target_find(struct tag_map_object *pObj)
{
	Target *pTarget = pTargetListStart, *pFound = NULL;

	while (pTarget && !pFound)	
	{
		if (pTarget->tpObj == pObj)		
			pFound = pTarget;
		pTarget = pTarget->tpNext;
	}

	return pFound;
}

void mfd_target_append(Target *pTarget)
{
	pTarget->tpPrev = pTargetListEnd;
	if (pTargetListEnd)
		pTargetListEnd->tpNext = pTarget;
	pTargetListEnd = pTarget;
	if (!pTargetListStart)
		pTargetListStart = pTarget;
}

void mfd_target_detach(Target *pTarget)
{
	if (pTarget == pTargetListEnd)
		pTargetListEnd = pTarget->tpPrev;
	if (pTarget == pTargetListStart)
		pTargetListStart = pTarget->tpNext;
	if (pTarget->tpPrev)
	{
		pTarget->tpPrev->tpNext = pTarget->tpNext;
		pTarget->tpPrev = NULL;
	}
	if (pTarget->tpNext)
	{
		pTarget->tpNext->tpPrev = pTarget->tpPrev;
		pTarget->tpNext = NULL;
	}
}

void		MFD_register_target(struct tag_map_object *pObj)
{
	Target *pNew = mfd_target_find(pObj);
	int i;

	if (!pNew)
	{
		pNew = (Target *)malloc(sizeof(Target));
		memset(pNew, 0x00, sizeof(Target));
		pNew->tpObj = pObj;
		for (i = 0; i < 8; i++)
			HUD_reset_layer(&pNew->thlBracets[i], &pNew->tgbBracet);
		for (i = 0; i < 2; i++)
		{
			pNew->thlBracets[0 + i].szSize.cy = pNew->thlBracets[4 + i].szSize.cy = 1;
			pNew->thlBracets[2 + i].szSize.cx = pNew->thlBracets[6 + i].szSize.cx = 1;
		}
		HUD_reset_layer(&pNew->thlHighlight, &pNew->tgbHighlight);
		G2D_init_clrbase(&pNew->tgbHighlight, 2, 2);
		G2D_init_clrbase(&pNew->tgbBracet, 2, 2);

		mfd_target_append(pNew);
/*
		{
			Enemy *pEnemy = mfd_find_enemy(pObj);
			printf("got new enemy %d/%d/%d (%s)\n", 
						 pObj->class, pObj->subclass, pObj->type,
						 (pEnemy ? pEnemy->eName : "Unknown"));
		}
*/
	}
	else
		pNew->tBoxIsSet = false;		
	pTargetLastAccess = pNew;
}

void		MFD_register_bmpbox(const bounding_box *pBox, 
														const bounding_box *pMaxBox)
{
	if (pTargetLastAccess && !pTargetLastAccess->tBoxIsSet)
	{
		pTargetLastAccess->tOrigBox			= *pBox;
		pTargetLastAccess->tBox.left		= max(pBox->left, pMaxBox->left);
		pTargetLastAccess->tBox.top			= max(pBox->top, pMaxBox->top);
		pTargetLastAccess->tBox.right		= min(pBox->right, pMaxBox->right);
		pTargetLastAccess->tBox.bottom	= min(pBox->bottom, pMaxBox->bottom);
		if (pTargetLastAccess->tBox.left < pTargetLastAccess->tBox.right && 
				pTargetLastAccess->tBox.right <= (uint16)getRectWidth(mfd_ss_items[C_MFD_TARGET].haArea) &&
				pTargetLastAccess->tBox.top < pTargetLastAccess->tBox.bottom && 
				pTargetLastAccess->tBox.bottom <= (uint16)getRectHeight(mfd_ss_items[C_MFD_TARGET].haArea))
		{
			pTargetLastAccess->tBoxIsSet = true;
			if (!(pTargetLastAccess->tEventFired))
			{
				Evt_fire_event(Evt_get_event(Evt_get_class("MFD"), "TargetNew"),
											 NULL, false);
				pTargetLastAccess->tEventFired = true;
			}
		}
		pTargetLastAccess = NULL;
	}
}

void		MFD_target_reset_targets()
{
	Target *pLast = pTargetListEnd;

	while (pLast)
	{
		pTargetListEnd = pLast->tpPrev;
		mfd_target_destroy(pLast);
		pLast = pTargetListEnd;
	}
	pTargetListStart = NULL;
	pTargetLastAccess = NULL;
}					
													
bool mfd_rinit_targeting(size szScreen)
{
	int i = 0;

	if (!bTrgInit)
	{
		mfd_ss_items[C_MFD_TARGET].haArea.top = 0;
		mfd_ss_items[C_MFD_TARGET].haArea.bottom = szScreen.cy;
		mfd_ss_items[C_MFD_TARGET].haArea.left = 0;
		mfd_ss_items[C_MFD_TARGET].haArea.right = szScreen.cx;

		while (ss_enemies[i].eSubClass != -1)
		{
			if (ss_enemies[i].eSubChunk != -1)
				mfd_load_bitmap(&ss_enemies[i].eG2DB, CHUNK_MFD_ICON_CRITTERS, 
																							ss_enemies[i].eSubChunk);

			i++;
		}
		
		Timer_register_function("MFD.Targeter", 10, mfd_timer_targeting, NULL);
		bTrgInit = true;
	}

	return bTrgInit;
}

void	mfd_rdeinit_targeting(void)
{
	int i = 0;
	
	if (bTrgInit)
	{
		MFD_target_reset_targets();

		while (ss_enemies[i].eSubClass != -1)
		{
			G2D_reset_base(&ss_enemies[i].eG2DB, true);
			i++;
		}
		bTrgInit = false;
	}
}

void	mfd_timer_targeting(uint32 now, uint32 last, void *pData)
{
	Target *pTarget = pTargetListStart;

	UNUSED(pData);
	while (pTarget)
	{
		if (!pTarget->tStartTime)		
			pTarget->tStartTime = last;
		pTarget->tLastTime = now;

		pTarget = pTarget->tpNext;
	}
}

bool	mfd_mouse_targeting(P_HUD_HOOK_MOUSE)
{
	Target *pTarget = pTargetListStart, *pTarget2 = pTargetListStart;
	bool rCode = false;

	if (bTrgInit)
	{
		switch (action)
		{
			case C_HUD_MOUSE_SELECT:
			{
				while (pTarget && bStart && !rCode)
				{
/*					if (pTarget->tBoxIsSet) */
					{
						if (isPointinRect(pTarget->tBox, x, y))
						{
							while (pTarget2)
							{
								pTarget2->tSelected = false;
								pTarget2 = pTarget2->tpNext;
							}
							pTarget->tSelected = true;
							rCode = true;
						}
					}			
					pTarget = pTarget->tpNext;
				}
				break;
			}
			default:	break;
		}
	}

	return rCode;
}

void	mfd_render_targeting(P_HUD_HOOK_RENDER)
{
	Target *pTarget = pTargetListStart;
	Enemy *pEnemy;
	size szScreen;
	int nTrgVer = Player_get_ver_targeter(0), i;
	byte cRed, cGreen, cBlue;

	if (nTrgVer >= 5 && bTrgInit)
	{
		szScreen.cx = (uint16)getRectWidth(mfd_ss_items[C_MFD_TARGET].haArea);
		szScreen.cy = (uint16)getRectHeight(mfd_ss_items[C_MFD_TARGET].haArea);
		while (pTarget)
		{
			pTarget->tRendered = pTarget->tBoxIsSet;
			if (pTarget->tBoxIsSet)
			{
				cRed = (pTarget->tSelected ? 0xFF : 0x00);
				cGreen = (pTarget->tSelected ? 0x00 : 0x00);
				cBlue = (pTarget->tSelected ? 0x00 : 0xFF);
				
				pEnemy = mfd_find_enemy(pTarget->tpObj);

				pTarget->thlHighlight.ptPos.x = pTarget->tBox.left;
				pTarget->thlHighlight.ptPos.y = pTarget->tBox.top;
				pTarget->thlHighlight.szSize.cx = (uint16)getRectWidth(pTarget->tBox);
				pTarget->thlHighlight.szSize.cy = (uint16)getRectHeight(pTarget->tBox);

				for (i = 0; i < 2; i++)
				{
					pTarget->thlBracets[0 + i].ptPos.y = pTarget->tBox.top;
					pTarget->thlBracets[4 + i].ptPos.y = pTarget->tBox.bottom - 1;
					pTarget->thlBracets[2 + i].ptPos.x = pTarget->tBox.right - 1;
					pTarget->thlBracets[6 + i].ptPos.x = pTarget->tBox.left;
					pTarget->thlBracets[0 + i].szSize.cx = pTarget->thlBracets[4 + i].szSize.cx = 
							pTarget->thlHighlight.szSize.cx / 3;
					pTarget->thlBracets[2 + i].szSize.cy = pTarget->thlBracets[6 + i].szSize.cy = 
							pTarget->thlHighlight.szSize.cy / 3;
				}
				if ((pTarget->tLastTime - pTarget->tStartTime) > C_MFD_TRGT_SPEED)
				{	/* target is fully enclosed */
					pTarget->tgbHighlight.bOp.clr = rgbt(cRed, cGreen, cBlue, 0x10);
					pTarget->tgbBracet.bOp.clr = rgbt(cRed, cGreen, cBlue, 0xFF);
					pTarget->thlBracets[0].ptPos.x = pTarget->thlBracets[5].ptPos.x = pTarget->tBox.left;
					pTarget->thlBracets[1].ptPos.x = pTarget->thlBracets[4].ptPos.x =
							pTarget->tBox.right - pTarget->thlBracets[1].szSize.cx;

					pTarget->thlBracets[2].ptPos.y = pTarget->thlBracets[7].ptPos.y = pTarget->tBox.top;
					pTarget->thlBracets[3].ptPos.y = pTarget->thlBracets[6].ptPos.y = 
							pTarget->tBox.bottom - pTarget->thlBracets[3].szSize.cy;
				}
				else
				{	/* calculate the shifting bracets according to time */
					pTarget->tgbHighlight.bOp.clr = rgbt(cRed, cGreen, cBlue, 
																							 (pTarget->tLastTime - pTarget->tStartTime) * 
																								0x10 / C_MFD_TRGT_SPEED);
					pTarget->tgbBracet.bOp.clr = rgbt(cRed, cGreen, cBlue, 
																					 (pTarget->tLastTime - pTarget->tStartTime) * 
																						0xFF / C_MFD_TRGT_SPEED);
					pTarget->thlBracets[0].ptPos.x = 
							pTarget->thlBracets[5].ptPos.x = pTarget->tBox.left *
										(pTarget->tLastTime - pTarget->tStartTime) / C_MFD_TRGT_SPEED;
					pTarget->thlBracets[1].ptPos.x = pTarget->thlBracets[4].ptPos.x = szScreen.cx - 
							(szScreen.cx - pTarget->tBox.right) *
										(pTarget->tLastTime - pTarget->tStartTime) / C_MFD_TRGT_SPEED;

					pTarget->thlBracets[2].ptPos.y = 
							pTarget->thlBracets[7].ptPos.y = pTarget->tBox.top *
										(pTarget->tLastTime - pTarget->tStartTime) / C_MFD_TRGT_SPEED;
					pTarget->thlBracets[3].ptPos.y = pTarget->thlBracets[6].ptPos.y = szScreen.cy - 
							(szScreen.cy - pTarget->tBox.bottom) *
										(pTarget->tLastTime - pTarget->tStartTime) / C_MFD_TRGT_SPEED;
				}
				pTarget->thlBracets[1].szSize.cx = pTarget->thlBracets[4].szSize.cx = 
						min(szScreen.cx - pTarget->thlBracets[1].ptPos.x, pTarget->thlBracets[1].szSize.cx);
				pTarget->thlBracets[3].szSize.cy = pTarget->thlBracets[6].szSize.cy = 
						min(szScreen.cy - pTarget->thlBracets[3].ptPos.y, pTarget->thlBracets[3].szSize.cy);
				G2Dc_fill_area(&pTarget->tgbHighlight);
				G2Dc_fill_area(&pTarget->tgbBracet);

				appendHudLayer(pLayer, pTarget->thlHighlight);
				for (i = 0; i < 8; i++)
					appendHudLayer(pLayer, pTarget->thlBracets[i]);

				pTarget->tBoxIsSet = false; /* next render step will fill again */
			}

			pTarget = pTarget->tpNext;
		}
	}
}

/*                      */
/* bigmap               */
/*                      */

/* begin section bigmap */

#define C_MFD_BIGMAP_WIDTH		88.0
#define	C_MFD_BIGMAP_HEIGHT		53.0
#define	C_MFD_BIGMAP_TOP			18.0

/* map needs 256 * SPEED msecs to fade in/out */
#define C_BIGMAP_SPEED	 2
												  
/* some general Map data (counts for both BigMap and SMap) */
#define	C_MAP_AMOUNT		16

enum
{
	C_MFD_MAP_STATION,
	C_MFD_MAP_LEVEL0,
	C_MFD_MAP_LEVEL1,
	C_MFD_MAP_LEVEL2,
	C_MFD_MAP_LEVEL3,
	C_MFD_MAP_LEVEL4,
	C_MFD_MAP_LEVEL5,
	C_MFD_MAP_LEVEL6,
	C_MFD_MAP_LEVEL7,
	C_MFD_MAP_LEVEL8,
	C_MFD_MAP_LEVEL9,
	C_MFD_MAP_UNKNOWN,
	C_MFD_MAP_DELTA,
	C_MFD_MAP_ALPHA,
	C_MFD_MAP_BETA,
	C_MFD_MAP_GAMMA,

	C_MFD_MAP_SIDES
};

uint16	nMapSides[C_MAP_AMOUNT] =	
{
	C_MFD_MAP_LEVEL0, C_MFD_MAP_LEVEL1, C_MFD_MAP_LEVEL2, C_MFD_MAP_LEVEL3,
	C_MFD_MAP_LEVEL4, C_MFD_MAP_LEVEL5, C_MFD_MAP_LEVEL6, C_MFD_MAP_LEVEL7,
	C_MFD_MAP_LEVEL8, C_MFD_MAP_LEVEL9, C_MFD_MAP_UNKNOWN,	C_MFD_MAP_DELTA,
	C_MFD_MAP_ALPHA, C_MFD_MAP_BETA, C_MFD_MAP_UNKNOWN, C_MFD_MAP_UNKNOWN
};

char *strMapLevels[C_MAP_AMOUNT] =
{
	"Reactor",
	"Medical",
	"Research",
	"Maintenance",
	"Storage",
	"Flight Deck",
	"Executive",
	"Engineering",
	"Security",
	"Bridge",
	"",
	"Delta Grove",
	"Alpha Grove",
	"Beta Grove",
	"",
	""
};

byte			 nMapLevelRadius[3] = { 1, 3, 6 };
byte			 nMapLevelFlags[3] = 
{	/* this is the 'feature' list of the different versions */
	F_DRAWMAP_MODEL | F_DRAWMAP_ELEV | F_DRAWMAP_SEC,
	F_DRAWMAP_MODEL | F_DRAWMAP_RADIUS | F_DRAWMAP_ELEV | F_DRAWMAP_SEC | 
		F_DRAWMAP_CRIT,
	F_DRAWMAP_MODEL | F_DRAWMAP_RADIUS | F_DRAWMAP_ELEV | F_DRAWMAP_SEC | 
		F_DRAWMAP_CRIT | F_DRAWMAP_RAD | F_DRAWMAP_BIO

};

Vector		 vecBigmap = { 1.0, 0.0, 0.0 };
int				 iBigmap = 0, nStepBigmap = 0;
byte			 transBigmap = 0x00;

g2dbase		 gbBigmap = D_G2D_NULLAREA;
hudLayer	 hlBigmap = D_HUD_NULLLAYER(&gbBigmap);

void mfd_map_refresh()
{
	Vector vecPlayer = Player_get_position(0);
	map_tile *pTile;
	int nDist, nTileX, nTileY, nMapVer = Player_get_ver_navunit(0),
			nSquareX, nSquareY, nStartX, nEndX, nStartY, nEndY;

	if (nMapVer)
	{
		nStartY = (int)(vecPlayer.y / TILE_GRID_SIZE - 0.5) - nMapLevelRadius[nMapVer - 1];
		nStartY = max(0, nStartY);
		nEndY = (int)(vecPlayer.y / TILE_GRID_SIZE + 0.5) + nMapLevelRadius[nMapVer - 1];
		nEndY = min((MAP_HEIGHT - 1), nEndY);
		nStartX = (int)(vecPlayer.x / TILE_GRID_SIZE - 0.5) - nMapLevelRadius[nMapVer - 1];
		nStartX = max(0, nStartX);
		nEndX = (int)(vecPlayer.x / TILE_GRID_SIZE + 0.5) + nMapLevelRadius[nMapVer - 1];
		nEndX = min((MAP_WIDTH - 1), nEndX);

		nSquareY = TILE_GRID_SIZE / 2;
		for (nTileY = 0; nTileY < MAP_HEIGHT; nTileY++)
		{	/* we cant use Start/End for the loop because of calculation of Square */
			if (nTileY >= nStartY && nTileY <= nEndY)
			{	/* only check those that are within range */
				nSquareX = TILE_GRID_SIZE / 2;
				for (nTileX = 0; nTileX < MAP_WIDTH; nTileX++)
				{
					if (nTileX >= nStartX && nTileX <= nEndX)
					{	/* same check here */
						pTile = Map_get_tile((uint16)(nSquareX), (uint16)(nSquareY));
						if (!(pTile->flags & TF_AUTOMAPPED))
						{	/* not yet mapped */
							nDist = (int)sqrt(((int)(nSquareX) - (int)vecPlayer.x) *
																 ((int)(nSquareX) - (int)vecPlayer.x) +
																((int)(nSquareY) - (int)vecPlayer.y) *
																 ((int)(nSquareY) - (int)vecPlayer.y));

							if (pTile->type != TILE_SOLID &&	/* solids can't be mapped */
									(pTile->flags & TF_RENDERED) && /* seen it AND within radius */
									nDist <= (nMapLevelRadius[nMapVer - 1] * TILE_GRID_SIZE))		 
								pTile->flags |= TF_AUTOMAPPED;	/* map it */
						}
					}
					nSquareX += TILE_GRID_SIZE;
				}
			}
			nSquareY += TILE_GRID_SIZE;
		}
	}
}

bool mfd_rinit_bigmap(size szScreen)
{
	Matrix matPlayer = Player_get_orientation(0);

	mfd_ss_items[C_MFD_BIGMAP].haArea.top = 
			(int)((double)(szScreen.cy) * C_MFD_BIGMAP_TOP / 100.0);
	mfd_ss_items[C_MFD_BIGMAP].haArea.bottom = 
			mfd_ss_items[C_MFD_BIGMAP].haArea.top + 
			(int)((double)(szScreen.cy) * C_MFD_BIGMAP_HEIGHT / 100.0);

	mfd_ss_items[C_MFD_BIGMAP].haArea.left = (szScreen.cx / 2) - 
			((int)((double)(szScreen.cx) * C_MFD_BIGMAP_WIDTH / 100.0) / 2);
	mfd_ss_items[C_MFD_BIGMAP].haArea.right = 
			mfd_ss_items[C_MFD_BIGMAP].haArea.left +
			(int)((double)(szScreen.cx) * C_MFD_BIGMAP_WIDTH / 100.0);

	vecBigmap.x = (mfdConfig.mfdBigMapAlignNorth ? 1.0 : matPlayer.x.x);
	vecBigmap.y = (mfdConfig.mfdBigMapAlignNorth ? 0.0 : matPlayer.x.y);
	vecBigmap.z = 0.0;
	vec_normalize(vecBigmap);

	hlBigmap.ptPos.x = mfd_ss_items[C_MFD_BIGMAP].haArea.left;
	hlBigmap.ptPos.y = mfd_ss_items[C_MFD_BIGMAP].haArea.top;
	hlBigmap.szSize.cx = (uint16)getRectWidth(mfd_ss_items[C_MFD_BIGMAP].haArea);
	hlBigmap.szSize.cy = (uint16)getRectHeight(mfd_ss_items[C_MFD_BIGMAP].haArea);
	G2D_init_clrbase(&gbBigmap, hlBigmap.szSize.cx, hlBigmap.szSize.cy);
	Timer_register_function("MFD.Bigmap", 10, mfd_timer_bigmap, NULL);

	return (gbBigmap.bImg.bit != NULL);
}

void mfd_rdeinit_bigmap()
{
	Timer_unregister_function("MFD.Bigmap");
	G2D_reset_base(&gbBigmap, true);
}

void MFD_show_bigmap(bool show)
{
	iBigmap = (show ? 1 : -1);
	Evt_fire_event(Evt_get_event(Evt_get_class("MFD"), "BigmapFade"),
								 (void *)iBigmap, false);
}

void MFD_set_zoom_bigmap(int nLevel)
{
	mfdConfig.mfdBigMapZoom = (byte)min(255, max(0, nLevel));
	Evt_fire_event(Evt_get_event(Evt_get_class("MFD"), "BigmapZoom"),
								 &mfdConfig.mfdBigMapZoom, false);
}

int MFD_get_zoom_bigmap()
{
	return mfdConfig.mfdBigMapZoom;
}

bool MFD_is_bigmap_visible()
{
	return (iBigmap > 0 || (iBigmap == 0 && transBigmap > 0x00));
}

void	mfd_timer_bigmap(uint32 now, uint32 last, void *pData)
{
	Matrix matPlayer = Player_get_orientation(0);
	double diff, dTurn = 250.0 / (double)(min(250, now - last));
	int newStep = nStepBigmap + (iBigmap * (now - last));

	UNUSED(pData);
	if (iBigmap)
	{
		nStepBigmap = min(256 * C_BIGMAP_SPEED - 1, max(0x00, newStep));
		transBigmap = (byte)(nStepBigmap / C_BIGMAP_SPEED % 0x100);
		if ((transBigmap == 0xFF && iBigmap > 0) || 
				(transBigmap == 0x00 && iBigmap < 0))
			iBigmap = 0;
	}
	if (!mfdConfig.mfdBigMapAlignNorth)
		diff = matPlayer.x.x - vecBigmap.x;
	else
		diff = 1.0 - vecBigmap.x;
	vecBigmap.x += (diff / dTurn);
	if (!mfdConfig.mfdBigMapAlignNorth)
		diff = matPlayer.x.y - vecBigmap.y;
	else
		diff = 0.0 - vecBigmap.y;
	vecBigmap.y += (diff / dTurn);
	vec_normalize(vecBigmap);

	mfd_map_refresh();
}

bool	mfd_mouse_bigmap(P_HUD_HOOK_MOUSE)
{
	UNUSED(action);
	UNUSED(bStart);
	UNUSED(x);
	UNUSED(y);

	return false;
}

void	mfd_render_bigmap(P_HUD_HOOK_RENDER)
{
	Vector vecPlayer = Player_get_position(0);
	point ptCenter;
	int nMapVer = Player_get_ver_navunit(0);
	byte nFlags = 0;

	if (transBigmap > 0x00 && nMapVer)
	{
		memset(gbBigmap.bImg.clr, 0x00, 
					 gbBigmap.bSizeImg.cx * gbBigmap.bSizeImg.cy * sizeof(pcolor));

		ptCenter.x = (int)(vecPlayer.x);
		ptCenter.y = (int)(vecPlayer.y);
		gbBigmap.bOp.clr = rgbt(0x00, 0xFF, 0x00, transBigmap);
		nFlags = nMapLevelFlags[nMapVer - 1];
		if (nDetailLevel >= C_MFD_DETAIL_WHOA)
			nFlags |= F_DRAWMAP_NORMBG;
		G2D_draw_map(&gbBigmap, 
								 vecBigmap.x, vecBigmap.y,
								 ptCenter,
								 mfdConfig.mfdBigMapZoom, 
								 nMapLevelRadius[nMapVer - 1], nFlags);
		if (nDetailLevel >= C_MFD_DETAIL_NICE)
		{	
			gbBigmap.bOp.clr = rgbt(0xFF, 0x00, 0x00, transBigmap);
			G2Dc_enclose_area(&gbBigmap, 
												(uint8)(100 - (transBigmap * 95 / 0xFF)), 
												(uint8)(100 - (transBigmap * 90 / 0xFF)));
		}
		appendHudLayer(pLayer, hlBigmap);
	}
}

/*                      */
/* compass              */
/*                      */

#define C_MFD_COMPASS_WIDTH		 100.0
#define	C_MFD_COMPASS_HEIGHT		 5.0
#define	C_MFD_COMPASS_TOP				12.0

#define C_MFD_COMPASS_FONT	0

char		*strCompassTicks[32] = { "N", "<", "-", ">", "NW", "<", "-", ">", 
																 "W", "<", "-", ">", "SW", "<", "-", ">", 
																 "S", "<", "-", ">", "SE", "<", "-", ">", 
																 "E", "<", "-", ">", "NE", "<", "-", ">" };

bool		 bCompassActive = false;

pcolor	 palCompass[3];
g2dbase	 gbCompassTicks[32];
hudLayer hlCompassTicks[32];

bool mfd_rinit_compass(size szScreen)
{
	ss_bitmap_header *pHead;
	int i, nWidth, nHeight;

	mfd_ss_items[C_MFD_COMPASS].haArea.top = 
			(int)((double)(szScreen.cy) * C_MFD_COMPASS_TOP / 100.0);
	mfd_ss_items[C_MFD_COMPASS].haArea.bottom = 
			mfd_ss_items[C_MFD_COMPASS].haArea.top + 
			(int)((double)(szScreen.cy) * C_MFD_COMPASS_HEIGHT / 100.0);
	mfd_ss_items[C_MFD_COMPASS].haArea.left = szScreen.cx -
			(int)((double)(szScreen.cx) * C_MFD_COMPASS_WIDTH / 100.0);
	mfd_ss_items[C_MFD_COMPASS].haArea.right = 
			mfd_ss_items[C_MFD_COMPASS].haArea.left +
			(int)((double)(szScreen.cx) * C_MFD_COMPASS_WIDTH / 100.0);

	palCompass[0] = rgbt(0x00, 0x00, 0x00, 0x00);
	palCompass[1] = rgbt(0xFF, 0xFF, 0x00, 0x20);
	palCompass[2] = rgbt(0xFF, 0xFF, 0x00, 0xFF);
	for (i = 0; i < 32; i++)
	{
		Font_get_size(C_MFD_COMPASS_FONT, strCompassTicks[i], &nWidth, &nHeight);
		pHead = G2D_write_string(&strCompassTicks[i], 1, C_MFD_COMPASS_FONT,
														 F_WSTR_DRAW_BORDER, 0, nWidth + 2);
		convertHeadToArea(*pHead, gbCompassTicks[i]);
		gbCompassTicks[i].pPal = palCompass;
		gbCompassTicks[i].nPalSize = 0x02;
		HUD_reset_layer(&hlCompassTicks[i], &gbCompassTicks[i]);
		hlCompassTicks[i].ptPos.y = mfd_ss_items[C_MFD_COMPASS].haArea.top + 
																(uint16)getRectHeight(mfd_ss_items[C_MFD_COMPASS].haArea) / 2 -
																gbCompassTicks[i].bSizeImg.cy / 2;
		hlCompassTicks[i].szSize.cy = gbCompassTicks[i].bSizeImg.cy;
		free(pHead);
	}

	return true;
}

void	mfd_rdeinit_compass(void)
{
	int i;

	for (i = 0; i < 32; i++)
	{
		G2D_reset_base(&gbCompassTicks[i], true);
		HUD_reset_layer(&hlCompassTicks[i], &gbCompassTicks[i]);
	}
}

void MFD_show_compass(bool bActive)
{
	bCompassActive = bActive;
	MFD_set_HW_activated(C_MFD_HW_COMPASS, bCompassActive);
	Evt_fire_event(Evt_get_event(Evt_get_class("MFD"), "CompassShow"),
								 &bCompassActive, false);
}

bool MFD_is_compass_active()
{
	return bCompassActive;
}

double mfd_get_angle(double cosine, double sine)
{
	double ang = asin(sine) * (180.0 / M_PI);

/* ff1: i copied this from another project of mine; that's years ago*/
/*      and i hope it works. if not, please; someone enlighten me ;)*/
	if(cosine < 0.0)
		ang = 180.0 - ang;
	else
	{
		if (sine < 0.0)
			ang = 360.0 + ang;
	}	

	return ang;
}					

int mfd_angle_to_screen(double front, int width, double target)
{
	double step = (double)(width) / Option_fov, /* pixel per grade */
				 diff = (front - target);
	
	if (diff >= 180.0)
		diff = diff - 360.0;
	if (diff < -180.0)
		diff = diff + 360.0;

	return (int)(((double)(width)) + (step * diff) + 0.5);
}

void	mfd_render_compass(P_HUD_HOOK_RENDER)
{
	Matrix matPlayer = Player_get_orientation(0);
	double angle = mfd_get_angle(matPlayer.x.x, matPlayer.x.y);
	size szArea;
	int nMapVer = Player_get_ver_navunit(0);
	int i, steps_level[3] = { 4, 2, 1 }, iOutput;

	if (bCompassActive && nMapVer)
	{
		szArea.cx = (uint16)getRectWidth(mfd_ss_items[C_MFD_COMPASS].haArea);
		szArea.cy = (uint16)getRectHeight(mfd_ss_items[C_MFD_COMPASS].haArea);

		for (i = 0; i < 32; i += steps_level[nMapVer - 1])
		{
			iOutput = mfd_angle_to_screen(angle, szArea.cx,
																		11.25 * (double)(i)) - 
																		(gbCompassTicks[i].bSizeImg.cx / 2);
			if ((iOutput + gbCompassTicks[i].bSizeImg.cx) > 0 && 
					iOutput < szArea.cx)
			{
				hlCompassTicks[i].ptPos.x = max(iOutput, 0);
				if (iOutput < 0)
					gbCompassTicks[i].bBoxArea.left = abs(iOutput);
				else
					gbCompassTicks[i].bBoxArea.left = 0;
				gbCompassTicks[i].bBoxArea.right = gbCompassTicks[i].bSizeImg.cx;
				gbCompassTicks[i].bBoxArea.right = 
						min(szArea.cx - iOutput, gbCompassTicks[i].bBoxArea.right);
				hlCompassTicks[i].szSize.cx = (uint16)getRectWidth(gbCompassTicks[i].bBoxArea);
				appendHudLayer(pLayer, hlCompassTicks[i]);
			}
		}
	}
}

/*                      */
/* HW buttons	          */
/*                      */

#define C_MFD_HW_WIDTH		 5.0
#define	C_MFD_HW_HEIGHT		45.0
#define	C_MFD_HW_TOP			25.0

typedef struct
{
	bool							 hwInstalled;
	bool							 hwActivated;

	hudLayer					 hwLayer;
	g2dbase						 hwBase[2];
	
} mfd_hw_socket;

char *strHWNames[C_MFD_HW_SOCKETS] =
{
	"BioMonitor",
	"???",
	"SensAround",
	"Lantern",
	"Shield",
	"InfraRed",
	"Compass",
	"eMail",
	"Turbo Booster",
	"Jump Jets"
};

mfd_hw_socket			hwSockets[C_MFD_HW_SOCKETS];

pcolor						palHWButtons[0x100];
g2dbase					  gbSocketEmpty = D_G2D_NULLAREA;

bool mfd_rinit_HW_but(size szScreen)
{
	size szBuffer, szSocket, szResult;
	int i, j, offy;
	uint16 subchunk = 0x0001;

	Pal_copy_palette(palHWButtons, RGB_trans, 0x40);
	palHWButtons[0] = 0x00000000;

	memset(hwSockets, 0x00, sizeof(mfd_hw_socket) * C_MFD_HW_SOCKETS);
	mfd_load_bitmap(&gbSocketEmpty, CHUNK_SCREEN_SIDEBAR, 0x0000);
	for (i = 0; i < C_MFD_HW_SOCKETS; i++)
	{
		for (j = 0; j < 2; j++)
		{
			mfd_load_bitmap(&hwSockets[i].hwBase[j], CHUNK_SCREEN_SIDEBAR, subchunk++);
			hwSockets[i].hwBase[j].pPal = palHWButtons;
		}
		HUD_reset_layer(&hwSockets[i].hwLayer, NULL);
	}

	mfd_rinit_LHW_but(szScreen);
	mfd_rinit_RHW_but(szScreen);

	offy = getRectHeight(mfd_ss_items[C_MFD_LHW].haArea) / 5;
	szBuffer.cx = (uint16)(getRectWidth(mfd_ss_items[C_MFD_LHW].haArea));
	szBuffer.cy = (uint16)(getRectHeight(mfd_ss_items[C_MFD_LHW].haArea));
	szSocket.cx = (uint16)(getRectWidth(gbSocketEmpty.bBoxArea));
	szSocket.cy = (uint16)(getRectHeight(gbSocketEmpty.bBoxArea));
	G2D_zoom_rect(&szSocket, &szBuffer, &szResult);	

	for (i = 0; i < C_MFD_HW_IR; i++)
	{
		hwSockets[i].hwLayer.ptPos.x = 
					mfd_ss_items[C_MFD_LHW].haArea.left +	szBuffer.cx / 2 - szResult.cx / 2;
		hwSockets[i].hwLayer.szSize.cx = szResult.cx;
		hwSockets[i].hwLayer.ptPos.y = 
				mfd_ss_items[C_MFD_LHW].haArea.top + offy * i + (offy / 2 - szResult.cy / 2);
		hwSockets[i].hwLayer.szSize.cy = szResult.cy;
	}
	for (i = C_MFD_HW_IR; i < C_MFD_HW_SOCKETS; i++)
	{
		hwSockets[i].hwLayer.ptPos.x = 
					mfd_ss_items[C_MFD_RHW].haArea.left +	szBuffer.cx / 2 - szResult.cx / 2;
		hwSockets[i].hwLayer.szSize.cx = szResult.cx;
		hwSockets[i].hwLayer.ptPos.y = mfd_ss_items[C_MFD_RHW].haArea.top + 
				offy * (i - C_MFD_HW_IR) + (offy / 2 - szResult.cy / 2);
		hwSockets[i].hwLayer.szSize.cy = szResult.cy;
	}

	return true;
}

void mfd_rdeinit_HW_but(void)
{
	int i;

	G2D_reset_base(&gbSocketEmpty, true);
	for (i = 0; i < C_MFD_HW_SOCKETS; i++)
	{
		G2D_reset_base(&hwSockets[i].hwBase[0], true);
		G2D_reset_base(&hwSockets[i].hwBase[1], true);
	}
}

void		MFD_set_HW_installed(int hw, bool installed)
{
	if (hw >= 0 && hw < C_MFD_HW_SOCKETS)
	{
		hwSockets[hw].hwInstalled = installed;
		if (!installed)
			MFD_set_HW_activated(hw, false);
		else
			Evt_fire_event(Evt_get_event(Evt_get_class("MFD"), "HWInstall"),
										 NULL, false);
	}
}

bool		MFD_is_HW_installed(int hw)
{
	bool rCode = false;

	if (hw >= 0 && hw < C_MFD_HW_SOCKETS)
		rCode = hwSockets[hw].hwInstalled;

	return rCode;
}

void		MFD_set_HW_activated(int hw, bool activated)
{
	assert(hw >= 0 && hw < C_MFD_HW_SOCKETS);
	assert(MFD_is_HW_installed(hw));

	hwSockets[hw].hwActivated = activated;
	Evt_fire_event(Evt_get_event(Evt_get_class("MFD"), "HWActivate"),
								 (void *)activated, false);
}

bool		MFD_is_HW_activated(int hw)
{
	bool rCode = false;

	if (hw >= 0 && hw < C_MFD_HW_SOCKETS)
		rCode = hwSockets[hw].hwActivated;

	return rCode;
}

/*                      */
/* LHW buttons          */
/*                      */

bool mfd_rinit_LHW_but(size szScreen)
{
	bool bVisible = mfd_ss_items[C_MFD_LHW].haVisible;

	mfd_ss_items[C_MFD_LHW].haVisible = false;
											
	mfd_ss_items[C_MFD_LHW].haArea.left = 0;
	mfd_ss_items[C_MFD_LHW].haArea.right = 
			mfd_ss_items[C_MFD_LHW].haArea.left +
			(int)((double)(szScreen.cx) * C_MFD_HW_WIDTH / 100.0);
	mfd_ss_items[C_MFD_LHW].haArea.top = 
				(int)((double)(szScreen.cy) * C_MFD_HW_TOP / 100.0);
	mfd_ss_items[C_MFD_LHW].haArea.bottom = 
			mfd_ss_items[C_MFD_LHW].haArea.top +
				(int)((double)(szScreen.cy) * C_MFD_HW_HEIGHT / 100.0);
	mfd_ss_items[C_MFD_LHW].haVisible = bVisible;

	return true;
}

bool	mfd_mouse_LHW_but(P_HUD_HOOK_MOUSE)
{
	bounding_box boxTest;
	size szBuffer, szSocket, szResult;
	int nSelectedHW = -1, i, /* starty = 0, */ offy;
	bool rCode = false;

	offy = getRectHeight(mfd_ss_items[C_MFD_LHW].haArea) / 5;
	szBuffer.cx = (uint16)(getRectWidth(mfd_ss_items[C_MFD_LHW].haArea));
	szBuffer.cy = (uint16)(getRectHeight(mfd_ss_items[C_MFD_LHW].haArea));
	szSocket.cx = gbSocketEmpty.bSizeImg.cx;
	szSocket.cy = gbSocketEmpty.bSizeImg.cy;
	G2D_zoom_rect(&szSocket, &szBuffer, &szResult);	

	boxTest.left = szBuffer.cx / 2 - szResult.cx / 2;
	boxTest.right = boxTest.left + szResult.cx;
	for (i = C_MFD_HW_BIO; i <= C_MFD_HW_SHIELD && nSelectedHW < 0; i++)
	{
		boxTest.top = offy * i + (offy / 2 - szResult.cy / 2);
		boxTest.bottom = boxTest.top + szResult.cy;
		if (isPointinRect(boxTest, x, y))
			nSelectedHW = i;
	}

	if (nSelectedHW >= 0)
	{
		boxTest.left += mfd_ss_items[C_MFD_LHW].haArea.left;
		boxTest.right += mfd_ss_items[C_MFD_LHW].haArea.left;
		boxTest.top += mfd_ss_items[C_MFD_LHW].haArea.top;
		boxTest.bottom += mfd_ss_items[C_MFD_LHW].haArea.top;
		switch (action)
		{
			case C_HUD_MOUSE_HOVER:
			{
				if (bStart)
				{
					if (hwSockets[nSelectedHW].hwInstalled)
						HUD_set_description(strHWNames[nSelectedHW],
																rgbt(0x00, 0xFF, 0x00, 0xFF),
																rgbt(0x00, 0x80, 0x00, 0x80),
																&boxTest);
					else
						HUD_set_description("Empty",
																rgbt(0x00, 0xFF, 0x00, 0xFF),
																rgbt(0x00, 0x80, 0x00, 0x80),
																&boxTest);
				}
				break;
			}
			case C_HUD_MOUSE_ACTIVATE:
			case C_HUD_MOUSE_SELECT:
			{
				if (bStart && hwSockets[nSelectedHW].hwInstalled)
				{
					switch (nSelectedHW)
					{
						case C_MFD_HW_BIO: /*	 mfd_set_id(F_SMFD_DATA, F_SMFD_DBIO, 0x0000); */
						case C_MFD_HW_FULLSCREEN:
						{
							break;
						}
						case C_MFD_HW_SENSA:
						{											
							MFD_show_sensaround((MFD_is_sensaround_active() ? false : true));
							break;
						}
						case C_MFD_HW_LIGHT:
						case C_MFD_HW_SHIELD:
						default:	break;
					}
				}
				break;
			}
			case C_HUD_MOUSE_GRAB:
			{	/* show settings of HW */
				if (bStart && hwSockets[nSelectedHW].hwInstalled)
				{

				}
				break;
			}
			default:	break;
		}
		rCode = true;
	}

	return rCode;
}

void	mfd_render_LHW_but(P_HUD_HOOK_RENDER)
{
	int i;

	for (i = C_MFD_HW_BIO; i <= C_MFD_HW_SHIELD; i++)
	{																		
		hwSockets[i].hwLayer.pNext = NULL;
		*pLayer = &(hwSockets[i].hwLayer);
		if (hwSockets[i].hwInstalled)
		{
			if (hwSockets[i].hwActivated)
				hwSockets[i].hwLayer.pG2DB = &(hwSockets[i].hwBase[1]);
			else
				hwSockets[i].hwLayer.pG2DB = &(hwSockets[i].hwBase[0]);
		}
		else
		{
			hwSockets[i].hwLayer.pG2DB = &(gbSocketEmpty);
		}
		pLayer = &(hwSockets[i].hwLayer.pNext);
	}
	*pLayer = NULL;
}

/*                      */
/* RHW buttons          */
/*                      */

bool mfd_rinit_RHW_but(size szScreen)
{
	bool bVisible = mfd_ss_items[C_MFD_LHW].haVisible;

	mfd_ss_items[C_MFD_RHW].haVisible = false;
											
	mfd_ss_items[C_MFD_RHW].haArea.right = szScreen.cx;
	mfd_ss_items[C_MFD_RHW].haArea.left = 
			mfd_ss_items[C_MFD_RHW].haArea.right -
			(int)((double)(szScreen.cx) * C_MFD_HW_WIDTH / 100.0);
	mfd_ss_items[C_MFD_RHW].haArea.top = 
				(int)((double)(szScreen.cy) * C_MFD_HW_TOP / 100.0);
	mfd_ss_items[C_MFD_RHW].haArea.bottom = 
			mfd_ss_items[C_MFD_RHW].haArea.top +
				(int)((double)(szScreen.cy) * C_MFD_HW_HEIGHT / 100.0);
	mfd_ss_items[C_MFD_RHW].haVisible = bVisible;

	return true;
}

bool	mfd_mouse_RHW_but(P_HUD_HOOK_MOUSE)
{
	bounding_box boxTest;
	size szBuffer, szSocket, szResult;
	int nSelectedHW = -1, i, /* starty = 0, */ offy;
	bool rCode = false;

	offy = getRectHeight(mfd_ss_items[C_MFD_RHW].haArea) / 5;
	szBuffer.cx = (uint16)(getRectWidth(mfd_ss_items[C_MFD_RHW].haArea));
	szBuffer.cy = (uint16)(getRectHeight(mfd_ss_items[C_MFD_RHW].haArea));
	szSocket.cx = gbSocketEmpty.bSizeImg.cx;
	szSocket.cy = gbSocketEmpty.bSizeImg.cy;
	G2D_zoom_rect(&szSocket, &szBuffer, &szResult);	

	boxTest.left = szBuffer.cx / 2 - szResult.cx / 2;
	boxTest.right = boxTest.left + szResult.cx;
	for (i = C_MFD_HW_IR; i <= C_MFD_HW_JUMP && nSelectedHW < 0; i++)
	{
		boxTest.top = offy * (i - C_MFD_HW_IR) + (offy / 2 - szResult.cy / 2);
		boxTest.bottom = boxTest.top + szResult.cy;
		if (isPointinRect(boxTest, x, y))
			nSelectedHW = i;
	}

	if (nSelectedHW >= 0)
	{
		boxTest.left += mfd_ss_items[C_MFD_RHW].haArea.left;
		boxTest.right += mfd_ss_items[C_MFD_RHW].haArea.left;
		boxTest.top += mfd_ss_items[C_MFD_RHW].haArea.top;
		boxTest.bottom += mfd_ss_items[C_MFD_RHW].haArea.top;
		switch (action)
		{
			case C_HUD_MOUSE_HOVER:
			{
				if (bStart)
				{
					if (hwSockets[nSelectedHW].hwInstalled)
						HUD_set_description(strHWNames[nSelectedHW],
																rgbt(0x00, 0xFF, 0x00, 0xFF),
																rgbt(0x00, 0x80, 0x00, 0x80),
																&boxTest);
					else
						HUD_set_description("Empty",
																rgbt(0x00, 0xFF, 0x00, 0xFF),
																rgbt(0x00, 0x80, 0x00, 0x80),
																&boxTest);
				}
				break;
			}
			case C_HUD_MOUSE_ACTIVATE:
			case C_HUD_MOUSE_SELECT:
			{
				if (bStart && hwSockets[nSelectedHW].hwInstalled)
				{
					switch (nSelectedHW)
					{
						case C_MFD_HW_IR:
						{

							break;
						}
						case C_MFD_HW_COMPASS:
						{												
							MFD_show_compass((MFD_is_compass_active() ? false : true));
							break;
						}
						case C_MFD_HW_EMAIL:
						case C_MFD_HW_TURBO:
						case C_MFD_HW_JUMP:

						default:	break;
					}
				}
				break;
			}
			case C_HUD_MOUSE_GRAB:
			{	/* show settings of HW */
				if (bStart && hwSockets[nSelectedHW].hwInstalled)
				{

				}
				break;
			}
			default:	break;
		}
		rCode = true;
	}

	return rCode;
}

void	mfd_render_RHW_but(P_HUD_HOOK_RENDER)
{
	int i;

	for (i = C_MFD_HW_IR; i <= C_MFD_HW_JUMP; i++)
	{
		hwSockets[i].hwLayer.pNext = NULL;
		*pLayer = &(hwSockets[i].hwLayer);
		if (hwSockets[i].hwInstalled)
		{
			if (hwSockets[i].hwActivated)
				hwSockets[i].hwLayer.pG2DB = &(hwSockets[i].hwBase[1]);
			else
				hwSockets[i].hwLayer.pG2DB = &(hwSockets[i].hwBase[0]);
		}
		else
		{
			hwSockets[i].hwLayer.pG2DB = &(gbSocketEmpty);
		}
		pLayer = &(hwSockets[i].hwLayer.pNext);
	}
	*pLayer = NULL;
}

/*                      */
/* main mfd functions   */
/*                      */

/* begin section mfd */

#define	C_MFD_MFD_HEIGHT			25.0
#define	C_MFD_MFD_TOP					72.0
#define C_MFD_MFDBUT_SIZE			 2.0
#define	C_MFD_MFD_SPACING			 1.0
#define C_MFD_CMFD_WIDTH			45.0
#define	C_MFD_SIDES						 6
#define	C_MFD_CENTERS					 7

typedef struct tag_MFD_card
{
  char								 cCardName[31];
	uint32							 cCardId;
} MFD_card;

/* MFD_event describes what should be shown when an event occurs						*/
typedef struct tag_MFD_event
{
	char								 eUserName[31];						/* name out of config				*/

	char								 eClassName[C_MAX_EVT_CLASSNAME + 1];
	char								 eEventName[C_MAX_EVT_EVENTNAME + 1];
	HEVENT							 ehEvent;									/* hEvent										*/

	byte								 eControlMode;						/* control mode to set			*/

	uint32							 eCards[3];								/* cards to show						*/
} MFD_event;

/* MFD_priority has both the events and cards priorised											*/
#define	C_MFD_SHOW_MAX_PRIORITY	20

#define	C_MFD_SHOW_TYPE_NONE		0
#define	C_MFD_SHOW_TYPE_EVENT		1
#define	C_MFD_SHOW_TYPE_LAYER		2

typedef struct tag_MFD_priority
{
	int									 prType;									/* C_MFD_SHOW_TYPE_?				*/
	uint32							 prId;										/* depends on seType				*/
} MFD_priority;

enum
{
  C_MFD_EVENT_WEAPEMPTY,
  C_MFD_EVENT_WEAPHEAT,
 
	C_MFD_EVENT_TARGETNEW,
	C_MFD_EVENT_TARGETKILLED,

	C_MFD_EVENT_BODYHEALTH,
	C_MFD_EVENT_BODYENERGY,
	C_MFD_EVENT_BODYHAZARD,
	C_MFD_EVENT_BODYFATIGUE,

	C_MFD_EVENT_LOGMAIL,
	C_MFD_EVENT_LOGLOG,

	C_MFD_EVENT_AMOUNT
};

MFD_event gMFDEvents[C_MFD_EVENT_AMOUNT] =
{
	{ "weapon.empty",			"",			"",						NULL, C_CON_MODE_AMOUNT, { 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF } },
	{ "weapon.overheat",	"",			"",						NULL, C_CON_MODE_AMOUNT, { 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF } },

	{ "target.new",				"MFD",	"TargetNew",	NULL, C_CON_MODE_AMOUNT, { 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF } },
	{ "target.killed",		"",			"",						NULL, C_CON_MODE_AMOUNT, { 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF } },

	{ "body.health",			"",			"",						NULL, C_CON_MODE_AMOUNT, { 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF } },
	{ "body.energy",			"",			"",						NULL, C_CON_MODE_AMOUNT, { 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF } },
	{ "body.hazard",			"",			"",						NULL, C_CON_MODE_AMOUNT, { 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF } },
	{ "body.fatigue",			"",			"",						NULL, C_CON_MODE_AMOUNT, { 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF } },

	{ "log.mail",					"",			"",						NULL, C_CON_MODE_AMOUNT, { 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF } },
	{ "log.log",					"",			"",						NULL, C_CON_MODE_AMOUNT, { 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF } }
};

MFD_card gMFDCCards[] =
{
	{ "none",				F_MFD_NONE			},
	{ "weapons",		F_CMFD_WEAPON		},
	{ "hardware",		F_CMFD_HW				},
	{ "inventory",	F_CMFD_INVENT		},
	{ "software",		F_CMFD_SW				},
	{ "email",			0/*F_CMFD_MEDIA | F_CMFD_MLISTE*/ },
	{ "sensaround", F_CMFD_SENSA		},
	{ "",						F_MFD_UNKNOWN		}
};

MFD_card gMFDSCards[] =
{										
	{ "none",				F_MFD_NONE			},
	{ "weapon",			F_SMFD_WEAPON		},
	{ "item",				F_SMFD_ITEM			},
	{ "map",				mfd_set_id(F_SMFD_MAP, 0x0400, F_SMFD_MMAP)			},
	{ "data",				F_SMFD_DATA			},
/* 	{ "email",			F_SMFD_ITEM | F_SMFD_IMAIL }, */
	{ "target",			F_SMFD_TARGET		},
	{ "sensaround", F_SMFD_SENSA		},
	{ "",						F_MFD_UNKNOWN		}
};

MFD_card *gMFDCards[3] = { gMFDSCards, gMFDCCards, gMFDSCards };

MFD_priority gMFDPriorities[C_MFD_SHOW_MAX_PRIORITY][3];

typedef struct tag_MFD_layer *		HLAYER;

typedef struct tag_MFD_hooks
{
	bool						(*mhRuntimeInit)(void);
	void						(*mhInit)(P_MFD_HOOK_INIT);
	bool						(*mhMouseEvent)(P_MFD_HOOK_MOUSE);
	void						(*mhRenderFunc)(P_MFD_HOOK_RENDER);
	void						(*mhDeinit)(P_MFD_HOOK_DEINIT);
	void						(*mhRuntimeDeinit)(void);
} MFD_hooks;

#define	C_LAYER_STATUS_ACTIVE		0x00000001	
#define	C_LAYER_STATUS_SEMI			0x00000002	
#define	C_LAYER_STATUS_DELAYED	0x00000100 
#define	C_LAYER_STATUS_PUSHED		0x00000200	
#define	C_LAYER_STATUS_FIRED		0x00000003	
#define	C_LAYER_STATUS_FADEOUT	0x80000000

#define	mfd_get_layer_type(l)	 (0x000000FF & l)

typedef struct tag_MFD_layer
{
	struct tag_MFD_layer	*lpPrev;
	struct tag_MFD_layer	*lpNext;
	/* status of the layer */
	uint32								 lStatus;
	uint32								 lStatusTick;
	byte						 			 lShade;
	/* identifications */
	int										 lSide;	/* 0 = left, 1 = center, 2 = right */
	uint32								 lId;
	struct tag_MFD_hooks	 lHooks;
	/* basic layer data */
	g2dbase								 lgbBg;
	hudLayer							 lhlBg;

	void									*lpData;	/* layer specific data										*/
} MFD_layer;

typedef struct tag_MFD_CardRef
{
	HLAYER					crLayer[3];
} MFD_CardRef;

bool		mfd_set_fired_id(int nSide, uint32 nId);

/* layer event system				*/
void	mfd_layer_init(void);
uint32 mfd_layer_find_card(char *strCard, MFD_card *pCards);
int		mfd_layer_get_priority(byte bSide, int nType, uint32 nId);
void	mfd_layer_event(HEVENTCLASS hClass, HEVENT hEvent, void *pData);
void *mfd_layer_request_data(uint32 nId, uint32 fMask);

/* Media Data */

short nMediaRanges[C_MFD_MEDIA_AMOUNT][2] =
{
	{ CHUNK_TEXT_EMAIL_BASE, CHUNK_TEXT_EMAIL_LAST },
	{ CHUNK_TEXT_LOG_BASE, CHUNK_TEXT_LOG_LAST },
	{ CHUNK_TEXT_DATA_BASE, CHUNK_TEXT_DATA_LAST },
	{ CHUNK_TEXT_PAPER_BASE, CHUNK_TEXT_PAPER_LAST },
	{ CHUNK_TEXT_VMAIL_BASE, CHUNK_TEXT_VMAIL_LAST }
};

typedef struct tag_MFD_MediaData
{
	MFD_CardRef				 mdRef;
	
	short							 mdType;
	int								 mdFlagI;
	bool							 mdFlagT;
	int								 mdFlagC;
	int								 mdPics[2];

	bool							 mdSenderSubject;
	int								 mdLines;
	char						 **mdpStrings;
} MFD_MediaData;

typedef struct tag_MFD_CMedia
{
	MFD_MediaData			*cmpData;

	pcolor						 cmpalTxt[0x03];
	g2dbase						 cmgbTxt;
	hudLayer					 cmhlTxt;

	bool							 cmbScrolling;
	int								 cmScrollStart;
	int								 cmScrollOffset;
} MFD_CMedia;

MFD_MediaData *mfd_mediadata_create(short nId);
void mfd_mediadata_destroy(MFD_MediaData *pData);

/* prototypes for the cards */
void	mfd_init_svoid(P_MFD_HOOK_INIT);
void	mfd_timer_svoid(uint32 now, uint32 last, void *pData);
void	mfd_render_svoid(P_MFD_HOOK_RENDER);
void	mfd_deinit_svoid(P_MFD_HOOK_DEINIT);

void	mfd_init_sweapon(P_MFD_HOOK_INIT);
void	mfd_timer_sweapon(uint32 now, uint32 last, void *pData);
bool	mfd_mouse_sweapon(P_MFD_HOOK_MOUSE);
void	mfd_render_sweapon(P_MFD_HOOK_RENDER);
void	mfd_deinit_sweapon(P_MFD_HOOK_DEINIT);

void	mfd_init_sitem(P_MFD_HOOK_INIT);
void	mfd_timer_sitem(uint32 now, uint32 last, void *pData);
bool	mfd_mouse_sitem(P_MFD_HOOK_MOUSE);
void	mfd_render_sitem(P_MFD_HOOK_RENDER);
void	mfd_deinit_sitem(P_MFD_HOOK_DEINIT);

bool	mfd_rinit_smap(void);
void	mfd_init_smap(P_MFD_HOOK_INIT);
void	mfd_timer_smap(uint32 now, uint32 last, void *pData);
bool	mfd_mouse_smap(P_MFD_HOOK_MOUSE);
void	mfd_render_smap(P_MFD_HOOK_RENDER);
void	mfd_deinit_smap(P_MFD_HOOK_DEINIT);
void	mfd_rdeinit_smap(void);

void	mfd_init_starget(P_MFD_HOOK_INIT);
void	mfd_timer_starget(uint32 now, uint32 last, void *pData);
bool	mfd_mouse_starget(P_MFD_HOOK_MOUSE);
void	mfd_render_starget(P_MFD_HOOK_RENDER);
void	mfd_deinit_starget(P_MFD_HOOK_DEINIT);

void	mfd_init_sdata(P_MFD_HOOK_INIT);
void	mfd_timer_sdata(uint32 now, uint32 last, void *pData);
bool	mfd_mouse_sdata(P_MFD_HOOK_MOUSE);
void	mfd_render_sdata(P_MFD_HOOK_RENDER);
void	mfd_deinit_sdata(P_MFD_HOOK_DEINIT);


void	mfd_init_cvoid(P_MFD_HOOK_INIT);
void	mfd_render_cvoid(P_MFD_HOOK_RENDER);
void	mfd_deinit_cvoid(P_MFD_HOOK_DEINIT);

void	mfd_init_cweapon(P_MFD_HOOK_INIT);
bool	mfd_mouse_cweapon(P_MFD_HOOK_MOUSE);
void	mfd_render_cweapon(P_MFD_HOOK_RENDER);
void	mfd_deinit_cweapon(P_MFD_HOOK_DEINIT);

void	mfd_init_chardware(P_MFD_HOOK_INIT);
bool	mfd_mouse_chardware(P_MFD_HOOK_MOUSE);
void	mfd_render_chardware(P_MFD_HOOK_RENDER);
void	mfd_deinit_chardware(P_MFD_HOOK_DEINIT);

void	mfd_init_cinventory(P_MFD_HOOK_INIT);
bool	mfd_mouse_cinventory(P_MFD_HOOK_MOUSE);
void	mfd_render_cinventory(P_MFD_HOOK_RENDER);
void	mfd_deinit_cinventory(P_MFD_HOOK_DEINIT);

void	mfd_init_csoftware(P_MFD_HOOK_INIT);
bool	mfd_mouse_csoftware(P_MFD_HOOK_MOUSE);
void	mfd_render_csoftware(P_MFD_HOOK_RENDER);
void	mfd_deinit_csoftware(P_MFD_HOOK_DEINIT);

void	mfd_init_cmedia(P_MFD_HOOK_INIT);
bool	mfd_mouse_cmedia(P_MFD_HOOK_MOUSE);
void	mfd_render_cmedia(P_MFD_HOOK_RENDER);
void	mfd_deinit_cmedia(P_MFD_HOOK_DEINIT);

uint32		 dwLayerFadeTick[3] = { 0, 0, 0 };
HLAYER		 hLayerFirst[3] = { NULL, NULL, NULL };
HLAYER		 hLayerLast[3] = { NULL, NULL, NULL };

uint32		 dwLayerFireTime = 3000;	/* for # msec fired layer stay up */

MFD_hooks	mfdSideCards[C_MFD_SIDES] = 
{												
	{ NULL,	mfd_init_svoid,		NULL,								mfd_render_svoid,		mfd_deinit_svoid, NULL },
	{ NULL, mfd_init_sweapon, mfd_mouse_sweapon,	mfd_render_sweapon,	mfd_deinit_sweapon, NULL },
	{ NULL, mfd_init_sitem,		mfd_mouse_sitem,		mfd_render_sitem,		mfd_deinit_sitem, NULL },
	{ mfd_rinit_smap, mfd_init_smap, mfd_mouse_smap, mfd_render_smap,	mfd_deinit_smap, mfd_rdeinit_smap },
	{ NULL, mfd_init_starget, mfd_mouse_starget,	mfd_render_starget,	mfd_deinit_starget, NULL },
	{ NULL, mfd_init_sdata,		mfd_mouse_sdata,		mfd_render_sdata,		mfd_deinit_sdata, NULL }
};

MFD_hooks	mfdCenterCards[C_MFD_CENTERS] =
{
	{ NULL, mfd_init_cvoid,	NULL,	mfd_render_cvoid,	mfd_deinit_cvoid, NULL },
	{ NULL, mfd_init_cweapon,	mfd_mouse_cweapon, mfd_render_cweapon, mfd_deinit_cweapon, NULL },
	{ NULL, mfd_init_chardware,	mfd_mouse_chardware, mfd_render_chardware, mfd_deinit_chardware, NULL },
	{ NULL, mfd_init_cinventory,	mfd_mouse_cinventory, mfd_render_cinventory, mfd_deinit_cinventory, NULL },
	{ NULL, mfd_init_csoftware,	mfd_mouse_csoftware, mfd_render_csoftware, mfd_deinit_csoftware, NULL },
	{ NULL, mfd_init_cmedia, mfd_mouse_cmedia, mfd_render_cmedia, mfd_deinit_cmedia, NULL }
};

MFD_hooks *pMFDHooks[3] = { mfdSideCards, mfdCenterCards, mfdSideCards };

g2dbase				 gbMFDBut[3][3] =
{
	{ D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA },
	{ D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA },
	{ D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA }
};
hudLayer			 hlMFDBut[3][3] =
{
	{ D_HUD_NULLLAYER(&gbMFDBut[0][0]), D_HUD_NULLLAYER(&gbMFDBut[0][1]), D_HUD_NULLLAYER(&gbMFDBut[0][2]) },
	{ D_HUD_NULLLAYER(&gbMFDBut[1][0]), D_HUD_NULLLAYER(&gbMFDBut[1][1]), D_HUD_NULLLAYER(&gbMFDBut[1][2]) },
	{ D_HUD_NULLLAYER(&gbMFDBut[2][0]), D_HUD_NULLLAYER(&gbMFDBut[2][1]), D_HUD_NULLLAYER(&gbMFDBut[2][2]) }
};
pcolor				 palMFDBut[0x100], palMFDButSel[2];
bounding_box	 boxSMFDBut[5];
bounding_box	 boxCMFDBut[4];

bool	mfd_set_layer(HLAYER hLayer, struct tag_MFD_hooks *pCard, uint32 id);

void mfd_append_mfd_layer_after(HLAYER hLayer, HLAYER hBefore)
{
	HLAYER hReal = hBefore;

	if (!hReal)
		hReal = hLayerLast[hLayer->lSide];
	hLayer->lpPrev = hReal;
	if (hReal)
	{	/* somewhere to insert */
		if (hReal == hLayerLast[hLayer->lSide])
			hLayerLast[hLayer->lSide] = hLayer;
		hLayer->lpNext = hReal->lpNext;
		hReal->lpNext = hLayer;
		if (hLayer->lpNext)
			hLayer->lpNext->lpPrev = hLayer;
	}
	else
	{	/* nope - nowhere to go. we're the first */
		hLayer->lpNext = NULL;
		hLayerFirst[hLayer->lSide] = hLayerLast[hLayer->lSide] = hLayer;
	}	
}

void mfd_insert_mfd_layer_before(HLAYER hLayer, HLAYER hAfter)
{
	HLAYER hReal = hAfter;

	if (!hReal)
		hReal = hLayerFirst[hLayer->lSide];
	hLayer->lpNext = hReal;
	if (hReal)
	{	/* somewhere to insert */
		if (hReal == hLayerFirst[hLayer->lSide])
			hLayerFirst[hLayer->lSide] = hLayer;
		hLayer->lpPrev = hReal->lpPrev;
		hReal->lpPrev = hLayer;
		if (hLayer->lpPrev)
			hLayer->lpPrev->lpNext = hLayer;
	}
	else
	{	/* nope - nowhere to go. we're the first */
		hLayer->lpPrev = NULL;
		hLayerFirst[hLayer->lSide] = hLayerLast[hLayer->lSide] = hLayer;
	}
}

HLAYER mfd_add_mfd_layer(int nSide)
{
	HLAYER hLayer = NULL;

	if (nSide < C_MFD_SIDE_LEFT || nSide > C_MFD_SIDE_RIGHT)	
		Cant_happen("mfd_add_mfd_layer() nSide out of range!");
	hLayer = (HLAYER)malloc(sizeof(MFD_layer));
	/* reset mem */
	memset(hLayer, 0x00, sizeof(MFD_layer));
	/* init data */
	hLayer->lSide = nSide;
	G2D_reset_base(&hLayer->lgbBg, false);
	HUD_reset_layer(&hLayer->lhlBg, &hLayer->lgbBg);
	/* hook into list */
	mfd_append_mfd_layer_after(hLayer, hLayerLast[nSide]);
/*
	printf(" add layer(%d): %p prev: %p next: %p\n", 
			   hLayer->lSide, hLayer, hLayer->lpPrev, hLayer->lpNext);
*/
	return hLayer;
}

void	mfd_init_basiclayer(P_MFD_HOOK_INIT, pcolor clrBg)
{
	HSHADE shShade = shade_get(hLayer->lShade);
	int nIds[3] = { C_MFD_LEFT, C_MFD_CENTER, C_MFD_RIGHT };

	G2D_init_clrbase(&hLayer->lgbBg, 4, 4);
	hLayer->lgbBg.bOp.clr = rgbt(get_red(clrBg), get_green(clrBg), get_blue(clrBg),
															 shade(shShade, 0x00, get_trans(clrBg)));
	G2Dc_fill_area(&hLayer->lgbBg);
	hLayer->lgbBg.bOp.clr = clrBg;
	
	hLayer->lhlBg.ptPos.x = mfd_ss_items[nIds[hLayer->lSide]].haArea.left;
	hLayer->lhlBg.ptPos.y = mfd_ss_items[nIds[hLayer->lSide]].haArea.top;
	hLayer->lhlBg.szSize.cx = (uint16)getRectWidth(mfd_ss_items[nIds[hLayer->lSide]].haArea);
	hLayer->lhlBg.szSize.cy = (uint16)getRectHeight(mfd_ss_items[nIds[hLayer->lSide]].haArea);
}

void	mfd_prerender_basiclayer(HLAYER hLayer)
{
	HSHADE shShade = shade_get(hLayer->lShade);

	G2D_set_shades(&hLayer->lgbBg, 
								 (byte)shade(shShade, 0x00, get_trans(hLayer->lgbBg.bOp.clr)));
}

void	mfd_deinit_basiclayer(P_MFD_HOOK_DEINIT)
{
	HUD_reset_layer(&hLayer->lhlBg, &hLayer->lgbBg);
	G2D_reset_base(&hLayer->lgbBg, true);
}

void mfd_detach_mfd_layer(HLAYER hLayer)
{
	if (hLayer->lpPrev)
		hLayer->lpPrev->lpNext = hLayer->lpNext;
	if (hLayer->lpNext)
		hLayer->lpNext->lpPrev = hLayer->lpPrev;
	if (hLayerLast[hLayer->lSide] == hLayer)
		hLayerLast[hLayer->lSide] = hLayer->lpPrev;
	if (hLayerFirst[hLayer->lSide] == hLayer)
		hLayerFirst[hLayer->lSide] = hLayer->lpNext;
	hLayer->lpNext = hLayer->lpPrev = NULL;
}

void mfd_remove_mfd_layer(HLAYER hLayer)
{
	if (hLayer->lHooks.mhDeinit)
		hLayer->lHooks.mhDeinit(hLayer);
/*
printf("free layer(%d): %p prev: %p next: %p\n", 
			 hLayer->lSide, hLayer, hLayer->lpPrev, hLayer->lpNext);
*/
	mfd_detach_mfd_layer(hLayer);
	free(hLayer);
}

void mfd_update_layer_flags(HLAYER hLayer, uint32 dwFlags, bool bAdd)
{
	bool bNew = false;

	if (hLayer)
	{
		if (bAdd)
		{
			if ((hLayer->lStatus & dwFlags) != dwFlags)
				bNew = true;
			hLayer->lStatus |= dwFlags;
		}
		else
		{
			if (hLayer->lStatus & dwFlags)
				bNew = true;
			hLayer->lStatus &= ~dwFlags;
		}
		if (bNew)
			hLayer->lStatusTick = 0;
	}
}

void mfd_update_mfd_layer(byte nSide, uint32 dwTick)
{
	HSHADE shShade;
	HLAYER hLayer, hLayerBuf;
	byte bShade, bTarget;
	bool bActiveFound = false, bFiredFound = false, bSemiFound = false;

	/* update topmost layer (clear any flags) */
	if (hLayerLast[nSide] && 
			(hLayerLast[nSide]->lStatus & C_LAYER_STATUS_PUSHED ||
			 hLayerLast[nSide]->lStatus & C_LAYER_STATUS_DELAYED))
		hLayerLast[nSide]->lStatus = 
				mfd_get_layer_type(hLayerLast[nSide]->lStatus);
	/* now browse the layer and shade them */
	hLayer = hLayerLast[nSide];
	bShade = (byte)min(0xFF, ((dwTick - dwLayerFadeTick[nSide]) / 2));
	shShade = shade_get(bShade);
	while (hLayer)
	{
		hLayerBuf = hLayer->lpPrev;
		/* update tick */
		if (!hLayer->lStatusTick)										
			hLayer->lStatusTick = dwTick;
		/* timeout fired layer */
		if (mfd_get_layer_type(hLayer->lStatus) == C_LAYER_STATUS_FIRED && 
				(dwTick - hLayer->lStatusTick) > dwLayerFireTime)
			mfd_update_layer_flags(hLayer, C_LAYER_STATUS_FADEOUT, true);
		/* update fired */
		if (bFiredFound)
		{
			if (mfd_get_layer_type(hLayer->lStatus) == C_LAYER_STATUS_ACTIVE ||
					mfd_get_layer_type(hLayer->lStatus) == C_LAYER_STATUS_SEMI)
				mfd_update_layer_flags(hLayer, C_LAYER_STATUS_PUSHED, true);
			else
				mfd_update_layer_flags(hLayer, C_LAYER_STATUS_FADEOUT, true);
		}
		else
		{
			if (mfd_get_layer_type(hLayer->lStatus) == C_LAYER_STATUS_FIRED &&
					!(hLayer->lStatus & C_LAYER_STATUS_FADEOUT))
				bFiredFound = true;
			if (mfd_get_layer_type(hLayer->lStatus) == C_LAYER_STATUS_ACTIVE)
				mfd_update_layer_flags(hLayer, C_LAYER_STATUS_PUSHED, false);
		}
		/* update semi */
		if (bSemiFound)
		{
			if (mfd_get_layer_type(hLayer->lStatus) == C_LAYER_STATUS_ACTIVE)
				mfd_update_layer_flags(hLayer, C_LAYER_STATUS_DELAYED, true);
			if (mfd_get_layer_type(hLayer->lStatus) == C_LAYER_STATUS_SEMI ||
					mfd_get_layer_type(hLayer->lStatus) == C_LAYER_STATUS_FIRED)
				mfd_update_layer_flags(hLayer, C_LAYER_STATUS_FADEOUT, true);
		}
		else
		{
			if (mfd_get_layer_type(hLayer->lStatus) == C_LAYER_STATUS_SEMI &&
					!(hLayer->lStatus & C_LAYER_STATUS_FADEOUT))
				bSemiFound = true;
			if (mfd_get_layer_type(hLayer->lStatus) == C_LAYER_STATUS_ACTIVE)
				mfd_update_layer_flags(hLayer, C_LAYER_STATUS_DELAYED, false);
		}
		/* update active */
		if (bActiveFound)
		{
			mfd_update_layer_flags(hLayer, C_LAYER_STATUS_FADEOUT, true);
		}
		else
		{
			if (mfd_get_layer_type(hLayer->lStatus) == C_LAYER_STATUS_ACTIVE &&
					!(hLayer->lStatus & C_LAYER_STATUS_FADEOUT))
				bActiveFound = true;
		}

		if (hLayer->lStatus & C_LAYER_STATUS_FADEOUT)
			bTarget = 0x00;
		else
		{
			bTarget = 0xFF;
			if (hLayer->lStatus & C_LAYER_STATUS_PUSHED)
				bTarget = 0x20;
			if (hLayer->lStatus & C_LAYER_STATUS_DELAYED)
				bTarget = 0x00;
		}
		hLayer->lShade = shade(shShade, hLayer->lShade, bTarget);
		if (!hLayer->lShade && (hLayer->lStatus & C_LAYER_STATUS_FADEOUT))
			mfd_remove_mfd_layer(hLayer);
		hLayer = hLayerBuf;
	}
	dwLayerFadeTick[nSide] = dwTick;
}

bool mfd_mouse_mfd_layer(byte action, bool bStart, uint16 x, uint16 y, byte nSide)
{
	bool rCode = false;

	if (hLayerLast[nSide])
	{
		if (hLayerLast[nSide]->lHooks.mhMouseEvent)
		{
			rCode = hLayerLast[nSide]->lHooks.mhMouseEvent(hLayerLast[nSide], 
																										 action, bStart, x, y);
		}
	}

	return rCode;
}

void mfd_render_mfd_layer(hudLayer **pLayer, byte nSide)
{
	HLAYER hLayer;
	hudLayer **pCurLayer = pLayer;

	hLayer = hLayerFirst[nSide];
	while (hLayer)
	{
		if (hLayer->lHooks.mhRenderFunc)
			hLayer->lHooks.mhRenderFunc(hLayer, pCurLayer);
		while (pCurLayer && *pCurLayer && (*pCurLayer)->pNext)
			pCurLayer = &((*pCurLayer)->pNext);
		
		hLayer = hLayer->lpNext;
	}
}

bool	mfd_set_layer(HLAYER hLayer, struct tag_MFD_hooks *pCard, uint32 id)
{
	memcpy(&(hLayer->lHooks), pCard, sizeof(MFD_hooks));
	hLayer->lId = id;
	if (hLayer->lHooks.mhInit)
		hLayer->lHooks.mhInit(hLayer);

	return true;
}

uint32	MFD_get_active_id(int nSide)
{
	HLAYER hLayer;
	uint32 id = F_MFD_UNKNOWN;
	
	assert(nSide >= C_MFD_SIDE_LEFT && nSide <= C_MFD_SIDE_RIGHT);
	hLayer = hLayerLast[nSide];
	while (hLayer && id == F_MFD_UNKNOWN)
	{
		if (mfd_get_layer_type(hLayer->lStatus) == C_LAYER_STATUS_ACTIVE ||
				mfd_get_layer_type(hLayer->lStatus) == C_LAYER_STATUS_SEMI)
			id = hLayer->lId;
		hLayer = hLayer->lpPrev;
	}

	return id;
}

void *mfd_layer_request_data(uint32 nId, uint32 fMask)
{	
	HLAYER hLayer = NULL;
	void *pData = NULL;
	int i;

	for (i = C_MFD_SIDE_LEFT; i <= C_MFD_SIDE_RIGHT && !pData; i++)
	{
		hLayer = hLayerLast[i];
		while (hLayer && !pData)
		{
			if ((hLayer->lId & fMask) == (nId & fMask))
				pData = hLayer->lpData;
			hLayer = hLayer->lpPrev;
		}
	}

	return pData;
}

void mfd_layer_addref_data(MFD_CardRef *pRef, HLAYER hLayer, bool bInit)
{
	if (bInit)	
	{ /* currently, init is a simple memset */
		memset(pRef, 0x00, sizeof(MFD_CardRef));
	}
	assert(pRef->crLayer[hLayer->lSide] == NULL);
	pRef->crLayer[hLayer->lSide] = hLayer;
	printf("mfd_layer_addref_data() AddRef %p from Layer %p\n",
				 pRef, hLayer);
}

bool mfd_layer_release_data(MFD_CardRef *pRef, HLAYER hLayer)
{
	bool bFree = false;

	assert(pRef->crLayer[hLayer->lSide] == hLayer);
	pRef->crLayer[hLayer->lSide] = NULL;
	bFree = (!pRef->crLayer[0] && !pRef->crLayer[1] && !pRef->crLayer[2]);
	printf("mfd_layer_release_data() Releasing %p Layer: %p %s\n",
				 pRef, hLayer, (bFree ? "(last)" : ""));
	if (bFree)
	{	/* for deinit currently nothing has to be done */

	}

	return bFree;
}

void mfd_layer_fadeout_data(MFD_CardRef *pRef)
{
	int i;

	for (i = 0; i < 3; i++)
		if (pRef->crLayer[i])
			mfd_update_layer_flags(pRef->crLayer[i], C_LAYER_STATUS_FADEOUT, true);
}

bool mfd_rinit_mfds(size szScreen)
{
	int i, nIndex;
	bool bEmpty, rCode = true;

	/* set sizes for mfd buttons	*/
	mfd_ss_items[C_MFD_LBUT].haArea.left = 0;
	mfd_ss_items[C_MFD_LBUT].haArea.right = 
			(int)((double)(szScreen.cx) * C_MFD_MFDBUT_SIZE / 100.0);
	mfd_ss_items[C_MFD_RBUT].haArea.right = szScreen.cx;
	mfd_ss_items[C_MFD_RBUT].haArea.left = 
			mfd_ss_items[C_MFD_RBUT].haArea.right - 
			(int)((double)(szScreen.cx) * C_MFD_MFDBUT_SIZE / 100.0);
	mfd_ss_items[C_MFD_CBUT].haArea.bottom = szScreen.cy;
	mfd_ss_items[C_MFD_CBUT].haArea.top = 
			mfd_ss_items[C_MFD_CBUT].haArea.bottom - 
			(int)((double)(szScreen.cy) * C_MFD_MFDBUT_SIZE / 100.0);

	mfd_ss_items[C_MFD_CBUT].haArea.left = (szScreen.cx / 2) - 
			((int)((double)(szScreen.cx) * C_MFD_CMFD_WIDTH / 100.0) / 2);
	mfd_ss_items[C_MFD_CBUT].haArea.right = 
			mfd_ss_items[C_MFD_CBUT].haArea.left + 
			(int)((double)(szScreen.cx) * C_MFD_CMFD_WIDTH / 100.0);
	/* set sizes for mfd areas */
	mfd_ss_items[C_MFD_CBUT].haArea.left = 
	mfd_ss_items[C_MFD_CENTER].haArea.left = (szScreen.cx / 2) - 
			((int)((double)(szScreen.cx) * C_MFD_CMFD_WIDTH / 100.0) / 2);
	mfd_ss_items[C_MFD_CBUT].haArea.right =
	mfd_ss_items[C_MFD_CENTER].haArea.right = 
			mfd_ss_items[C_MFD_CENTER].haArea.left + 
			(int)((double)(szScreen.cx) * C_MFD_CMFD_WIDTH / 100.0);

	mfd_ss_items[C_MFD_LEFT].haArea.left = 
			mfd_ss_items[C_MFD_LBUT].haArea.right +
			(int)((double)(szScreen.cx) * C_MFD_MFD_SPACING / 100.0);
	mfd_ss_items[C_MFD_LEFT].haArea.right = 
			mfd_ss_items[C_MFD_CENTER].haArea.left - 
			(int)((double)(szScreen.cx) * C_MFD_MFD_SPACING / 100.0);
	
	mfd_ss_items[C_MFD_RIGHT].haArea.left = 
			mfd_ss_items[C_MFD_CENTER].haArea.right +
			(int)((double)(szScreen.cx) * C_MFD_MFD_SPACING / 100.0);
	mfd_ss_items[C_MFD_RIGHT].haArea.right = 
			mfd_ss_items[C_MFD_RBUT].haArea.left - 
			(int)((double)(szScreen.cx) * C_MFD_MFD_SPACING / 100.0);

	mfd_ss_items[C_MFD_LBUT].haArea.top =
	mfd_ss_items[C_MFD_RBUT].haArea.top =
	mfd_ss_items[C_MFD_LEFT].haArea.top =
	mfd_ss_items[C_MFD_CENTER].haArea.top =
	mfd_ss_items[C_MFD_RIGHT].haArea.top =
			(int)((double)(szScreen.cy) * C_MFD_MFD_TOP / 100.0);

	mfd_ss_items[C_MFD_LBUT].haArea.bottom =
	mfd_ss_items[C_MFD_RBUT].haArea.bottom =
	mfd_ss_items[C_MFD_LEFT].haArea.bottom =
	mfd_ss_items[C_MFD_CENTER].haArea.bottom =
	mfd_ss_items[C_MFD_RIGHT].haArea.bottom = 
			mfd_ss_items[C_MFD_CENTER].haArea.top +
			(int)((double)(szScreen.cy) * C_MFD_MFD_HEIGHT / 100.0);

	for (i = 0; i < 3; i++)
	{
		hlMFDBut[i][0].ptPos.x = mfd_ss_items[C_MFD_LBUT + i].haArea.left;
		hlMFDBut[i][0].ptPos.y = mfd_ss_items[C_MFD_LBUT + i].haArea.top;

		hlMFDBut[i][0].szSize.cx = (uint16)getRectWidth(mfd_ss_items[C_MFD_LBUT + i].haArea);
		hlMFDBut[i][0].szSize.cy = (uint16)getRectHeight(mfd_ss_items[C_MFD_LBUT + i].haArea);
	}

	mfd_load_bitmap(&gbMFDBut[0][0], 0x0259, 0x001A);
	mfd_load_bitmap(&gbMFDBut[1][0], 0x0259, 0x001B);
	mfd_load_bitmap(&gbMFDBut[2][0], 0x0259, 0x001A);

	palMFDButSel[0] = rgbt(0x00, 0x00, 0x00, 0x00);
	palMFDButSel[1] = rgbt(0x00, 0xFF, 0x00, 0x20);
	Pal_copy_palette(palMFDBut, RGB_trans, 0x40);
	palMFDBut[0] = rgbt(0x00, 0x00, 0x00, 0x00);

	for (i = 0; i < 3; i++)
	{
		G2D_init_bitbase(&gbMFDBut[i][1], 2, 2);
		memset(gbMFDBut[i][1].bImg.bit, 0x01, 2 * 2);
		gbMFDBut[i][1].pPal = palMFDButSel;
		gbMFDBut[i][1].nPalSize = 2;
	}

	nIndex = -1;
	bEmpty = true;
	for (i = 0; i < gbMFDBut[0][0].bSizeImg.cy; i++)
	{
		if (!gbMFDBut[0][0].bImg.bit[gbMFDBut[0][0].bSizeImg.cx * i])
			bEmpty = true;
		else
		{
			if (bEmpty)
			{
				bEmpty = false;
				nIndex++;
				boxSMFDBut[nIndex].left = 0;
				boxSMFDBut[nIndex].top = 1 +
						((i * getRectHeight(mfd_ss_items[C_MFD_LBUT].haArea))) / 
								(gbMFDBut[0][0].bSizeImg.cy);
				boxSMFDBut[nIndex].right = getRectWidth(mfd_ss_items[C_MFD_LBUT].haArea);
			}
			boxSMFDBut[nIndex].bottom = 2 +
						((i * getRectHeight(mfd_ss_items[C_MFD_LBUT].haArea))) / 
								(gbMFDBut[0][0].bSizeImg.cy);
		}
	}
	nIndex = -1;
	bEmpty = true;
	for (i = 0; i < gbMFDBut[1][0].bSizeImg.cx; i++)
	{
		if (!gbMFDBut[1][0].bImg.bit[i + gbMFDBut[1][0].bSizeImg.cx * 
																		(gbMFDBut[1][0].bSizeImg.cy - 1)])
			bEmpty = true;
		else
		{
			if (bEmpty)
			{
				bEmpty = false;
				nIndex++;
				boxCMFDBut[nIndex].top = 0;
				boxCMFDBut[nIndex].bottom = getRectHeight(mfd_ss_items[C_MFD_CBUT].haArea);
				boxCMFDBut[nIndex].left = 1 +
						(i * getRectWidth(mfd_ss_items[C_MFD_CBUT].haArea)) / 
								(gbMFDBut[1][0].bSizeImg.cx);
			}
			boxCMFDBut[nIndex].right = 2 +
						(i * getRectWidth(mfd_ss_items[C_MFD_CBUT].haArea)) / 
								(gbMFDBut[1][0].bSizeImg.cx);

		}
	}

	mfd_layer_init();

	for (i = 0; i < C_MFD_SIDES && rCode; i++)
	{
		if (mfdSideCards[i].mhRuntimeInit)
			rCode = mfdSideCards[i].mhRuntimeInit();
	}
	for (i = 0; i < C_MFD_CENTERS && rCode; i++)
	{
		if (mfdCenterCards[i].mhRuntimeInit)
			rCode = mfdCenterCards[i].mhRuntimeInit();
	}

	Timer_register_function("MFD.Layer", 50, mfd_timer_mfd, NULL);
	Timer_register_function("MFD.SWeapon", 50, mfd_timer_sweapon, NULL);
	Timer_register_function("MFD.SItem", 50, mfd_timer_sitem, NULL);
	Timer_register_function("MFD.STarget", 50, mfd_timer_starget, NULL);
	Timer_register_function("MFD.SData", 50, mfd_timer_sdata, NULL);

	return rCode;
}

void	mfd_rdeinit_mfds(void)
{
	int i, j;

	Timer_unregister_function("MFD.SData");
	Timer_unregister_function("MFD.STarget");
	Timer_unregister_function("MFD.SItem");
	Timer_unregister_function("MFD.SWeapon");
	Timer_unregister_function("MFD.Layer");

	for (i = C_MFD_SIDES; i > 0; i--)
	{
		if (mfdSideCards[i - 1].mhRuntimeDeinit)
			mfdSideCards[i - 1].mhRuntimeDeinit();
	}
	for (i = C_MFD_CENTERS; i > 0; i--)
	{
		if (mfdCenterCards[i - 1].mhRuntimeDeinit)
			mfdCenterCards[i - 1].mhRuntimeDeinit();
	}

	for (i = 0; i < 3; i++)
	{
		while (hLayerFirst[i])
			mfd_remove_mfd_layer(hLayerFirst[i]);
		for (j = 0; j < 3; j++)
			G2D_reset_base(&gbMFDBut[i][j], true);
	}
}

void mfd_layer_init(void)
{
	HEVENTCLASS hClass;
	FILE *pCfgFile = NULL;
	char pLineBuffer[256], **pStringList;
	int i, j, nCurIndizes[4] = { 0, 0, 0, 0 }, 
			nCurSection = -1, nPrevSection, nLine = 0, nLength; 
	bool bParsed;

	for (i = 0; i < C_MFD_EVENT_AMOUNT; i++)
	{
		if (strlen(gMFDEvents[i].eClassName) && 
				strlen(gMFDEvents[i].eEventName))
		{
			hClass = Evt_get_class(gMFDEvents[i].eClassName);
			gMFDEvents[i].ehEvent = Evt_get_event(hClass, gMFDEvents[i].eEventName);
			Evt_register(hClass, gMFDEvents[i].ehEvent, mfd_layer_event);
		}
		else
			gMFDEvents[i].ehEvent = NULL;
	}
	memset(gMFDPriorities, 0x00, 
				 sizeof(MFD_priority) * C_MFD_SHOW_MAX_PRIORITY * 3);
	if (strlen(Option_mfd_cfg))
		pCfgFile = fopen(Option_mfd_cfg, "r+t");
	if (pCfgFile)
	{
		memset(pLineBuffer, 0x00, 256);
		fgets(pLineBuffer, 255, pCfgFile);
		while (!feof(pCfgFile))
		{
			nLine++;
			nLength = strlen(pLineBuffer);
			if (nLength > 0 && (pLineBuffer[nLength - 1] == 0x0A || 
													pLineBuffer[nLength - 1] == 0x0D))
				pLineBuffer[nLength - 1] = 0x00;
			if (nLength > 1 && (pLineBuffer[nLength - 2] == 0x0A || 
													pLineBuffer[nLength - 2] == 0x0D))
				pLineBuffer[nLength - 2] = 0x00;
			nLength = strlen(pLineBuffer);
			if (nLength && pLineBuffer[0] != ';')
			{
				nPrevSection = nCurSection;
				bParsed = false;
				if (!strncmp(pLineBuffer, "[events]", 8))
					nCurSection = 3;
				if (!strncmp(pLineBuffer, "[left]", 6))
					nCurSection = C_MFD_SIDE_LEFT;
				if (!strncmp(pLineBuffer, "[center]", 8))
					nCurSection = C_MFD_SIDE_CENTER;
				if (!strncmp(pLineBuffer, "[right]", 7))
					nCurSection = C_MFD_SIDE_RIGHT;
				if (nCurSection != nPrevSection)
					bParsed = true;
				if (nCurSection != -1 && !bParsed)
				{	/* an usable section? */
					if (nCurSection == 3)
					{	/* event section */
						for (i = 0; i < C_MFD_EVENT_AMOUNT && !bParsed; i++)
						{
							if (!strncmp(pLineBuffer, gMFDEvents[i].eUserName,
													 strlen(gMFDEvents[i].eUserName)))
							{	/* find event entry */
								pStringList = Str_extract(pLineBuffer, ";");
								j = 0;
								while (pStringList[j++]); /* count tokens*/
								if (j > 2) /* left card */
									gMFDEvents[i].eCards[0] = mfd_layer_find_card(pStringList[1], gMFDSCards);
								if (j > 3) /* center card */
									gMFDEvents[i].eCards[1] = mfd_layer_find_card(pStringList[2], gMFDCCards);
								if (j > 4) /* right card */
									gMFDEvents[i].eCards[2] = mfd_layer_find_card(pStringList[3], gMFDSCards);
								if (j > 5)
								{	/* Control Mode */
									gMFDEvents[i].eControlMode = atoi(pStringList[4]);
									if (gMFDEvents[i].eControlMode &&
											gMFDEvents[i].eControlMode <= C_CON_MODE_USERSTART)
										gMFDEvents[i].eControlMode += (C_CON_MODE_USERSTART - 1);
									else
										gMFDEvents[i].eControlMode = C_CON_MODE_AMOUNT;
								}
								bParsed = true;
								for (j = 0; pStringList[j]; j++)
									free(pStringList[j]);
								free(pStringList);
							}
						}
					}
					else
					{	/* side priorities */
						gMFDPriorities[nCurIndizes[nCurSection]][nCurSection].prId = 
									mfd_layer_find_card(pLineBuffer, gMFDCards[nCurSection]);
						if (gMFDPriorities[nCurIndizes[nCurSection]][nCurSection].prId ==
								F_MFD_UNKNOWN)
						{	/* it's not a card -- check for events */
							for (i = 0; i < C_MFD_EVENT_AMOUNT && !bParsed; i++)
							{
								if (!strncmp(pLineBuffer, gMFDEvents[i].eUserName,
														 strlen(gMFDEvents[i].eUserName)))
								{	/* find event entry */
									gMFDPriorities[nCurIndizes[nCurSection]][nCurSection].prId = i;
									gMFDPriorities[nCurIndizes[nCurSection]][nCurSection].prType =
											C_MFD_SHOW_TYPE_EVENT;
									bParsed = true;
								}
							}
						}
						else
						{
							gMFDPriorities[nCurIndizes[nCurSection]][nCurSection].prType =
									C_MFD_SHOW_TYPE_LAYER;
							bParsed = true;
						}
					}
					if (!bParsed)
						printf("mfd_layer_init(): couldn't parse MFD cfg Line %d: '%s'\n",
									 nLine, pLineBuffer);
					else
						nCurIndizes[nCurSection]++;
				}
			}
			memset(pLineBuffer, 0x00, 256);
			fgets(pLineBuffer, 255, pCfgFile);
		}
	}
	else
	{	/* well - they have to use my settings then ;) */
		if (strlen(Option_mfd_cfg))
			printf("mfd_layer_init(): MFD Config File '%s' not found, "
						 "defaulting to standards\n", Option_mfd_cfg);
		/* standards: all cards = none as top priority; */
		/* weapon.empty shows weapon;weapons;--					*/
		gMFDEvents[0].eCards[0] = F_SMFD_WEAPON;
		gMFDEvents[0].eCards[1] = F_CMFD_WEAPON;
		gMFDPriorities[0][0].prType = gMFDPriorities[0][1].prType = 
			gMFDPriorities[0][2].prType = C_MFD_SHOW_TYPE_LAYER;
		gMFDPriorities[0][0].prId = gMFDPriorities[0][2].prId = F_MFD_NONE;
		gMFDPriorities[0][1].prId = F_MFD_NONE;
		gMFDPriorities[1][0].prType = 
			gMFDPriorities[1][1].prType = C_MFD_SHOW_TYPE_EVENT;
		gMFDPriorities[1][0].prId = gMFDPriorities[1][1].prId = 0;
	}
}

void mfd_layer_deinit()
{	/* ff1: uhh... */
	printf("mfd_layer_deinit()\n");
}

uint32 mfd_layer_find_card(char *strCard, MFD_card *pCards)
{
	uint32 dwCardId = F_MFD_UNKNOWN;
	int i = 0;

	do
	{
		if (!strcmp(pCards[i].cCardName, strCard))
			dwCardId = pCards[i].cCardId;
	} 
	while (pCards[i++].cCardName[0] && dwCardId == F_MFD_UNKNOWN);

	return dwCardId;
}

int		mfd_layer_get_priority(byte bSide, int nType, uint32 nId)
{
	int rCode = C_MFD_SHOW_MAX_PRIORITY, i;

	for (i = 0; i < C_MFD_SHOW_MAX_PRIORITY && 
							rCode == C_MFD_SHOW_MAX_PRIORITY; i++)
	{
		if (gMFDPriorities[i][bSide].prType == nType)
		{
			switch (nType)
			{
				case C_MFD_SHOW_TYPE_EVENT:
				{
					if (gMFDPriorities[i][bSide].prId == nId)
						rCode = i;
					break;
				}
				case C_MFD_SHOW_TYPE_LAYER:
				{
					if (mfd_get_card(gMFDPriorities[i][bSide].prId) == 
							mfd_get_card(nId))
						rCode = i;
					break;
				}
				default:	break;
			}
		}
	}

	return rCode;
}

void	mfd_layer_event(HEVENTCLASS hClass, HEVENT hEvent, void *pData)
{
	int i, j, nEventPriority[3], nCardPriority[3];
	bool bFound = false;

	UNUSED(hClass);
	UNUSED(pData);
	
	if (hEvent != NULL)	
	{	/* a specific event */
		for (j = 0; j < 3; j++)											/* get priorities of cards */
			nCardPriority[j] = mfd_layer_get_priority((byte)j, 
													C_MFD_SHOW_TYPE_LAYER, MFD_get_active_id(j));
		for (i = 0; i < C_MFD_EVENT_AMOUNT && !bFound; i++)
		{ /* search all events */
			if (gMFDEvents[i].ehEvent == hEvent)
			{
				for (j = 0; j < 3; j++)									/* get priorities of event */
					nEventPriority[j] = mfd_layer_get_priority((byte)j, 
													C_MFD_SHOW_TYPE_EVENT, i);
/*
printf("found event: <%s>\n", gMFDEvents[i].eUserName);
*/
				/* now check if event is allowed to show card and do so */
				if (gMFDEvents[i].eCards[0] != F_MFD_UNKNOWN &&
						nEventPriority[0] < nCardPriority[0])
						mfd_set_fired_id(0, gMFDEvents[i].eCards[0]);
				if (gMFDEvents[i].eCards[1] != F_MFD_UNKNOWN &&
						nEventPriority[1] < nCardPriority[1])
						mfd_set_fired_id(1, gMFDEvents[i].eCards[1]);
				if (gMFDEvents[i].eCards[2] != F_MFD_UNKNOWN &&
						nEventPriority[2] < nCardPriority[2])
						mfd_set_fired_id(2, gMFDEvents[i].eCards[2]);
				if (gMFDEvents[i].eControlMode < C_CON_MODE_AMOUNT)
					Con_activate_user_mode(gMFDEvents[i].eControlMode);

				bFound = true; /* found the event */
			}
		}
	}
}

/*                      */
/* LMFD buttons         */
/*                      */

char *strSMFDBut[5] =
{
	"Weapon",
	"Item",
	"Map",
	"Target",
	"Data"
};

uint32 nSMFDButIds[5] =
{
	F_SMFD_WEAPON,
	F_SMFD_ITEM,
	F_SMFD_MAP,
	F_SMFD_TARGET,
	F_SMFD_DATA
};

bool	mfd_mouse_LMFD_but(P_HUD_HOOK_MOUSE)
{
	bounding_box boxArea;
	bool rCode = false;
	int i, nIndex = -1;

	for (i = 0; i < 5 && nIndex < 0; i++)
		if (isPointinRect(boxSMFDBut[i], x, y))
			nIndex = i;
	if (nIndex >= 0)
	{
		switch (action)
		{
			case C_HUD_MOUSE_HOVER:	
			{
				if (bStart)
				{
					memcpy(&boxArea, &(boxSMFDBut[nIndex]), sizeof(bounding_box));
					boxArea.left += mfd_ss_items[C_MFD_LBUT].haArea.left;
					boxArea.right += mfd_ss_items[C_MFD_LBUT].haArea.left;
					boxArea.top += mfd_ss_items[C_MFD_LBUT].haArea.top;
					boxArea.bottom += mfd_ss_items[C_MFD_LBUT].haArea.top;
					HUD_set_description(strSMFDBut[nIndex], 
															rgbt(0x00, 0x00, 0xFF, 0xFF),
															rgbt(0x00, 0x00, 0x80, 0x80),
															&boxArea);
				}
				break;
			}
			case C_HUD_MOUSE_SELECT:
			case C_HUD_MOUSE_ACTIVATE:
			{
				if (bStart)
				{
/* sound_play_fx_2d(280); */
					MFD_set_active_id(C_MFD_SIDE_LEFT, nSMFDButIds[nIndex], true);
				}
				break;
			}
			case C_HUD_MOUSE_GRAB:
			{	/* hmm... anything planned? */
	
				break;
			}
			default:
							break;
		}
		rCode = true;
	}

	return rCode;
}

void	mfd_render_LMFD_but(P_HUD_HOOK_RENDER)
{
	int nCurSel = -1;

	appendHudLayer(pLayer, hlMFDBut[C_MFD_SIDE_LEFT][0]);

	nCurSel = mfd_get_card(MFD_get_active_id(C_MFD_SIDE_LEFT));
	if (nCurSel >= F_SMFD_WEAPON && nCurSel <= F_SMFD_DATA)
	{
		nCurSel -= F_SMFD_WEAPON;
		hlMFDBut[C_MFD_SIDE_LEFT][1].pG2DB = &gbMFDBut[C_MFD_SIDE_LEFT][1];
		hlMFDBut[C_MFD_SIDE_LEFT][1].ptPos.x = 
				mfd_ss_items[C_MFD_LBUT].haArea.left + boxSMFDBut[nCurSel].left;
		hlMFDBut[C_MFD_SIDE_LEFT][1].ptPos.y = 
				mfd_ss_items[C_MFD_LBUT].haArea.top + boxSMFDBut[nCurSel].top;
		hlMFDBut[C_MFD_SIDE_LEFT][1].szSize.cx = (uint16)getRectWidth(boxSMFDBut[nCurSel]);
		hlMFDBut[C_MFD_SIDE_LEFT][1].szSize.cy = (uint16)getRectHeight(boxSMFDBut[nCurSel]);
		appendHudLayer(pLayer, hlMFDBut[C_MFD_SIDE_LEFT][1]);
	}
	else
		hlMFDBut[C_MFD_SIDE_LEFT][1].pG2DB = NULL;
}

/*                      */
/* CMFD buttons         */
/*                      */

char *strCMFDBut[4] =
{
	"Weapons",
	"Hardware",
	"Inventory",
	"Software"
};

uint32 nCMFDButIds[4] =
{
	F_CMFD_WEAPON,
	F_CMFD_HW,
	F_CMFD_INVENT,
	F_CMFD_SW
};

bool	mfd_mouse_CMFD_but(P_HUD_HOOK_MOUSE)
{
	bounding_box boxArea;
	bool rCode = false;
	int i, nIndex = -1;

	for (i = 0; i < 4 && nIndex < 0; i++)
		if (isPointinRect(boxCMFDBut[i], x, y))
			nIndex = i;
	if (nIndex >= 0)
	{
		switch (action)
		{
			case C_HUD_MOUSE_HOVER:	
			{
				if (bStart)
				{
					memcpy(&boxArea, &(boxCMFDBut[nIndex]), sizeof(bounding_box));
					boxArea.left += mfd_ss_items[C_MFD_CBUT].haArea.left;
					boxArea.right += mfd_ss_items[C_MFD_CBUT].haArea.left;
					boxArea.top += mfd_ss_items[C_MFD_CBUT].haArea.top;
					boxArea.bottom += mfd_ss_items[C_MFD_CBUT].haArea.top;
					HUD_set_description(strCMFDBut[nIndex], 
															rgbt(0x00, 0x00, 0xFF, 0xFF),
															rgbt(0x00, 0x00, 0x80, 0x80),
															&boxArea);
				}
				break;
			}
			case C_HUD_MOUSE_SELECT:
			case C_HUD_MOUSE_ACTIVATE:
			{
				if (bStart)
					MFD_set_active_id(C_MFD_SIDE_CENTER, nCMFDButIds[nIndex], true);
				break;
			}
			case C_HUD_MOUSE_GRAB:
			{	/* hmm... anything planned? */
	
				break;
			}
			default:
							break;
		}
		rCode = true;
	}

	return rCode;
}

void	mfd_render_CMFD_but(P_HUD_HOOK_RENDER)
{
	int nCurSel = -1;

	appendHudLayer(pLayer, hlMFDBut[C_MFD_SIDE_CENTER][0]);

	nCurSel = mfd_get_card(MFD_get_active_id(C_MFD_SIDE_CENTER));
	if (nCurSel >= F_CMFD_WEAPON && nCurSel <= F_CMFD_SW)
	{
		nCurSel -= F_CMFD_WEAPON;
		hlMFDBut[C_MFD_SIDE_CENTER][1].ptPos.x = 
				mfd_ss_items[C_MFD_CBUT].haArea.left + boxCMFDBut[nCurSel].left;
		hlMFDBut[C_MFD_SIDE_CENTER][1].ptPos.y = 
				mfd_ss_items[C_MFD_CBUT].haArea.top + boxCMFDBut[nCurSel].top;
		hlMFDBut[C_MFD_SIDE_CENTER][1].szSize.cx = (uint16)getRectWidth(boxCMFDBut[nCurSel]);
		hlMFDBut[C_MFD_SIDE_CENTER][1].szSize.cy = (uint16)getRectHeight(boxCMFDBut[nCurSel]);
		hlMFDBut[C_MFD_SIDE_CENTER][1].pG2DB = &gbMFDBut[C_MFD_SIDE_CENTER][1];
		appendHudLayer(pLayer, hlMFDBut[C_MFD_SIDE_CENTER][1]);
	}
	else
		hlMFDBut[C_MFD_SIDE_CENTER][1].pG2DB = NULL;
}

/*                      */
/* RMFD buttons         */
/*                      */

bool	mfd_mouse_RMFD_but(P_HUD_HOOK_MOUSE)
{
	bounding_box boxArea;
	bool rCode = false;
	int i, nIndex = -1;

	for (i = 0; i < 5 && nIndex < 0; i++)
		if (isPointinRect(boxSMFDBut[i], x, y))
			nIndex = i;
	if (nIndex >= 0)
	{
		switch (action)
		{
			case C_HUD_MOUSE_HOVER:	
			{
				if (bStart)
				{
					memcpy(&boxArea, &(boxSMFDBut[nIndex]), sizeof(bounding_box));
					boxArea.left += mfd_ss_items[C_MFD_RBUT].haArea.left;
					boxArea.right += mfd_ss_items[C_MFD_RBUT].haArea.left;
					boxArea.top += mfd_ss_items[C_MFD_RBUT].haArea.top;
					boxArea.bottom += mfd_ss_items[C_MFD_RBUT].haArea.top;
					HUD_set_description(strSMFDBut[nIndex], 
															rgbt(0x00, 0x00, 0xFF, 0xFF),
															rgbt(0x00, 0x00, 0x80, 0x80),
															&boxArea);
				}
				rCode = true;
				break;
			}
			case C_HUD_MOUSE_SELECT:
			case C_HUD_MOUSE_ACTIVATE:
			{
				if (bStart)
					MFD_set_active_id(C_MFD_SIDE_RIGHT, nSMFDButIds[nIndex], true);
				break;
			}
			case C_HUD_MOUSE_GRAB:
			{	/* hmm... anything planned? */
	
				break;
			}
			default:
							break;
		}
	}

	return rCode;
}

void	mfd_render_RMFD_but(P_HUD_HOOK_RENDER)
{
	int nCurSel = -1;

	appendHudLayer(pLayer, hlMFDBut[C_MFD_SIDE_RIGHT][0]);

	nCurSel = mfd_get_card(MFD_get_active_id(C_MFD_SIDE_RIGHT));
	if (nCurSel >= F_SMFD_WEAPON && nCurSel <= F_SMFD_DATA)
	{
		nCurSel -= F_SMFD_WEAPON;
		hlMFDBut[C_MFD_SIDE_RIGHT][1].ptPos.x = 
				mfd_ss_items[C_MFD_RBUT].haArea.left + boxSMFDBut[nCurSel].left;
		hlMFDBut[C_MFD_SIDE_RIGHT][1].ptPos.y = 
				mfd_ss_items[C_MFD_RBUT].haArea.top + boxSMFDBut[nCurSel].top;
		hlMFDBut[C_MFD_SIDE_RIGHT][1].szSize.cx = (uint16)getRectWidth(boxSMFDBut[nCurSel]);
		hlMFDBut[C_MFD_SIDE_RIGHT][1].szSize.cy = (uint16)getRectHeight(boxSMFDBut[nCurSel]);
		hlMFDBut[C_MFD_SIDE_RIGHT][1].pG2DB = &gbMFDBut[C_MFD_SIDE_RIGHT][1];
		appendHudLayer(pLayer, hlMFDBut[C_MFD_SIDE_RIGHT][1]);
	}
	else
		hlMFDBut[C_MFD_SIDE_RIGHT][1].pG2DB = NULL;
}


/*                      */
/* side MFDS            */
/*                      */

HLAYER	mfd_show_id(int nSide, uint32 nId, uint32 nShowType)
{
	HLAYER hLayer = NULL;
	bool bOk = true;

	if (!hLayerLast[nSide] || hLayerLast[nSide]->lId != nId)
	{
		hLayer = mfd_add_mfd_layer(nSide);
		if (nSide == C_MFD_SIDE_CENTER)
		{
			switch (mfd_get_card(nId))
			{
				case F_MFD_NONE:		mfd_set_layer(hLayer, &mfdCenterCards[0], nId);
											break;
				case F_CMFD_WEAPON:	mfd_set_layer(hLayer, &mfdCenterCards[1], nId);
											break;
				case F_CMFD_HW:			mfd_set_layer(hLayer, &mfdCenterCards[2], nId);
											break;
				case F_CMFD_INVENT:	mfd_set_layer(hLayer, &mfdCenterCards[3], nId);
											break;
				case F_CMFD_SW:			mfd_set_layer(hLayer, &mfdCenterCards[4], nId);
											break;
				case F_CMFD_MEDIA:	mfd_set_layer(hLayer, &mfdCenterCards[5], nId);
											break;
				default:	
							printf("mfd_show_id: unknown card id %08lX for side %d\n\r", 
										 nId, nSide);
							bOk = false;
							break;
			}
		}
		else
		{
			switch (mfd_get_card(nId))
			{
				case F_MFD_NONE:		mfd_set_layer(hLayer, &mfdSideCards[0], nId);
											break;
				case F_SMFD_WEAPON:	mfd_set_layer(hLayer, &mfdSideCards[1], nId);
											break;
				case F_SMFD_ITEM:		mfd_set_layer(hLayer, &mfdSideCards[2], nId);
											break;
				case F_SMFD_MAP:		mfd_set_layer(hLayer, &mfdSideCards[3], nId);
											break;
				case F_SMFD_TARGET:	mfd_set_layer(hLayer, &mfdSideCards[4], nId);
											break;
				case F_SMFD_DATA:		mfd_set_layer(hLayer, &mfdSideCards[5], nId);
											break;
				default:	
							printf("mfd_show_id: unknown card id %08lX for side %d\n\r", 
										 nId, nSide);
							bOk = false;
							break;
			}
		}
		if (bOk)
			hLayer->lStatus = nShowType;
		else
			mfd_remove_mfd_layer(hLayer);
	}
	else
	{
		if (mfd_get_layer_type(nShowType) != C_LAYER_STATUS_FIRED ||
				mfd_get_layer_type(hLayerLast[nSide]->lStatus) != 
						C_LAYER_STATUS_ACTIVE)
		{
			mfd_update_layer_flags(hLayerLast[nSide], C_LAYER_STATUS_FADEOUT, false);
		}
	}

	return hLayer;
}

bool		mfd_set_fired_id(int nSide, uint32 nId)
{
	bool rCode = false;

	if (mfd_show_id(nSide, nId, C_LAYER_STATUS_FIRED))
	{
		if (nSide == C_MFD_SIDE_CENTER)
			Evt_fire_event(Evt_get_event(Evt_get_class("MFD"), "CenterFired"),
										 (void *)nId, false);
		else
			Evt_fire_event(Evt_get_event(Evt_get_class("MFD"), "SideFired"),
										 (void *)nId, false);
		rCode = true;
	}

	return rCode;
}

bool		MFD_set_active_id(int nSide, uint32 nId, bool bFixed)
{
	bool rCode = false;

	if (mfd_show_id(nSide, nId, 
									(bFixed ? C_LAYER_STATUS_ACTIVE : C_LAYER_STATUS_SEMI)))
	{
		if (nSide == C_MFD_SIDE_CENTER)
			Evt_fire_event(Evt_get_event(Evt_get_class("MFD"), "CenterShow"),
										 (void *)nId, false);
		else
			Evt_fire_event(Evt_get_event(Evt_get_class("MFD"), "SideShow"),
										 (void *)nId, false);
		rCode = true;
	}

	return rCode;
}

void	mfd_timer_mfd(uint32 now, uint32 last, void *pData)
{
	byte i;

	UNUSED(pData);
	UNUSED(last);
	for (i = 0; i < 3; i++)
		mfd_update_mfd_layer(i, now);
}

bool	mfd_mouse_lmfd(P_HUD_HOOK_MOUSE)
{
	return mfd_mouse_mfd_layer(action, bStart, x, y, C_MFD_SIDE_LEFT);
}

bool	mfd_mouse_rmfd(P_HUD_HOOK_MOUSE)
{
	return mfd_mouse_mfd_layer(action, bStart, x, y, C_MFD_SIDE_RIGHT);
}

bool	mfd_mouse_cmfd(P_HUD_HOOK_MOUSE)
{
	return mfd_mouse_mfd_layer(action, bStart, x, y, C_MFD_SIDE_CENTER);
}

void	mfd_render_lmfd(P_HUD_HOOK_RENDER)
{
	mfd_render_mfd_layer(pLayer, C_MFD_SIDE_LEFT);
}

void	mfd_render_cmfd(P_HUD_HOOK_RENDER)
{
	mfd_render_mfd_layer(pLayer, C_MFD_SIDE_CENTER);
}

void	mfd_render_rmfd(P_HUD_HOOK_RENDER)
{
	mfd_render_mfd_layer(pLayer, C_MFD_SIDE_RIGHT);
}

/* begin section smfd */

/* begin section svoid */

typedef struct tag_MFD_side_void
{
	int						 svStep;
	bool					 svExpand;
	size					 svszMax;

	pcolor				 svpalTriOp[0x100];
	g2dbase				 svgbTriOp;
	hudLayer			 svhlTriOp;
} MFD_side_void;

void	mfd_timer_svoid(uint32 now, uint32 last, void *pData)
{
	HLAYER hLayer = (HLAYER)pData;
	MFD_side_void *pVoid = (MFD_side_void *)hLayer->lpData;
	pVoid->svStep += ((int)(now - last) / (pVoid->svExpand ? 5 : -5));
	if (pVoid->svStep > pVoid->svszMax.cx)
	{
		pVoid->svExpand = false;
		pVoid->svStep = pVoid->svszMax.cx;
	}
	if (pVoid->svStep < 0)
	{
		pVoid->svExpand = true;
		pVoid->svStep = 1;
	}
}

void	mfd_init_svoid(P_MFD_HOOK_INIT)
{
	MFD_side_void *pVoid = (MFD_side_void *)malloc(sizeof(MFD_side_void));
	char strTimer[50];

	mfd_init_basiclayer(hLayer, rgbt(0x00, 0x80, 0x00, 0x10));
	memset(pVoid, 0x00, sizeof(MFD_side_void));
	
	HUD_reset_layer(&pVoid->svhlTriOp, &pVoid->svgbTriOp);
	mfd_load_bitmap(&pVoid->svgbTriOp, CHUNK_MFD_GENERAL_STUFF, 
																		 SUBCHUNK_MFD_GEN_TRIOP);	
	pVoid->svgbTriOp.pPal = pVoid->svpalTriOp;
	pVoid->svgbTriOp.nPalSize = 0xFF;
	G2D_zoom_rect(&pVoid->svgbTriOp.bSizeImg, 
								&(hLayer->lhlBg.szSize), &pVoid->svszMax);	
	pVoid->svhlTriOp.szSize.cy = pVoid->svszMax.cy;
	pVoid->svhlTriOp.ptPos.y = hLayer->lhlBg.ptPos.y +
														 hLayer->lhlBg.szSize.cy / 2 - 
														 pVoid->svhlTriOp.szSize.cy / 2;

	hLayer->lpData = pVoid;
	sprintf(strTimer, "MFD.SVoid %p", hLayer);
	Timer_register_function(strTimer, 50, mfd_timer_svoid, hLayer);
}

void	mfd_render_svoid(P_MFD_HOOK_RENDER)
{
	MFD_side_void *pVoid = (MFD_side_void *)hLayer->lpData;
	HSHADE shShade = shade_get(hLayer->lShade);
	int i;

	if (nDetailLevel > C_MFD_DETAIL_BORING && hLayer->lShade)
	{
		mfd_prerender_basiclayer(hLayer);
		appendHudLayer(pLayer, hLayer->lhlBg);

		if (nDetailLevel > C_MFD_DETAIL_NICE)
		{
			pVoid->svhlTriOp.szSize.cx = pVoid->svStep;
			pVoid->svhlTriOp.ptPos.x = hLayer->lhlBg.ptPos.x +
																 hLayer->lhlBg.szSize.cx / 2 -
																 pVoid->svhlTriOp.szSize.cx / 2;

			pVoid->svpalTriOp[0] = 0x00000000;
			for(i = 1; i < 0x100; i++)
				pVoid->svpalTriOp[i] = rgbt(0x00, 0xFF, 0x00, 
																		shade(shShade, 0x00, 0x40));

			appendHudLayer(pLayer, pVoid->svhlTriOp);
		}
	}
}

void mfd_deinit_svoid(P_MFD_HOOK_INIT)
{
	char strTimer[50]	;

	sprintf(strTimer, "MFD.SVoid %p", hLayer);
	Timer_unregister_function(strTimer);
	free(hLayer->lpData);
	hLayer->lpData = NULL;
	mfd_deinit_basiclayer(hLayer);
}


/* begin section sweapon */

void	mfd_init_sweapon(P_MFD_HOOK_INIT)
{
	mfd_init_basiclayer(hLayer, rgbt(0x40, 0x00, 0x00, 0x18));
}

void	mfd_timer_sweapon(uint32 now, uint32 last, void *pData)
{
	UNUSED(now);
	UNUSED(last);
	UNUSED(pData);
}

bool	mfd_mouse_sweapon(P_MFD_HOOK_MOUSE)
{
	UNUSED(hLayer);
	UNUSED(action);
	UNUSED(bStart);
	UNUSED(x);
	UNUSED(y);

	return false;
}

void	mfd_render_sweapon(P_MFD_HOOK_RENDER)
{
	if (hLayer->lShade)
	{
		if (nDetailLevel > C_MFD_DETAIL_WHOA)
		{
			mfd_prerender_basiclayer(hLayer);
			appendHudLayer(pLayer, hLayer->lhlBg);
		}
	}
}

void mfd_deinit_sweapon(P_MFD_HOOK_DEINIT)
{
	mfd_deinit_basiclayer(hLayer);
}

/* begin section sitem */

short nSMediaPicBases[C_MFD_MEDIA_AMOUNT] =
{
	CHUNK_MFD_EMAIL_PORTRAITS,
	CHUNK_MFD_EMAIL_PORTRAITS,
	CHUNK_MFD_EMAIL_PORTRAITS,
	0,
	CHUNK_MFD_EMAIL_PORTRAITS
};

typedef struct tag_MFD_SMedia
{
	MFD_MediaData			*smpData;

	pcolor						 smpalPic[0x100];
	pcolor						 smpalTxt[0x03];
	g2dbase						 smgbPic;
	g2dbase						 smgbSender;
	g2dbase						 smgbSubject;
	hudLayer					 smhlPic;
	hudLayer					 smhlSender;
	hudLayer					 smhlSubject;
} MFD_SMedia;

typedef struct tag_MFD_SFragment
{
	pcolor						 smpalTxt[0x03];
	g2dbase						 smgbTxt;
} MFD_SFragment;

void mfd_sitem_init_media(P_MFD_HOOK_INIT)
{
	ss_bitmap_header *pHead = NULL;
	MFD_CMedia *pCMedia = NULL;
	MFD_SMedia *pMedia = (MFD_SMedia *)malloc(sizeof(MFD_SMedia));
	int nIndex = SIDE2I(hLayer->lSide);
	bool bFirst = false;

	memset(pMedia, 0x00, sizeof(MFD_SMedia));
	pCMedia = 
			(MFD_CMedia *)mfd_layer_request_data(mfd_set_id(F_CMFD_MEDIA, 0x00,
																											mfd_get_info(hLayer->lId)),
																					 F_MFD_MASK_CARD | F_MFD_MASK_INFO);
	if (!pCMedia)
	{	/* we have to load up the data for ourselves. */
		pMedia->smpData = mfd_mediadata_create((short)mfd_get_info_s(hLayer->lId));
		bFirst = true;
	}
	else
		pMedia->smpData = pCMedia->cmpData;
	mfd_layer_addref_data(&pMedia->smpData->mdRef, hLayer, bFirst);

	HUD_reset_layer(&pMedia->smhlPic, &pMedia->smgbPic);
	HUD_reset_layer(&pMedia->smhlSender, &pMedia->smgbSender);
	HUD_reset_layer(&pMedia->smhlSubject, &pMedia->smgbSubject);
	if (pMedia->smpData->mdPics[nIndex])
	{	/* init picture */
		mfd_load_bitmap(&pMedia->smgbPic, 
										nSMediaPicBases[pMedia->smpData->mdType],
										pMedia->smpData->mdPics[nIndex]);
		pMedia->smgbPic.pPal = pMedia->smpalPic;
		pMedia->smgbPic.nPalSize = 0xFF;
		G2D_zoom_rect(&pMedia->smgbPic.bSizeImg, &hLayer->lhlBg.szSize,
									&pMedia->smhlPic.szSize);
		pMedia->smhlPic.ptPos.x = hLayer->lhlBg.ptPos.x + 
				hLayer->lhlBg.szSize.cx / 2 - pMedia->smhlPic.szSize.cx / 2;
		pMedia->smhlPic.ptPos.y = hLayer->lhlBg.ptPos.y + 
				hLayer->lhlBg.szSize.cy / 2 - pMedia->smhlPic.szSize.cy / 2;
	}
	else
		G2D_reset_base(&pMedia->smgbPic, false);
	if (hLayer->lSide == C_MFD_SIDE_LEFT && strlen(pMedia->smpData->mdpStrings[2]))
	{	/* any Sender for the left side? */
		pHead = G2D_write_string(&pMedia->smpData->mdpStrings[2], 
														 1, 1, 
														 F_WSTR_DRAW_BORDER, 0, 
														 hLayer->lhlBg.szSize.cx);
		if (pHead)
		{
			convertHeadToArea(*pHead, pMedia->smgbSender);
			pMedia->smgbSender.pPal = pMedia->smpalTxt;
			pMedia->smgbSender.nPalSize = 0x02;
			pMedia->smhlSender.ptPos = hLayer->lhlBg.ptPos;
			pMedia->smhlSender.szSize.cx = pHead->width;
			pMedia->smhlSender.szSize.cy = pHead->height;
			free(pHead);
			pHead = NULL;
		}
		else
			G2D_reset_base(&pMedia->smgbSender, false);
	}
	else
		G2D_reset_base(&pMedia->smgbSender, false);
	if (hLayer->lSide == C_MFD_SIDE_LEFT && strlen(pMedia->smpData->mdpStrings[3]))
	{	/* any Subject for the left side? */
		pHead = G2D_write_string(&pMedia->smpData->mdpStrings[3], 
														 1, 1, 
														 F_WSTR_DRAW_BORDER, 0, 
														 hLayer->lhlBg.szSize.cx);
		if (pHead)
		{
			convertHeadToArea(*pHead, pMedia->smgbSubject);
			pMedia->smgbSubject.pPal = pMedia->smpalTxt;
			pMedia->smgbSubject.nPalSize = 0x02;
			pMedia->smhlSubject.ptPos.x = hLayer->lhlBg.ptPos.x;
			pMedia->smhlSubject.ptPos.y = hLayer->lhlBg.ptPos.y + 
					hLayer->lhlBg.szSize.cy - pHead->height - 1;
			pMedia->smhlSubject.szSize.cx = pHead->width;
			pMedia->smhlSubject.szSize.cy = pHead->height;
			free(pHead);
			pHead = NULL;
		}
		else
			G2D_reset_base(&pMedia->smgbSubject, false);
	}
	else
		G2D_reset_base(&pMedia->smgbSubject, false);
	hLayer->lpData = pMedia;
}

void mfd_sitem_render_media(P_MFD_HOOK_RENDER)
{
	HSHADE shShade = shade_get(hLayer->lShade);
	MFD_SMedia *pMedia = (MFD_SMedia *)hLayer->lpData;

	if (pMedia->smgbPic.bImg.bit)
	{
		Pal_copy_palette(pMedia->smpalPic, RGB_trans, hLayer->lShade);
		pMedia->smpalPic[0] = rgbt(0x00, 0x00, 0x00, 0x00);
		appendHudLayer(pLayer, pMedia->smhlPic);
	}
	if (hLayer->lSide == C_MFD_SIDE_LEFT)
	{	/* only left side shows Sender & Subject */
		if (pMedia->smpData->mdFlagC)
		{	/* make up color according to flag */
			pMedia->smpalTxt[1] = rgbt(get_red(RGB_trans[pMedia->smpData->mdFlagC]),
																 get_green(RGB_trans[pMedia->smpData->mdFlagC]),
																 get_blue(RGB_trans[pMedia->smpData->mdFlagC]),
																 shade(shShade, 0x00, 0x20));
			pMedia->smpalTxt[2] = rgbt(get_red(RGB_trans[pMedia->smpData->mdFlagC]),
																 get_green(RGB_trans[pMedia->smpData->mdFlagC]),
																 get_blue(RGB_trans[pMedia->smpData->mdFlagC]),
																 shade(shShade, 0x00, 0xFF));
		}
		else
		{
			pMedia->smpalTxt[1] = rgbt(0x00, 0xFF, 0x00, shade(shShade, 0x00, 0x20));
			pMedia->smpalTxt[2] = rgbt(0x00, 0xFF, 0x00, shade(shShade, 0x00, 0xFF));
		}
		if (pMedia->smgbSender.bImg.bit)
			appendHudLayer(pLayer, pMedia->smhlSender);
		if (pMedia->smgbSubject.bImg.bit)
			appendHudLayer(pLayer, pMedia->smhlSubject);
	}
}

void mfd_sitem_deinit_media(P_MFD_HOOK_DEINIT)
{
	MFD_SMedia *pMedia = (MFD_SMedia *)hLayer->lpData;
	
	if (mfd_layer_release_data(&pMedia->smpData->mdRef, hLayer))
	{
		mfd_mediadata_destroy(pMedia->smpData);
		pMedia->smpData = NULL;
	}
	G2D_reset_base(&pMedia->smgbPic, true);
	G2D_reset_base(&pMedia->smgbSender, true);
	G2D_reset_base(&pMedia->smgbSubject, true);

	free(pMedia);
	hLayer->lpData = NULL;
}

void	mfd_init_sitem(P_MFD_HOOK_INIT)
{
	mfd_init_basiclayer(hLayer, rgbt(0x00, 0xFF, 0xFF, 0x10));

	switch (mfd_get_det(hLayer->lId))
	{
		case F_SMFD_IMEDIA:  mfd_sitem_init_media(hLayer);	break;
		default:	break;
	}
}

void	mfd_timer_sitem(uint32 now, uint32 last, void *pData)
{
	UNUSED(now);
	UNUSED(last);
	UNUSED(pData);
}

bool	mfd_mouse_sitem(P_MFD_HOOK_MOUSE)
{
	UNUSED(hLayer);
	UNUSED(action);
	UNUSED(bStart);
	UNUSED(x);
	UNUSED(y);

	return false;
}

void	mfd_render_sitem(P_MFD_HOOK_RENDER)
{
	if (nDetailLevel > C_MFD_DETAIL_BORING && hLayer->lShade)
	{
		if (nDetailLevel > C_MFD_DETAIL_WHOA)
		{
			mfd_prerender_basiclayer(hLayer);
			appendHudLayer(pLayer, hLayer->lhlBg);
		}

		switch (mfd_get_det(hLayer->lId))
		{
			case F_SMFD_IMEDIA:  mfd_sitem_render_media(hLayer, pLayer);	break;
			default:	break;
		}
	}
}

void mfd_deinit_sitem(P_MFD_HOOK_DEINIT)
{
	switch (mfd_get_det(hLayer->lId))
	{
		case F_SMFD_IMEDIA:  mfd_sitem_deinit_media(hLayer);	break;
		default:	break;
	}
	mfd_deinit_basiclayer(hLayer);
}

/* begin section smap */

typedef struct tag_MFD_SMap
{
	Vector					 smOrient;
	bool						 smZooming;
	uint16					 smZoomStart;
	g2dbase					 smgbMap;
	hudLayer				 smhlMap;
} MFD_SMap;

typedef struct tag_MFD_SMapConfig
{
	g2dbase					 smcCheckBox;		
} MFD_SMapConfig;

uint32	nSMapOutline[2] = { 0, 0 };
bool		bSMapZooming[2] = { false, false };
uint16	nSMapZoomStart[2] = { 0, 0 };

bounding_box boxSMapCommands[2][2];

uint32	nSMapCommands[4][2] =
{
	{ 0, 0 },
	{ F_SMFD_MSIDE, F_SMFD_MCONFIG },
	{ F_SMFD_MMAP,	F_SMFD_MCONFIG },
	{ F_SMFD_MMAP,	F_SMFD_MSIDE	 }
};

pcolor				 palSMapSides[2][0x100];
g2dbase				 gbSMapSides[C_MFD_MAP_SIDES] =
{
	D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA, 
	D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA, 
	D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA, 
	D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA
};
g2dbase				 gbSMapSideDisplay[2][2] =
{
	{ D_G2D_NULLAREA, D_G2D_NULLAREA }, 
	{ D_G2D_NULLAREA, D_G2D_NULLAREA }
};

hudLayer			 hlSMapDisplay[2][2] =
{
	{ D_HUD_NULLLAYER(&gbSMapSideDisplay[0][0]), D_HUD_NULLLAYER(&gbSMapSideDisplay[0][1]) }, 
	{ D_HUD_NULLLAYER(&gbSMapSideDisplay[1][0]), D_HUD_NULLLAYER(&gbSMapSideDisplay[1][1]) }
};

char *strSMapCommands[] =
{
	"",
	"Map",
	"Side",
	"Config",
	NULL
};

char *strSMapConfigs[] =
{
	"Orient North",
	"Critter",
	"Security",
	"Radius",
	NULL
};

pcolor palSMapCommands[3];

void mfd_smap_init_map(P_MFD_HOOK_INIT)
{
	Matrix matPlayer = Player_get_orientation(0);
	MFD_SMap *pMap = (MFD_SMap *)malloc(sizeof(MFD_SMap));
	int nIndex = SIDE2I(hLayer->lSide);
	char strTimer[50];

	memset(pMap, 0x00, sizeof(MFD_SMap));
	pMap->smOrient.x = (mfdConfig.mfdSMapAlignNorth[nIndex] ? 1.0 : matPlayer.x.x);
	pMap->smOrient.y = (mfdConfig.mfdSMapAlignNorth[nIndex] ? 0.0 : matPlayer.x.y);
	pMap->smOrient.z = 0.0;
	vec_normalize(pMap->smOrient);
	G2D_init_clrbase(&pMap->smgbMap, 
									 hLayer->lhlBg.szSize.cx, hLayer->lhlBg.szSize.cy);
	HUD_reset_layer(&pMap->smhlMap, &pMap->smgbMap);
	pMap->smhlMap.ptPos = hLayer->lhlBg.ptPos;
	pMap->smhlMap.szSize = hLayer->lhlBg.szSize;

	hLayer->lpData = pMap;
	sprintf(strTimer, "MFD.SMap %p", hLayer);
	Timer_register_function(strTimer, 50, mfd_timer_smap, hLayer);
}

void mfd_smap_deinit_map(P_MFD_HOOK_DEINIT)
{
	MFD_SMap *pMap = (MFD_SMap *)hLayer->lpData;
	char strTimer[50];
	
	sprintf(strTimer, "MFD.SMap %p", hLayer);
	Timer_unregister_function(strTimer);
	G2D_reset_base(&pMap->smgbMap, true);
	free(pMap);
	hLayer->lpData = NULL;
}

bool mfd_rinit_smap(void)
{
	int i, nIds[2] = { C_MFD_LEFT, C_MFD_RIGHT };
	bool rCode = true;

	for (i = 0; i < C_MFD_MAP_SIDES && rCode; i++)
	{
		if (!gbSMapSides[i].bImg.bit)
			rCode = mfd_load_bitmap(&gbSMapSides[i], (uint16)CHUNK_MFD_GENERAL_STUFF,
																							 (uint16)(i + SUBCHUNK_MFD_GEN_SMAP_BASE));
	}
	for (i = 0; i < C_MFD_MAP_SIDES; i++)
	{
		if (i != C_MFD_MAP_STATION)
			G2D_resize_base(&gbSMapSides[i], 
											gbSMapSides[C_MFD_MAP_STATION].bSizeImg.cx,
											gbSMapSides[C_MFD_MAP_STATION].bSizeImg.cy);
	}
	for (i = 0; i < 2; i++)
	{
		hlSMapDisplay[i][0].szSize.cx = getRectWidth(mfd_ss_items[nIds[i]].haArea);
		hlSMapDisplay[i][0].szSize.cy = getRectHeight(mfd_ss_items[nIds[i]].haArea);
		memcpy(&gbSMapSideDisplay[i][0], &gbSMapSides[C_MFD_MAP_STATION], sizeof(g2dbase));
		gbSMapSideDisplay[i][0].pPal = palSMapSides[i];
		gbSMapSideDisplay[i][0].nPalSize = 0xFF;
	}

	return rCode;
}

void	mfd_init_smap(P_MFD_HOOK_INIT)
{
	int nIndex = SIDE2I(hLayer->lSide);

	if (!mfd_get_det(hLayer->lId))
		hLayer->lId |= mfdConfig.mfdSMapLastView[nIndex];
	mfdConfig.mfdSMapLastView[nIndex] = mfd_get_det(hLayer->lId);

	if (!mfd_get_info(hLayer->lId))
		mfd_set_info_s(hLayer->lId, mfdConfig.mfdSMapZoom[nIndex]);
	mfdConfig.mfdSMapZoom[nIndex] = mfd_get_info_s(hLayer->lId);

	mfd_init_basiclayer(hLayer, rgbt(0xFF, 0xFF, 0xFF, 0x18));

	memset(&boxSMapCommands[nIndex][0], 0x00, sizeof(bounding_box));
	memset(&boxSMapCommands[nIndex][1], 0x00, sizeof(bounding_box));
	nSMapOutline[nIndex] = 0;
	palSMapCommands[0] = rgbt(0x00, 0x00, 0x00, 0x00);

	switch (mfd_get_det(hLayer->lId))
	{
		case F_SMFD_MMAP:			mfd_smap_init_map(hLayer);	break;
		case F_SMFD_MSIDE:		break;
		case F_SMFD_MCONFIG:	break;
		default:	break;
	}
}

void	mfd_timer_smap(uint32 now, uint32 last, void *pData)
{
	Matrix matPlayer = Player_get_orientation(0);
	HLAYER hLayer = (HLAYER)pData;
	MFD_SMap *pMap = NULL;
	double diff, dTurn = 250.0 / (double)(min(250, now - last));
	int nIndex = SIDE2I(hLayer->lSide);

	switch (mfd_get_det(hLayer->lId))
	{
		case F_SMFD_MMAP:
		{
			pMap = (MFD_SMap *)hLayer->lpData;

			if (!mfdConfig.mfdSMapAlignNorth[nIndex])
				diff = matPlayer.x.x - pMap->smOrient.x;
			else
				diff = 1.0 - pMap->smOrient.x;
			pMap->smOrient.x += (diff / dTurn);
			if (!mfdConfig.mfdSMapAlignNorth[nIndex])
				diff = matPlayer.x.y - pMap->smOrient.y;
			else
				diff = 0.0 - pMap->smOrient.y;
			pMap->smOrient.y += (diff / dTurn);
			vec_normalize(pMap->smOrient);
			break;
		}
		default:					break;
	}
}

bool	mfd_mouse_smap(P_MFD_HOOK_MOUSE)
{
	int nIndex = SIDE2I(hLayer->lSide);
	uint16 nIds[2] = { C_MFD_LEFT, C_MFD_RIGHT };
	bool rCode = false;

	if (hLayer->lShade)
	{
		switch (action)
		{
			case C_HUD_MOUSE_HOVER:
			{	
				if (bStart && bSMapZooming[SIDE2I(hLayer->lSide)])
				{
					mfdConfig.mfdSMapZoom[nIndex] = 
							min(255, max(0, mfdConfig.mfdSMapZoom[nIndex] +
													 (nSMapZoomStart[nIndex] - y)));
					mfd_set_info_s(hLayer->lId, mfdConfig.mfdSMapZoom[nIndex]);
					nSMapZoomStart[nIndex] = y;
					rCode = true;	
				}
				break;
			}
			case C_HUD_MOUSE_GRAB:
			{
				if (mfd_get_det(hLayer->lId) == F_SMFD_MMAP || !bStart)
				{
					if (bStart)
					{
						HUD_lock_cursor(&(mfd_ss_items[nIds[nIndex]].haArea));
						nSMapZoomStart[nIndex] = y;
					}
					else
					{
						HUD_unlock_cursor();
						nSMapZoomStart[nIndex] = 0;
					}
					bSMapZooming[nIndex] = bStart;
					rCode = true;
				}
				break;
			}
			case C_HUD_MOUSE_SELECT:
			case C_HUD_MOUSE_ACTIVATE:
			{
				if (bStart)
				{
					if (isPointinRect(boxSMapCommands[nIndex][0], x, y))
					{
						mfdConfig.mfdSMapLastView[nIndex] = 
									nSMapCommands[mfd_get_det(hLayer->lId)][0];
						hLayer->lId = mfd_set_id(F_SMFD_MAP, 
																		 mfdConfig.mfdSMapLastView[nIndex],
																		 mfd_get_info_s(hLayer->lId));
						rCode = true;
					}
					if (isPointinRect(boxSMapCommands[nIndex][1], x, y))
					{
						mfdConfig.mfdSMapLastView[nIndex] =
								nSMapCommands[mfd_get_info(hLayer->lId)][1];
						hLayer->lId = mfd_set_id(F_SMFD_MAP, 
																		 mfdConfig.mfdSMapLastView[nIndex],
																		 mfd_get_det(hLayer->lId));
						rCode = true;
					}
					if (mfd_get_det(hLayer->lId) != F_SMFD_MSIDE)
						rCode = true;
				}
				break;
			}
			default:	break;
		}
	}

	return rCode;
}

void	mfd_render_smap(P_MFD_HOOK_RENDER)
{
	Vector vecPlayer = Player_get_position(0);
	MFD_SMap *pMap = NULL;
	point ptCenter;
	int nIndex = SIDE2I(hLayer->lSide), nIds[2] = { C_MFD_LEFT, C_MFD_RIGHT },
			nMapVer = Player_get_ver_navunit(0);
	byte nFlags = 0;

	if (hLayer->lShade && nMapVer)
	{
		mfd_prerender_basiclayer(hLayer);
		appendHudLayer(pLayer, hLayer->lhlBg);

		switch (mfd_get_det(hLayer->lId))
		{
			case F_SMFD_MMAP:
			{
				pMap = (MFD_SMap *)hLayer->lpData;
				ptCenter.x = (int)(vecPlayer.x);
				ptCenter.y = (int)(vecPlayer.y);
				pMap->smgbMap.bOp.clr = rgbt(0x00, 0x00, 0x00, 0x00);
				G2Dc_fill_area(&pMap->smgbMap);
				pMap->smgbMap.bOp.clr = rgbt(0x00, 0xFF, 0x00, hLayer->lShade);
				nFlags = mfdConfig.mfdSMapFlags[nIndex] & nMapLevelFlags[nMapVer - 1];
				if (nDetailLevel >= C_MFD_DETAIL_WHOA)
					nFlags |= F_DRAWMAP_NORMBG;
				G2D_draw_map(&pMap->smgbMap,
										 pMap->smOrient.x, pMap->smOrient.y,
										 ptCenter,
										 (byte)mfd_get_info_s(hLayer->lId), 
										 nMapLevelRadius[nMapVer - 1], 
										 nFlags);
				/* nice effect */
/*
				gbBuffer.bOp.clr = rgbt(0x00, 0xFF, 0x00, shade(shShade, 0x00, 0x40));
				G2Dr_do_static(&gbBuffer, 					
											 (uint16)(nSMapOutline[SIDE2I(hLayer->lSide)] % 
																(getRectHeight(gbScreen->bBoxArea))), 
											 9, 100, 0, 10);
*/
				appendHudLayer(pLayer, pMap->smhlMap);
				break;
			}
			case F_SMFD_MSIDE:
			{
				Pal_copy_palette(palSMapSides[nIndex], RGB_trans, hLayer->lShade);
				palSMapSides[nIndex][0] = rgb(0x00, 0x00, 0x00);
				G2D_zoom_rect(&gbSMapSideDisplay[nIndex][0].bSizeImg, 
											&hLayer->lhlBg.szSize, 
											&hlSMapDisplay[nIndex][0].szSize);
				hlSMapDisplay[nIndex][0].ptPos.x = mfd_ss_items[nIds[nIndex]].haArea.left + 
																					 hLayer->lhlBg.szSize.cx / 2 -
																					 hlSMapDisplay[nIndex][0].szSize.cx / 2;
				hlSMapDisplay[nIndex][0].ptPos.y = mfd_ss_items[nIds[nIndex]].haArea.top + 
																					 hLayer->lhlBg.szSize.cy / 2 -
																					 hlSMapDisplay[nIndex][0].szSize.cy / 2;

				appendHudLayer(pLayer, hlSMapDisplay[nIndex][0]);
	
				memcpy(&(gbSMapSideDisplay[nIndex][1]), 
							 &(gbSMapSides[nMapSides[Option_start_level]]), sizeof(g2dbase));
				gbSMapSideDisplay[nIndex][1].pPal = palSMapSides[nIndex];
				gbSMapSideDisplay[nIndex][1].nPalSize = 0xFF;
				hlSMapDisplay[nIndex][1].ptPos = hlSMapDisplay[nIndex][0].ptPos;
				hlSMapDisplay[nIndex][1].szSize = hlSMapDisplay[nIndex][0].szSize;

				appendHudLayer(pLayer, hlSMapDisplay[nIndex][1]);
/* !!!!! level name */
				break;
			}
			case F_SMFD_MCONFIG:
			{

				break;
			}
			default:break;
		}
	}
#if 0
		nPart = gbBuffer.bSizeImg.cx / 2 - 5;

		Font_get_size(1, strSMapCommands[mfd_get_info_s(nSMapCommands[nCommand][0])],
									&nWidth, &nHeight);
		nWidth += 2;
		nHeight += 4;
		pStrings[0] = strSMapCommands[mfd_get_info_s(nSMapCommands[nCommand][0])];
		pHead = G2D_write_string(pStrings, 1, 1, F_WSTR_DRAW_BORDER, 0, nWidth);
		palSMapCommands[1] = rgbt(0x80, 0x80, 0x00, shade(shShade, 0x00, 0x80));
		palSMapCommands[2] = rgbt(0xFF, 0xFF, 0x00, shade(shShade, 0x00, 0xFF));
		if (pHead)
		{
			gbBuffer.bBoxArea.left = 2;
			gbBuffer.bBoxArea.bottom -= 2;
			gbBuffer.bBoxArea.right = gbBuffer.bBoxArea.left + nPart;
			gbBuffer.bBoxArea.top = gbBuffer.bBoxArea.bottom - nHeight;
			gbBuffer.bOp.clr = rgbt(0xFF, 0xFF, 0x00, shade(shShade, 0x00, 0x80));
/*			G2Dr_enclose_area(&gbBuffer, 100, 20);*/
			boxSMapCommands[SIDE2I(hLayer->lSide)][0] = gbBuffer.bBoxArea;

			gbBuffer.bBoxArea.top++;
			gbBuffer.bBoxArea.left = 2 + (nPart / 2 - nWidth / 2);
			gbBuffer.bBoxArea.right = gbBuffer.bBoxArea.left + nWidth;
			G2Dr_head_to_base(&gbBuffer, pHead, palSMapCommands);
			Bmp_free_bitmap(pHead);
		}
		Font_get_size(1, strSMapCommands[mfd_get_info_s(nSMapCommands[nCommand][1])], 
									&nWidth, &nHeight);
		nWidth += 2;
		nHeight += 4;
		pStrings[0] = strSMapCommands[mfd_get_info_s(nSMapCommands[nCommand][1])];
		pHead = G2D_write_string(pStrings, 1, 1, F_WSTR_DRAW_BORDER, 0, nWidth);
		if (pHead)
		{
			gbBuffer.bBoxArea.top--;
			gbBuffer.bBoxArea.right = gbBuffer.bSizeImg.cx - 2;
			gbBuffer.bBoxArea.left = gbBuffer.bSizeImg.cx - nPart;
			gbBuffer.bOp.clr = rgbt(0xFF, 0xFF, 0x00, shade(shShade, 0x00, 0x80));
/*			G2Dr_enclose_area(&gbBuffer, 100, 20); */
			boxSMapCommands[SIDE2I(hLayer->lSide)][1] = gbBuffer.bBoxArea;

			gbBuffer.bBoxArea.top++;
			gbBuffer.bBoxArea.right = gbBuffer.bSizeImg.cx - 2;
			gbBuffer.bBoxArea.left = gbBuffer.bBoxArea.right - nPart + 
															 (nPart / 2 - nWidth / 2);
			G2Dr_head_to_base(&gbBuffer, pHead, palSMapCommands);
			Bmp_free_bitmap(pHead);
		}

		resetArea(gbBuffer);
		/* free buffer */
		G2D_color_to_pixel(gbScreen, &gbBuffer, RGB_trans);
		free(gbBuffer.bImg.clr);
	}
#endif 
}

void mfd_deinit_smap(P_MFD_HOOK_DEINIT)
{
	switch (mfd_get_det(hLayer->lId))
	{
		case F_SMFD_MMAP:			mfd_smap_deinit_map(hLayer);	break;
		case F_SMFD_MSIDE:		break;
		case F_SMFD_MCONFIG:	break;
		default:	break;
	}
	mfd_deinit_basiclayer(hLayer);
}

void mfd_rdeinit_smap(void)
{
	int i;

	for (i = 0; i < C_MFD_MAP_SIDES; i++)
	{
		G2D_reset_base(&gbSMapSides[i], true);
	}
}

/* begin section starget */

typedef struct tag_MFD_STarget
{
	pcolor					 stpalNotInstalled[0x03];
	g2dbase					 stgbNotInstalled;
	hudLayer				 sthlNotInstalled;

} MFD_STarget;

void	mfd_init_starget(P_MFD_HOOK_INIT)
{
	MFD_STarget *pTarget = (MFD_STarget *)malloc(sizeof(MFD_STarget));
	ss_bitmap_header *pHead = NULL;
	char *pStrings[2] = { "Targeting system not installed", NULL };

	mfd_init_basiclayer(hLayer, rgbt(0x00, 0x00, 0x00, 0x18));
	memset(pTarget, 0x00, sizeof(MFD_STarget));
	pHead = G2D_write_string(pStrings, 1, 0,
													 F_WSTR_DRAW_BORDER | F_WSTR_CENTER_ALIGN, 
													 0, hLayer->lhlBg.szSize.cx);
	convertHeadToArea(*pHead, pTarget->stgbNotInstalled);
	pTarget->stgbNotInstalled.pPal = pTarget->stpalNotInstalled;
	pTarget->stgbNotInstalled.nPalSize = 0x02;
	HUD_reset_layer(&pTarget->sthlNotInstalled, &pTarget->stgbNotInstalled);
	pTarget->sthlNotInstalled.szSize.cx = pHead->width;
	pTarget->sthlNotInstalled.szSize.cy = pHead->height;
	pTarget->sthlNotInstalled.ptPos.x = hLayer->lhlBg.ptPos.x +
			hLayer->lhlBg.szSize.cx / 2 - pTarget->sthlNotInstalled.szSize.cx / 2;
	pTarget->sthlNotInstalled.ptPos.y = hLayer->lhlBg.ptPos.y +
			hLayer->lhlBg.szSize.cy / 2 - pTarget->sthlNotInstalled.szSize.cy / 2;

	free(pHead);
	hLayer->lpData = pTarget;
}

void	mfd_timer_starget(uint32 now, uint32 last, void *pData)
{
	UNUSED(now);
	UNUSED(last);
	UNUSED(pData);
}

bool	mfd_mouse_starget(P_MFD_HOOK_MOUSE)
{
	UNUSED(hLayer);
	UNUSED(action);
	UNUSED(bStart);
	UNUSED(x);
	UNUSED(y);

	return false;
}

void	mfd_render_starget(P_MFD_HOOK_RENDER)
{
	MFD_STarget *pTarget = (MFD_STarget *)hLayer->lpData;
	HSHADE shShade = shade_get(hLayer->lShade);
	int nTrgVer = Player_get_ver_targeter(0);

	if (hLayer->lShade)
	{
		if (nDetailLevel > C_MFD_DETAIL_WHOA)
		{
			mfd_prerender_basiclayer(hLayer);
			appendHudLayer(pLayer, hLayer->lhlBg);
		}
		if (nTrgVer)
		{

		}
		else
		{
			pTarget->stpalNotInstalled[1] = rgbt(0x00, 0x00, 0x00, shade(shShade, 0x00, 0x40));
			pTarget->stpalNotInstalled[2] = rgbt(0xFF, 0x00, 0x00, shade(shShade, 0x00, 0x40));
			appendHudLayer(pLayer, pTarget->sthlNotInstalled);
		}
	}
}

void mfd_deinit_starget(P_MFD_HOOK_DEINIT)
{
	MFD_STarget *pTarget = (MFD_STarget *)hLayer->lpData;

	mfd_deinit_basiclayer(hLayer);
	G2D_reset_base(&pTarget->stgbNotInstalled, true);
	free(pTarget);
	hLayer->lpData = NULL;
}


/* begin section sdata */

typedef struct tag_MFD_SBio
{	/* Biomonitor - No Data/Ref because data is taken from player */
	pcolor				 sbpalBody[0x100];
	g2dbase				 sbgbBody;
	hudLayer			 sbhlBody;
} MFD_SBio;

typedef struct tag_MFD_SCont
{	/* Container - No Data/Ref beause taken from Object */
	int						 scActContent[4];
	pcolor				 scpalContent[0x100];
	g2dbase				 scgbContent[4];
	hudLayer			 schlContent[4];
} MFD_SCont;

typedef struct tag_MFD_SKeypadData
{	/* KeypadData - pressed digits and tryamount */
	MFD_CardRef		 skdRef;
	char					 skdDigits[3];
	int						 skdTries;	
} MFD_SKeypadData;

typedef struct tag_MFD_SKeypad
{	/* Keypad with Data pointer */
	MFD_SKeypadData	*skpData;

	pcolor				 skpalKeypad[0x100];
	pcolor				 skpalDigits[2][3];
	g2dbase				 skgbKeypadFrames[3];
	g2dbase				 skgbDigits[12];
	g2dbase				 skgbCode;
	hudLayer			 skhlKeypadFrame;
	hudLayer			 skhlDigits[12];
	hudLayer			 skhlCode;
} MFD_SKeypad;

void mfd_sdata_init_keypad(P_MFD_HOOK_INIT)
{
	MFD_SKeypad *pKeypad = NULL, *pKeypadOther;
	bool bFirstData = false;
	int i;

	pKeypad = (MFD_SKeypad *)malloc(sizeof(MFD_SKeypad));
	memset(pKeypad, 0x00, sizeof(MFD_SKeypad));
	for (i = 0; i < 3; i++)
	{
		mfd_load_bitmap(&pKeypad->skgbKeypadFrames[i], CHUNK_MFD_ICON_INVENTORY,
																									 SUBCHUNK_MFD_ICON_SKEYPAD_BASE + i);
		pKeypad->skgbKeypadFrames[i].pPal = pKeypad->skpalKeypad;
		pKeypad->skgbKeypadFrames[i].nPalSize = 0xFF;
	}
	HUD_reset_layer(&pKeypad->skhlKeypadFrame, &pKeypad->skgbKeypadFrames[0]);
	G2D_zoom_rect(&pKeypad->skgbKeypadFrames[0].bSizeImg, &hLayer->lhlBg.szSize, 
								&pKeypad->skhlKeypadFrame.szSize);
	pKeypad->skhlKeypadFrame.ptPos.x = hLayer->lhlBg.ptPos.x +
			hLayer->lhlBg.szSize.cx / 2 - pKeypad->skhlKeypadFrame.szSize.cx / 2;
	pKeypad->skhlKeypadFrame.ptPos.y = hLayer->lhlBg.ptPos.y +
			hLayer->lhlBg.szSize.cy / 2 - pKeypad->skhlKeypadFrame.szSize.cy / 2;
	
	pKeypadOther = (MFD_SKeypad *)mfd_layer_request_data(hLayer->lId, F_MFD_MASK_FULL);
	if (!pKeypadOther)
	{
		pKeypad->skpData = (MFD_SKeypadData *)malloc(sizeof(MFD_SKeypadData));
		memset(pKeypad->skpData, 0x00, sizeof(MFD_SKeypadData));
		for (i = 0; i < 3; i++)
			pKeypad->skpData->skdDigits[i] = -1;
		bFirstData = true;
	}
	else
		pKeypad->skpData = pKeypadOther->skpData;
	mfd_layer_addref_data(&pKeypad->skpData->skdRef, hLayer, bFirstData);
	hLayer->lpData = pKeypad;
}

void mfd_sdata_deinit_keypad(P_MFD_HOOK_DEINIT)
{
	int i;

	MFD_SKeypad *pKeypad = (MFD_SKeypad *)hLayer->lpData;

	if (mfd_layer_release_data(&pKeypad->skpData->skdRef, hLayer))
		free(pKeypad->skpData);
	for (i = 0; i < 3; i++)
		G2D_reset_base(&pKeypad->skgbKeypadFrames[i], true);
	free(pKeypad);
	hLayer->lpData = NULL;
}

void mfd_sdata_init_cont(P_MFD_HOOK_INIT)
{
	MFD_SCont *pCont = (MFD_SCont *)malloc(sizeof(MFD_SCont));
	int i;

	memset(pCont, 0x00, sizeof(MFD_SCont));
	for (i = 0; i < 4; i++)
	{
		pCont->scActContent[i] = -1;
		HUD_reset_layer(&pCont->schlContent[i], &pCont->scgbContent[i]);
		G2D_reset_base(&pCont->scgbContent[i], false);
	}

	hLayer->lpData = pCont;
}

void mfd_sdata_render_cont(P_MFD_HOOK_RENDER)
{
	MFD_SCont *pCont = (MFD_SCont *)hLayer->lpData;
	container *pData = Cnr_get_container_i(mfd_get_info_s(hLayer->lId));
	map_object *pObject = NULL;
	object_prop *pProp = NULL;
	size szDest, szBox;
	int i;

	Pal_copy_palette(pCont->scpalContent, RGB_trans, hLayer->lShade);
	pCont->scpalContent[0x00] = rgbt(0x00, 0x00, 0x00, 0x00);
	szDest.cx = hLayer->lhlBg.szSize.cx / 3;
	szDest.cy = hLayer->lhlBg.szSize.cy / 3;
	szBox.cx = hLayer->lhlBg.szSize.cx / 2;
	szBox.cy = hLayer->lhlBg.szSize.cy / 2;
	for (i = 0; i < 4; i++)
	{
		if (pData->content[i])
		{	/* it could be possible that the content changes, so get it here */
			if (pData->content[i] != pCont->scActContent[i])
			{
				G2D_reset_base(&pCont->scgbContent[i], true);
				pObject = Obj_get_object(pData->content[i]);
				pProp = Obj_get_property(pObject->object_id);

				mfd_load_bitmap(&pCont->scgbContent[i], CHUNK_OBJECT_SPRITES, 
																								pProp->baseframe);
				pCont->scgbContent[i].pPal = pCont->scpalContent;
				pCont->scgbContent[i].nPalSize = 0xFF;

				G2D_zoom_rect(&pCont->scgbContent[i].bSizeImg, &szDest, 
											&pCont->schlContent[i].szSize);
				pCont->schlContent[i].ptPos.x = hLayer->lhlBg.ptPos.x +
						szBox.cx / 2 - pCont->schlContent[i].szSize.cx / 2;
				pCont->schlContent[i].ptPos.y = hLayer->lhlBg.ptPos.y +
						szBox.cy / 2 - pCont->schlContent[i].szSize.cy / 2;
				if (i % 2)
					pCont->schlContent[i].ptPos.x += szBox.cx;
				if (i >= 2)
					pCont->schlContent[i].ptPos.y += szBox.cy;

				pCont->scActContent[i] = pData->content[i];
			}
			appendHudLayer(pLayer, pCont->schlContent[i]);
		}
		else
		{
			G2D_reset_base(&pCont->scgbContent[i], true);
		}
	}
}

void mfd_sdata_deinit_cont(P_MFD_HOOK_DEINIT)
{
	MFD_SCont *pCont = (MFD_SCont *)hLayer->lpData;
	int i;

	for (i = 0; i < 4; i++)
		G2D_reset_base(&pCont->scgbContent[i], true);
	free(pCont);

	hLayer->lpData = NULL;
}

void mfd_sdata_init_bio(P_MFD_HOOK_INIT)
{
	MFD_SBio *pBio = NULL;

	pBio = (MFD_SBio *)malloc(sizeof(MFD_SBio));
	memset(pBio, 0x00, sizeof(MFD_SBio));
	mfd_load_bitmap(&pBio->sbgbBody, CHUNK_MFD_GENERAL_STUFF, 
																	 SUBCHUNK_MFD_GEN_BODY);
	pBio->sbgbBody.pPal = pBio->sbpalBody;
	HUD_reset_layer(&pBio->sbhlBody, &pBio->sbgbBody);
	G2D_zoom_rect(&pBio->sbgbBody.bSizeImg, &hLayer->lhlBg.szSize, 
								&pBio->sbhlBody.szSize);
	pBio->sbhlBody.ptPos.x = hLayer->lhlBg.ptPos.x;
	pBio->sbhlBody.ptPos.y = hLayer->lhlBg.ptPos.y +
			hLayer->lhlBg.szSize.cy / 2 - pBio->sbhlBody.szSize.cy / 2;
	hLayer->lpData = pBio;
}

void mfd_sdata_deinit_bio(P_MFD_HOOK_DEINIT)
{
	MFD_SBio *pBio = (MFD_SBio *)hLayer->lpData;

	G2D_reset_base(&pBio->sbgbBody, true);
	free(pBio);
	hLayer->lpData = NULL;
}

void	mfd_init_sdata(P_MFD_HOOK_INIT)
{
	mfd_init_basiclayer(hLayer, rgbt(0x00, 0x40, 0x00, 0x18));
	switch (mfd_get_det(hLayer->lId))
	{
		case F_SMFD_DKEYPAD:	mfd_sdata_init_keypad(hLayer);	break;
		case F_SMFD_DCONT:		mfd_sdata_init_cont(hLayer);		break;
		case F_SMFD_DBIO:			mfd_sdata_init_bio(hLayer);			break;
		default:	break;
	}
}

void	mfd_timer_sdata(uint32 now, uint32 last, void *pData)
{
	UNUSED(now);
	UNUSED(last);
	UNUSED(pData);
}

bool	mfd_mouse_sdata(P_MFD_HOOK_MOUSE)
{
	UNUSED(hLayer);
	UNUSED(action);
	UNUSED(bStart);
	UNUSED(x);
	UNUSED(y);

	return false;
}

void	mfd_render_sdata(P_MFD_HOOK_RENDER)
{
	MFD_SKeypad *pKeypad = NULL;
	MFD_SBio *pBio = NULL;

	if (hLayer->lShade)
	{
		if (nDetailLevel > C_MFD_DETAIL_WHOA)
		{
			mfd_prerender_basiclayer(hLayer);
			appendHudLayer(pLayer, hLayer->lhlBg);
		}
		switch (mfd_get_det(hLayer->lId))
		{
			case F_SMFD_DKEYPAD:
			{
				pKeypad = (MFD_SKeypad *)hLayer->lpData;
				Pal_copy_palette(pKeypad->skpalKeypad, RGB_trans, hLayer->lShade);
				pKeypad->skpalKeypad[0] = rgbt(0x00, 0x00, 0x00, 0x00);
				appendHudLayer(pLayer, pKeypad->skhlKeypadFrame);
				break;
			}
			case F_SMFD_DCONT: mfd_sdata_render_cont(hLayer, pLayer); break;
			case F_SMFD_DBIO:
			{
				pBio = (MFD_SBio *)hLayer->lpData;
				Pal_copy_palette(pBio->sbpalBody, RGB_trans, hLayer->lShade);
				pBio->sbpalBody[0] = rgbt(0x00, 0x00, 0x00, 0x00);
				appendHudLayer(pLayer, pBio->sbhlBody);
				break;
			}
			default:	break;
		}
	}
}

void mfd_deinit_sdata(P_MFD_HOOK_DEINIT)
{
	switch (mfd_get_det(hLayer->lId))
	{
		case F_SMFD_DKEYPAD:	mfd_sdata_deinit_keypad(hLayer);	break;
		case F_SMFD_DCONT:		mfd_sdata_deinit_cont(hLayer);		break;
		case F_SMFD_DBIO:			mfd_sdata_deinit_bio(hLayer);			break;
		default:	break;
	}
	mfd_deinit_basiclayer(hLayer);
}

/* begin section cvoid */

void mfd_init_cvoid(P_MFD_HOOK_INIT)
{
	mfd_init_basiclayer(hLayer, rgbt(0x00, 0x00, 0x40, 0x18));
}

void	mfd_render_cvoid(P_MFD_HOOK_RENDER)
{
	if (hLayer->lShade)
	{
		if (nDetailLevel > C_MFD_DETAIL_WHOA)
		{
			mfd_prerender_basiclayer(hLayer);
			appendHudLayer(pLayer, hLayer->lhlBg);
		}
	}
}

void mfd_deinit_cvoid(P_MFD_HOOK_INIT)
{
	mfd_deinit_basiclayer(hLayer);
}

/* begin section cweapon */

void mfd_init_cweapon(P_MFD_HOOK_INIT)
{
	mfd_init_basiclayer(hLayer, rgbt(0x40, 0x00, 0x00, 0x18));
}

bool	mfd_mouse_cweapon(P_MFD_HOOK_MOUSE)
{
	UNUSED(hLayer);
	UNUSED(action);
	UNUSED(bStart);
	UNUSED(x);
	UNUSED(y);

	return false;
}

void	mfd_render_cweapon(P_MFD_HOOK_RENDER)
{
	if (hLayer->lShade)
	{
		if (nDetailLevel > C_MFD_DETAIL_WHOA)
		{
			mfd_prerender_basiclayer(hLayer);
			appendHudLayer(pLayer, hLayer->lhlBg);
		}
	}
}

void mfd_deinit_cweapon(P_MFD_HOOK_INIT)
{
	mfd_deinit_basiclayer(hLayer);
}

/* begin section chardware */

void mfd_init_chardware(P_MFD_HOOK_INIT)
{
	mfd_init_basiclayer(hLayer, rgbt(0x00, 0x00, 0x40, 0x18));
}

bool	mfd_mouse_chardware(P_MFD_HOOK_MOUSE)
{
	UNUSED(hLayer);
	UNUSED(action);
	UNUSED(bStart);
	UNUSED(x);
	UNUSED(y);

	return false;
}

void	mfd_render_chardware(P_MFD_HOOK_RENDER)
{
	if (hLayer->lShade)
	{
		if (nDetailLevel > C_MFD_DETAIL_WHOA)
		{
			mfd_prerender_basiclayer(hLayer);
			appendHudLayer(pLayer, hLayer->lhlBg);
		}
	}
}

void mfd_deinit_chardware(P_MFD_HOOK_INIT)
{
	mfd_deinit_basiclayer(hLayer);
}

/* begin section cinventory */

void mfd_init_cinventory(P_MFD_HOOK_INIT)
{
	mfd_init_basiclayer(hLayer, rgbt(0x40, 0x40, 0x00, 0x18));
}

bool	mfd_mouse_cinventory(P_MFD_HOOK_MOUSE)
{
	UNUSED(hLayer);
	UNUSED(action);
	UNUSED(bStart);
	UNUSED(x);
	UNUSED(y);

	return false;
}

void	mfd_render_cinventory(P_MFD_HOOK_RENDER)
{
	if (hLayer->lShade)
	{
		if (nDetailLevel > C_MFD_DETAIL_WHOA)
		{
			mfd_prerender_basiclayer(hLayer);
			appendHudLayer(pLayer, hLayer->lhlBg);
		}
	}
}

void mfd_deinit_cinventory(P_MFD_HOOK_INIT)
{
	mfd_deinit_basiclayer(hLayer);
}


/* begin section csoftware */

void mfd_init_csoftware(P_MFD_HOOK_INIT)
{
	mfd_init_basiclayer(hLayer, rgbt(0x00, 0x40, 0x00, 0x18));
}

bool	mfd_mouse_csoftware(P_MFD_HOOK_MOUSE)
{
	UNUSED(hLayer);
	UNUSED(action);
	UNUSED(bStart);
	UNUSED(x);
	UNUSED(y);

	return false;
}

void	mfd_render_csoftware(P_MFD_HOOK_RENDER)
{
	if (hLayer->lShade)
	{
		if (nDetailLevel > C_MFD_DETAIL_WHOA)
		{
			mfd_prerender_basiclayer(hLayer);
			appendHudLayer(pLayer, hLayer->lhlBg);
		}
	}
}

void mfd_deinit_csoftware(P_MFD_HOOK_INIT)
{
	mfd_deinit_basiclayer(hLayer);
}

/* begin section cmedia */

int nMediaTextOffsets[C_MFD_MEDIA_AMOUNT] = { 4, 4, 4, 0, 4 };

void	mfd_init_cmedia(P_MFD_HOOK_INIT)
{
	ss_bitmap_header *pHead = NULL;
	MFD_SMedia *pSMedia = NULL;
	MFD_CMedia *pMedia = (MFD_CMedia *)malloc(sizeof(MFD_CMedia));
	int nBodyLines = 0;
	bool bFirst = false;

	mfd_init_basiclayer(hLayer, rgbt(0x00, 0x00, 0x00, 0x18));
	memset(pMedia, 0x00, sizeof(MFD_CMedia));
	pSMedia = 
			(MFD_SMedia *)mfd_layer_request_data(mfd_set_id(F_SMFD_IMEDIA, 0x00,
																											mfd_get_info(hLayer->lId)),
																					 F_MFD_MASK_CARD | F_MFD_MASK_INFO);
	if (!pSMedia)
	{	/* we have to load up the data for ourselves. */
		pMedia->cmpData = mfd_mediadata_create((short)mfd_get_info_s(hLayer->lId));
		bFirst = true;
	}
	else
		pMedia->cmpData = pSMedia->smpData;
	mfd_layer_addref_data(&pMedia->cmpData->mdRef, hLayer, bFirst);

	HUD_reset_layer(&pMedia->cmhlTxt, &pMedia->cmgbTxt);

	while ((nMediaTextOffsets[pMedia->cmpData->mdType] + nBodyLines) < pMedia->cmpData->mdLines && 
				 strlen(pMedia->cmpData->mdpStrings[nMediaTextOffsets[pMedia->cmpData->mdType] + nBodyLines]))
		nBodyLines++;
	pHead = G2D_write_string(&pMedia->cmpData->mdpStrings[nMediaTextOffsets[pMedia->cmpData->mdType]], 
													 nBodyLines, 1, 
													 F_WSTR_DRAW_BORDER, 0, 
													 hLayer->lhlBg.szSize.cx);
	if (pHead)
	{
		convertHeadToArea(*pHead, pMedia->cmgbTxt);
		pMedia->cmgbTxt.pPal = pMedia->cmpalTxt;
		pMedia->cmgbTxt.nPalSize = 0x02;
		pMedia->cmhlTxt.ptPos = hLayer->lhlBg.ptPos;
		pMedia->cmhlTxt.szSize.cx = pHead->width;
		free(pHead);
		pHead = NULL;
	}
	else
		G2D_reset_base(&pMedia->cmgbTxt, false);

	hLayer->lpData = pMedia;
}

bool	mfd_mouse_cmedia(P_MFD_HOOK_MOUSE)
{
	MFD_CMedia *pMedia = (MFD_CMedia *)hLayer->lpData;
	bool rCode = false;

	UNUSED(x);

	switch (action)
	{
		case C_HUD_MOUSE_HOVER:	
		{
			if (pMedia->cmbScrolling)
				pMedia->cmScrollOffset = max((y - pMedia->cmScrollStart), 0);
			else
				HUD_set_description("Select or Grab and Drag",
														rgbt(0xFF, 0xFF, 0xFF, 0x80),
														rgbt(0x80, 0x80, 0x80, 0x20),
														&(mfd_ss_items[C_MFD_CENTER].haArea));
			rCode = true;
			break;
		}
		case C_HUD_MOUSE_SELECT:
		{
			if (bStart)
			{
				pMedia->cmScrollOffset += (hLayer->lhlBg.szSize.cy - hLayer->lhlBg.szSize.cy / 10);
				if (pMedia->cmgbTxt.bBoxArea.bottom == pMedia->cmgbTxt.bSizeImg.cy)
				{
					mfd_layer_fadeout_data(&pMedia->cmpData->mdRef);
					if (pMedia->cmpData->mdFlagI)
						MFD_show_media_relative((byte)(pMedia->cmpData->mdType), 
																		(short)(pMedia->cmpData->mdFlagI));
				}
			}
			break;
		}
		case C_HUD_MOUSE_ACTIVATE:
		{	/* here it's rather deactivate... */
			if (bStart)
			{
				mfd_layer_fadeout_data(&pMedia->cmpData->mdRef);
				if (pMedia->cmpData->mdFlagI)
					MFD_show_media_relative((byte)(pMedia->cmpData->mdType), 
																	(short)(pMedia->cmpData->mdFlagI));
			}
			break;
		}
		case C_HUD_MOUSE_GRAB:
		{
			if (bStart)
			{
				pMedia->cmbScrolling = true;
				pMedia->cmScrollStart = (y - pMedia->cmScrollOffset);				
				HUD_lock_cursor(&mfd_ss_items[C_MFD_CENTER].haArea);
			}
			else
			{
				pMedia->cmbScrolling = false;
				pMedia->cmScrollStart = 0;				
				HUD_unlock_cursor();
			}
			rCode = true;
			break;
		}
		default:	break;
	}

	return rCode;
}

void	mfd_render_cmedia(P_MFD_HOOK_RENDER)
{
	HSHADE shShade = shade_get(hLayer->lShade);
	MFD_CMedia *pMedia = (MFD_CMedia *)hLayer->lpData;

	if (hLayer->lShade)
	{
		if (nDetailLevel >= C_MFD_DETAIL_WHOA)
		{	/* nice background */
			mfd_prerender_basiclayer(hLayer);
			appendHudLayer(pLayer, hLayer->lhlBg);
		}
		if (pMedia->cmgbTxt.bImg.bit)
		{
			pMedia->cmpalTxt[1] = rgbt(0x00, 0xFF, 0x00, shade(shShade, 0x00, 0x40));
			pMedia->cmpalTxt[2] = rgbt(0x00, 0xFF, 0x00, shade(shShade, 0x00, 0xFF));
			if (pMedia->cmgbTxt.bSizeImg.cy > hLayer->lhlBg.szSize.cy)
			{
				if ((pMedia->cmgbTxt.bSizeImg.cy - pMedia->cmScrollOffset) < hLayer->lhlBg.szSize.cy)
					pMedia->cmScrollOffset = pMedia->cmgbTxt.bSizeImg.cy - hLayer->lhlBg.szSize.cy;
			}
			else
				pMedia->cmScrollOffset = 0;

			pMedia->cmgbTxt.bBoxArea.top = pMedia->cmScrollOffset;
			pMedia->cmgbTxt.bBoxArea.bottom = min(pMedia->cmgbTxt.bBoxArea.top + 
																								hLayer->lhlBg.szSize.cy,
																						pMedia->cmgbTxt.bSizeImg.cy);
			pMedia->cmhlTxt.szSize.cy = getRectHeight(pMedia->cmgbTxt.bBoxArea);
			appendHudLayer(pLayer, pMedia->cmhlTxt);
		}
	}
}

void mfd_deinit_cmedia(P_MFD_HOOK_DEINIT)
{
	MFD_CMedia *pMedia = (MFD_CMedia *)hLayer->lpData;

	if (mfd_layer_release_data(&pMedia->cmpData->mdRef, hLayer))
	{
		mfd_mediadata_destroy(pMedia->cmpData);
		pMedia->cmpData = NULL;
	}
	else
	{
		/* !!!!! let the other fade out */
	}
	G2D_reset_base(&pMedia->cmgbTxt, true);
	if (pMedia->cmbScrolling)
		HUD_unlock_cursor();
	free(pMedia);
	hLayer->lpData = NULL;
	mfd_deinit_basiclayer(hLayer);
}

/*                      */
/* top line items       */
/*                      */

/* begin section topline */

#define	C_MFD_TOPITEMS_TOP			 2.0
#define	C_MFD_TOPITEMS_SPACE		 1.0

/*                      */
/* body status          */
/*                      */

/* begin section bodystat */

#define	C_BODYSTAT_MAXFILL		66.6
#define	C_BODYSTAT_MAXTIME		5000
/* hearbeats */
#define	C_BODYSTAT_BPM_MIN		40.0	
#define	C_BODYSTAT_BPM_MAX	 200.0

enum
{
	C_BODYSTAT_BIO,
	C_BODYSTAT_RAD,
	C_BODYSTAT_POWER,
	C_BODYSTAT_FATIGUE,
	C_BODYSTAT_CHI,

	C_BODYSTAT_AMOUNT
};

typedef struct tag_MFD_StatEntry
{
	struct tag_MFD_StatEntry		*pNext;
	
	byte												 seShade;
	uint32											 seTick;

	double											 seValue;
	uint16											 sePosX;
} MFD_StatEntry;

typedef struct tag_MFD_StatList
{
	struct tag_MFD_StatEntry		*pFirst;
	struct tag_MFD_StatEntry		*pLast;
	int													 slAmount;

	bool												 slVisible;
	pcolor											 slColor;
	double											 slMaxValue;
	uint32											 slTimeCur;
	uint32											 slTimeMin;
	uint32											 slTimeMax;

	double										 (*slGetValue)(int nPlayer);
	void											 (*slRender)(int nIndex);
} MFD_StatList;

void	bodystat_render_abs(int nIndex);
void	bodystat_render_fatigue(int nIndex);
void	bodystat_render_chi(int nIndex);

pcolor				 palBodyStatEntry[C_BODYSTAT_AMOUNT][0x100];

g2dbase				 gbBodyStatBg = D_G2D_NULLAREA;
g2dbase				 gbBodyStatEntry[C_BODYSTAT_AMOUNT] =
{ 
	D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA, 
	D_G2D_NULLAREA, D_G2D_NULLAREA 
};

hudLayer			 hlBodyStatBg = D_HUD_NULLLAYER(&gbBodyStatBg);
hudLayer			 hlBodyStatEntry[C_BODYSTAT_AMOUNT] =
{
	D_HUD_NULLLAYER(&gbBodyStatEntry[0]), D_HUD_NULLLAYER(&gbBodyStatEntry[1]), 
	D_HUD_NULLLAYER(&gbBodyStatEntry[2]), D_HUD_NULLLAYER(&gbBodyStatEntry[3]), 
	D_HUD_NULLLAYER(&gbBodyStatEntry[4])
};

MFD_StatList	 slBodyStat[C_BODYSTAT_AMOUNT] =
{
	{ NULL, NULL, 0, true, 0xD0C7EB33, 200.0, 2000, 5000, 1000,	
		Player_get_bio_contamination, bodystat_render_abs },
	{ NULL, NULL, 0, true, 0xD09393F3, 200.0, 2500, 5000, 1000, 
		Player_get_rad_contamination, bodystat_render_abs },
	{ NULL, NULL, 0, true, 0xD000FBFB, 300.0, 3000, 5000, 1000, 
		Player_get_power_usage, bodystat_render_abs },
	{ NULL, NULL, 0, true, 0xD0F34B4F, 100.0, 5000, 5000, 1000, 
		Player_get_fatigue, bodystat_render_fatigue },
	{ NULL, NULL, 0, true, 0xD0E38BFB, 100.0, 8000, 5000, 1000, 
		Player_get_chi, bodystat_render_chi }
};

MFD_StatEntry *bodystat_add_entry(MFD_StatList *pList, 
																	uint32 dwTick, double dValue)
{
	MFD_StatEntry *pEntry = (MFD_StatEntry *)malloc(sizeof(MFD_StatEntry));

	memset(pEntry, 0x00, sizeof(MFD_StatEntry));
	pEntry->seShade = 0xFF;
	pEntry->seTick = dwTick;
	pEntry->seValue = dValue;
	if (pList->pLast)
	{
		pList->pLast->pNext = pEntry;
		pEntry->sePosX = (pList->pLast->sePosX + 1) % hlBodyStatBg.szSize.cx;
	}
	else
	{
		pEntry->sePosX = 0;
		pList->pFirst = pEntry;
	}
	pList->pLast = pEntry;
	pList->slAmount++;

	return pEntry;
}

void bodystat_rem_first(MFD_StatList *pList)
{
	MFD_StatEntry *pEntry = pList->pFirst;

	if (pList->pFirst)
	{
		pList->slAmount--;
		assert(pList->slAmount >= 0);
		pList->pFirst = pEntry->pNext;
		if (!pList->pFirst)
			pList->pLast = NULL;
		free(pEntry);
	}
}

void bodystat_update_list(MFD_StatList *pList, uint32 dwNow)
{
	MFD_StatEntry *pEntry = pList->pFirst;
	uint32 dwMaxTime, dwFirstTime = 0, dwSlope;
	int nNextShade, nOverFlow = 0, i;

	dwMaxTime = ((double)pList->slTimeCur * C_BODYSTAT_MAXFILL / 100.0);
	dwMaxTime = min(C_BODYSTAT_MAXTIME, dwMaxTime);
	if (pEntry)
	{
		dwFirstTime = dwNow - pEntry->seTick - dwMaxTime;
		dwFirstTime = min(dwMaxTime, max(dwFirstTime, 1));
	}
	while (pEntry && ((int)(dwNow - pEntry->seTick) >= (int)dwMaxTime))
	{
		nOverFlow++;
		pEntry = pEntry->pNext;
	}
	dwSlope = (0xFF * dwFirstTime) / (dwMaxTime / 2);
	pEntry = pList->pFirst;
	for (i = 0; i < nOverFlow; i++)
	{
		nNextShade = 0xFF - (int)((dwSlope * 
															(dwNow - pEntry->seTick - dwMaxTime)) / 
															dwFirstTime);
		assert(nNextShade < 0x100);
		if (nNextShade < 1)
		{
			bodystat_rem_first(pList);
			pEntry = pList->pFirst;
		}
		else
		{
			pEntry->seShade = (byte)nNextShade;
			pEntry = pEntry->pNext;
		}
	}
}

void	bodystat_render_abs(int nIndex)
{
	MFD_StatEntry *pEntry;
	int y, nLastY = 0, nDir;

	pEntry = slBodyStat[nIndex].pFirst;
	while (pEntry)
	{
		y = 0;
		if (pEntry->seValue > 0.0)
		{ /* ff1: dunno what i'm calculating here, but looks good ;) */
			y = log(1.0 + pEntry->seValue / 10.0) * 100.0 / slBodyStat[nIndex].slMaxValue * 
							gbBodyStatEntry[nIndex].bSizeImg.cy;
			y = min(y, gbBodyStatEntry[nIndex].bSizeImg.cy);
		}
		if (y > 0)
		{
			gbBodyStatEntry[nIndex].bImg.bit[pEntry->sePosX + 
																			 gbBodyStatEntry[nIndex].bSizeImg.cx *
																			 (gbBodyStatEntry[nIndex].bSizeImg.cy - y)] = 
						pEntry->seShade;
		}
		if (nLastY != y && pEntry != slBodyStat[nIndex].pFirst)
		{	/* draw a vertical line */
			nDir = sgn(y - nLastY);
			while (nLastY != y)
			{
				if (nLastY)
					gbBodyStatEntry[nIndex].bImg.bit[pEntry->sePosX + 
																					 gbBodyStatEntry[nIndex].bSizeImg.cx *
																					 (gbBodyStatEntry[nIndex].bSizeImg.cy - nLastY)] = 
								pEntry->seShade;
				nLastY += nDir;
			}
		}
		nLastY = y;
		pEntry = pEntry->pNext;
	}	
}

/*   0: about 40 beats per minute; */
/* ~90: -> about 10 beats per width of orig ss ;) */

void	bodystat_render_fatigue(int nIndex)
{
	MFD_StatEntry *pEntry;
	double dFatigue;
	int y, nLastY = 0, nDir, nLastBeat = -1, nBeatExpected, nFirstEntry = 0;
	bool bBeat = false, bNextBeat = false;

	assert(nIndex == C_BODYSTAT_FATIGUE);

	pEntry = slBodyStat[nIndex].pFirst;
	while (pEntry)
	{
		bBeat = false;
		y = 0;
		if (pEntry == slBodyStat[nIndex].pFirst)
			nFirstEntry = pEntry->seTick;
		dFatigue = pEntry->seValue * (double)sgn(pEntry->seValue);
		nBeatExpected = dFatigue * 60000.0 / C_BODYSTAT_BPM_MAX +
										(1.0 - dFatigue) * 60000.0 / C_BODYSTAT_BPM_MIN;

		if (pEntry->seValue < 0.0)									/* pre stored beat */
			bBeat = true;
		else
		{	/* expected any beat? */
			if ((nLastBeat < 0 && (uint32)(nFirstEntry + nBeatExpected) <= pEntry->seTick) ||
					(nLastBeat >= 0 && (uint32)(nLastBeat + nBeatExpected) <= pEntry->seTick))
				bBeat = true;
		}
		if (bBeat)
		{	/* got a beat; store it */
			dFatigue += 0.0001;												/* fatigue 0 also has beats ;) */
			pEntry->seValue = dFatigue * -1.0;
			nLastBeat = pEntry->seTick;
			y = gbBodyStatEntry[nIndex].bSizeImg.cy / -4;
		}
		if (bNextBeat)
			y = gbBodyStatEntry[nIndex].bSizeImg.cy / +5;
		bNextBeat = bBeat;
		gbBodyStatEntry[nIndex].bImg.bit[pEntry->sePosX + 
																		 gbBodyStatEntry[nIndex].bSizeImg.cx * 
																		 (gbBodyStatEntry[nIndex].bSizeImg.cy / 2 + y)] = 
					pEntry->seShade;
		if (nLastY != y && pEntry != slBodyStat[nIndex].pFirst)
		{	/* draw a vertical line */
			nDir = sgn(y - nLastY);
			while (nLastY != y)
			{
				if (nLastY)
					gbBodyStatEntry[nIndex].bImg.bit[pEntry->sePosX + 
																					 gbBodyStatEntry[nIndex].bSizeImg.cx *
																					 (gbBodyStatEntry[nIndex].bSizeImg.cy / 2 + nLastY)] = 
								pEntry->seShade;
				nLastY += nDir;
			}
		}
		nLastY = y;
		pEntry = pEntry->pNext;
	}	
}

/* berserk: x / 2; genious: x * 2 */
void	bodystat_render_chi(int nIndex)
{
	MFD_StatEntry *pEntry;
	int y;

	assert(nIndex == C_BODYSTAT_CHI);

	pEntry = slBodyStat[nIndex].pFirst;
	while (pEntry)
	{
		y = sin((double)((double)pEntry->sePosX) / (2.0 * pEntry->seValue) / M_PI) * 
				(double)(gbBodyStatEntry[nIndex].bSizeImg.cy / 3);
		gbBodyStatEntry[nIndex].bImg.bit[pEntry->sePosX + 
																		 gbBodyStatEntry[nIndex].bSizeImg.cx * 
																		 (gbBodyStatEntry[nIndex].bSizeImg.cy / 2 + y)] = 
					pEntry->seShade;
		pEntry = pEntry->pNext;
	}	
}

bool mfd_rinit_bodystat(size szScreen)
{
	int i;
	bool bVisible = mfd_ss_items[C_MFD_STATS].haVisible;

	mfd_ss_items[C_MFD_STATS].haVisible = false;
											
	mfd_ss_items[C_MFD_STATS].haArea.left = 
			(int)((double)(szScreen.cx) * C_MFD_TOPITEMS_SPACE / 100.0);
	mfd_ss_items[C_MFD_STATS].haArea.right = 
			mfd_ss_items[C_MFD_VIEWPOSE].haArea.left - 
			(int)((double)(szScreen.cx) * C_MFD_TOPITEMS_SPACE / 100.0);
	mfd_ss_items[C_MFD_STATS].haArea.top = 
				(int)((double)(szScreen.cy) * C_MFD_TOPITEMS_TOP / 100.0);
	mfd_ss_items[C_MFD_STATS].haArea.bottom = 
			mfd_ss_items[C_MFD_VIEWPOSE].haArea.bottom;

	hlBodyStatBg.ptPos.x = mfd_ss_items[C_MFD_STATS].haArea.left;
	hlBodyStatBg.ptPos.y = mfd_ss_items[C_MFD_STATS].haArea.top;
	hlBodyStatBg.szSize.cx = (uint16)getRectWidth(mfd_ss_items[C_MFD_STATS].haArea);
	hlBodyStatBg.szSize.cy = (uint16)getRectHeight(mfd_ss_items[C_MFD_STATS].haArea);
	G2D_init_clrbase(&gbBodyStatBg, hlBodyStatBg.szSize.cx, hlBodyStatBg.szSize.cy);
	gbBodyStatBg.bOp.clr = rgbt(0x00, 0xFF, 0xFF, 0x10);
	G2Dc_fill_area(&gbBodyStatBg);

	for (i = 0; i < C_BODYSTAT_AMOUNT; i++)
	{
		hlBodyStatEntry[i].ptPos = hlBodyStatBg.ptPos;
		hlBodyStatEntry[i].szSize = hlBodyStatBg.szSize;
		G2D_init_bitbase(&gbBodyStatEntry[i], hlBodyStatEntry[i].szSize.cx,
																					hlBodyStatEntry[i].szSize.cy);
		memset(gbBodyStatEntry[i].bImg.bit, 0x00,
					 hlBodyStatEntry[i].szSize.cx * hlBodyStatEntry[i].szSize.cy);
		Pal_shade_palette(palBodyStatEntry[i], 0, 0x100, 
											rgbt(get_red(slBodyStat[i].slColor),
													 get_green(slBodyStat[i].slColor),
													 get_blue(slBodyStat[i].slColor),
													 0x00),
											slBodyStat[i].slColor);
		gbBodyStatEntry[i].pPal = palBodyStatEntry[i];
		gbBodyStatEntry[i].nPalSize = 0xFF;
	}

	Timer_register_function("MFD.BodyStat", 50, mfd_timer_bodystat, NULL);

	mfd_ss_items[C_MFD_STATS].haVisible = bVisible;

	return true;
}

void	mfd_rdeinit_bodystat(void)
{
	int i;

	Timer_unregister_function("MFD.BodyStat");
	G2D_reset_base(&gbBodyStatBg, true);

	for (i = 0; i < C_BODYSTAT_AMOUNT; i++)
	{
		while (slBodyStat[i].pFirst)
			bodystat_rem_first(&slBodyStat[i]);
		slBodyStat[i].slTimeCur = slBodyStat[i].slTimeMin;
	}
}

void	mfd_timer_bodystat(uint32 now, uint32 last, void *pData)
{
	MFD_StatEntry *pEntry;
	double dTimePerPixel;
	int i, nTimePassed, nPixelToDraw, nPixel;

	UNUSED(pData);
	UNUSED(last);

	for (i = 0; i < C_BODYSTAT_AMOUNT; i++)
	{
		nPixelToDraw = 0;
		dTimePerPixel = (double)(slBodyStat[i].slTimeCur) / 
										(double)(hlBodyStatBg.szSize.cx);
		if (slBodyStat[i].pLast)
		{
			nTimePassed = now - slBodyStat[i].pLast->seTick;
			nPixelToDraw = nTimePassed / dTimePerPixel;
		}
		else
			nPixelToDraw = 1;

		for (nPixel = 0; nPixel < nPixelToDraw; nPixel++)
			pEntry = bodystat_add_entry(&slBodyStat[i], now, 
																	slBodyStat[i].slGetValue(0));
		bodystat_update_list(&slBodyStat[i], now);
	}
}

void	mfd_render_bodystat(P_HUD_HOOK_RENDER)
{
	int i;

	if (nDetailLevel >= C_MFD_DETAIL_WHOA)
		appendHudLayer(pLayer, hlBodyStatBg);

	G2Dc_fill_area(&gbBodyStatBg);
	for (i = 0; i < C_BODYSTAT_AMOUNT; i++)
	{
		if (slBodyStat[i].slRender && slBodyStat[i].slVisible)
		{
			memset(gbBodyStatEntry[i].bImg.bit, 0x00,
						 hlBodyStatEntry[i].szSize.cx * hlBodyStatEntry[i].szSize.cy);
			slBodyStat[i].slRender(i);
			appendHudLayer(pLayer, hlBodyStatEntry[i]);
		}
	}
}

/*                      */
/* viewpose             */
/*                      */

/* begin section cviewpose */

#define C_MFD_VIEWPOSE_WIDTH		15.0
#define	C_MFD_VIEWPOSE_SHIELDS	 4
#define	C_MFD_VIEWPOSE_SUITS		 3

pcolor				 palViewPoseGrid[0x100];

g2dbase				 gbViewPoseGrid = D_G2D_NULLAREA;
							 /* crough, kneel, stand | left, center, right*/
g2dbase				 gbViewPoseModel[C_MFD_VIEWPOSE_SUITS][3][3];
g2dbase				 gbViewPoseShield[C_MFD_VIEWPOSE_SHIELDS][3][3];
g2dbase				 gbViewPoseKnob[2] = { D_G2D_NULLAREA, D_G2D_NULLAREA };
g2dbase				 gbViewPoseGrip[2] = { D_G2D_NULLAREA, D_G2D_NULLAREA };

hudLayer			 hlViewPoseGrid = D_HUD_NULLLAYER(&gbViewPoseGrid);
hudLayer			 hlViewPoseModel = D_HUD_NULLLAYER(NULL);
hudLayer			 hlViewPoseShield = D_HUD_NULLLAYER(NULL);
hudLayer			 hlViewPoseKnob = D_HUD_NULLLAYER(&gbViewPoseKnob[0]);
hudLayer			 hlViewPoseGrip = D_HUD_NULLLAYER(&gbViewPoseGrip[0]);

bool					 bViewPoseGripped = false, bViewPoseKnobbed = false;
bool					 bViewPoseLocked[2] = { false, false };

bool mfd_rinit_viewpose(size szScreen)
{
	int nShield, nSuit, nModel, nSide, i,
			width = (int)((double)(szScreen.cx) * C_MFD_VIEWPOSE_WIDTH / 100.0),
			height = width / 2;
	bool bVisible = mfd_ss_items[C_MFD_VIEWPOSE].haVisible, rCode = false;

	mfd_ss_items[C_MFD_VIEWPOSE].haVisible = false;

	mfd_ss_items[C_MFD_VIEWPOSE].haArea.left = szScreen.cx / 2 - width / 2;
	mfd_ss_items[C_MFD_VIEWPOSE].haArea.right = 
			mfd_ss_items[C_MFD_VIEWPOSE].haArea.left + width;
	mfd_ss_items[C_MFD_VIEWPOSE].haArea.top = 
				(int)((double)(szScreen.cy) * C_MFD_TOPITEMS_TOP / 100.0);
	mfd_ss_items[C_MFD_VIEWPOSE].haArea.bottom = 
			mfd_ss_items[C_MFD_VIEWPOSE].haArea.top + height;

	Pal_copy_palette(palViewPoseGrid, RGB_trans, 0x40);

	rCode = mfd_load_bitmap(&gbViewPoseGrid, CHUNK_SCREEN_POSTURE_CONTROL,
																					 SUBCHUNK_SCR_POST_OPEN);
	if (rCode)
	{
		gbViewPoseGrid.pPal = palViewPoseGrid;
		gbViewPoseGrid.nPalSize = 0xFF;
		/* load up paperdolls */
		for (nSuit = 0; nSuit < C_MFD_VIEWPOSE_SUITS; nSuit++)
			for (nModel = 0; nModel < 3; nModel++)
				for (nSide = 0; nSide < 3; nSide++)
					rCode = mfd_load_bitmap(&gbViewPoseModel[nSuit][nModel][nSide],
																	CHUNK_SCREEN_POSTURE_SUIT + nSuit,
																	3 * (2 - nModel) + (2 - nSide)) && rCode;
		for (nShield = 0; nShield < C_MFD_VIEWPOSE_SHIELDS; nShield++)
			for (nModel = 0; nModel < 3; nModel++)
				for (nSide = 0; nSide < 3; nSide++)
					rCode = mfd_load_bitmap(&gbViewPoseShield[nShield][nModel][nSide],
																	CHUNK_SCREEN_POSTURE_SHIELD + nShield,
																	3 * (2 - nModel) + (2 - nSide)) && rCode;
		/* load knobs & grips */
		for (i = 0; i < 2; i++)
		{
			rCode = mfd_load_bitmap(&gbViewPoseKnob[i], CHUNK_SCREEN_POSTURE_CONTROL,
															SUBCHUNK_SCR_POST_KNOB_FREE + i) && rCode;
			rCode = mfd_load_bitmap(&gbViewPoseGrip[i], CHUNK_SCREEN_POSTURE_CONTROL,
															SUBCHUNK_SCR_POST_FIX + i) && rCode;
		}

		hlViewPoseGrid.ptPos.x = mfd_ss_items[C_MFD_VIEWPOSE].haArea.left;
		hlViewPoseGrid.ptPos.y = mfd_ss_items[C_MFD_VIEWPOSE].haArea.top;
		hlViewPoseGrid.szSize.cx = (uint16)getRectWidth(mfd_ss_items[C_MFD_VIEWPOSE].haArea);
		hlViewPoseGrid.szSize.cy = (uint16)getRectHeight(mfd_ss_items[C_MFD_VIEWPOSE].haArea);
		hlViewPoseShield.szSize.cx = hlViewPoseModel.szSize.cx = 
				hlViewPoseGrid.szSize.cx / 2 - 2;
		hlViewPoseShield.ptPos.x = hlViewPoseModel.ptPos.x = 
				mfd_ss_items[C_MFD_VIEWPOSE].haArea.left + hlViewPoseGrid.szSize.cx / 2;
		hlViewPoseShield.ptPos.y = hlViewPoseModel.ptPos.y = 
				mfd_ss_items[C_MFD_VIEWPOSE].haArea.top;

		hlViewPoseGrip.szSize.cx = hlViewPoseGrid.szSize.cx / 5;
		hlViewPoseGrip.szSize.cy = hlViewPoseGrid.szSize.cy / 4;
		hlViewPoseGrip.ptPos.x = hlViewPoseGrid.ptPos.x + hlViewPoseGrid.szSize.cx / 4;

		hlViewPoseKnob.szSize.cx = hlViewPoseGrid.szSize.cx / 6;
		hlViewPoseKnob.szSize.cy = hlViewPoseGrid.szSize.cy / 4;
		hlViewPoseKnob.ptPos.x = hlViewPoseGrid.ptPos.x;
	}

	mfd_ss_items[C_MFD_VIEWPOSE].haVisible = bVisible;

	return rCode;
}

void	mfd_rdeinit_viewpose(void)
{
	int nShield, nSuit, nModel, nSide, i;

	G2D_reset_base(&gbViewPoseGrid, true);
	for (nSuit = 0; nSuit < C_MFD_VIEWPOSE_SUITS; nSuit++)
		for (nModel = 0; nModel < 3; nModel++)
			for (nSide = 0; nSide < 3; nSide++)
				G2D_reset_base(&gbViewPoseModel[nSuit][nModel][nSide], true);
	for (nShield = 0; nShield < C_MFD_VIEWPOSE_SHIELDS; nShield++)
		for (nModel = 0; nModel < 3; nModel++)
			for (nSide = 0; nSide < 3; nSide++)
				G2D_reset_base(&gbViewPoseShield[nShield][nModel][nSide], true);
	for (i = 0; i < 2; i++)
	{
		G2D_reset_base(&gbViewPoseKnob[i], true);
		G2D_reset_base(&gbViewPoseGrip[i], true);
	}
}

bool	mfd_mouse_viewpose(P_HUD_HOOK_MOUSE)
{
	bounding_box boxLock;
	double dHeight, dLean;
	bool rCode = false, bViewFix = false, bView = false, bPose = false;

	dHeight = 1.0 - (double)y / (double)hlViewPoseGrid.szSize.cy;
	dLean = (double)(x - hlViewPoseGrid.szSize.cx / 2) * 2.0 / 
					(double)(hlViewPoseGrid.szSize.cx / 2);
	dLean -= 1.0;
	switch (action)
	{
		case C_HUD_MOUSE_HOVER:	
		{
			if (x >= hlViewPoseGrid.szSize.cx / 2)
			{	
				if (bViewPoseLocked[1])
				{	/* dragging the model */
					bPose = true;
				}
			}
			else
			{
				if (x >= hlViewPoseGrid.szSize.cx / 4)
				{
					bViewPoseGripped = true;
					bViewPoseKnobbed = false;
					if (bViewPoseLocked[0])
					{	/* dragging the grip */
						bView = true;
					}
				}
				else
				{
					bViewPoseKnobbed = true;
					bViewPoseGripped = false;
				}
			}

			break;
		}
		case C_HUD_MOUSE_SELECT:
		{
			if (bStart)
			{
				if (x >= hlViewPoseGrid.szSize.cx / 2)
				{	/* clicked on model */
					boxLock.left = hlViewPoseGrid.szSize.cx / 2;
					boxLock.top = 0;
					boxLock.right = hlViewPoseGrid.szSize.cx;
					boxLock.bottom = hlViewPoseGrid.szSize.cy;
					bPose = bViewPoseLocked[1] = true;
				}
				else
				{
					if (x >= hlViewPoseGrid.szSize.cx / 4)
					{	/* clicked on grip */
						boxLock.left = hlViewPoseGrid.szSize.cx / 4;
						boxLock.top = 0;
						boxLock.right = hlViewPoseGrid.szSize.cx / 2;
						boxLock.bottom = hlViewPoseGrid.szSize.cy;
						bView = bViewPoseGripped = bViewPoseLocked[0] = true;
						bViewPoseKnobbed = false;
					}
					else
					{	/* clicked on knob */
						bViewFix = bViewPoseKnobbed = true;
						bViewPoseGripped = false;
					}
				}
				if (bViewPoseLocked[0] || bViewPoseLocked[1])
				{
					boxLock.left += mfd_ss_items[C_MFD_VIEWPOSE].haArea.left;
					boxLock.right += mfd_ss_items[C_MFD_VIEWPOSE].haArea.left;
					boxLock.top += mfd_ss_items[C_MFD_VIEWPOSE].haArea.top;
					boxLock.bottom += mfd_ss_items[C_MFD_VIEWPOSE].haArea.top;
					HUD_lock_cursor(&boxLock);
				}
			}
			else
			{
				if (bViewPoseLocked[0] || bViewPoseLocked[1])
				{
					bViewPoseLocked[0] = bViewPoseLocked[1] = false;
					HUD_unlock_cursor();
				}
			}
			rCode = true;
			break;
		}
		case C_HUD_MOUSE_ACTIVATE:
		case C_HUD_MOUSE_GRAB:
		default:	break;
	}
	if (bPose)
	{
/*
		printf("Posing to <-> %g | %g\n",
					 dLean, dHeight);
*/
	}
	if (bView)
	{
/*
		printf("Viewing %g\n",
					 dHeight * 2.0 - 1.0);
*/
	}
	if (bViewFix)
	{

	}

	return rCode;
}

void	mfd_render_viewpose(P_HUD_HOOK_RENDER)
{	
	Matrix matPlayer = Player_get_orientation(0);
	double dLean = Player_get_lean(0), dSqueeze;
	int nShield = Player_get_shield(0), nLean;
	uint16 nHeight = (uint16)((double)(hlViewPoseGrid.szSize.cy) * Player_get_posture_height(0) + 0.5);

	if (nDetailLevel >= C_MFD_DETAIL_WHOA)
		palViewPoseGrid[0] = rgbt(0x00, 0xFF, 0x00, 0x10);
	else
		palViewPoseGrid[0] = rgbt(0x00, 0x00, 0x00, 0x00);

	if (nHeight)
		nHeight--;
	nLean = 1.0 + dLean + sgn(dLean) * 0.5;
	dSqueeze = dLean * sgn(dLean);


	/* calculate Model position and size */
	hlViewPoseModel.szSize.cx = hlViewPoseGrid.szSize.cx / 3;
	hlViewPoseModel.szSize.cx -= (uint16)(((double)(hlViewPoseModel.szSize.cx) / 2.0) * dSqueeze);

	hlViewPoseModel.ptPos.x = hlViewPoseGrid.ptPos.x + hlViewPoseGrid.szSize.cx * 3 / 4 - hlViewPoseModel.szSize.cx / 2;
	hlViewPoseModel.ptPos.x += (double)(hlViewPoseGrid.szSize.cx / 2 - hlViewPoseGrid.szSize.cx / 3) * dLean;

	hlViewPoseModel.szSize.cy = nHeight;
	hlViewPoseModel.ptPos.y = hlViewPoseGrid.ptPos.y + hlViewPoseGrid.szSize.cy - 
														hlViewPoseModel.szSize.cy;
	hlViewPoseModel.pG2DB = &gbViewPoseModel[Player_get_biosuit(0)]
																					[nHeight * 3 / hlViewPoseGrid.szSize.cy]
																					[nLean];
	appendHudLayer(pLayer, hlViewPoseGrid);
	appendHudLayer(pLayer, hlViewPoseModel);
	if (nShield)
	{	/* apply any shield */
		hlViewPoseShield.ptPos = hlViewPoseModel.ptPos;
		hlViewPoseShield.szSize = hlViewPoseModel.szSize;
		hlViewPoseShield.pG2DB = &gbViewPoseShield[nShield - 1]
																							[nHeight * 3 / hlViewPoseGrid.szSize.cy]
																							[nLean];
		appendHudLayer(pLayer, hlViewPoseShield);
	}
	else
		hlViewPoseShield.pG2DB = NULL;
	/* draw Grip */
	hlViewPoseGrip.ptPos.y = hlViewPoseGrid.ptPos.y +
													 (hlViewPoseGrid.szSize.cy / 2) -
													 (hlViewPoseGrid.szSize.cy / 2) * matPlayer.z.z -
													 hlViewPoseGrip.szSize.cy / 2;
	hlViewPoseGrip.pG2DB = &gbViewPoseGrip[bViewPoseGripped];
	appendHudLayer(pLayer, hlViewPoseGrip);
	/* draw Knob */
	hlViewPoseKnob.ptPos.y = hlViewPoseGrid.ptPos.y + 
													 (hlViewPoseGrid.szSize.cy / 3 + hlViewPoseGrip.ptPos.y * 3) / 
													 hlViewPoseGrid.szSize.cy * 
													 (hlViewPoseGrid.szSize.cy / 3);
	hlViewPoseKnob.pG2DB = &gbViewPoseKnob[bViewPoseKnobbed];
	appendHudLayer(pLayer, hlViewPoseKnob);
}

/*                      */
/* health               */
/*                      */

/* begin section health */

#define	D_HEALTH_MAXTICKS		23

pcolor			 palHealthBar[0x100];

g2dbase			 gbHealthBar[2] = { D_G2D_NULLAREA, D_G2D_NULLAREA };
g2dbase			 gbHealthTicker[3] = { D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA };
g2dbase			 gbHealthIcon[2][3] = 
{ 
	{ D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA },
	{ D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA }
};

hudLayer		 hlHealthBar[2] = 
{ 
	D_HUD_NULLLAYER(&gbHealthBar[0]), D_HUD_NULLLAYER(&gbHealthBar[1])
};
hudLayer		 hlHealthTicker[2][D_HEALTH_MAXTICKS];
hudLayer		 hlHealthIcon[2][3] = 
{
	{ D_HUD_NULLLAYER(&gbHealthIcon[0][0]), D_HUD_NULLLAYER(&gbHealthIcon[0][1]), D_HUD_NULLLAYER(&gbHealthIcon[0][2]) },
	{ D_HUD_NULLLAYER(&gbHealthIcon[1][0]), D_HUD_NULLLAYER(&gbHealthIcon[1][1]), D_HUD_NULLLAYER(&gbHealthIcon[1][2]) }
};

bool mfd_rinit_health(size szScreen)
{
	int i;
	uint16 nWidth, nHeight;
	bool bVisible = mfd_ss_items[C_MFD_HEALTH].haVisible, rCode = true;

	mfd_ss_items[C_MFD_HEALTH].haVisible = false;
											
	mfd_ss_items[C_MFD_HEALTH].haArea.right = szScreen.cx - 
			(int)((double)(szScreen.cx) * C_MFD_TOPITEMS_SPACE / 100.0);
	mfd_ss_items[C_MFD_HEALTH].haArea.left = 
			mfd_ss_items[C_MFD_VIEWPOSE].haArea.right + 
			(int)((double)(szScreen.cx) * C_MFD_TOPITEMS_SPACE / 100.0);
	mfd_ss_items[C_MFD_HEALTH].haArea.top = 
				(int)((double)(szScreen.cy) * C_MFD_TOPITEMS_TOP / 100.0);
	mfd_ss_items[C_MFD_HEALTH].haArea.bottom = 
			mfd_ss_items[C_MFD_VIEWPOSE].haArea.bottom;
	mfd_ss_items[C_MFD_HEALTH].haVisible = bVisible;

	nWidth = (uint16)getRectWidth(mfd_ss_items[C_MFD_HEALTH].haArea);
	nHeight = (uint16)getRectHeight(mfd_ss_items[C_MFD_HEALTH].haArea);

	for (i = 0; i < D_HEALTH_MAXTICKS; i++)
		memset(&hlHealthTicker[0][i], 0x00, sizeof(hudArea));

	/* init palette */
	Pal_copy_palette(palHealthBar, RGB_trans, 0x20);
	/* load up images */
	rCode = mfd_load_bitmap(&gbHealthBar[0], CHUNK_SCREEN_STATUS, SUBCHUNK_SCR_STAT_HEALTHBAR);
	rCode = rCode && mfd_load_bitmap(&gbHealthBar[1], CHUNK_SCREEN_STATUS, SUBCHUNK_SCR_STAT_POWERBAR);
	for (i = 0; i < 3 && rCode; i++)
	{
		rCode = mfd_load_bitmap(&gbHealthTicker[i], CHUNK_SCREEN_STATUS, SUBCHUNK_SCR_STAT_TICKER_FIRST + i);
		rCode = rCode && mfd_load_bitmap(&gbHealthIcon[0][i], CHUNK_SCREEN_STATUS, SUBCHUNK_SCR_STAT_HEALTH_FIRST + i);
		rCode = rCode && mfd_load_bitmap(&gbHealthIcon[1][i], CHUNK_SCREEN_STATUS, SUBCHUNK_SCR_STAT_POWER_FIRST + i);
	}
	if (rCode)
	{	/* build up hudAreas */
		gbHealthBar[0].pPal = gbHealthBar[1].pPal = palHealthBar;
		gbHealthBar[0].nPalSize = gbHealthBar[1].nPalSize = 0xFF;

		hlHealthBar[0].ptPos.x = hlHealthBar[1].ptPos.x = mfd_ss_items[C_MFD_HEALTH].haArea.left;
		hlHealthBar[0].ptPos.y = hlHealthBar[1].ptPos.y = mfd_ss_items[C_MFD_HEALTH].haArea.top;
		hlHealthBar[0].szSize.cx = hlHealthBar[1].szSize.cx = nWidth;
		hlHealthBar[0].szSize.cy = hlHealthBar[1].szSize.cy = nHeight / 2;
		hlHealthBar[1].ptPos.y += nHeight - hlHealthBar[1].szSize.cy;
	
		/* !!!!! realign icons better */
		for (i = 0; i < 3; i++)
		{	
			hlHealthIcon[0][i].szSize.cx = hlHealthIcon[0][i].szSize.cy = nHeight / 3;
			hlHealthIcon[0][i].ptPos.x = hlHealthBar[0].ptPos.x + nWidth - (nHeight * 2 / 5);
			hlHealthIcon[0][i].ptPos.y = hlHealthBar[0].ptPos.y + ((nHeight / 4) + hlHealthIcon[0][i].szSize.cy / 8) / 2;
			
			hlHealthIcon[1][i].szSize.cx = hlHealthIcon[1][i].szSize.cy = nHeight / 3;
			hlHealthIcon[1][i].ptPos.x = hlHealthBar[1].ptPos.x + nWidth - (nHeight * 2 / 5);
			hlHealthIcon[1][i].ptPos.y = (nHeight / 2) + hlHealthIcon[0][i].ptPos.y;
		}
		for (i = 0; i < D_HEALTH_MAXTICKS; i++)
		{
			hlHealthTicker[0][i].pG2DB = hlHealthTicker[1][i].pG2DB = &(gbHealthTicker[i / 8]);

			hlHealthTicker[0][i].szSize.cx = hlHealthTicker[0][i].szSize.cy = nHeight / 3;
			hlHealthTicker[0][i].ptPos.x = hlHealthBar[0].ptPos.x + 
																		 (nWidth * i / (D_HEALTH_MAXTICKS + 3));
			hlHealthTicker[0][i].ptPos.y = hlHealthBar[0].ptPos.y + 
																		 ((nHeight / 4) + hlHealthTicker[0][i].szSize.cy / 8) / 2;

			hlHealthTicker[1][i].szSize.cx = hlHealthTicker[1][i].szSize.cy = nHeight / 3;
			hlHealthTicker[1][i].ptPos.x = hlHealthBar[1].ptPos.x + (nWidth * i / (D_HEALTH_MAXTICKS + 3));
			hlHealthTicker[1][i].ptPos.y = (nHeight / 2) + hlHealthTicker[0][i].ptPos.y;
		}
	}

	return rCode;
}

void	mfd_rdeinit_health(void)
{
	int i;

	G2D_reset_base(&gbHealthBar[0], true);
	G2D_reset_base(&gbHealthBar[1], true);
	for (i = 0; i < 3; i++)
	{
		G2D_reset_base(&gbHealthTicker[i], true);
		G2D_reset_base(&gbHealthIcon[0][i], true);
		G2D_reset_base(&gbHealthIcon[1][i], true);
	}
}

void	mfd_render_health(P_HUD_HOOK_RENDER)
{
	int nHealthTicks = (D_HEALTH_MAXTICKS * Player_get_health(0) + 0.5), 
			nPowerTicks = (D_HEALTH_MAXTICKS * Player_get_power(0) + 0.5), 
			nHealthStatus, nPowerStatus, i;

	/* show background */
	if (nDetailLevel >= C_MFD_DETAIL_WHOA)
	{	
		appendHudLayer(pLayer, hlHealthBar[0]);
		appendHudLayer(pLayer, hlHealthBar[1]);
	}
	/* show icons */
	nHealthStatus = nHealthTicks / ((D_HEALTH_MAXTICKS / 3) + 1);
	nPowerStatus = nPowerTicks / ((D_HEALTH_MAXTICKS / 3) + 1);
	appendHudLayer(pLayer, hlHealthIcon[0][nHealthStatus]);
	appendHudLayer(pLayer, hlHealthIcon[1][nPowerStatus]);
	/* show ticker */
	for (i = 0; i < nHealthTicks; i++)
		appendHudLayer(pLayer, hlHealthTicker[0][i]);
	for (i = 0; i < nPowerTicks; i++)
		appendHudLayer(pLayer, hlHealthTicker[1][i]);
}

/*                      */
/* sensaround           */
/*                      */

/* begin section sensaround */

/*
#define	D_SA_DEBUG
*/
#define	C_SA_TIME				500						/* 500 msec */
#define	C_SA_MAXDOTS		 10

bool			 bSAActive = false;
uint32		 nSAStep = 0;
HLAYER		 hSALayer[3] = { NULL, NULL, NULL };
camera		 camSensAround[3];
pcolor		 palSensAround[3];
g2dbase		 gbSensAround[3] = { D_G2D_NULLAREA, D_G2D_NULLAREA, D_G2D_NULLAREA };
g2dbase		 gbSensAroundProc[C_SA_MAXDOTS];
hudLayer	 hlSensAround[3] = { D_HUD_NULLLAYER(&gbSensAround[0]), 
															 D_HUD_NULLLAYER(&gbSensAround[1]),
															 D_HUD_NULLLAYER(&gbSensAround[2]) };
hudLayer	 hlSensAroundProc[C_SA_MAXDOTS];

void	sa_prerender(void);

void	sa_timer_center(uint32 now, uint32 last, void *pData);
void	sa_render(P_MFD_HOOK_RENDER);
void	sa_deinit(P_MFD_HOOK_DEINIT);

MFD_hooks	mfdSACards =											
{	NULL, NULL, NULL,									sa_render,					sa_deinit, NULL };

bool mfd_rinit_SA(size szScreen)
{
	ss_bitmap_header *pHead;
	double Screen_scale = 1. / (2 * tan (Option_fov * M_PI / 360.));
	int i, nCards[3] = { C_MFD_LEFT, C_MFD_CENTER, C_MFD_RIGHT },
			nWidth, nHeight;	
	char strBuf[50], *pStrings[2] = { NULL, NULL };

	UNUSED(szScreen);
	memset(camSensAround, 0x00, sizeof(camera) * 3);
	for (i = 0; i < 3; i++)
	{
		hlSensAround[i].ptPos.x = mfd_ss_items[nCards[i]].haArea.left;
		hlSensAround[i].ptPos.y = mfd_ss_items[nCards[i]].haArea.top;
		hlSensAround[i].szSize.cx = camSensAround[i].window_width = 
				getRectWidth(mfd_ss_items[nCards[i]].haArea);
		hlSensAround[i].szSize.cy = camSensAround[i].window_height = 
				getRectHeight(mfd_ss_items[nCards[i]].haArea);
		camSensAround[i].screen_z = hlSensAround[i].szSize.cx * Screen_scale;
		camSensAround[i].flags = 0; /* jim - not main, has framebuf */
		Clip_init_camera_frustrum(&camSensAround[i]);
	}

	palSensAround[0] = rgbt(0x00, 0x00, 0x00, 0x00);
	sprintf(strBuf, "Processing");
	pStrings[0] = strBuf;
	for (i = 0; i < C_SA_MAXDOTS; i++)
	{
		Font_get_size(0, strBuf, &nWidth, &nHeight);
		pHead = G2D_write_string(pStrings, 1, 0, F_WSTR_DRAW_BORDER,
														 0, nWidth + 2);
		if (pHead)
		{
			convertHeadToArea(*pHead, gbSensAroundProc[i]);
			gbSensAroundProc[i].pPal = palSensAround;
			gbSensAroundProc[i].nPalSize = 0x02;
			free(pHead);
		}
		else
			G2D_reset_base(&gbSensAroundProc[i], false);
		HUD_reset_layer(&hlSensAroundProc[i], &gbSensAroundProc[i]);
		hlSensAroundProc[i].ptPos = hlSensAround[C_MFD_SIDE_CENTER].ptPos;
		hlSensAroundProc[i].szSize = gbSensAroundProc[i].bSizeImg;

		strcat(strBuf, ".");
	}

	Timer_register_function("MFD.Sensaround", 20, sa_timer_center, NULL);

	return true;
}

void mfd_rdeinit_SA(void)
{
	int i;

	Timer_unregister_function("MFD.Sensaround");
	MFD_show_sensaround(0);
	for (i = 0; i < C_SA_MAXDOTS; i++)
	{
		G2D_reset_base(&gbSensAroundProc[i], true);
	}
}

void MFD_show_sensaround(bool bShow)
{
	int i, nFlags[3] = { F_SMFD_SENSA, F_CMFD_SENSA, F_SMFD_SENSA },
			nSAVer = Player_get_ver_sensaround(0);

	bSAActive = bShow;
	Evt_fire_event(Evt_get_event(Evt_get_class("MFD"), "SensaroundShow"),
								 &bSAActive, false);
	if(bSAActive)										/* turning sa on*/
	{
		MFD_set_HW_activated(C_MFD_HW_SENSA, true); 
		nSAStep = 0;													/* reset counter */
		for (i = (nSAVer == C_MFD_SENS_LVL3 ? 0 : 1); 
				 i < (nSAVer == C_MFD_SENS_LVL3 ? 3 : 2); i++)
		{
			if (gbSensAround[i].bImg.pxl == NULL)			/* no mem leak... */
			{ /* init memory for areas */
				if (!G2D_init_pxlbase(&gbSensAround[i], hlSensAround[i].szSize.cx,
															hlSensAround[i].szSize.cy))
				{
					Cant_happen("MFD_show_sensaround: Couldn't init buffer");
				}
				memset(gbSensAround[i].bImg.pxl, 0x00, sizeof(pixel) *
							 gbSensAround[i].bSizeImg.cx * gbSensAround[i].bSizeImg.cy);
				camSensAround[i].framebuffer = gbSensAround[i].bImg.pxl;
				hSALayer[i] = mfd_add_mfd_layer(i);
				mfd_set_layer(hSALayer[i], &mfdSACards, 
											mfd_set_id_s(nFlags[i], 0, nSAVer));
			}
			else
			{	/* stop fadeout */
				mfd_detach_mfd_layer(hSALayer[i]);
				mfd_append_mfd_layer_after(hSALayer[i], NULL);
			}
			hSALayer[i]->lStatus = C_LAYER_STATUS_SEMI;
			hSALayer[i]->lStatusTick = 0;
		}
	}
	else
	{
		MFD_set_HW_activated(C_MFD_HW_SENSA, false);
		for (i = 0; i < 3; i++)
			mfd_update_layer_flags(hSALayer[i], C_LAYER_STATUS_FADEOUT, true);
	}
}

bool MFD_is_sensaround_active()
{
	return bSAActive;
}

void	sa_prerender(void)
{
	Matrix matPlayer = Player_get_orientation(0);
	Vector vecBuffer, vecPlayer = Player_get_position(0);
	int i;
	
	for (i = 0; i < 3; i++)
	{
		memcpy(&camSensAround[i].position, &vecPlayer, sizeof(Vector));
		memcpy(&camSensAround[i].orientation, &(matPlayer), sizeof(Matrix));
		if (gbSensAround[i].bImg.pxl)
			memset(gbSensAround[i].bImg.pxl, 0x00, sizeof(pixel) *
						 gbSensAround[i].bSizeImg.cx * gbSensAround[i].bSizeImg.cy);
	}
#ifndef D_SA_DEBUG
	/* left eye */
	vecBuffer = camSensAround[0].orientation.z;
	camSensAround[0].orientation.z = camSensAround[0].orientation.x;
	vec_multiply(camSensAround[0].orientation.z, -1);
	camSensAround[0].orientation.x = vecBuffer;
	/* back eye */
	vec_multiply(camSensAround[1].orientation.z, -1);
	vec_multiply(camSensAround[1].orientation.x, -1);
	/* right eye */
	vecBuffer = camSensAround[2].orientation.x;
	camSensAround[2].orientation.x = camSensAround[2].orientation.z;
	vec_multiply(camSensAround[2].orientation.x, -1);
	camSensAround[2].orientation.z = vecBuffer;
#endif
	for (i = 0; i < 3; i++)
		if (camSensAround[i].framebuffer)
			Ren_scene(&camSensAround[i]);
}

void	sa_timer_center(uint32 now, uint32 last, void *pData)
{
	UNUSED(pData);
	if (Player_get_ver_sensaround(0) == 1)
	{
		if (!nSAStep || nSAStep >= C_SA_TIME)
		{
			sa_prerender();
			nSAStep = 0;
		}
		if (last != 0)
			nSAStep += (now - last);
	}
}

void	sa_render(P_MFD_HOOK_RENDER)
{
	HSHADE shShade = shade_get(hLayer->lShade);
	int nDots, nSAVer = Player_get_ver_sensaround(0);

	if (hLayer->lShade && hSALayer[hLayer->lSide] &&
			gbSensAround[hLayer->lSide].bImg.pxl)
	{
		if (hLayer->lSide == C_MFD_SIDE_CENTER && nSAVer > 1)
			sa_prerender();

		gbSensAround[hLayer->lSide].bOp.clr = rgbt(0x00, 0x00, 0x00, hLayer->lShade);
		appendHudLayer(pLayer, hlSensAround[hLayer->lSide]);

		if (hLayer->lSide == C_MFD_SIDE_CENTER && nSAVer == 1)
		{	/* print 'processing.....' */
			palSensAround[1] = rgbt(0x00, 0x00, 0xFF, shade(shShade, 0x00, 0x40));
			palSensAround[2] = rgbt(0x00, 0x00, 0xFF, shade(shShade, 0x00, 0xFF));
			nDots = nSAStep * C_SA_MAXDOTS / C_SA_TIME;
			nDots = min(nDots, C_SA_MAXDOTS - 1);
			appendHudLayer(pLayer, hlSensAroundProc[nDots]);
		}
	}
}

void	sa_deinit(P_MFD_HOOK_DEINIT)
{
	camSensAround[hLayer->lSide].framebuffer = NULL;
	G2D_reset_base(&gbSensAround[hLayer->lSide], true);

	hSALayer[hLayer->lSide] = NULL;
	if (!hSALayer[0] && !hSALayer[1] && !hSALayer[2])
		MFD_show_sensaround(false);
}


/*                      */
/* video mail           */
/*                      */

/* begin section vidmail */

/*
#define MFD_DUMP_TGA
*/

#define	C_MFD_VIDMAIL_TOP		 25.0
#define	C_MFD_VIDMAIL_WIDTH	 62.5
#define	C_MFD_VIDMAIL_HEIGHT 41.7

#define C_VIDMAIL_PARTS				7
#define C_VIDMAIL_FRAMETIME	100

typedef struct tag_VidMailFrameset
{
/* resfile data */
	byte					 vmfTag;
	byte					 vmfFrom;
	byte					 vmfTo;
	uint16				 vmfTime;
/* private data */

} VidMailFrameset;

typedef struct tag_VidMailHeader
{
/* resfile data */
	uint16				 vmhWidth;
	uint16				 vmhHeight;
	uint16				 vmhChunkID;
	
	byte					 vmhUnk1[6];
	uint16				 vmhUnk2;

	VidMailFrameset	*vmhFrames;

	uint16				 vmhEnd;
/* private data */
	uint16				 vmhThisChunkID;

	byte					 vmhVidMailID;
	byte					 vmhCurPart;
	byte					 vmhFramesets;

	byte					 vmhCurFrame;
	uint32				 vmhMailTime;
} VidMailHeader;

const uint16 listVidMailIDs[C_VIDMAIL_AMOUNT][C_VIDMAIL_PARTS] = 
{
	{ 0x0A56, 0x0A57, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 },
	{ 0x0A56, 0x0A4C, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 },
	{ 0x0A56, 0x0A4F, 0x0A50, 0x0000, 0x0000, 0x0000, 0x0000 }, 
	{ 0x0A56, 0x0A4D, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 }, 
	{ 0x0A56, 0x0A4E, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 }, 
	{ 0x0A56, 0x0A51, 0x0A52, 0x0A53, 0x0A54, 0x0A55, 0x0000 } 
};

const uint16 listVidMailTexts[C_VIDMAIL_AMOUNT] =
{ 4, 1, 3, 0, 2, 5 };

VidMailHeader vmhCurVidMail = 
{ 0, 0, 0, { 0, 0, 0, 0, 0, 0 }, 0, NULL, 0, 0, 0, 0, 0, 0, 0 };

pcolor					 palVidMail[0x100];
g2dbase					 gbVidMailBackBuffer = D_G2D_NULLAREA;
hudLayer				 hlVidMail = D_HUD_NULLLAYER(&gbVidMailBackBuffer);

ss_bitmap_header **bmpVidMailFrames = NULL;

void	vidmail_reset_frames(void);
void	vidmail_reset_header(void);
void	vidmail_reset_backbuffer(void);
void	vidmail_init_backbuffer(void);
void	vidmail_prebuffer_frames(void);
bool	vidmail_load_header(byte id, byte part);
int		vidmail_get_frameset(int frame);
void	vidmail_render_frame(int frame);

bool mfd_rinit_vidmail(size szScreen)
{
	bool bVisible = mfd_ss_items[C_MFD_VIDMAIL].haVisible;

	mfd_ss_items[C_MFD_VIDMAIL].haVisible = false;
	mfd_ss_items[C_MFD_VIDMAIL].haArea.left = szScreen.cx / 2 -
			(int)((double)(szScreen.cx) * C_MFD_VIDMAIL_WIDTH / 100.0) / 2;
	mfd_ss_items[C_MFD_VIDMAIL].haArea.top = 
			(int)((double)(szScreen.cy) * C_MFD_VIDMAIL_TOP / 100.0);
	mfd_ss_items[C_MFD_VIDMAIL].haArea.right = 
			szScreen.cx - mfd_ss_items[C_MFD_VIDMAIL].haArea.left;
	mfd_ss_items[C_MFD_VIDMAIL].haArea.bottom = 
			mfd_ss_items[C_MFD_VIDMAIL].haArea.top + 
			(int)((double)(szScreen.cy) * C_MFD_VIDMAIL_HEIGHT / 100.0);

	Pal_copy_palette(palVidMail, RGB_trans, 0xFF);

	hlVidMail.ptPos.x = mfd_ss_items[C_MFD_VIDMAIL].haArea.left;
	hlVidMail.ptPos.y = mfd_ss_items[C_MFD_VIDMAIL].haArea.top;
	hlVidMail.szSize.cx = (uint16)getRectWidth(mfd_ss_items[C_MFD_VIDMAIL].haArea);
	hlVidMail.szSize.cy = (uint16)getRectHeight(mfd_ss_items[C_MFD_VIDMAIL].haArea);

	mfd_ss_items[C_MFD_VIDMAIL].haVisible = bVisible;
	Timer_register_function("MFD.Vidmail", 100,
													mfd_timer_vidmail, NULL);

	return true;
}

void	mfd_rdeinit_vidmail(void)
{
	Timer_unregister_function("MFD.Vidmail");
	/* clean up any residing frames */
	vidmail_reset_backbuffer();
	vidmail_reset_header();
}

void	vidmail_reset_frames(void)
{
	int i, amount;

	/* clear frames */
	if (vmhCurVidMail.vmhFrames != NULL &&
			bmpVidMailFrames != NULL)
	{
		amount = vmhCurVidMail.vmhFrames[vmhCurVidMail.vmhFramesets - 1].vmfTo + 1;
		for (i = 0; i < amount; i++)
		{
			free(bmpVidMailFrames[i]->bits);
			free(bmpVidMailFrames[i]);
		}
		free(bmpVidMailFrames);
		bmpVidMailFrames = NULL;
	}
}

void	vidmail_reset_header(void)
{
	vidmail_reset_frames();
	/* clear header */
	free(vmhCurVidMail.vmhFrames);
	memset(&vmhCurVidMail, 0x00, sizeof(VidMailHeader));
}

void	vidmail_reset_backbuffer(void)
{
	G2D_reset_base(&gbVidMailBackBuffer, true);		/* clear back buffer */	
}

void	vidmail_init_backbuffer(void)
{
	vidmail_reset_backbuffer();
	G2D_init_bitbase(&gbVidMailBackBuffer, vmhCurVidMail.vmhWidth, 
																				 vmhCurVidMail.vmhHeight);
	memset(gbVidMailBackBuffer.bImg.bit, 0x00, vmhCurVidMail.vmhWidth *
																						 vmhCurVidMail.vmhHeight);
	gbVidMailBackBuffer.pPal = palVidMail;
	gbVidMailBackBuffer.nPalSize = 0xFF;
}

void	vidmail_prebuffer_frames(void)
{
	int i, amount;

	if (vmhCurVidMail.vmhFrames != NULL)
	{
		amount = vmhCurVidMail.vmhFrames[vmhCurVidMail.vmhFramesets - 1].vmfTo + 1;
		bmpVidMailFrames = malloc(sizeof(ss_bitmap_header *) * amount);
		memset(bmpVidMailFrames, 0x00, sizeof(ss_bitmap_header *) * amount);
		for (i = 0; i < amount; i++)
		{
			if (!(vmhCurVidMail.vmhChunkID == 0x0A4A && i == 0))
				bmpVidMailFrames[i] = Bmp_load_bitmap(vmhCurVidMail.vmhChunkID, i);
			else
			{ /* lousy hack ;) */
				bmpVidMailFrames[i] = Bmp_create(vmhCurVidMail.vmhWidth,
																				 vmhCurVidMail.vmhHeight);
				memset(bmpVidMailFrames[i]->bits, 0x00, 
							 vmhCurVidMail.vmhWidth * vmhCurVidMail.vmhHeight);
			}
		}
		Res_free_chunk_by_id(vmhCurVidMail.vmhChunkID);
	}
}

bool	vidmail_load_header(byte id, byte part)
{
	VidMailFrameset *vmfBuf;
	int i;			 
	unsigned int size;
	block_handle bh = Res_load_subchunk(listVidMailIDs[id][part], 
																			0, &size);
	byte *pBuf = *bh;
	bool rCode = false;
									 
	if (bh != NULL && *bh != NULL)
	{
		/* be nice */
		vidmail_reset_header();	
		/* init private data */
		vmhCurVidMail.vmhVidMailID = id;
		vmhCurVidMail.vmhCurPart = part;
		vmhCurVidMail.vmhThisChunkID = listVidMailIDs[id][part];
		/* read header part 1 */
		vmhCurVidMail.vmhWidth =		pBuf[0x00] + (pBuf[0x01] << 8);
		vmhCurVidMail.vmhHeight =		pBuf[0x02] + (pBuf[0x03] << 8);
		vmhCurVidMail.vmhChunkID =	pBuf[0x04] + (pBuf[0x05] << 8);
		memcpy(vmhCurVidMail.vmhUnk1, &(pBuf[6]), 6);
		vmhCurVidMail.vmhUnk2 =			pBuf[0x0C] + (pBuf[0x0D] << 8);
		i = 0x0E;
		/* read subtable(s) */
		while (pBuf[i] == 0x04)
		{
			/* new buffer */
			vmfBuf = malloc(sizeof(VidMailFrameset) * 
											(vmhCurVidMail.vmhFramesets + 1));
			memset(vmfBuf, 0x00, sizeof(VidMailFrameset) * 
														(vmhCurVidMail.vmhFramesets + 1));
			memcpy(vmfBuf, vmhCurVidMail.vmhFrames, 
						 sizeof(VidMailFrameset) * vmhCurVidMail.vmhFramesets);
			/* reassign buffer */
			free(vmhCurVidMail.vmhFrames);
			vmhCurVidMail.vmhFrames = vmfBuf;
			/* load up subentry */
			vmfBuf = &(vmhCurVidMail.vmhFrames[vmhCurVidMail.vmhFramesets]);

			vmfBuf->vmfTag =	pBuf[i    ];
			vmfBuf->vmfFrom = pBuf[i + 1];
			vmfBuf->vmfTo =		pBuf[i + 2];
			vmfBuf->vmfTime = pBuf[i + 3] + (pBuf[i + 4] << 8);

			i += 5;
			vmhCurVidMail.vmhFramesets++;
		}
		/* read end tag - part 2 */
		vmhCurVidMail.vmhEnd = pBuf[i] + (pBuf[i + 1] << 8);
		/* be nice :) */
		Res_free_chunk_by_id(id);

		rCode = true;
	}

	return rCode;
}

int vidmail_get_frameset(int frame)
{
	int i, index = -1;

	for (i = 0; i < vmhCurVidMail.vmhFramesets && index == -1; i++)
	{
		if (frame >= vmhCurVidMail.vmhFrames[i].vmfFrom &&
				frame <= vmhCurVidMail.vmhFrames[i].vmfTo)
			index = i;
	}

	return index;
}

void	vidmail_render_frame(int frame)
{
	byte *pIn, *pOut;
	int i, j, bytes; /* , index = -1; */

#ifdef MFD_DUMP_TGA	
	char fname[20];

	sprintf(fname, "%04X-%02X", vmhCurVidMail.vmhChunkID, frame);
	Exp_write_tga8(bmpVidMailFrames[frame]->width, bmpVidMailFrames[frame]->height,
								 bmpVidMailFrames[frame]->bits, 
								 fname, RGB_trans, "TriOptimum VidMail");
#endif

	bytes = vmhCurVidMail.vmhWidth * vmhCurVidMail.vmhHeight;
	memset(gbVidMailBackBuffer.bImg.bit, 0x00, bytes);
	for (i = 0; i <= frame; i++)					
	{
/*
		index = vidmail_get_frameset(i);
		if (index >= 0 &&
				(i + vmhCurVidMail.vmhFrames[index].vmfTime / 
				 C_VIDMAIL_FRAMETIME) >= frame)
*/
		{
/*
printf("target: %02d | rendering: %02d\n", frame, i);
*/
			pOut = gbVidMailBackBuffer.bImg.bit;
			pIn = bmpVidMailFrames[i]->bits;
			for(j = 0; j < bytes; j++)
			{
				if (*pIn)
					*pOut = *pIn;
				pOut++;
				pIn++;
			}
		}
	}
/*
printf("---\n");
*/
}

bool	MFD_show_vidmail(byte id)
{
	bool rCode = false;

	if (id < C_VIDMAIL_AMOUNT)
	{
		Evt_fire_event(Evt_get_event(Evt_get_class("MFD"), "VidmailShow"),
									 &id, false);
		mfd_show_item(C_MFD_VIDMAIL, false);
		vidmail_load_header(id, 0);

		vidmail_init_backbuffer();
		vidmail_prebuffer_frames();
		mfd_show_item(C_MFD_VIDMAIL, true);
		MFD_show_media_relative(C_MFD_MEDIA_VMAIL, listVidMailTexts[id]);
		rCode = true;
	}

	return rCode;
}

void	mfd_timer_vidmail(uint32 now, uint32 last, void *pData)
{
	int index;

	UNUSED(pData);
	if (gbVidMailBackBuffer.bImg.bit != NULL)
	{
		index = vidmail_get_frameset(vmhCurVidMail.vmhCurFrame);
		if (index >= 0)
		{
			vidmail_render_frame(vmhCurVidMail.vmhCurFrame);
		}
		vmhCurVidMail.vmhMailTime += (now - last);
/*
		vmhCurVidMail.vmhCurFrame = (byte)(vmhCurVidMail.vmhMailTime / 
																			 C_VIDMAIL_FRAMETIME);
*/
		vmhCurVidMail.vmhCurFrame++;
		index = vidmail_get_frameset(vmhCurVidMail.vmhCurFrame);
		if (index < 0)
		{
			if (vmhCurVidMail.vmhCurPart < (C_VIDMAIL_PARTS - 1) &&
					listVidMailIDs[vmhCurVidMail.vmhVidMailID]
												[vmhCurVidMail.vmhCurPart + 1] > 0)
			{
/*
printf("next\n");
*/
				vidmail_load_header(vmhCurVidMail.vmhVidMailID,
														(byte)(vmhCurVidMail.vmhCurPart + 1));
				vidmail_prebuffer_frames();
			}
			else
			{
/*
printf("finished\n");
*/
				/* movie finished */
			}
		}
	}
}

bool	mfd_mouse_vidmail(P_HUD_HOOK_MOUSE)
{
	bool rCode = false;

	UNUSED(x);
	UNUSED(y);

	if (bStart && bmpVidMailFrames)
	{
		switch (action)
		{
			case C_HUD_MOUSE_SELECT:
			case C_HUD_MOUSE_ACTIVATE:
			case C_HUD_MOUSE_GRAB:
			{
				vidmail_reset_backbuffer();
				vidmail_reset_header();
				break;
			}
			default:	break;
		}
		rCode = true;
	}

	return rCode;
}

void	mfd_render_vidmail(P_HUD_HOOK_RENDER)
{
	if (gbVidMailBackBuffer.bImg.bit)
	{
/*
		nStep = vmhCurVidMail.vmhMailTime / 2;
		nStep = min(nStep, 0xFF);
		Pal_copy_palette(palVidMail, RGB_trans, );
*/
		appendHudLayer(pLayer, hlVidMail);
	}
}

bool	mfd_destroy_vidmail(P_HUD_HOOK_DEST)
{
	return true;
}

/* begin section media */

char *strMediaEvents[C_MFD_MEDIA_AMOUNT] =
{
	"PlayEMail",
	"PlayLog",
	NULL,
	NULL,
	"PlayVMail"
};

MFD_MediaData *mfd_mediadata_create(short nId)
{
	MFD_MediaData *pData = (MFD_MediaData *)malloc(sizeof(MFD_MediaData));
	HEVENT hEvent = NULL;
	int nChunks = 0, nIndex = 0, nConvertId = nId;
	char **pChunks = NULL, *strBuf = NULL;

	memset(pData, 0x00, sizeof(MFD_MediaData));
	while (pData->mdType < C_MFD_MEDIA_AMOUNT &&
				 nId < nMediaRanges[pData->mdType][0])
		pData->mdType++;
	assert(pData->mdType < C_MFD_MEDIA_AMOUNT);
	assert(nId >= nMediaRanges[pData->mdType][0]);

	pData->mdpStrings = Str_load_string_chunk(nId, &(pData->mdLines));
	if (pData->mdLines && nSMediaPicBases[pData->mdType])
	{
		pData->mdSenderSubject = true;
		pChunks = Str_extract(pData->mdpStrings[0], " ");
		while (pChunks[nChunks])
		{
			if (strlen(pChunks[nChunks]))
			{
				if (pChunks[nChunks][0] == 'i' || pChunks[nChunks][0] == 'I')
					pData->mdFlagI = strtol(&pChunks[nChunks][1], NULL, 16);
				if (pChunks[nChunks][0] == 't' || pChunks[nChunks][0] == 'T')
					pData->mdFlagT = true;
				if (pChunks[nChunks][0] == 'c' || pChunks[nChunks][0] == 'C')
					pData->mdFlagC = strtol(&pChunks[nChunks][1], NULL, 16);
				if (!strpbrk(pChunks[nChunks], "ictICT") && nIndex < 2)
				{
					pData->mdPics[nIndex++] = atoi(pChunks[nChunks]);
					if ((strBuf = strchr(pChunks[nChunks], ',')) != NULL &&
							strBuf[1] != 0x00)
					{
						pData->mdPics[nIndex++] = atoi(&strBuf[1]);
					}
				}
			}
			free(pChunks[nChunks++]);
		}
		free(pChunks);
	}
	if (strMediaEvents[pData->mdType])
	{
		hEvent = Evt_get_event(Evt_get_class("MFD"), strMediaEvents[pData->mdType]);
		Evt_fire_event(hEvent, (void *)(nConvertId - nMediaRanges[pData->mdType][0]),
									 false);
	}

	return pData;
}

void mfd_mediadata_destroy(MFD_MediaData *pData)
{
	if (pData->mdpStrings)
		Str_free_string_chunk(pData->mdpStrings);

	free(pData);
}

void		MFD_show_media_relative(byte nMedia, short nId)
{
	assert(nMedia < C_MFD_MEDIA_AMOUNT);

	MFD_show_media((short)(nMediaRanges[nMedia][0] + nId));
}

void		MFD_show_media(short nId)
{
	MFD_CMedia *pMedia = NULL;
	HLAYER hLayerCenter = NULL;
	
	hLayerCenter = mfd_show_id(C_MFD_SIDE_CENTER, 
														 mfd_set_id_s(F_CMFD_MEDIA, 0x00, nId),
														 C_LAYER_STATUS_SEMI);
	pMedia = (MFD_CMedia *)hLayerCenter->lpData;
	if (pMedia)
	{
		if (pMedia->cmpData->mdPics[0] || pMedia->cmpData->mdSenderSubject)
			mfd_show_id(C_MFD_SIDE_LEFT,
									mfd_set_id_s(F_SMFD_ITEM, F_SMFD_IMEDIA, nId),
									C_LAYER_STATUS_SEMI);
		if (pMedia->cmpData->mdPics[1])
			mfd_show_id(C_MFD_SIDE_RIGHT,
									mfd_set_id_s(F_SMFD_ITEM, F_SMFD_IMEDIA, nId),
									C_LAYER_STATUS_SEMI);
	}
}

void MFD_show_fragment(short nId)
{
	UNUSED(nId);
#if 0
	MFD_media *pMedia = NULL;
	HLAYER hLayer = NULL;
	
	if ((SUBCHUNK_TXT_DATA_FIRST + nId) <= SUBCHUNK_TXT_DATA_LAST)
	{
		hLayer = mfd_show_id(C_MFD_SIDE_RIGHT, 
												 mfd_set_id_s(F_SMFD_ITEM, F_SMFD_IFRAG, nId),
												 C_LAYER_STATUS_ACTIVE);
		pMedia = mfd_media_create();
		hLayer->lpData = pMedia;
		pMedia->mhLayer[C_MFD_SIDE_RIGHT] = hLayer;
		pMedia->mpStrings = Str_load_string_chunk(CHUNK_TEXT_DATA,
																							&(pMedia->mLines));
/*
		{
			int i;

			printf("+++ orig\n");
			for (i = 0; i < pMedia->mLines && i < 15; i++)
			{
				printf("<%s>\n", pMedia->mpStrings[i]);
			}
			printf("---\n");
		}
*/
	}
	else
	{
		Cant_happen("MFD_show_fragment() Chunk Id out of range!");
	}
#endif
}


/*
 * End of src/mfd.c
 */

/*
 * Follows some emacs magic to get VC++-formatted code to display right (and
 *  hopefully vice versa)
 *
 * Local Variables:
 *  mode: c
 *  tab-width: 2
 *  indent-tabs-mode: nil
 * End:
 */
